/******************************************************************************
 *
 * Source file for harness-console (main) module.
 * Copyright (c) 2002-2006 Oak Ridge National Laboratory.
 *
 * For more information see the following files in the source distribution top-
 * level directory or package data directory (usually /usr/local/share/package):
 *
 * - README    for general package information.
 * - INSTALL   for package install information.
 * - COPYING   for package license information and copying conditions.
 * - AUTHORS   for package authors information.
 * - ChangeLog for package changes information.
 *
 ******************************************************************************/

/******************************************************************************
 *
 * Macros
 *
 ******************************************************************************/

/******************************************************************************
 *
 * Includes
 *
 ******************************************************************************/

/* Main module header file. */
#include "harness-console.h"

/******************************************************************************
 *
 * Data Types
 *
 ******************************************************************************/

/* Harness-console plug-in module data type. */
typedef struct {
  pthread_mutex_t  mutex;        /* mutex                       */
  struct {
    unsigned int   count;        /* handle array count          */
    unsigned int  *array;        /* handle array                */
  }                instances;    /* instances                   */
  unsigned int     harness_rmix; /* harness-rmix plug-in handle */
  rmix_localref_t *localref;     /* local object reference      */
} harness_console_data_t;

/******************************************************************************
 *
 * Function Prototypes
 *
 ******************************************************************************/

/*
 * Initializes the harness-console plug-in.
 *
 * handle = The plug-in instance handle.
 * return = 0 on success or -1 on any error with errno set appropriately.
 */
HARNESS_PLUGINS_INIT(harness_console_init);

/*
 * Finalizes the harness-console plug-in.
 *
 * handle = The plug-in instance handle.
 * return = 0 on success or -1 on any error with errno set appropriately.
 */
HARNESS_PLUGINS_FINI(harness_console_fini);

/******************************************************************************
 *
 * Data
 *
 ******************************************************************************/

/* Harness-console plug-in module data. */
harness_console_data_t harness_console_data = {
  PTHREAD_MUTEX_INITIALIZER, /* mutex                       */
  {
    0,                       /* handle array count          */
    NULL                     /* handle array                */
  },                         /* instances                   */
  0,                         /* harness-rmix plug-in handle */
  NULL                       /* local object reference      */
};

/* Harness-console plug-in "object". */
harness_console_t harness_console = {
  {
    LIBHARNESS_CONSOLE_VERSION_CURRENT,  /* current version         */
    LIBHARNESS_CONSOLE_VERSION_REVISION, /* current revision        */
    LIBHARNESS_CONSOLE_VERSION_AGE,      /* version age             */
    LIBHARNESS_CONSOLE_VERSION_FIRST     /* first supported version */
  },                                     /* plug-in version         */
  /**********************************************/
  /* PUT YOUR PUBLIC PLUG-IN DATA POINTERS HERE */
  /**********************************************/
  /**************************************************/
  /* PUT YOUR PUBLIC PLUG-IN FUNCTION POINTERS HERE */
  /**************************************************/
  harness_console_instances
};

/******************************************************************************
 *
 * Functions
 *
 ******************************************************************************/

/*
 * Initializes the harness-console plug-in.
 *
 * handle = The plug-in instance handle.
 * return = 0 on success or -1 on any error with errno set appropriately.
 */
#undef  __FUNC__
#define __FUNC__ "harness_console_init"
HARNESS_PLUGINS_INIT(harness_console_init) {
  int           error;
  void         *array;
  unsigned int  index;
  HARNESS_LOG((HARNESS_INFO(harness-console startup)))
  /* Lock harness-console plug-in mutex. */
  if (0 != (error = pthread_mutex_lock(&harness_console_data.mutex))) {
    errno = error;
    HARNESS_LOG((HARNESS_WARN(unable to lock harness-console plug-in mutex)))
    errno = error;
    harness_syserr();
    errno = error;
    return -1;
  }
  /* Search for handle in handle array. */
  for (index = 0; index < harness_console_data.instances.count; index++) {
    if (harness_console_data.instances.array[index] == handle) {
      errno = EALREADY;
      HARNESS_LOG((HARNESS_WARN(handle is already in handle array)))
      errno = EALREADY;
      harness_syserr();
      /* Unlock harness-console plug-in mutex. */
      if (0 != (error = pthread_mutex_unlock(&harness_console_data.mutex))) {
        errno = error;
        HARNESS_LOG((
          HARNESS_WARN(unable to unlock harness-console plug-in mutex)))
        errno = error;
        harness_syserr();
      }
      errno = EALREADY;
      return -1;
    }
  }
  /* Increase handle array. */
  index = harness_console_data.instances.count;
  harness_console_data.instances.count++;
  /* Reallocate handle array. */
  if (NULL == (array = realloc(harness_console_data.instances.array,
                               harness_console_data.instances.count *
                        sizeof(harness_console_data.instances.array[0])))) {
    int errno2 = errno;
    HARNESS_LOG((HARNESS_WARN(unable to reallocate handle array)))
    /* Decrease handle array. */
    harness_console_data.instances.count--;
    /* Unlock harness-console plug-in mutex. */
    if (0 != (error = pthread_mutex_unlock(&harness_console_data.mutex))) {
      errno = error;
      HARNESS_LOG((
        HARNESS_WARN(unable to unlock harness-console plug-in mutex)))
      errno = error;
      harness_syserr();
    }
    errno2 = errno;
    return -1;
  }
  harness_console_data.instances.array = array;
  /* Save handle array entry. */
  harness_console_data.instances.array[index] = handle;
  /* Check for first initialization. */
  if (1 == harness_console_data.instances.count) {
    /* Load harness-rmix plug-in. */
    if (0 != harness_plugins_load(&harness_console_data.harness_rmix,
                                  "libharness-rmix.0",
                                  HARNESS_PLUGINS_EXPORT)) {
      int errno2 = errno;
      HARNESS_LOG((HARNESS_WARN(unable to load harness-rmix plug-in)))
      /* Deallocate handle array. */
      free(harness_console_data.instances.array);
      harness_console_data.instances.array = NULL;
      harness_console_data.instances.count = 0;
      /* Unlock harness-console plug-in mutex. */
      if (0 != (error = pthread_mutex_unlock(&harness_console_data.mutex))) {
        errno = error;
        HARNESS_LOG((
          HARNESS_WARN(unable to unlock harness-console plug-in mutex)))
        errno = error;
        harness_syserr();
      }
      errno = errno2;
      return -1;
    }
    /* Export harness. */
    if (0 != harness_rmix_export4(&harness_console_data.localref,
                                  HARNESS_CONSOLE_EXPORT_RPC)) {
      int errno2 = errno;
      HARNESS_LOG((HARNESS_WARN(unable to export harness)))
      /* Unload harness-rmix plug-in. */
      if (0 != harness_plugins_unload(harness_console_data.harness_rmix)) {
        HARNESS_LOG((HARNESS_WARN(unable to unload harness-rmix plug-in)))
      }
      /* Deallocate handle array. */
      free(harness_console_data.instances.array);
      harness_console_data.instances.array = NULL;
      harness_console_data.instances.count = 0;
      /* Unlock harness-console plug-in mutex. */
      if (0 != (error = pthread_mutex_unlock(&harness_console_data.mutex))) {
        errno = error;
        HARNESS_LOG((
          HARNESS_WARN(unable to unlock harness-console plug-in mutex)))
        errno = error;
        harness_syserr();
      }
      errno = errno2;
      return -1;
    }
  }
  /* Unlock harness-console plug-in mutex. */
  if (0 != (error = pthread_mutex_unlock(&harness_console_data.mutex))) {
    errno = error;
    HARNESS_LOG((
      HARNESS_WARN(unable to unlock harness-console plug-in mutex)))
    errno = error;
    harness_syserr();
    errno = error;
    return -1;
  }
  return 0;
}

/*
 * Finalizes the harness-console plug-in.
 *
 * handle = The plug-in instance handle.
 * return = 0 on success or -1 on any error with errno set appropriately.
 */
#undef  __FUNC__
#define __FUNC__ "harness_console_fini"
HARNESS_PLUGINS_FINI(harness_console_fini) {
  int           error, result, errno2;
  void         *array;
  unsigned int  index;
  result = 0;
  errno2 = 0;
  HARNESS_LOG((HARNESS_INFO(harness-console shutdown)))
  /* Lock harness-console plug-in mutex. */
  if (0 != (error = pthread_mutex_lock(&harness_console_data.mutex))) {
    errno = error;
    HARNESS_LOG((HARNESS_WARN(unable to lock harness-console plug-in mutex)))
    errno = error;
    harness_syserr();
    errno = error;
    return -1;
  }
  /* Search for handle in handle array. */
  for (index = 0; index < harness_console_data.instances.count; index++) {
    if (harness_console_data.instances.array[index] == handle) {
      break;
    }
  }
  /* Check if handle is not in handle array. */
  if (index == harness_console_data.instances.count) {
    errno = ESRCH;
    HARNESS_LOG((HARNESS_WARN(handle is not in handle array)))
    errno = ESRCH;
    harness_syserr();
    /* Unlock harness-console plug-in mutex. */
    if (0 != (error = pthread_mutex_unlock(&harness_console_data.mutex))) {
      errno = error;
      HARNESS_LOG((
        HARNESS_WARN(unable to unlock harness-console plug-in mutex)))
      errno = error;
      harness_syserr();
    }
    errno = ESRCH;
    return -1;
  }
  /* Remove handle from handle array. */
  harness_console_data.instances.count--;
  memmove(harness_console_data.instances.array + index,
          harness_console_data.instances.array + index + 1,
         (harness_console_data.instances.count   - index) *
   sizeof(harness_console_data.instances.array[0]));
  /* Reallocate handle array. */
  if (0 == harness_console_data.instances.count) {
    free(harness_console_data.instances.array);
    harness_console_data.instances.array = NULL;
  } else if (NULL ==
             (array = realloc(harness_console_data.instances.array,
                              harness_console_data.instances.count *
                       sizeof(harness_console_data.instances.array[0])))) {
    HARNESS_LOG((HARNESS_WARN(unable to reallocate handle array)))
    errno = 0;
  } else {
    harness_console_data.instances.array = array;
  }
  /* Check for last finalization. */
  if (0 == harness_console_data.instances.count) {
    /* Unexport harness. */
    if (0 != harness_rmix_unexport(&harness_console_data.localref)) {
      errno2 = errno;
      HARNESS_LOG((HARNESS_WARN(unable to unexport harness)))
      result = -1;
    }
    /* Unload harness-rmix plug-in. */
    if (0 != harness_plugins_unload(harness_console_data.harness_rmix)) {
      errno2 = errno;
      HARNESS_LOG((HARNESS_WARN(unable to unload harness-rmix plug-in)))
      result = -1;
    }
  }
  /* Unlock harness-console plug-in mutex. */
  if (0 != (error = pthread_mutex_unlock(&harness_console_data.mutex))) {
    errno = error;
    HARNESS_LOG((
      HARNESS_WARN(unable to unlock harness-console plug-in mutex)))
    errno = error;
    harness_syserr();
    if (0 != errno2) {
      errno = errno2;
    } else {
      errno = error;
    }
    return -1;
  }
  if (0 != errno2) {
    errno = errno2;
  }
  return result;
}

/*
 * Returns the number of loaded instances of this harness-console plug-in
 * without locking its plug-in mutex.
 *
 * return = The number of loaded instances of this harness-console plug-in.
 */
#undef  __FUNC__
#define __FUNC__ "harness_console_instances"
unsigned int harness_console_instances () {
  return harness_console_data.instances.count;
}

/******************************************************************************
 *
 * END OF FILE
 *
 ******************************************************************************/
