/******************************************************************************
 *
 * Header file for rmix interface (object interface) module.
 * Copyright (c) 2002-2006 Oak Ridge National Laboratory.
 *
 * For more information see the following files in the source distribution top-
 * level directory or package data directory (usually /usr/local/share/package):
 *
 * - README    for general package information.
 * - INSTALL   for package install information.
 * - COPYING   for package license information and copying conditions.
 * - AUTHORS   for package authors information.
 * - ChangeLog for package changes information.
 *
 ******************************************************************************/

/* Avoid to include the content of this header file twice. */
#ifndef RMIX_INTERFACE_H
#define RMIX_INTERFACE_H

/******************************************************************************
 *
 * Macros
 *
 ******************************************************************************/

/* None at this time. */

/******************************************************************************
 *
 * Includes
 *
 ******************************************************************************/

/* None at this time. */

/******************************************************************************
 *
 * Data Types
 *
 ******************************************************************************/

/* Object interface data type. */
typedef struct {
  unsigned int   count;   /* object method descriptor list size */
  rmix_method_t *methods; /* object method descriptor list      */
} rmix_interface_t;

/******************************************************************************
 *
 * Function Prototypes
 *
 ******************************************************************************/

/*
 * Creates an object interface.
 *
 * interface = The object interface to create (return).
 * return    = 0 on success or -1 on any error with errno set appropriately.
 */
int rmix_interface_create (rmix_interface_t **interface);

/*
 * Creates an object interface.
 *
 * interface = The object interface to create (return).
 * count     = The object method descriptor list size.
 * methods   = The object method descriptor list.
 * return    = 0 on success or -1 on any error with errno set appropriately.
 */
int rmix_interface_create2 (rmix_interface_t    **interface,
                            unsigned int          count,
                            const rmix_method_t  *methods);

/*
 * Destroys an object interface.
 *
 * interface = The object interface to destroy (also return).
 * return    = 0 on success or -1 on any error with errno set appropriately.
 */
int rmix_interface_destroy (rmix_interface_t **interface);

/*
 * Duplicates an object interface.
 *
 * interface1 = The object interface to create (return).
 * interface2 = The object interface to duplicate.
 * return     = 0 on success or -1 on any error with errno set appropriately.
 */
int rmix_interface_duplicate (rmix_interface_t       **interface1,
                              const rmix_interface_t  *interface2);

/*
 * Appends an object interface to another object interface.
 *
 * interface1 = The object interface.
 * interface2 = The object interface to append.
 * return     = 0 on success or -1 on any error with errno set appropriately.
 */
int rmix_interface_append (rmix_interface_t       *interface1,
                           const rmix_interface_t *interface2);

/*
 * Inserts an object interface to another object interface.
 *
 * interface1 = The object interface.
 * index      = The object method descriptor list index to insert.
 * interface2 = The object interface to insert.
 * return     = 0 on success or -1 on any error with errno set appropriately.
 */
int rmix_interface_insert (rmix_interface_t       *interface1,
                           unsigned int            index,
                           const rmix_interface_t *interface2);

/*
 * Removes an object interface from another object interface.
 *
 * interface1 = The object interface.
 * interface2 = The object interface to remove.
 * return     = 0 on success or -1 on any error with errno set appropriately.
 */
int rmix_interface_remove (rmix_interface_t       *interface1,
                           const rmix_interface_t *interface2);

/*
 * Finds an object interface in another object interface.
 *
 * interface1 = The object interface.
 * index      = The object method descriptor list index (return).
 * interface2 = The object interface to find.
 * return     = 0 on success or -1 on any error with errno set appropriately.
 */
int rmix_interface_find (const rmix_interface_t *interface1,
                         unsigned int           *index,
                         const rmix_interface_t *interface2);

/*
 * Splits an object interface in two object interfaces.
 *
 * interface = The object interface.
 * head      = The object interface head part.
 * tail      = The object interface tail part.
 * index     = The object method descriptor list index to split.
 * return    = 0 on success or -1 on any error with errno set appropriately.
 */
int rmix_interface_split (const rmix_interface_t  *interface,
                          rmix_interface_t       **head,
                          rmix_interface_t       **tail,
                          unsigned int             index);

/*
 * Compares two object interfaces.
 *
 * diff       = The difference between both object interfaces (return).
 *              diff > 0 if interface1 < interface2,
 *              diff < 0 if interface1 > interface2, and
 *              diff = 0 if interface1 = interface2
 * interface1 = The object interface to compare against.
 * interface2 = The object interface to compare with.
 * return     = 0 on success or -1 on any error with errno set appropriately.
 */
int rmix_interface_compare (int                    *diff,
                            const rmix_interface_t *interface1,
                            const rmix_interface_t *interface2);

/*
 * Appends an object method descriptor to the object interface.
 *
 * interface = The object interface.
 * name      = The object method name to append.
 * output    = The object method output signature to append.
 * input     = The object method input  signature to append.
 * method    = The object method function to append.
 * return    = 0 on success or -1 on any error with errno set appropriately.
 */
int rmix_interface_methods_append (rmix_interface_t   *interface,
                                   const char         *name,
                                   const char         *output,
                                   const char         *input,
                                   rmix_method_call_t  call);

/*
 * Appends an object method descriptor to the object interface.
 *
 * interface = The object interface.
 * method    = The object method descriptor to append.
 * return    = 0 on success or -1 on any error with errno set appropriately.
 */
int rmix_interface_methods_append2 (rmix_interface_t    *interface,
                                    const rmix_method_t *method);

/*
 * Inserts an object method descriptor to the object interface.
 *
 * interface = The object interface.
 * index     = The object method descriptor list index to insert.
 * name      = The object method name to insert.
 * output    = The object method output signature to insert.
 * input     = The object method input  signature to insert.
 * method    = The object method function to insert.
 * return    = 0 on success or -1 on any error with errno set appropriately.
 */
int rmix_interface_methods_insert (rmix_interface_t   *interface,
                                   unsigned int        index,
                                   const char         *name,
                                   const char         *output,
                                   const char         *input,
                                   rmix_method_call_t  method);

/*
 * Inserts an object method descriptor to the object interface.
 *
 * interface = The object interface.
 * index     = The object method descriptor list index to insert.
 * method    = The object method descriptor to insert.
 * return    = 0 on success or -1 on any error with errno set appropriately.
 */
int rmix_interface_methods_insert2 (rmix_interface_t    *interface,
                                    unsigned int         index,
                                    const rmix_method_t *method);

/*
 * Removes an object method descriptor from the object interface.
 *
 * interface = The object interface.
 * name      = The object method name to remove.
 * return    = 0 on success or -1 on any error with errno set appropriately.
 */
int rmix_interface_methods_remove (rmix_interface_t *interface,
                                   const char       *name);

/*
 * Removes an object method descriptor from the object interface.
 *
 * interface = The object interface.
 * index     = The object method descriptor list index to remove.
 * return    = 0 on success or -1 on any error with errno set appropriately.
 */
int rmix_interface_methods_remove2 (rmix_interface_t *interface,
                                    unsigned int      index);

/*
 * Finds an object method descriptor in the object interface.
 *
 * interface = The object interface.
 * index     = The object method descriptor list index (return).
 * name      = The object method name to find.
 * return    = 0 on success or -1 on any error with errno set appropriately.
 */
int rmix_interface_methods_find (const rmix_interface_t *interface,
                                 unsigned int           *index,
                                 const char             *name);

/*
 * Finds an object method descriptor in the object interface.
 *
 * interface = The object interface.
 * index     = The object method descriptor list index (return).
 * method    = The object method descriptor to find.
 * return    = 0 on success or -1 on any error with errno set appropriately.
 */
int rmix_interface_methods_find2 (const rmix_interface_t *interface,
                                  unsigned int           *index,
                                  const rmix_method_t    *method);

/*
 * Gets the object method name of the object interface.
 *
 * interface = The object interface.
 * name      = The object method name (return).
 * index     = The object method descriptor list index.
 * return    = 0 on success or -1 on any error with errno set appropriately.
 */
int rmix_interface_methods_name (const rmix_interface_t  *interface,
                                 const char             **name,
                                 unsigned int             index);

/*
 * Gets the object method output signature of the object interface.
 *
 * interface = The object interface.
 * output    = The object method output signature (return).
 * index     = The object method descriptor list index.
 * return    = 0 on success or -1 on any error with errno set appropriately.
 */
int rmix_interface_methods_output (const rmix_interface_t  *interface,
                                   const char             **output,
                                   unsigned int             index);

/*
 * Gets the object method input signature of the object interface.
 *
 * interface = The object interface.
 * input     = The object method input signature (return).
 * index     = The object method descriptor list index.
 * return    = 0 on success or -1 on any error with errno set appropriately.
 */
int rmix_interface_methods_input (const rmix_interface_t  *interface,
                                  const char             **input,
                                  unsigned int             index);

/*
 * Gets the object method function of the object interface.
 *
 * interface = The object interface.
 * method    = The object method function (return).
 * index     = The object method descriptor list index.
 * return    = 0 on success or -1 on any error with errno set appropriately.
 */
int rmix_interface_methods_call (const rmix_interface_t *interface,
                                 rmix_method_call_t     *call,
                                 unsigned int            index);

/******************************************************************************
 *
 * Data
 *
 ******************************************************************************/

/* None at this time. */

#endif /* RMIX_INTERFACE_H */

/******************************************************************************
 *
 * END OF FILE
 *
 ******************************************************************************/
