/******************************************************************************
 *
 * Source file for rmix remoteref (invocation reference) module.
 * Copyright (c) 2002-2006 Oak Ridge National Laboratory.
 *
 * For more information see the following files in the source distribution top-
 * level directory or package data directory (usually /usr/local/share/package):
 *
 * - README    for general package information.
 * - INSTALL   for package install information.
 * - COPYING   for package license information and copying conditions.
 * - AUTHORS   for package authors information.
 * - ChangeLog for package changes information.
 *
 ******************************************************************************/

/******************************************************************************
 *
 * Macros
 *
 ******************************************************************************/

/* None at this time. */

/******************************************************************************
 *
 * Includes
 *
 ******************************************************************************/

/* Main module header file. */
#include "rmix.h"

/******************************************************************************
 *
 * Data Types
 *
 ******************************************************************************/

/* None at this time. */

/******************************************************************************
 *
 * Data
 *
 ******************************************************************************/

/* None at this time. */

/******************************************************************************
 *
 * Function Prototypes
 *
 ******************************************************************************/

/* None at this time. */

/******************************************************************************
 *
 * Functions
 *
 ******************************************************************************/

/*
 * Creates a invocation reference.
 *
 * invokeref  = The invocation reference to create (return).
 * invocation = The invocation handle.
 * provider   = The provider handle.
 * remoteref  = The remote object reference.
 * interface  = The object interface.
 * method     = The method index.
 * return     = 0 on success or -1 on any error with errno set appropriately.
 */
#undef  __FUNC__
#define __FUNC__ "rmix_invokeref_create"
int rmix_invokeref_create (rmix_invokeref_t       **invokeref,
                           unsigned int             invocation,
                           unsigned int             provider,
                           const rmix_remoteref_t  *remoteref,
                           const rmix_interface_t  *interface,
                           unsigned int             method) {
#ifdef DEBUG
  /* Check invokeref parameter. */
  if (NULL == invokeref) {
    errno = EINVAL;
    RMIX_LOG((RMIX_WARN(invokeref parameter is null)))
    errno = EINVAL;
    return -1;
  }
  /* Check remoteref parameter. */
  if (NULL == remoteref) {
    errno = EINVAL;
    RMIX_LOG((RMIX_WARN(remoteref parameter is null)))
    errno = EINVAL;
    return -1;
  }
  /* Check interface parameter. */
  if (NULL == interface) {
    errno = EINVAL;
    RMIX_LOG((RMIX_WARN(interface parameter is null)))
    errno = EINVAL;
    return -1;
  }
#endif /* DEBUG */
  /* Allocate invokation reference. */
  if (NULL == (*invokeref = malloc(sizeof(rmix_invokeref_t)))) {
    int errno2 = errno;
    RMIX_LOG((RMIX_WARN(unable to allocate invokation reference)))
    errno = errno2;
    return -1;
  }
  /* Duplicate invocation handle. */
  (*invokeref)->invocation = invocation;
  /* Duplicate provider handle. */
  (*invokeref)->provider = provider;
  /* Duplicate remote object reference. */
  if (0 != rmix_remoteref_duplicate(&(*invokeref)->remoteref, remoteref)) {
    int errno2 = errno;
    RMIX_LOG((RMIX_WARN(unable to duplicate remote object reference)))
    /* Deallocate invokation reference. */
    free(*invokeref);
    errno = errno2;
    return -1;
  }
  /* Duplicate remote object interface. */
  if (0 != rmix_interface_duplicate(&(*invokeref)->interface, interface)) {
    int errno2 = errno;
    RMIX_LOG((RMIX_WARN(unable to duplicate remote object interface)))
    /* Destroy remote object reference duplicate. */
    if (0 != rmix_remoteref_destroy(&(*invokeref)->remoteref)) {
      RMIX_LOG((RMIX_WARN(unable to destroy remote object reference duplicate)))
    }
    /* Deallocate invokation reference. */
    free(*invokeref);
    errno = errno2;
    return -1;
  }
  /* Duplicate remote object method index. */
  (*invokeref)->method = method;
  return 0;
}

/*
 * Destroys an invokation reference.
 *
 * invokeref = The invokation reference to destroy (also return).
 * return    = 0 on success or -1 on any error with errno set appropriately.
 */
#undef  __FUNC__
#define __FUNC__ "rmix_invokeref_destroy"
int rmix_invokeref_destroy (rmix_invokeref_t **invokeref) {
  int result, errno2;
#ifdef DEBUG
  /* Check invokeref parameter. */
  if (NULL == invokeref) {
    errno = EINVAL;
    RMIX_LOG((RMIX_WARN(invokeref parameter is null)))
    errno = EINVAL;
    return -1;
  }
#endif /* DEBUG */
  /* Check invokeref parameter. */
  if (NULL == *invokeref) {
    return 0;
  }
  result = 0;
  errno2 = 0;
  /* Destroy remote object reference duplicate. */
  if (0 != rmix_remoteref_destroy(&(*invokeref)->remoteref)) {
    errno2 = errno;
    RMIX_LOG((RMIX_WARN(unable to destroy remote object reference duplicate)))
    result = -1;
  }
  /* Destroy remote object interface duplicate. */
  if (0 != rmix_interface_destroy(&(*invokeref)->interface)) {
    errno2 = errno;
    RMIX_LOG((RMIX_WARN(unable to destroy remote object interface duplicate)))
    result = -1;
  }
  /* Deallocate invokation reference. */
  free(*invokeref);
  *invokeref = NULL;
  errno = errno2;
  return result;
}

/*
 * Duplicates an invocation reference.
 *
 * invokeref1 = The invocation reference to create (return).
 * invokeref2 = The invocation reference to duplicate.
 * return     = 0 on success or -1 on any error with errno set appropriately.
 */
#undef  __FUNC__
#define __FUNC__ "rmix_invokeref_duplicate"
int rmix_invokeref_duplicate (rmix_invokeref_t       **invokeref1,
                              const rmix_invokeref_t  *invokeref2) {
#ifdef DEBUG
  /* Check invokeref1 parameter. */
  if (NULL == invokeref1) {
    errno = EINVAL;
    RMIX_LOG((RMIX_WARN(invokeref1 parameter is null)))
    errno = EINVAL;
    return -1;
  }
  /* Check invokeref2 parameter. */
  if (NULL == invokeref2) {
    errno = EINVAL;
    RMIX_LOG((RMIX_WARN(invokeref2 parameter is null)))
    errno = EINVAL;
    return -1;
  }
#endif /* DEBUG */
  /* Create invocation reference duplicate. */
  if (0 != rmix_invokeref_create(invokeref1, invokeref2->invocation,
                                 invokeref2->provider, invokeref2->remoteref,
                                 invokeref2->interface, invokeref2->method)) {
    int errno2 = errno;
    RMIX_LOG((RMIX_WARN(unable to create invocation reference duplicate)))
    errno = errno2;
    return -1;
  }
  return 0;
}

/*
 * Compares two invocation references.
 *
 * diff       = The difference between both invocation references (return).
 *              diff > 0 if invokeref1 < invokeref2,
 *              diff < 0 if invokeref1 > invokeref2, and
 *              diff = 0 if invokeref1 = invokeref2
 * invokeref1 = The invocation reference to compare against.
 * invokeref2 = The invocation reference to compare with.
 * return     = 0 on success or -1 on any error with errno set appropriately.
 */
#undef  __FUNC__
#define __FUNC__ "rmix_invokeref_compare"
int rmix_invokeref_compare (int                    *diff,
                            const rmix_invokeref_t *invokeref1,
                            const rmix_invokeref_t *invokeref2) {
#ifdef DEBUG
  /* Check diff parameter. */
  if (NULL == diff) {
    errno = EINVAL;
    RMIX_LOG((RMIX_WARN(diff parameter is null)))
    errno = EINVAL;
    return -1;
  }
  /* Check invokeref1 parameter. */
  if (NULL == invokeref1) {
    errno = EINVAL;
    RMIX_LOG((RMIX_WARN(invokeref1 parameter is null)))
    errno = EINVAL;
    return -1;
  }
  /* Check invokeref2 parameter. */
  if (NULL == invokeref2) {
    errno = EINVAL;
    RMIX_LOG((RMIX_WARN(invokeref2 parameter is null)))
    errno = EINVAL;
    return -1;
  }
#endif /* DEBUG */
  /* Compare and check for equal provider
     invocation handles, provider handles and method indices. */
  if ((0 == (*diff = invokeref2->invocation - invokeref1->invocation))&&
      (0 == (*diff = invokeref2->provider   - invokeref1->provider  ))&&
      (0 == (*diff = invokeref2->method     - invokeref1->method  ))) {
    /* Compare remote object references. */
    if (0 != rmix_remoteref_compare(diff, invokeref1->remoteref,
                                          invokeref2->remoteref)) {
      int errno2 = errno;
      RMIX_LOG((RMIX_WARN(unable to compare remote object references)))
      errno = errno2;
      return -1;
    }
    /* Check for equal remote object references. */
    if (0 == *diff) {
      /* Compare object interfaces. */
      if (0 != rmix_interface_compare(diff, invokeref1->interface,
                                            invokeref2->interface)) {
        int errno2 = errno;
        RMIX_LOG((RMIX_WARN(unable to compare object interfaces)))
        errno = errno2;
        return -1;
      }
    }
  }
  return 0;
}

/******************************************************************************
 *
 * END OF FILE
 *
 ******************************************************************************/
