/******************************************************************************
 *
 * Header file for rmix method (object method descriptor) module.
 * Copyright (c) 2002-2006 Oak Ridge National Laboratory.
 *
 * For more information see the following files in the source distribution top-
 * level directory or package data directory (usually /usr/local/share/package):
 *
 * - README    for general package information.
 * - INSTALL   for package install information.
 * - COPYING   for package license information and copying conditions.
 * - AUTHORS   for package authors information.
 * - ChangeLog for package changes information.
 *
 ******************************************************************************/

/* Avoid to include the content of this header file twice. */
#ifndef RMIX_METHOD_H
#define RMIX_METHOD_H

/******************************************************************************
 *
 * Macros
 *
 ******************************************************************************/

/* Object method function method. */
/*
 * Invokes a method at a local object.
 *
 * object = The local object.
 * outary = The output values array. The values array and its containing
 *          values are allocated before the call with the exception of
 *          variable arrays and strings. They are allocated dynamically
 *          or explicitly set to NULL by this function. In the case of
 *          variable arrays, the length value is allocated before the call
 *          and is set to 0 if its variable array is NULL.
 * outcnt = The output values count.
 * inary  = The input values array. The values array and its containing
 *          values are not modified or deallocated by this call.
 * incnt  = The input values count.
 * return = 0 on success or -1 on any error with errno set appropriately.
 */
#define RMIX_METHOD_CALL(name) int name (void           *object,\
                                         void          **outary,\
                                         unsigned int    outcnt,\
                                         const void    **inary,\
                                         unsigned int    incnt)

/* Initializes an object method descriptor. */
#define RMIX_METHOD_INIT(name, output, input, method)\
          { name, output, input, method }

/******************************************************************************
 *
 * Includes
 *
 ******************************************************************************/

/* None at this time. */

/******************************************************************************
 *
 * Data Types
 *
 ******************************************************************************/

/* Object method function data type. */
typedef RMIX_METHOD_CALL((*rmix_method_call_t));

/* Object method descriptor data type. */
typedef struct {
  char               *name;   /* name             */
  char               *output; /* output signature */
  char               *input;  /* input  signature */
  rmix_method_call_t  call;   /* function         */
} rmix_method_t;

/******************************************************************************
 *
 * Function Prototypes
 *
 ******************************************************************************/

/*
 * Creates an object method descriptor.
 *
 * method = The object method descriptor to create (return).
 * name   = The object method name (cannot be NULL).
 * output = The object method output signature (cannot be NULL).
 * input  = The object method input  signature (cannot be NULL).
 * call   = The object method function (cannot be NULL).
 * return = 0 on success or -1 on any error with errno set appropriately.
 */
int rmix_method_create (rmix_method_t      **method,
                        const char          *name,
                        const char          *output,
                        const char          *input,
                        rmix_method_call_t   call);

/*
 * Destroys an object method descriptor.
 *
 * method = The object method descriptor to destroy (also return).
 * return = 0 on success or -1 on any error with errno set appropriately.
 */
int rmix_method_destroy (rmix_method_t **method);

/*
 * Duplicates an object method descriptor.
 *
 * method1 = The object method descriptor to create (return).
 * method2 = The object method descriptor to duplicate.
 * return  = 0 on success or -1 on any error with errno set appropriately.
 */
int rmix_method_duplicate (rmix_method_t       **method1,
                           const rmix_method_t  *method2);

/*
 * Compares two object method descriptors.
 *
 * diff    = The difference between both object method descriptors (return).
 *           diff > 0 if method1 < method2,
 *           diff < 0 if method1 > method2, and
 *           diff = 0 if method1 = method2
 * method1 = The object method descriptor to compare against.
 * method2 = The object method descriptor to compare with.
 * return  = 0 on success or -1 on any error with errno set appropriately.
 */
int rmix_method_compare (int                    *diff,
                         const rmix_method_t *method1,
                         const rmix_method_t *method2);

/******************************************************************************
 *
 * Data
 *
 ******************************************************************************/

/* None at this time. */

#endif /* RMIX_METHOD_H */

/******************************************************************************
 *
 * END OF FILE
 *
 ******************************************************************************/
