/******************************************************************************
 *
 * Header file for the RMIX parameters (export parameter list) module.
 * Copyright (c) 2002-2006 Oak Ridge National Laboratory.
 *
 * For more information see the following files in the source distribution top-
 * level directory or package data directory (usually /usr/local/share/package):
 *
 * - README    for general package information.
 * - INSTALL   for package install information.
 * - COPYING   for package license information and copying conditions.
 * - AUTHORS   for package authors information.
 * - ChangeLog for package changes information.
 *
 ******************************************************************************/

/* Avoid to include the content of this header file twice. */
#ifndef RMIX_PARAMETERS_H
#define RMIX_PARAMETERS_H

/******************************************************************************
 *
 * Macros
 *
 ******************************************************************************/

/* None at this time. */

/******************************************************************************
 *
 * Includes
 *
 ******************************************************************************/

/* None at this time. */

/******************************************************************************
 *
 * Data Types
 *
 ******************************************************************************/

/* The export parameter list data type. */
typedef struct {
  unsigned int       count; /* export parameter list count */
  rmix_parameter_t **array; /* export parameter list array */
} rmix_parameters_t;

/******************************************************************************
 *
 * Function Prototypes
 *
 ******************************************************************************/

/*
 * Creates an export parameter list.
 *
 * parameters = The export parameter list to create (return).
 * count      = The export parameter list count.
 * array      = The export parameter list array.
 * return     = 0 on success or -1 on any error with errno set appropriately.
 */
int rmix_parameters_create (rmix_parameters_t      **parameters,
                            unsigned int             count,
                            const rmix_parameter_t **array);

/*
 * Creates an export parameter list.
 *
 * parameters = The export parameter list to create (return).
 * count      = The export parameter list count.
 * names      = The export parameter list names array.
 * values     = The export parameter list values array.
 * return     = 0 on success or -1 on any error with errno set appropriately.
 */
int rmix_parameters_create2 (rmix_parameters_t **parameters,
                             unsigned int        count,
                             const char        **names,
                             const char        **values);

/*
 * Creates an export parameter list.
 *
 * parameters = The export parameter list to create (return).
 * count      = The export parameter list count.
 * pairs      = The export parameter list name-value pairs array. Each pair
 *              must be in the form <name>=<value>. Use a leading '\' for '='
 *              that are part of name or value and for other '\'s. For example,
 *              name\==\=value\\ means 'name=' and '=value\'.
 * return     = 0 on success or -1 on any error with errno set appropriately.
 */
int rmix_parameters_create3 (rmix_parameters_t **parameters,
                             unsigned int        count,
                             const char        **pairs);

/*
 * Creates an export parameter list.
 *
 * parameters = The export parameter list to create (return).
 * pairs      = The export parameter list name-value pairs string. Each pair
 *              must be in the form <name>=<value>. Use a leading '\' for '='
 *              and spaces that are part of name or value and for other '\'s.
 *              For example, 'name1\=\ =\=value1\\ name2\ =value2' means
 *              'name1= ' and '=value1\', and 'name 2' and 'value2'. The name-
 *              value pairs string may contain leading and trailing spaces, and
 *              multiple consecutive separating spaces.
 * return     = 0 on success or -1 on any error with errno set appropriately.
 */
int rmix_parameters_create4 (rmix_parameters_t **parameters,
                             const char         *pairs);

/*
 * Destroys an export parameter list.
 *
 * parameters = The export parameter list to destroy (also return).
 * return     = 0 on success or -1 on any error with errno set appropriately.
 */
int rmix_parameters_destroy (rmix_parameters_t **parameters);

/*
 * Duplicates an export parameter list.
 *
 * parameters1 = The export parameter list to create (return).
 * parameters2 = The export parameter list to duplicate.
 * return      = 0 on success or -1 on any error with errno set appropriately.
 */
int rmix_parameters_duplicate (rmix_parameters_t       **parameters1,
                               const rmix_parameters_t  *parameters2);

/*
 * Returns the string form of an export parameter list.
 *
 * pairs      = The export parameter list name-value pairs string. Each pair
 *              must be in the form <name>=<value>. Use a leading '\' for '='
 *              and spaces that are part of name or value and for other '\'s.
 *              For example, 'name1\=\ =\=value1\\ name2\ =value2' means
 *              'name1= ' and '=value1\', and 'name 2' and 'value2'. The name-
 *              value pairs string may contain leading and trailing spaces, and
 *              multiple consecutive separating spaces (return).
 * parameters = The export parameter list to create.
 * return     = 0 on success or -1 on any error with errno set appropriately.
 */
int rmix_parameters_string (char                    **pairs,
                            const rmix_parameters_t  *parameters);

/*
 * Compares two export parameter lists.
 *
 * diff        = The difference between both export parameter lists (return).
 *               diff > 0 if parameters1 < parameters2,
 *               diff < 0 if parameters1 > parameters2, and
 *               diff = 0 if parameters1 = parameters2.
 * parameters1 = The export parameter list to compare against.
 * parameters2 = The export parameter list to compare with.
 * return      = 0 on success or -1 on any error with errno set appropriately.
 */
int rmix_parameters_compare (int                     *diff,
                             const rmix_parameters_t *parameters1,
                             const rmix_parameters_t *parameters2);

/*
 * Searches for and returns the value of the parameter associated to name.
 *
 * parameters = The export parameter list to search.
 * value      = The export parameter value associated to name (return).
 * name       = The export parameter name to search for.
 * return     = 0 on success or -1 on any error with errno set appropriately.
 */
int rmix_parameters_search (const rmix_parameters_t  *parameters,
                            const char              **value,
                            const char               *name);

/*
 * Searches for and replaces the value of the parameter associated to name.
 *
 * parameters = The export parameter list to search.
 * name       = The export parameter name to search for.
 * value      = The export parameter value to replace.
 * return     = 0 on success or -1 on any error with errno set appropriately.
 */
int rmix_parameters_replace (rmix_parameters_t *parameters,
                             const char        *name,
                             const char        *value);

/*
 * Searches for and replaces the value of the parameter associated to name.
 * Inserts the name/value pair if not found.
 *
 * parameters = The export parameter list to search.
 * name       = The export parameter name to search for.
 * value      = The export parameter value to replace.
 * return     = 0 on success or -1 on any error with errno set appropriately.
 */
int rmix_parameters_insert (rmix_parameters_t *parameters,
                            const char        *name,
                            const char        *value);

/*
 * Searches for the value of the parameter associated to name. Inserts the
 * name/value pair only if not found.
 *
 * parameters = The export parameter list to search.
 * name       = The export parameter name to search for.
 * value      = The export parameter value to replace.
 * return     = 0 on success or -1 on any error with errno set appropriately.
 */
int rmix_parameters_insert2 (rmix_parameters_t *parameters,
                             const char        *name,
                             const char        *value);

/*
 * Searches for and removes the name/value pair.
 *
 * parameters = The export parameter list to search.
 * name       = The export parameter name to search for.
 * return     = 0 on success or -1 on any error with errno set appropriately.
 */
int rmix_parameters_remove (rmix_parameters_t *parameters,
                            const char        *name);

/*
 * Searches for and compares the value of the parameter associated to name.
 *
 * parameters = The export parameter list to search.
 * diff       = The difference between both values if found (return).
 *              diff > 0 if value < value in parameters,
 *              diff < 0 if value > value in parameters, and
 *              diff = 0 if value = value in parameters.
 * value      = The export parameter value to compare against.
 * name       = The export parameter name to search for.
 * return     = 0 on success or -1 on any error with errno set appropriately.
 */
int rmix_parameters_check (const rmix_parameters_t *parameters,
                           int                     *diff,
                           const char              *value,
                           const char              *name);

/******************************************************************************
 *
 * Data
 *
 ******************************************************************************/

/* None at this time. */

#endif /* RMIX_PARAMETERS_H */

/******************************************************************************
 *
 * END OF FILE
 *
 ******************************************************************************/
