/******************************************************************************
 *
 * Source file for rmix (main) module.
 * Copyright (c) 2002-2006 Oak Ridge National Laboratory.
 *
 * For more information see the following files in the source distribution top-
 * level directory or package data directory (usually /usr/local/share/package):
 *
 * - README    for general package information.
 * - INSTALL   for package install information.
 * - COPYING   for package license information and copying conditions.
 * - AUTHORS   for package authors information.
 * - ChangeLog for package changes information.
 *
 ******************************************************************************/

/******************************************************************************
 *
 * Macros
 *
 ******************************************************************************/

/******************************************************************************
 *
 * Includes
 *
 ******************************************************************************/

/* Main module header file. */
#include "rmix.h"

/******************************************************************************
 *
 * Data Types
 *
 ******************************************************************************/

/* Rmix module data type. */
typedef struct {
  pthread_mutex_t   mutex;       /* mutex                 */
  unsigned int      ready;       /* initialized flag      */
  unsigned int      closing;     /* closing flag          */
  void            (*onsyserr)(); /* system error callback */
} rmix_data_t;

/******************************************************************************
 *
 * Function Prototypes
 *
 ******************************************************************************/

/*
 * Initializes the remote object reference module.
 *
 * return = 0 on success or -1 on any error with errno set appropriately.
 */
int rmix_remoteref_init();

/*
 * Finalizes the remote object reference module.
 *
 * return = 0 on success or -1 on any error with errno set appropriately.
 */
int rmix_remoteref_fini();

/*
 * Initializes the object registry.
 *
 * return = 0 on success or -1 on any error with errno set appropriately.
 */
int rmix_objects_init ();

/*
 * Finalizes the object registry.
 *
 * return = 0 on success or -1 on any error with errno set appropriately.
 */
int rmix_objects_fini ();

/*
 * Initializes the plug-in loader.
 *
 * return = 0 on success or -1 on any error with errno set appropriately.
 */
int rmix_plugins_init ();

/*
 * Finalizes the plug-in loader.
 *
 * return = 0 on success or -1 on any error with errno set appropriately.
 */
int rmix_plugins_fini ();

/*
 * Initializes the thread id registry.
 *
 * return = 0 on success or -1 on any error with errno set appropriately.
 */
int rmix_threadids_init ();

/*
 * Finalizes the thread id registry.
 *
 * return = 0 on success or -1 on any error with errno set appropriately.
 */
int rmix_threadids_fini ();

/*
 * Sets the thread pool configuration and initializes the thread pool.
 *
 * pool_minimum = The new minimum thread count.
 * pool_maximum = The new maximum thread count.
 * pool_timeout = The new exiting thread timeout.
 * jobs_maximum = The new maximum job count.
 * jobs_timeout = The new waiting thread timeout.
 * return       = 0 on success or -1 on any error with errno set appropriately.
 */
int rmix_threads_init (unsigned int pool_minimum,
                       unsigned int pool_maximum,
                       unsigned int pool_timeout,
                       unsigned int jobs_maximum,
                       unsigned int jobs_timeout);

/*
 * Initializes the thread pool.
 *
 * return = 0 on success or -1 on any error with errno set appropriately.
 */
int rmix_threads_init2 ();

/*
 * Finalizes the thread pool.
 *
 * return = 0 on success or -1 on any error with errno set appropriately.
 */
int rmix_threads_fini ();

/*
 * Initializes the TCP/IPv4 pool.
 *
 * return = 0 on success or -1 on any error with errno set appropriately.
 */
int rmix_tcpip4_init ();

/*
 * Finalizes the TCP/IPv4 pool.
 *
 * return = 0 on success or -1 on any error with errno set appropriately.
 */
int rmix_tcpip4_fini ();

/******************************************************************************
 *
 * Data
 *
 ******************************************************************************/

/* Rmix module data. */
rmix_data_t rmix_data = {
  PTHREAD_MUTEX_INITIALIZER, /* mutex                 */
  0,                         /* initialized flag      */
  0,                         /* closing flag          */
  RMIX_ONSYSERR              /* system error callback */
};

/******************************************************************************
 *
 * Functions
 *
 ******************************************************************************/

/*
 * Initializes rmix.
 *
 * pool_minimum = The new minimum thread count.
 * pool_maximum = The new maximum thread count.
 * pool_timeout = The new exiting thread timeout.
 * jobs_maximum = The new maximum job count.
 * jobs_timeout = The new waiting thread timeout.
 * return       = 0 on success or -1 on any error with errno set appropriately.
 */
#undef  __FUNC__
#define __FUNC__ "rmix_init"
int rmix_init (unsigned int pool_minimum,
               unsigned int pool_maximum,
               unsigned int pool_timeout,
               unsigned int jobs_maximum,
               unsigned int jobs_timeout) {
  int      error;
  sigset_t sigmask;
  RMIX_LOG((RMIX_INFO(rmix startup)))
  /* Initialize signal mask. */
  if (-1 == sigemptyset(&sigmask)) {
    int errno2 = errno;
    RMIX_LOG((RMIX_WARN(unable to empty signal mask)))
    errno = errno2;
    rmix_syserr();
    return -1;
  }
  /* Add SIGPIPE signal to mask. */
  if (-1 == sigaddset(&sigmask, SIGPIPE)) {
    int errno2 = errno;
    RMIX_LOG((RMIX_WARN(unable to add signal to signal mask)))
    errno = errno2;
    rmix_syserr();
    return -1;
  }
  /* Lock rmix mutex. */
  if (0 != (error = pthread_mutex_lock(&rmix_data.mutex))) {
    errno = error;
    RMIX_LOG((RMIX_WARN(unable to lock rmix mutex)))
    errno = error;
    rmix_syserr();
    errno = error;
    return -1;
  }
  /* Check for initialized rmix. */
  if (1 == rmix_data.ready) {
    errno = EACCES;
    RMIX_LOG((RMIX_WARN(rmix already initialized)))
    /* Unlock rmix mutex. */
    if (0 != (error = pthread_mutex_unlock(&rmix_data.mutex))) {
      errno = error;
      RMIX_LOG((RMIX_WARN(unable to unlock rmix mutex)))
      errno = error;
      rmix_syserr();
    }
    errno = EACCES;
    return -1;
  }
  /* Block SIGPIPE signal. */
  if (-1 == pthread_sigmask(SIG_BLOCK, &sigmask, NULL)) {
    int errno2 = errno;
    RMIX_LOG((RMIX_WARN(unable to set thread signal mask)))
    errno = errno2;
    rmix_syserr();
    /* Unlock rmix mutex. */
    if (0 != (error = pthread_mutex_unlock(&rmix_data.mutex))) {
      errno = error;
      RMIX_LOG((RMIX_WARN(unable to unlock rmix mutex)))
      errno = error;
      rmix_syserr();
    }
    errno = errno2;
    return -1;
  }
  /* Initialize remote object reference module. */
  if (0 != rmix_remoteref_init()) {
    int errno2 = errno;
    RMIX_LOG((RMIX_WARN(unable to initialize remote object reference module)))
    /* Unlock rmix mutex. */
    if (0 != (error = pthread_mutex_unlock(&rmix_data.mutex))) {
      errno = error;
      RMIX_LOG((RMIX_WARN(unable to unlock rmix mutex)))
      errno = error;
      rmix_syserr();
    }
    errno = errno2;
    return -1;
  }
  /* Initialize object registry. */
  if (0 != rmix_objects_init()) {
    int errno2 = errno;
    RMIX_LOG((RMIX_WARN(unable to initialize object registry)))
    /* Finalize remote object reference module. */
    if (0 != rmix_remoteref_fini()) {
      RMIX_LOG((RMIX_WARN(unable to finalize remote object reference module)))
    }
    /* Unlock rmix mutex. */
    if (0 != (error = pthread_mutex_unlock(&rmix_data.mutex))) {
      errno = error;
      RMIX_LOG((RMIX_WARN(unable to unlock rmix mutex)))
      errno = error;
      rmix_syserr();
    }
    errno = errno2;
    return -1;
  }
  /* Initialize plug-in loader. */
  if (0 != rmix_plugins_init()) {
    int errno2 = errno;
    RMIX_LOG((RMIX_WARN(unable to initialize plug-in loader)))
    /* Finalize object registry. */
    if (0 != rmix_objects_fini()) {
      RMIX_LOG((RMIX_WARN(unable to finalize object registry)))
    }
    /* Finalize remote object reference module. */
    if (0 != rmix_remoteref_fini()) {
      RMIX_LOG((RMIX_WARN(unable to finalize remote object reference module)))
    }
    /* Unlock rmix mutex. */
    if (0 != (error = pthread_mutex_unlock(&rmix_data.mutex))) {
      errno = error;
      RMIX_LOG((RMIX_WARN(unable to unlock rmix mutex)))
      errno = error;
      rmix_syserr();
    }
    errno = errno2;
    return -1;
  }
  /* Initialize thread id registry. */
  if (0 != rmix_threadids_init()) {
    int errno2 = errno;
    RMIX_LOG((RMIX_WARN(unable to initialize thread id registry)))
    /* Finalize plug-in loader. */
    if (0 != rmix_plugins_fini()) {
      RMIX_LOG((RMIX_WARN(unable to finalize plug-in loader)))
    }
    /* Finalize object registry. */
    if (0 != rmix_objects_fini()) {
      RMIX_LOG((RMIX_WARN(unable to finalize object registry)))
    }
    /* Finalize remote object reference module. */
    if (0 != rmix_remoteref_fini()) {
      RMIX_LOG((RMIX_WARN(unable to finalize remote object reference module)))
    }
    /* Unlock rmix mutex. */
    if (0 != (error = pthread_mutex_unlock(&rmix_data.mutex))) {
      errno = error;
      RMIX_LOG((RMIX_WARN(unable to unlock rmix mutex)))
      errno = error;
      rmix_syserr();
    }
    errno = errno2;
    return -1;
  }
  /* Initialize thread pool. */
  if (0 != rmix_threads_init(pool_minimum, pool_maximum, pool_timeout,
                             jobs_maximum, jobs_timeout)) {
    int errno2 = errno;
    RMIX_LOG((RMIX_WARN(unable to initialize thread pool)))
    /* Finalize thread id registry. */
    if (0 != rmix_threadids_fini()) {
      RMIX_LOG((RMIX_WARN(unable to finalize thread id registry)))
    }
    /* Finalize plug-in loader. */
    if (0 != rmix_plugins_fini()) {
      RMIX_LOG((RMIX_WARN(unable to finalize plug-in loader)))
    }
    /* Finalize object registry. */
    if (0 != rmix_objects_fini()) {
      RMIX_LOG((RMIX_WARN(unable to finalize object registry)))
    }
    /* Finalize remote object reference module. */
    if (0 != rmix_remoteref_fini()) {
      RMIX_LOG((RMIX_WARN(unable to finalize remote object reference module)))
    }
    /* Unlock rmix mutex. */
    if (0 != (error = pthread_mutex_unlock(&rmix_data.mutex))) {
      errno = error;
      RMIX_LOG((RMIX_WARN(unable to unlock rmix mutex)))
      errno = error;
      rmix_syserr();
    }
    errno = errno2;
    return -1;
  }
  /* Initialize TCP/IPv4 pool. */
  if (0 != rmix_tcpip4_init()) {
    int errno2 = errno;
    RMIX_LOG((RMIX_WARN(unable to initialize TCP/IPv4 pool)))
    /* Finalize thread pool. */
    if (0 != rmix_threads_fini()) {
      RMIX_LOG((RMIX_WARN(unable to finalize thread pool)))
    }
    /* Finalize thread id registry. */
    if (0 != rmix_threadids_fini()) {
      RMIX_LOG((RMIX_WARN(unable to finalize thread id registry)))
    }
    /* Finalize plug-in loader. */
    if (0 != rmix_plugins_fini()) {
      RMIX_LOG((RMIX_WARN(unable to finalize plug-in loader)))
    }
    /* Finalize object registry. */
    if (0 != rmix_objects_fini()) {
      RMIX_LOG((RMIX_WARN(unable to finalize object registry)))
    }
    /* Finalize remote object reference module. */
    if (0 != rmix_remoteref_fini()) {
      RMIX_LOG((RMIX_WARN(unable to finalize remote object reference module)))
    }
    /* Unlock rmix mutex. */
    if (0 != (error = pthread_mutex_unlock(&rmix_data.mutex))) {
      errno = error;
      RMIX_LOG((RMIX_WARN(unable to unlock rmix mutex)))
      errno = error;
      rmix_syserr();
    }
    errno = errno2;
    return -1;
  }
  /* Mark rmix initialized. */
  rmix_data.ready = 1;
  /* Unlock rmix mutex. */
  if (0 != (error = pthread_mutex_unlock(&rmix_data.mutex))) {
    errno = error;
    RMIX_LOG((RMIX_WARN(unable to unlock rmix mutex)))
    errno = error;
    rmix_syserr();
    /* Mark rmix finalized. */
    rmix_data.ready = 0;
    errno = error;
    return -1;
  }
  return 0;
}

/*
 * Initializes rmix.
 *
 * return = 0 on success or -1 on any error with errno set appropriately.
 */
#undef  __FUNC__
#define __FUNC__ "rmix_init2"
int rmix_init2 () {
  int      error;
  sigset_t sigmask;
  RMIX_LOG((RMIX_INFO(rmix startup)))
  /* Initialize signal mask. */
  if (-1 == sigemptyset(&sigmask)) {
    int errno2 = errno;
    RMIX_LOG((RMIX_WARN(unable to empty signal mask)))
    errno = errno2;
    rmix_syserr();
    return -1;
  }
  /* Add SIGPIPE signal to mask. */
  if (-1 == sigaddset(&sigmask, SIGPIPE)) {
    int errno2 = errno;
    RMIX_LOG((RMIX_WARN(unable to add signal to signal mask)))
    errno = errno2;
    rmix_syserr();
    return -1;
  }
  /* Lock rmix mutex. */
  if (0 != (error = pthread_mutex_lock(&rmix_data.mutex))) {
    errno = error;
    RMIX_LOG((RMIX_WARN(unable to lock rmix mutex)))
    errno = error;
    rmix_syserr();
    return -1;
  }
  /* Block SIGPIPE signal. */
  if (-1 == pthread_sigmask(SIG_BLOCK, &sigmask, NULL)) {
    int errno2 = errno;
    RMIX_LOG((RMIX_WARN(unable to set thread signal mask)))
    errno = errno2;
    rmix_syserr();
    /* Unlock rmix mutex. */
    if (0 != (error = pthread_mutex_unlock(&rmix_data.mutex))) {
      errno = error;
      RMIX_LOG((RMIX_WARN(unable to unlock rmix mutex)))
      errno = error;
      rmix_syserr();
    }
    errno = errno2;
    return -1;
  }
  /* Check for initialized rmix. */
  if (1 == rmix_data.ready) {
    errno = EACCES;
    RMIX_LOG((RMIX_WARN(rmix already initialized)))
    /* Unlock rmix mutex. */
    if (0 != (error = pthread_mutex_unlock(&rmix_data.mutex))) {
      errno = error;
      RMIX_LOG((RMIX_WARN(unable to unlock rmix mutex)))
      errno = error;
      rmix_syserr();
    }
    errno = EACCES;
    return -1;
  }
  /* Initialize remote object reference module. */
  if (0 != rmix_remoteref_init()) {
    int errno2 = errno;
    RMIX_LOG((RMIX_WARN(unable to initialize remote object reference module)))
    /* Unlock rmix mutex. */
    if (0 != (error = pthread_mutex_unlock(&rmix_data.mutex))) {
      errno = error;
      RMIX_LOG((RMIX_WARN(unable to unlock rmix mutex)))
      errno = error;
      rmix_syserr();
    }
    errno = errno2;
    return -1;
  }
  /* Initialize object registry. */
  if (0 != rmix_objects_init()) {
    int errno2 = errno;
    RMIX_LOG((RMIX_WARN(unable to initialize object registry)))
    /* Finalize remote object reference module. */
    if (0 != rmix_remoteref_fini()) {
      RMIX_LOG((RMIX_WARN(unable to finalize remote object reference module)))
    }
    /* Unlock rmix mutex. */
    if (0 != (error = pthread_mutex_unlock(&rmix_data.mutex))) {
      errno = error;
      RMIX_LOG((RMIX_WARN(unable to unlock rmix mutex)))
      errno = error;
      rmix_syserr();
    }
    errno = errno2;
    return -1;
  }
  /* Initialize plug-in loader. */
  if (0 != rmix_plugins_init()) {
    int errno2 = errno;
    RMIX_LOG((RMIX_WARN(unable to initialize plug-in loader)))
    /* Finalize object registry. */
    if (0 != rmix_objects_fini()) {
      RMIX_LOG((RMIX_WARN(unable to finalize object registry)))
    }
    /* Finalize remote object reference module. */
    if (0 != rmix_remoteref_fini()) {
      RMIX_LOG((RMIX_WARN(unable to finalize remote object reference module)))
    }
    /* Unlock rmix mutex. */
    if (0 != (error = pthread_mutex_unlock(&rmix_data.mutex))) {
      errno = error;
      RMIX_LOG((RMIX_WARN(unable to unlock rmix mutex)))
      errno = error;
      rmix_syserr();
    }
    errno = errno2;
    return -1;
  }
  /* Initialize thread id registry. */
  if (0 != rmix_threadids_init()) {
    int errno2 = errno;
    RMIX_LOG((RMIX_WARN(unable to initialize thread id registry)))
    /* Finalize plug-in loader. */
    if (0 != rmix_plugins_fini()) {
      RMIX_LOG((RMIX_WARN(unable to finalize plug-in loader)))
    }
    /* Finalize object registry. */
    if (0 != rmix_objects_fini()) {
      RMIX_LOG((RMIX_WARN(unable to finalize object registry)))
    }
    /* Finalize remote object reference module. */
    if (0 != rmix_remoteref_fini()) {
      RMIX_LOG((RMIX_WARN(unable to finalize remote object reference module)))
    }
    /* Unlock rmix mutex. */
    if (0 != (error = pthread_mutex_unlock(&rmix_data.mutex))) {
      errno = error;
      RMIX_LOG((RMIX_WARN(unable to unlock rmix mutex)))
      errno = error;
      rmix_syserr();
    }
    errno = errno2;
    return -1;
  }
  /* Initialize thread pool. */
  if (0 != rmix_threads_init2()) {
    int errno2 = errno;
    RMIX_LOG((RMIX_WARN(unable to initialize thread pool)))
    /* Finalize thread id registry. */
    if (0 != rmix_threadids_fini()) {
      RMIX_LOG((RMIX_WARN(unable to finalize thread id registry)))
    }
    /* Finalize plug-in loader. */
    if (0 != rmix_plugins_fini()) {
      RMIX_LOG((RMIX_WARN(unable to finalize plug-in loader)))
    }
    /* Finalize object registry. */
    if (0 != rmix_objects_fini()) {
      RMIX_LOG((RMIX_WARN(unable to finalize object registry)))
    }
    /* Finalize remote object reference module. */
    if (0 != rmix_remoteref_fini()) {
      RMIX_LOG((RMIX_WARN(unable to finalize remote object reference module)))
    }
    /* Unlock rmix mutex. */
    if (0 != (error = pthread_mutex_unlock(&rmix_data.mutex))) {
      errno = error;
      RMIX_LOG((RMIX_WARN(unable to unlock rmix mutex)))
      errno = error;
      rmix_syserr();
    }
    errno = errno2;
    return -1;
  }
  /* Initialize TCP/IPv4 pool. */
  if (0 != rmix_tcpip4_init()) {
    int errno2 = errno;
    RMIX_LOG((RMIX_WARN(unable to initialize TCP/IPv4 pool)))
    /* Finalize thread pool. */
    if (0 != rmix_threads_fini()) {
      RMIX_LOG((RMIX_WARN(unable to finalize thread pool)))
    }
    /* Finalize thread id registry. */
    if (0 != rmix_threadids_fini()) {
      RMIX_LOG((RMIX_WARN(unable to finalize thread id registry)))
    }
    /* Finalize plug-in loader. */
    if (0 != rmix_plugins_fini()) {
      RMIX_LOG((RMIX_WARN(unable to finalize plug-in loader)))
    }
    /* Finalize object registry. */
    if (0 != rmix_objects_fini()) {
      RMIX_LOG((RMIX_WARN(unable to finalize object registry)))
    }
    /* Finalize remote object reference module. */
    if (0 != rmix_remoteref_fini()) {
      RMIX_LOG((RMIX_WARN(unable to finalize remote object reference module)))
    }
    /* Unlock rmix mutex. */
    if (0 != (error = pthread_mutex_unlock(&rmix_data.mutex))) {
      errno = error;
      RMIX_LOG((RMIX_WARN(unable to unlock rmix mutex)))
      errno = error;
      rmix_syserr();
    }
    errno = errno2;
    return -1;
  }
  /* Mark rmix initialized. */
  rmix_data.ready = 1;
  /* Unlock rmix mutex. */
  if (0 != (error = pthread_mutex_unlock(&rmix_data.mutex))) {
    errno = error;
    RMIX_LOG((RMIX_WARN(unable to unlock rmix mutex)))
    errno = error;
    rmix_syserr();
    /* Mark rmix finalized. */
    rmix_data.ready = 0;
    errno = error;
    return -1;
  }
  return 0;
}

/*
 * Finalizes rmix.
 *
 * return = 0 on success or -1 on any error with errno set appropriately.
 */
#undef  __FUNC__
#define __FUNC__ "rmix_fini"
int rmix_fini () {
  int error;
  RMIX_LOG((RMIX_INFO(rmix shutdown)))
  /* Lock rmix mutex. */
  if (0 != (error = pthread_mutex_lock(&rmix_data.mutex))) {
    errno = error;
    RMIX_LOG((RMIX_WARN(unable to lock rmix mutex)))
    errno = error;
    rmix_syserr();
    return -1;
  }
  /* Check for finalized rmix. */
  if (0 == rmix_data.ready) {
    errno = EACCES;
    RMIX_LOG((RMIX_WARN(rmix already finalized)))
    /* Unlock rmix mutex. */
    if (0 != (error = pthread_mutex_unlock(&rmix_data.mutex))) {
      errno = error;
      RMIX_LOG((RMIX_WARN(unable to unlock rmix mutex)))
      errno = error;
      rmix_syserr();
    }
    errno = EACCES;
    return -1;
  }
  /* Mark rmix closing. */
  rmix_data.closing = 1;
  /* Finalize thread pool. */
  if (0 != rmix_threads_fini()) {
    int errno2 = errno;
    RMIX_LOG((RMIX_WARN(unable to finalize thread pool)))
    /* Unlock rmix mutex. */
    if (0 != (error = pthread_mutex_unlock(&rmix_data.mutex))) {
      errno = error;
      RMIX_LOG((RMIX_WARN(unable to unlock rmix mutex)))
      errno = error;
      rmix_syserr();
    }
    errno = errno2;
    return -1;
  }
  /* Finalize thread id registry. */
  if (0 != rmix_threadids_fini()) {
    int errno2 = errno;
    RMIX_LOG((RMIX_WARN(unable to finalize thread id registry)))
    /* Unlock rmix mutex. */
    if (0 != (error = pthread_mutex_unlock(&rmix_data.mutex))) {
      errno = error;
      RMIX_LOG((RMIX_WARN(unable to unlock rmix mutex)))
      errno = error;
      rmix_syserr();
    }
    errno = errno2;
    return -1;
  }
  /* Finalize TCP/IPv4 pool. */
  if (0 != rmix_tcpip4_fini()) {
    int errno2 = errno;
    RMIX_LOG((RMIX_WARN(unable to finalize TCP/IPv4 pool)))
    /* Unlock rmix mutex. */
    if (0 != (error = pthread_mutex_unlock(&rmix_data.mutex))) {
      errno = error;
      RMIX_LOG((RMIX_WARN(unable to unlock rmix mutex)))
      errno = error;
      rmix_syserr();
    }
    errno = errno2;
    return -1;
  }
  /* Finalize plug-in loader. */
  if (0 != rmix_plugins_fini()) {
    int errno2 = errno;
    RMIX_LOG((RMIX_WARN(unable to finalize plug-in loader)))
    /* Unlock rmix mutex. */
    if (0 != (error = pthread_mutex_unlock(&rmix_data.mutex))) {
      errno = error;
      RMIX_LOG((RMIX_WARN(unable to unlock rmix mutex)))
      errno = error;
      rmix_syserr();
    }
    errno = errno2;
    return -1;
  }
  /* Finalize object registry. */
  if (0 != rmix_objects_fini()) {
    int errno2 = errno;
    RMIX_LOG((RMIX_WARN(unable to finalize object registry)))
    /* Unlock rmix mutex. */
    if (0 != (error = pthread_mutex_unlock(&rmix_data.mutex))) {
      errno = error;
      RMIX_LOG((RMIX_WARN(unable to unlock rmix mutex)))
      errno = error;
      rmix_syserr();
    }
    errno = errno2;
    return -1;
  }
  /* Finalize remote object reference module. */
  if (0 != rmix_remoteref_fini()) {
    int errno2 = errno;
    RMIX_LOG((RMIX_WARN(unable to finalize remote object reference module)))
    /* Unlock rmix mutex. */
    if (0 != (error = pthread_mutex_unlock(&rmix_data.mutex))) {
      errno = error;
      RMIX_LOG((RMIX_WARN(unable to unlock rmix mutex)))
      errno = error;
      rmix_syserr();
    }
    errno = errno2;
    return -1;
  }
  /* Mark rmix finalized. */
  rmix_data.ready   = 0; 
  rmix_data.closing = 0; 
  /* Unlock rmix mutex. */
  if (0 != (error = pthread_mutex_unlock(&rmix_data.mutex))) {
    errno = error;
    RMIX_LOG((RMIX_WARN(unable to unlock rmix mutex)))
    errno = error;
    rmix_syserr();
    errno = error;
    return -1;
  }
  return 0;
}

/*
 * Sets the callback for reporting system errors inside rmix to external
 * modules. This function may be called before initializing rmix. This function
 * does not aquire any locks.
 *
 * onsyserr = The callback for reporting system errors inside rmix to
 *            external modules.
 * return   = 0 on success or -1 on any error with errno set appropriately.
 */
#undef  __FUNC__
#define __FUNC__ "rmix_onsyserr_set"
int rmix_onsyserr_set (void (*onsyserr)()) {
  rmix_data.onsyserr = onsyserr;
  return 0;
}

/*
 * Gets the callback for reporting system errors inside rmix to external
 * modules. This function may be called before initializing rmix. This function
 * does not aquire any locks.
 *
 * onsyserr = The callback for reporting system errors inside rmix to
 *            external modules (return).
 * return   = 0 on success or -1 on any error with errno set appropriately.
 */
#undef  __FUNC__
#define __FUNC__ "rmix_onsyserr_get"
int rmix_onsyserr_get (void (**onsyserr)()) {
#ifdef DEBUG
  /* Check onsyserr parameter. */
  if (NULL == onsyserr) {
    errno = EINVAL;
    RMIX_LOG((RMIX_WARN(onsyserr parameter is null)))
    errno = EINVAL;
    return -1;
  }
#endif /* DEBUG */
  *onsyserr = rmix_data.onsyserr;
  return 0;
}

/*
 * Reports a system error inside rmix to external modules.
 */
#undef  __FUNC__
#define __FUNC__ "rmix_syserr"
void rmix_syserr () {
  void (*onsyserr)();
  /* Check for system error callback. */
  onsyserr = rmix_data.onsyserr;
  if (NULL != onsyserr) {
    int errno2 = errno;
    /* Execute system error callback. */
    RMIX_LOG((RMIX_INFO(reporting system error via callback)))
    errno = errno2;
    onsyserr();
  } else {
    RMIX_LOG((RMIX_INFO(not reporting system error via callback)))
  }
}

/*
 * Invokes a method at a remote object.
 *
 * remoteref = The remote object reference.
 * interface = The remote object interface.
 * method    = The index of the method to invoke.
 * outary    = The output values array. The values array and its containing
 *             values are allocated before the call with the exception of
 *             variable arrays and strings. They are allocated dynamically
 *             or explicitly set to NULL by this function. In the case of
 *             variable arrays, the length value is allocated before the call
 *             and is set to 0 if its variable array is NULL.
 * outcnt    = The output values count.
 * inary     = The input values array. The values array and its containing
 *             values are not modified or deallocated by this call.
 * incnt     = The input values count.
 * return    = 0 on success or -1 on any error with errno set appropriately.
 */
#undef  __FUNC__
#define __FUNC__ "rmix_invoke"
int rmix_invoke (const rmix_remoteref_t  *remoteref,
                 const rmix_interface_t  *interface,
                 unsigned int             method,
                 void                   **outary,
                 unsigned int             outcnt,
                 const void             **inary,
                 unsigned int             incnt) {
  unsigned int  provider;
  void         *invoke;
#ifdef DEBUG
  /* Check remoteref parameter. */
  if (NULL == remoteref) {
    errno = EINVAL;
    RMIX_LOG((RMIX_WARN(remoteref parameter is null)))
    errno = EINVAL;
    return -1;
  }
  /* Check interface parameter. */
  if (NULL == interface) {
    errno = EINVAL;
    RMIX_LOG((RMIX_WARN(interface parameter is null)))
    errno = EINVAL;
    return -1;
  }
  /* Check outary parameter. */
  if ((NULL == outary)&&(0 != outcnt)) {
    errno = EINVAL;
    RMIX_LOG((RMIX_WARN(outary parameter is null)))
    errno = EINVAL;
    return -1;
  }
  /* Check outcnt parameter. */
  if ((0 == outcnt)&&(NULL != outary)) {
    errno = EINVAL;
    RMIX_LOG((RMIX_WARN(outcnt parameter is zero)))
    errno = EINVAL;
    return -1;
  }
  /* Check inary parameter. */
  if ((NULL == inary)&&(0 != incnt)) {
    errno = EINVAL;
    RMIX_LOG((RMIX_WARN(inary parameter is null)))
    errno = EINVAL;
    return -1;
  }
  /* Check incnt parameter. */
  if ((0 == incnt)&&(NULL != inary)) {
    errno = EINVAL;
    RMIX_LOG((RMIX_WARN(incnt parameter is zero)))
    errno = EINVAL;
    return -1;
  }
#endif /* DEBUG */
  /* Load protocol provider plug-in. */
  if (0 != rmix_providers_load(&provider, remoteref->protocol)) {
    int errno2 = errno;
    RMIX_LOG((RMIX_WARN(unable to load protocol provider plug-in)))
    errno = errno2;
    return -1;
  }
  /* Resolve provider plug-in invoke function. */
  if (0 != rmix_providers_symbol2(&invoke, provider, "invoke")) {
    int errno2 = errno;
    RMIX_LOG((RMIX_WARN(unable to resolve provider plug-in invoke function)))
    /* Unload protocol provider plug-in. */
    if (0 != rmix_providers_unload(provider)) {
      RMIX_LOG((RMIX_WARN(unable to unoad protocol provider plug-in)))
    }
    errno = errno2;
    return -1;
  }
  /* Check provider plug-in invoke function. */
  if (NULL == invoke) {
    errno = ENOTSUP;
    RMIX_LOG((RMIX_WARN(provider plug-in invoke function does not exist)))
    /* Unload protocol provider plug-in. */
    if (0 != rmix_providers_unload(provider)) {
      RMIX_LOG((RMIX_WARN(unable to unoad protocol provider plug-in)))
    }
    errno = ENOTSUP;
    return -1;
  }
  /* Invoke method at a remote object. */
  if (0 != ((rmix_providers_invoke_t)invoke)(remoteref, interface, method,
                                             outary, outcnt, inary, incnt)) {
    int errno2 = errno;
    RMIX_LOG((RMIX_WARN(unable to invoke method at a remote object)))
    /* Unload protocol provider plug-in. */
    if (0 != rmix_providers_unload(provider)) {
      RMIX_LOG((RMIX_WARN(unable to unoad protocol provider plug-in)))
    }
    errno = errno2;
    return -1;
  }
  /* Unload protocol provider plug-in. */
  if (0 != rmix_providers_unload(provider)) {
    int errno2 = errno;
    RMIX_LOG((RMIX_WARN(unable to unoad protocol provider plug-in)))
    errno = errno2;
    return -1;
  }
  return 0;
}

/*
 * Invokes a method at a remote object without loading the provider plug-in.
 *
 * remoteref = The remote object reference.
 * provider  = The preloaded provider plug-in handle.
 * interface = The remote object interface.
 * method    = The index of the method to invoke.
 * outary    = The output values array. The values array and its containing
 *             values are allocated before the call with the exception of
 *             variable arrays and strings. They are allocated dynamically
 *             or explicitly set to NULL by this function. In the case of
 *             variable arrays, the length value is allocated before the call
 *             and is set to 0 if its variable array is NULL.
 * outcnt    = The output values count.
 * inary     = The input values array. The values array and its containing
 *             values are not modified or deallocated by this call.
 * incnt     = The input values count.
 * return    = 0 on success or -1 on any error with errno set appropriately.
 */
#undef  __FUNC__
#define __FUNC__ "rmix_invoke_p"
int rmix_invoke_p (const rmix_remoteref_t  *remoteref,
                   unsigned int             provider,
                   const rmix_interface_t  *interface,
                   unsigned int             method,
                   void                   **outary,
                   unsigned int             outcnt,
                   const void             **inary,
                   unsigned int             incnt) {
  void *invoke;
#ifdef DEBUG
  /* Check remoteref parameter. */
  if (NULL == remoteref) {
    errno = EINVAL;
    RMIX_LOG((RMIX_WARN(remoteref parameter is null)))
    errno = EINVAL;
    return -1;
  }
  /* Check interface parameter. */
  if (NULL == interface) {
    errno = EINVAL;
    RMIX_LOG((RMIX_WARN(interface parameter is null)))
    errno = EINVAL;
    return -1;
  }
  /* Check outary parameter. */
  if ((NULL == outary)&&(0 != outcnt)) {
    errno = EINVAL;
    RMIX_LOG((RMIX_WARN(outary parameter is null)))
    errno = EINVAL;
    return -1;
  }
  /* Check outcnt parameter. */
  if ((0 == outcnt)&&(NULL != outary)) {
    errno = EINVAL;
    RMIX_LOG((RMIX_WARN(outcnt parameter is zero)))
    errno = EINVAL;
    return -1;
  }
  /* Check inary parameter. */
  if ((NULL == inary)&&(0 != incnt)) {
    errno = EINVAL;
    RMIX_LOG((RMIX_WARN(inary parameter is null)))
    errno = EINVAL;
    return -1;
  }
  /* Check incnt parameter. */
  if ((0 == incnt)&&(NULL != inary)) {
    errno = EINVAL;
    RMIX_LOG((RMIX_WARN(incnt parameter is zero)))
    errno = EINVAL;
    return -1;
  }
#endif /* DEBUG */
  /* Resolve provider plug-in invoke function. */
  if (0 != rmix_providers_symbol2(&invoke, provider, "invoke")) {
    int errno2 = errno;
    RMIX_LOG((RMIX_WARN(unable to resolve provider plug-in invoke function)))
    errno = errno2;
    return -1;
  }
  /* Check provider plug-in invoke function. */
  if (NULL == invoke) {
    errno = ENOTSUP;
    RMIX_LOG((RMIX_WARN(provider plug-in invoke function does not exist)))
    errno = ENOTSUP;
    return -1;
  }
  /* Invoke method at a remote object. */
  if (0 != ((rmix_providers_invoke_t)invoke)(remoteref, interface, method,
                                             outary, outcnt, inary, incnt)) {
    int errno2 = errno;
    RMIX_LOG((RMIX_WARN(unable to invoke method at a remote object)))
    errno = errno2;
    return -1;
  }
  return 0;
}

/*
 * Invokes a method at a remote object one-way, i.e. there is no return value
 * and invokation returns immediately after sending message.
 *
 * remoteref = The remote object reference.
 * interface = The remote object interface.
 * method    = The index of the method to invoke.
 * inary     = The input values array. The values array and its containing
 *             values are not modified or deallocated by this call.
 * incnt     = The input values count.
 * return    = 0 on success or -1 on any error with errno set appropriately.
 */
#undef  __FUNC__
#define __FUNC__ "rmix_oneway"
int rmix_oneway (const rmix_remoteref_t  *remoteref,
                 const rmix_interface_t  *interface,
                 unsigned int             method,
                 const void             **inary,
                 unsigned int             incnt) {
  unsigned int  provider;
  void         *oneway;
#ifdef DEBUG
  /* Check remoteref parameter. */
  if (NULL == remoteref) {
    errno = EINVAL;
    RMIX_LOG((RMIX_WARN(remoteref parameter is null)))
    errno = EINVAL;
    return -1;
  }
  /* Check interface parameter. */
  if (NULL == interface) {
    errno = EINVAL;
    RMIX_LOG((RMIX_WARN(interface parameter is null)))
    errno = EINVAL;
    return -1;
  }
  /* Check inary parameter. */
  if ((NULL == inary)&&(0 != incnt)) {
    errno = EINVAL;
    RMIX_LOG((RMIX_WARN(inary parameter is null)))
    errno = EINVAL;
    return -1;
  }
  /* Check incnt parameter. */
  if ((0 == incnt)&&(NULL != inary)) {
    errno = EINVAL;
    RMIX_LOG((RMIX_WARN(incnt parameter is zero)))
    errno = EINVAL;
    return -1;
  }
#endif /* DEBUG */
  /* Load protocol provider plug-in. */
  if (0 != rmix_providers_load(&provider, remoteref->protocol)) {
    int errno2 = errno;
    RMIX_LOG((RMIX_WARN(unable to load protocol provider plug-in)))
    errno = errno2;
    return -1;
  }
  /* Resolve provider plug-in invoke function. */
  if (0 != rmix_providers_symbol2(&oneway, provider, "oneway")) {
    int errno2 = errno;
    RMIX_LOG((RMIX_WARN(unable to resolve provider plug-in oneway function)))
    /* Unload protocol provider plug-in. */
    if (0 != rmix_providers_unload(provider)) {
      RMIX_LOG((RMIX_WARN(unable to unoad protocol provider plug-in)))
    }
    errno = errno2;
    return -1;
  }
  /* Check provider plug-in oneway function. */
  if (NULL == oneway) {
    errno = ENOTSUP;
    RMIX_LOG((RMIX_WARN(provider plug-in oneway function does not exist)))
    /* Unload protocol provider plug-in. */
    if (0 != rmix_providers_unload(provider)) {
      RMIX_LOG((RMIX_WARN(unable to unoad protocol provider plug-in)))
    }
    errno = ENOTSUP;
    return -1;
  }
  /* Invoke method at a remote object. */
  if (0 != ((rmix_providers_oneway_t)oneway)
            (remoteref, interface, method, inary, incnt)) {
    int errno2 = errno;
    RMIX_LOG((RMIX_WARN(unable to invoke method at a remote object)))
    /* Unload protocol provider plug-in. */
    if (0 != rmix_providers_unload(provider)) {
      RMIX_LOG((RMIX_WARN(unable to unoad protocol provider plug-in)))
    }
    errno = errno2;
    return -1;
  }
  /* Unload protocol provider plug-in. */
  if (0 != rmix_providers_unload(provider)) {
    int errno2 = errno;
    RMIX_LOG((RMIX_WARN(unable to unoad protocol provider plug-in)))
    errno = errno2;
    return -1;
  }
  return 0;
}

/*
 * Invokes a method at a remote object one-way, i.e. there is no return value
 * and invokation returns immediately after sending message, without loading
 * the provider plug-in..
 *
 * provider  = The preloaded provider plug-in handle.
 * remoteref = The remote object reference.
 * interface = The remote object interface.
 * method    = The index of the method to invoke.
 * inary     = The input values array. The values array and its containing
 *             values are not modified or deallocated by this call.
 * incnt     = The input values count.
 * return    = 0 on success or -1 on any error with errno set appropriately.
 */
#undef  __FUNC__
#define __FUNC__ "rmix_oneway_p"
int rmix_oneway_p (const rmix_remoteref_t  *remoteref,
                   unsigned int             provider,
                   const rmix_interface_t  *interface,
                   unsigned int             method,
                   const void             **inary,
                   unsigned int             incnt) {
  void *oneway;
#ifdef DEBUG
  /* Check remoteref parameter. */
  if (NULL == remoteref) {
    errno = EINVAL;
    RMIX_LOG((RMIX_WARN(remoteref parameter is null)))
    errno = EINVAL;
    return -1;
  }
  /* Check interface parameter. */
  if (NULL == interface) {
    errno = EINVAL;
    RMIX_LOG((RMIX_WARN(interface parameter is null)))
    errno = EINVAL;
    return -1;
  }
  /* Check inary parameter. */
  if ((NULL == inary)&&(0 != incnt)) {
    errno = EINVAL;
    RMIX_LOG((RMIX_WARN(inary parameter is null)))
    errno = EINVAL;
    return -1;
  }
  /* Check incnt parameter. */
  if ((0 == incnt)&&(NULL != inary)) {
    errno = EINVAL;
    RMIX_LOG((RMIX_WARN(incnt parameter is zero)))
    errno = EINVAL;
    return -1;
  }
#endif /* DEBUG */
  /* Resolve provider plug-in invoke function. */
  if (0 != rmix_providers_symbol2(&oneway, provider, "oneway")) {
    int errno2 = errno;
    RMIX_LOG((RMIX_WARN(unable to resolve provider plug-in oneway function)))
    errno = errno2;
    return -1;
  }
  /* Check provider plug-in oneway function. */
  if (NULL == oneway) {
    errno = ENOTSUP;
    RMIX_LOG((RMIX_WARN(provider plug-in oneway function does not exist)))
    errno = ENOTSUP;
    return -1;
  }
  /* Invoke method at a remote object. */
  if (0 != ((rmix_providers_oneway_t)oneway)
            (remoteref, interface, method, inary, incnt)) {
    int errno2 = errno;
    RMIX_LOG((RMIX_WARN(unable to invoke method at a remote object)))
    errno = errno2;
    return -1;
  }
  return 0;
}

/*
 * Asynchronously invokes a method at a remote object, i.e. invocation returns
 * immediately after sending message. The output may be retrieved later using
 * the invocation reference. The provider gets not unloaded until the output
 * was retrieved.
 *
 * invokeref = The invocation reference (return).
 * remoteref = The remote object reference.
 * interface = The remote object interface.
 * method    = The index of the method to invoke.
 * inary     = The input values array. The values array and its containing
 *             values are not modified or deallocated by this call.
 * incnt     = The input values count.
 * return    = 0 on success or -1 on any error with errno set appropriately.
 */
#undef  __FUNC__
#define __FUNC__ "rmix_send"
int rmix_send (rmix_invokeref_t       **invokeref,
               const rmix_remoteref_t  *remoteref,
               const rmix_interface_t  *interface,
               unsigned int             method,
               const void             **inary,
               unsigned int             incnt) {
  void         *send;
  unsigned int  provider;
#ifdef DEBUG
  /* Check invokeref parameter. */
  if (NULL == invokeref) {
    errno = EINVAL;
    RMIX_LOG((RMIX_WARN(invokeref parameter is null)))
    errno = EINVAL;
    return -1;
  }
  /* Check remoteref parameter. */
  if (NULL == remoteref) {
    errno = EINVAL;
    RMIX_LOG((RMIX_WARN(remoteref parameter is null)))
    errno = EINVAL;
    return -1;
  }
  /* Check interface parameter. */
  if (NULL == interface) {
    errno = EINVAL;
    RMIX_LOG((RMIX_WARN(interface parameter is null)))
    errno = EINVAL;
    return -1;
  }
  /* Check inary parameter. */
  if ((NULL == inary)&&(0 != incnt)) {
    errno = EINVAL;
    RMIX_LOG((RMIX_WARN(inary parameter is null)))
    errno = EINVAL;
    return -1;
  }
  /* Check incnt parameter. */
  if ((0 == incnt)&&(NULL != inary)) {
    errno = EINVAL;
    RMIX_LOG((RMIX_WARN(incnt parameter is zero)))
    errno = EINVAL;
    return -1;
  }
#endif /* DEBUG */
  /* Load protocol provider plug-in. */
  if (0 != rmix_providers_load(&provider, remoteref->protocol)) {
    int errno2 = errno;
    RMIX_LOG((RMIX_WARN(unable to load protocol provider plug-in)))
    errno = errno2;
    return -1;
  }
  /* Resolve provider plug-in send function. */
  if (0 != rmix_providers_symbol2(&send, provider, "send")) {
    int errno2 = errno;
    RMIX_LOG((RMIX_WARN(unable to resolve provider plug-in send function)))
    /* Unload protocol provider plug-in. */
    if (0 != rmix_providers_unload(provider)) {
      RMIX_LOG((RMIX_WARN(unable to unoad protocol provider plug-in)))
    }
    errno = errno2;
    return -1;
  }
  /* Check provider plug-in send function. */
  if (NULL == send) {
    errno = ENOTSUP;
    RMIX_LOG((RMIX_WARN(provider plug-in send function does not exist)))
    /* Unload protocol provider plug-in. */
    if (0 != rmix_providers_unload(provider)) {
      RMIX_LOG((RMIX_WARN(unable to unoad protocol provider plug-in)))
    }
    errno = ENOTSUP;
    return -1;
  }
  /* Create invocation reference. */
   if (0 != rmix_invokeref_create(invokeref, 0, provider,
                                  remoteref, interface, method)) {
    int errno2 = errno;
    RMIX_LOG((RMIX_WARN(unable to create invocation reference)))
    /* Unload protocol provider plug-in. */
    if (0 != rmix_providers_unload(provider)) {
      RMIX_LOG((RMIX_WARN(unable to unoad protocol provider plug-in)))
    }
    errno = errno2;
    return -1;
  }
  /* Invoke method at a remote object. */
  if (0 != ((rmix_providers_send_t)send)
            (remoteref, interface, method, inary, incnt,
             &(*invokeref)->invocation)) {
    int errno2 = errno;
    RMIX_LOG((RMIX_WARN(unable to invoke method at a remote object)))
    /* Unload protocol provider plug-in. */
    if (0 != rmix_providers_unload(provider)) {
      RMIX_LOG((RMIX_WARN(unable to unoad protocol provider plug-in)))
    }
    /* Destroy invocation reference. */
    if (0 != rmix_invokeref_destroy(invokeref)) {
      RMIX_LOG((RMIX_WARN(unable to destroy invocation reference)))
    }
    errno = errno2;
    return -1;
  }
  return 0;
}

/*
 * Asynchronously invokes a method at a remote object, i.e. invocation returns
 * immediately after sending message, without loading the provider plug-in. The
 * output may be retrieved later using the invocation reference. The provider
 * gets not unloaded until the output was retrieved.
 *
 * invokeref = The invocation reference (return).
 * provider  = The preloaded provider plug-in handle.
 * remoteref = The remote object reference.
 * interface = The remote object interface.
 * method    = The index of the method to invoke.
 * inary     = The input values array. The values array and its containing
 *             values are not modified or deallocated by this call.
 * incnt     = The input values count.
 * return    = 0 on success or -1 on any error with errno set appropriately.
 */
#undef  __FUNC__
#define __FUNC__ "rmix_send_p"
int rmix_send_p (rmix_invokeref_t       **invokeref,
                 const rmix_remoteref_t  *remoteref,
                 unsigned int             provider,
                 const rmix_interface_t  *interface,
                 unsigned int             method,
                 const void             **inary,
                 unsigned int             incnt) {
  void *send;
#ifdef DEBUG
  /* Check invokeref parameter. */
  if (NULL == invokeref) {
    errno = EINVAL;
    RMIX_LOG((RMIX_WARN(invokeref parameter is null)))
    errno = EINVAL;
    return -1;
  }
  /* Check remoteref parameter. */
  if (NULL == remoteref) {
    errno = EINVAL;
    RMIX_LOG((RMIX_WARN(remoteref parameter is null)))
    errno = EINVAL;
    return -1;
  }
  /* Check interface parameter. */
  if (NULL == interface) {
    errno = EINVAL;
    RMIX_LOG((RMIX_WARN(interface parameter is null)))
    errno = EINVAL;
    return -1;
  }
  /* Check inary parameter. */
  if ((NULL == inary)&&(0 != incnt)) {
    errno = EINVAL;
    RMIX_LOG((RMIX_WARN(inary parameter is null)))
    errno = EINVAL;
    return -1;
  }
  /* Check incnt parameter. */
  if ((0 == incnt)&&(NULL != inary)) {
    errno = EINVAL;
    RMIX_LOG((RMIX_WARN(incnt parameter is zero)))
    errno = EINVAL;
    return -1;
  }
#endif /* DEBUG */
  /* Resolve provider plug-in send function. */
  if (0 != rmix_providers_symbol2(&send, provider, "send")) {
    int errno2 = errno;
    RMIX_LOG((RMIX_WARN(unable to resolve provider plug-in send function)))
    errno = errno2;
    return -1;
  }
  /* Check provider plug-in send function. */
  if (NULL == send) {
    errno = ENOTSUP;
    RMIX_LOG((RMIX_WARN(provider plug-in send function does not exist)))
    errno = ENOTSUP;
    return -1;
  }
  /* Create invocation reference. */
   if (0 != rmix_invokeref_create(invokeref, 0, provider,
                                  remoteref, interface, method)) {
    int errno2 = errno;
    RMIX_LOG((RMIX_WARN(unable to create invocation reference)))
    errno = errno2;
    return -1;
  }
  /* Invoke method at a remote object. */
  if (0 != ((rmix_providers_send_t)send)
            (remoteref, interface, method, inary, incnt,
             &(*invokeref)->invocation)) {
    int errno2 = errno;
    RMIX_LOG((RMIX_WARN(unable to invoke method at a remote object)))
    /* Destroy invocation reference. */
    if (0 != rmix_invokeref_destroy(invokeref)) {
      RMIX_LOG((RMIX_WARN(unable to destroy invocation reference)))
    }
    errno = errno2;
    return -1;
  }
  return 0;
}

/*
 * Retrieves the output (if any) of an asynchronous invocation using a
 * previously issued invokation reference and unloads the provider plug-in. The
 * invokation reference will be destroyed.
 *
 * invokeref = The invocation reference (also return).
 * outary    = The output values array. The values array and its containing
 *             values are allocated before the call with the exception of
 *             variable arrays and strings. They are allocated dynamically
 *             or explicitly set to NULL by this function. In the case of
 *             variable arrays, the length value is allocated before the call
 *             and is set to 0 if its variable array is NULL.
 * outcnt    = The output values count.
 * return    = 0 on success or -1 on any error with errno set appropriately.
 */
#undef  __FUNC__
#define __FUNC__ "rmix_retrieve"
int rmix_retrieve (rmix_invokeref_t **invokeref,
                   void             **outary,
                   unsigned int       outcnt) {
  void *retrieve;
#ifdef DEBUG
  /* Check invokeref parameter. */
  if (NULL == invokeref) {
    errno = EINVAL;
    RMIX_LOG((RMIX_WARN(invokeref parameter is null)))
    errno = EINVAL;
    return -1;
  }
  /* Check *invokeref parameter. */
  if (NULL == *invokeref) {
    errno = EINVAL;
    RMIX_LOG((RMIX_WARN(*invokeref parameter is null)))
    errno = EINVAL;
    return -1;
  }
  /* Check outary parameter. */
  if ((NULL == outary)&&(0 != outcnt)) {
    errno = EINVAL;
    RMIX_LOG((RMIX_WARN(outary parameter is null)))
    errno = EINVAL;
    return -1;
  }
  /* Check outcnt parameter. */
  if ((0 == outcnt)&&(NULL != outary)) {
    errno = EINVAL;
    RMIX_LOG((RMIX_WARN(outcnt parameter is zero)))
    errno = EINVAL;
    return -1;
  }
#endif /* DEBUG */
  /* Resolve provider plug-in retrieve function. */
  if (0 != rmix_providers_symbol2(&retrieve, (*invokeref)->provider,
                                  "retrieve")) {
    int errno2 = errno;
    RMIX_LOG((RMIX_WARN(unable to resolve provider plug-in retrieve function)))
    /* Unload protocol provider plug-in. */
    if (0 != rmix_providers_unload((*invokeref)->provider)) {
      RMIX_LOG((RMIX_WARN(unable to unoad protocol provider plug-in)))
    }
    /* Destroy invocation reference. */
    if (0 != rmix_invokeref_destroy(invokeref)) {
      RMIX_LOG((RMIX_WARN(unable to destroy invocation reference)))
    }
    errno = errno2;
    return -1;
  }
  /* Check provider plug-in retrieve function. */
  if (NULL == retrieve) {
    errno = ENOTSUP;
    RMIX_LOG((RMIX_WARN(provider plug-in retrieve function does not exist)))
    /* Unload protocol provider plug-in. */
    if (0 != rmix_providers_unload((*invokeref)->provider)) {
      RMIX_LOG((RMIX_WARN(unable to unoad protocol provider plug-in)))
    }
    /* Destroy invocation reference. */
    if (0 != rmix_invokeref_destroy(invokeref)) {
      RMIX_LOG((RMIX_WARN(unable to destroy invocation reference)))
    }
    errno = ENOTSUP;
    return -1;
  }
  /* Retrieve invocation output (if any). */
  if (0 != ((rmix_providers_retrieve_t)retrieve)
            ((*invokeref)->remoteref, (*invokeref)->interface, 
             (*invokeref)->method, outary, outcnt, (*invokeref)->invocation)) {
    int errno2 = errno;
    RMIX_LOG((RMIX_WARN(unable to invoke method at a remote object)))
    /* Unload protocol provider plug-in. */
    if (0 != rmix_providers_unload((*invokeref)->provider)) {
      RMIX_LOG((RMIX_WARN(unable to unoad protocol provider plug-in)))
    }
    /* Destroy invocation reference. */
    if (0 != rmix_invokeref_destroy(invokeref)) {
      RMIX_LOG((RMIX_WARN(unable to destroy invocation reference)))
    }
    errno = errno2;
    return -1;
  }
  /* Unload protocol provider plug-in. */
  if (0 != rmix_providers_unload((*invokeref)->provider)) {
    int errno2 = errno;
    RMIX_LOG((RMIX_WARN(unable to unoad protocol provider plug-in)))
    errno = errno2;
    return -1;
  }
  /* Destroy invocation reference. */
  if (0 != rmix_invokeref_destroy(invokeref)) {
    int errno2 = errno;
    RMIX_LOG((RMIX_WARN(unable to destroy invocation reference)))
    errno = errno2;
    return -1;
  }
  return 0;
}

/*
 * Retrieves the output (if any) of an asynchronous invocation using a
 * previously issued invokation reference without unloading the provider
 * plug-in. The invokation reference will be destroyed.
 *
 * invokeref = The invocation reference (also return).
 * outary    = The output values array. The values array and its containing
 *             values are allocated before the call with the exception of
 *             variable arrays and strings. They are allocated dynamically
 *             or explicitly set to NULL by this function. In the case of
 *             variable arrays, the length value is allocated before the call
 *             and is set to 0 if its variable array is NULL.
 * outcnt    = The output values count.
 * return    = 0 on success or -1 on any error with errno set appropriately.
 */
#undef  __FUNC__
#define __FUNC__ "rmix_retrieve_p"
int rmix_retrieve_p (rmix_invokeref_t **invokeref,
                     void             **outary,
                     unsigned int       outcnt) {
  void *retrieve;
#ifdef DEBUG
  /* Check invokeref parameter. */
  if (NULL == invokeref) {
    errno = EINVAL;
    RMIX_LOG((RMIX_WARN(invokeref parameter is null)))
    errno = EINVAL;
    return -1;
  }
  /* Check *invokeref parameter. */
  if (NULL == *invokeref) {
    errno = EINVAL;
    RMIX_LOG((RMIX_WARN(*invokeref parameter is null)))
    errno = EINVAL;
    return -1;
  }
  /* Check outary parameter. */
  if ((NULL == outary)&&(0 != outcnt)) {
    errno = EINVAL;
    RMIX_LOG((RMIX_WARN(outary parameter is null)))
    errno = EINVAL;
    return -1;
  }
  /* Check outcnt parameter. */
  if ((0 == outcnt)&&(NULL != outary)) {
    errno = EINVAL;
    RMIX_LOG((RMIX_WARN(outcnt parameter is zero)))
    errno = EINVAL;
    return -1;
  }
#endif /* DEBUG */
  /* Resolve provider plug-in retrieve function. */
  if (0 != rmix_providers_symbol2(&retrieve, (*invokeref)->provider,
                                  "retrieve")) {
    int errno2 = errno;
    RMIX_LOG((RMIX_WARN(unable to resolve provider plug-in retrieve function)))
    /* Destroy invocation reference. */
    if (0 != rmix_invokeref_destroy(invokeref)) {
      RMIX_LOG((RMIX_WARN(unable to destroy invocation reference)))
    }
    errno = errno2;
    return -1;
  }
  /* Check provider plug-in retrieve function. */
  if (NULL == retrieve) {
    errno = ENOTSUP;
    RMIX_LOG((RMIX_WARN(provider plug-in retrieve function does not exist)))
    /* Destroy invocation reference. */
    if (0 != rmix_invokeref_destroy(invokeref)) {
      RMIX_LOG((RMIX_WARN(unable to destroy invocation reference)))
    }
    errno = ENOTSUP;
    return -1;
  }
  /* Retrieve invocation output (if any). */
  if (0 != ((rmix_providers_retrieve_t)retrieve)
            ((*invokeref)->remoteref, (*invokeref)->interface, 
             (*invokeref)->method, outary, outcnt, (*invokeref)->invocation)) {
    int errno2 = errno;
    RMIX_LOG((RMIX_WARN(unable to invoke method at a remote object)))
    /* Destroy invocation reference. */
    if (0 != rmix_invokeref_destroy(invokeref)) {
      RMIX_LOG((RMIX_WARN(unable to destroy invocation reference)))
    }
    errno = errno2;
    return -1;
  }
  /* Destroy invocation reference. */
  if (0 != rmix_invokeref_destroy(invokeref)) {
    int errno2 = errno;
    RMIX_LOG((RMIX_WARN(unable to destroy invocation reference)))
    errno = errno2;
    return -1;
  }
  return 0;
}

/*
 * Exports a local object without forcing a specific object id.
 *
 * localref   = The local object reference of the exported object (return).
 * protocol   = The communication protocol. Overrides any additional PROTOCOL
 *              parameter. May only be NULL or "" if additional parameters
 *              contains PROTOCOL.
 * binding    = The binding  (protocol specific) (may be NULL, ""). Overrides
 *              any additional BINDING parameter.
 * parameters = Additional parameters (protocol specific) (may be NULL).
 * object     = The local object (may be NULL).
 * interface  = The local object interface.
 * return     = 0 on success or -1 on any error with errno set appropriately.
 */
#undef  __FUNC__
#define __FUNC__ "rmix_export"
int rmix_export (rmix_localref_t         **localref,
                 const char               *protocol,
                 const char               *binding,
                 const rmix_parameters_t  *parameters,
                 void                     *object,
                 const rmix_interface_t   *interface) {
  void         *export;
  unsigned int  objectid, provider;
#ifdef DEBUG
  /* Check localref parameter. */
  if (NULL == localref) {
    errno = EINVAL;
    RMIX_LOG((RMIX_WARN(localref parameter is null)))
    errno = EINVAL;
    return -1;
  }
  /* Check interface parameter. */
  if (NULL == interface) {
    errno = EINVAL;
    RMIX_LOG((RMIX_WARN(interface parameter is null)))
    errno = EINVAL;
    return -1;
  }
#endif /* DEBUG */
  /* Check protocol parameter. */
  if ((NULL == protocol)||('\0' == protocol[0])) {
    /* Check parameters parameter. */
    if (NULL == parameters) {
      errno = EINVAL;
      RMIX_LOG((
        RMIX_WARN(parameters parameter is null and no protocol specified)))
      errno = EINVAL;
      return -1;
    }
    /* Search for additional protocol parameter. */
    if (0 != rmix_parameters_search(parameters, &protocol, "PROTOCOL")) {
      if (ESRCH == errno) {
        errno = EINVAL;
        RMIX_LOG((RMIX_WARN(no protocol specified)))
        errno = EINVAL;
        return -1;
      } else {
        int errno2 = errno;
        RMIX_LOG((RMIX_WARN(unable to search for protocol parameter)))
        errno = errno2;
        return -1;
      }
    }
  }
  /* Check binding parameter. */
  if (((NULL == binding)||('\0' == binding[0]))&&(NULL != parameters)) {
    /* Search for additional binding parameter. */
    if ((0     != rmix_parameters_search(parameters, &binding, "BINDING"))&&
        (ESRCH != errno)) {
      int errno2 = errno;
      RMIX_LOG((RMIX_WARN(unable to search for binding parameter)))
      errno = errno2;
      return -1;
    }
  }
  /* Add object to object registry. */
  if (0 != rmix_objects_add(&objectid, object, interface)) {
    int errno2 = errno;
    RMIX_LOG((RMIX_WARN(unable to add object to object registry)))
    errno = errno2;
    return -1;
  }
  /* Load protocol provider plug-in. */
  if (0 != rmix_providers_load(&provider, protocol)) {
    int errno2 = errno;
    RMIX_LOG((RMIX_WARN(unable to load protocol provider plug-in)))
    /* Remove object from object registry. */
    if (0 != rmix_objects_remove(objectid)) {
      RMIX_LOG((RMIX_WARN(unable to remove object from object registry)))
    }
    errno = errno2;
    return -1;
  }
  /* Resolve provider plug-in export function. */
  if (0 != rmix_providers_symbol2(&export, provider, "export")) {
    int errno2 = errno;
    RMIX_LOG((RMIX_WARN(unable to resolve provider plug-in export function)))
    /* Unload protocol provider plug-in. */
    if (0 != rmix_providers_unload(provider)) {
      RMIX_LOG((RMIX_WARN(unable to unoad protocol provider plug-in)))
    }
    /* Remove object from object registry. */
    if (0 != rmix_objects_remove(objectid)) {
      RMIX_LOG((RMIX_WARN(unable to remove object from object registry)))
    }
    errno = errno2;
    return -1;
  }
  /* Check provider plug-in export function. */
  if (NULL == export) {
    errno = ENOTSUP;
    RMIX_LOG((RMIX_WARN(provider plug-in export function does not exist)))
    /* Unload protocol provider plug-in. */
    if (0 != rmix_providers_unload(provider)) {
      RMIX_LOG((RMIX_WARN(unable to unoad protocol provider plug-in)))
    }
    /* Remove object from object registry. */
    if (0 != rmix_objects_remove(objectid)) {
      RMIX_LOG((RMIX_WARN(unable to remove object from object registry)))
    }
    errno = ENOTSUP;
    return -1;
  }
  /* Export local object. */
  if (0 != ((rmix_providers_export_t)export)
            (localref, protocol, binding, parameters,
             provider, objectid, object, interface)) {
    int errno2 = errno;
    RMIX_LOG((RMIX_WARN(unable to export local object)))
    /* Unload protocol provider plug-in. */
    if (0 != rmix_providers_unload(provider)) {
      RMIX_LOG((RMIX_WARN(unable to unoad protocol provider plug-in)))
    }
    /* Remove object from object registry. */
    if (0 != rmix_objects_remove(objectid)) {
      RMIX_LOG((RMIX_WARN(unable to remove object from object registry)))
    }
    errno = errno2;
    return -1;
  }
  return 0;
}

/*
 * Exports a local object without forcing a specific object id and without
 * loading the provider plug-in.
 *
 * localref   = The local object reference of the exported object (return).
 * provider   = The preloaded provider plug-in handle.
 * protocol   = The communication protocol. Overrides any additional PROTOCOL
 *              parameter. May only be NULL or "" if additional parameters
 *              contains PROTOCOL.
 * binding    = The binding  (protocol specific) (may be NULL, ""). Overrides
 *              any additional BINDING parameter.
 * parameters = Additional parameters (protocol specific) (may be NULL).
 * object     = The local object (may be NULL).
 * interface  = The local object interface.
 * return     = 0 on success or -1 on any error with errno set appropriately.
 */
#undef  __FUNC__
#define __FUNC__ "rmix_export_p"
int rmix_export_p (rmix_localref_t         **localref,
                   unsigned int              provider,
                   const char               *protocol,
                   const char               *binding,
                   const rmix_parameters_t  *parameters,
                   void                     *object,
                   const rmix_interface_t   *interface) {
  void         *export;
  unsigned int  objectid;
#ifdef DEBUG
  /* Check localref parameter. */
  if (NULL == localref) {
    errno = EINVAL;
    RMIX_LOG((RMIX_WARN(localref parameter is null)))
    errno = EINVAL;
    return -1;
  }
  /* Check interface parameter. */
  if (NULL == interface) {
    errno = EINVAL;
    RMIX_LOG((RMIX_WARN(interface parameter is null)))
    errno = EINVAL;
    return -1;
  }
#endif /* DEBUG */
  /* Check protocol parameter. */
  if ((NULL == protocol)||('\0' == protocol[0])) {
    /* Check parameters parameter. */
    if (NULL == parameters) {
      errno = EINVAL;
      RMIX_LOG((
        RMIX_WARN(parameters parameter is null and no protocol specified)))
      errno = EINVAL;
      return -1;
    }
    /* Search for additional protocol parameter. */
    if (0 != rmix_parameters_search(parameters, &protocol, "PROTOCOL")) {
      if (ESRCH == errno) {
        errno = EINVAL;
        RMIX_LOG((RMIX_WARN(no protocol specified)))
        errno = EINVAL;
        return -1;
      } else {
        int errno2 = errno;
        RMIX_LOG((RMIX_WARN(unable to search for protocol parameter)))
        errno = errno2;
        return -1;
      }
    }
  }
  /* Check binding parameter. */
  if (((NULL == binding)||('\0' == binding[0]))&&(NULL != parameters)) {
    /* Search for additional binding parameter. */
    if ((0     != rmix_parameters_search(parameters, &binding, "BINDING"))&&
        (ESRCH != errno)) {
      int errno2 = errno;
      RMIX_LOG((RMIX_WARN(unable to search for binding parameter)))
      errno = errno2;
      return -1;
    }
  }
  /* Add object to object registry. */
  if (0 != rmix_objects_add(&objectid, object, interface)) {
    int errno2 = errno;
    RMIX_LOG((RMIX_WARN(unable to add object to object registry)))
    errno = errno2;
    return -1;
  }
  /* Resolve provider plug-in export function. */
  if (0 != rmix_providers_symbol2(&export, provider, "export")) {
    int errno2 = errno;
    RMIX_LOG((RMIX_WARN(unable to resolve provider plug-in export function)))
    /* Remove object from object registry. */
    if (0 != rmix_objects_remove(objectid)) {
      RMIX_LOG((RMIX_WARN(unable to remove object from object registry)))
    }
    errno = errno2;
    return -1;
  }
  /* Check provider plug-in export function. */
  if (NULL == export) {
    errno = ENOTSUP;
    RMIX_LOG((RMIX_WARN(provider plug-in export function does not exist)))
    /* Remove object from object registry. */
    if (0 != rmix_objects_remove(objectid)) {
      RMIX_LOG((RMIX_WARN(unable to remove object from object registry)))
    }
    errno = ENOTSUP;
    return -1;
  }
  /* Export local object. */
  if (0 != ((rmix_providers_export_t)export)
            (localref, protocol, binding, parameters,
             provider, objectid, object, interface)) {
    int errno2 = errno;
    RMIX_LOG((RMIX_WARN(unable to export local object)))
    /* Remove object from object registry. */
    if (0 != rmix_objects_remove(objectid)) {
      RMIX_LOG((RMIX_WARN(unable to remove object from object registry)))
    }
    errno = errno2;
    return -1;
  }
  return 0;
}

/*
 * Exports a local object while forcing a specific object id. This is very
 * useful for standard object exports.
 *
 * localref   = The local object reference of the exported object (return).
 * protocol   = The communication protocol. Overrides any additional PROTOCOL
 *              parameter. May only be NULL or "" if additional parameters
 *              contains PROTOCOL.
 * binding    = The binding  (protocol specific) (may be NULL, ""). Overrides
 *              any additional BINDING parameter.
 * parameters = Additional parameters (protocol specific) (may be NULL).
 * objectid   = The object id.
 * object     = The local object (may be NULL).
 * interface  = The local object interface.
 * return     = 0 on success or -1 on any error with errno set appropriately.
 */
#undef  __FUNC__
#define __FUNC__ "rmix_export2"
int rmix_export2 (rmix_localref_t         **localref,
                  const char               *protocol,
                  const char               *binding,
                  const rmix_parameters_t  *parameters,
                  unsigned int              objectid,
                  void                     *object,
                  const rmix_interface_t   *interface) {
  void         *export;
  unsigned int  provider;
#ifdef DEBUG
  /* Check localref parameter. */
  if (NULL == localref) {
    errno = EINVAL;
    RMIX_LOG((RMIX_WARN(localref parameter is null)))
    errno = EINVAL;
    return -1;
  }
  /* Check interface parameter. */
  if (NULL == interface) {
    errno = EINVAL;
    RMIX_LOG((RMIX_WARN(interface parameter is null)))
    errno = EINVAL;
    return -1;
  }
#endif /* DEBUG */
  /* Check protocol parameter. */
  if ((NULL == protocol)||('\0' == protocol[0])) {
    /* Check parameters parameter. */
    if (NULL == parameters) {
      errno = EINVAL;
      RMIX_LOG((
        RMIX_WARN(parameters parameter is null and no protocol specified)))
      errno = EINVAL;
      return -1;
    }
    /* Search for additional protocol parameter. */
    if (0 != rmix_parameters_search(parameters, &protocol, "PROTOCOL")) {
      if (ESRCH == errno) {
        errno = EINVAL;
        RMIX_LOG((RMIX_WARN(no protocol specified)))
        errno = EINVAL;
        return -1;
      } else {
        int errno2 = errno;
        RMIX_LOG((RMIX_WARN(unable to search for protocol parameter)))
        errno = errno2;
        return -1;
      }
    }
  }
  /* Check binding parameter. */
  if (((NULL == binding)||('\0' == binding[0]))&&(NULL != parameters)) {
    /* Search for additional binding parameter. */
    if ((0     != rmix_parameters_search(parameters, &binding, "BINDING"))&&
        (ESRCH != errno)) {
      int errno2 = errno;
      RMIX_LOG((RMIX_WARN(unable to search for binding parameter)))
      errno = errno2;
      return -1;
    }
  }
  /* Add object to object registry. */
  if (0 != rmix_objects_add2(objectid, object, interface)) {
    int errno2 = errno;
    RMIX_LOG((RMIX_WARN(unable to add object to object registry)))
    errno = errno2;
    return -1;
  }
  /* Load protocol provider plug-in. */
  if (0 != rmix_providers_load(&provider, protocol)) {
    int errno2 = errno;
    RMIX_LOG((RMIX_WARN(unable to load protocol provider plug-in)))
    /* Remove object from object registry. */
    if (0 != rmix_objects_remove(objectid)) {
      RMIX_LOG((RMIX_WARN(unable to remove object from object registry)))
    }
    errno = errno2;
    return -1;
  }
  /* Resolve provider plug-in export function. */
  if (0 != rmix_providers_symbol2(&export, provider, "export")) {
    int errno2 = errno;
    RMIX_LOG((RMIX_WARN(unable to resolve provider plug-in export function)))
    /* Unload protocol provider plug-in. */
    if (0 != rmix_providers_unload(provider)) {
      RMIX_LOG((RMIX_WARN(unable to unoad protocol provider plug-in)))
    }
    /* Remove object from object registry. */
    if (0 != rmix_objects_remove(objectid)) {
      RMIX_LOG((RMIX_WARN(unable to remove object from object registry)))
    }
    errno = errno2;
    return -1;
  }
  /* Check provider plug-in export function. */
  if (NULL == export) {
    errno = ENOTSUP;
    RMIX_LOG((RMIX_WARN(provider plug-in export function does not exist)))
    /* Unload protocol provider plug-in. */
    if (0 != rmix_providers_unload(provider)) {
      RMIX_LOG((RMIX_WARN(unable to unoad protocol provider plug-in)))
    }
    /* Remove object from object registry. */
    if (0 != rmix_objects_remove(objectid)) {
      RMIX_LOG((RMIX_WARN(unable to remove object from object registry)))
    }
    errno = ENOTSUP;
    return -1;
  }
  /* Export local object. */
  if (0 != ((rmix_providers_export_t)export)
            (localref, protocol, binding, parameters,
             provider, objectid, object, interface)) {
    int errno2 = errno;
    RMIX_LOG((RMIX_WARN(unable to export local object)))
    /* Unload protocol provider plug-in. */
    if (0 != rmix_providers_unload(provider)) {
      RMIX_LOG((RMIX_WARN(unable to unoad protocol provider plug-in)))
    }
    /* Remove object from object registry. */
    if (0 != rmix_objects_remove(objectid)) {
      RMIX_LOG((RMIX_WARN(unable to remove object from object registry)))
    }
    errno = errno2;
    return -1;
  }
  return 0;
}

/*
 * Exports a local object while forcing a specific object id and without
 * loading the provider plug-in. This is very useful for standard object
 * exports.
 *
 * localref   = The local object reference of the exported object (return).
 * provider   = The preloaded provider plug-in handle.
 * protocol   = The communication protocol. Overrides any additional PROTOCOL
 *              parameter. May only be NULL or "" if additional parameters
 *              contains PROTOCOL.
 * binding    = The binding  (protocol specific) (may be NULL, ""). Overrides
 *              any additional BINDING parameter.
 * parameters = Additional parameters (protocol specific) (may be NULL).
 * objectid   = The object id.
 * object     = The local object (may be NULL).
 * interface  = The local object interface.
 * return     = 0 on success or -1 on any error with errno set appropriately.
 */
#undef  __FUNC__
#define __FUNC__ "rmix_export2_p"
int rmix_export2_p (rmix_localref_t         **localref,
                    unsigned int              provider,
                    const char               *protocol,
                    const char               *binding,
                    const rmix_parameters_t  *parameters,
                    unsigned int              objectid,
                    void                     *object,
                    const rmix_interface_t   *interface) {
  void *export;
#ifdef DEBUG
  /* Check localref parameter. */
  if (NULL == localref) {
    errno = EINVAL;
    RMIX_LOG((RMIX_WARN(localref parameter is null)))
    errno = EINVAL;
    return -1;
  }
  /* Check interface parameter. */
  if (NULL == interface) {
    errno = EINVAL;
    RMIX_LOG((RMIX_WARN(interface parameter is null)))
    errno = EINVAL;
    return -1;
  }
#endif /* DEBUG */
  /* Check protocol parameter. */
  if ((NULL == protocol)||('\0' == protocol[0])) {
    /* Check parameters parameter. */
    if (NULL == parameters) {
      errno = EINVAL;
      RMIX_LOG((
        RMIX_WARN(parameters parameter is null and no protocol specified)))
      errno = EINVAL;
      return -1;
    }
    /* Search for additional protocol parameter. */
    if (0 != rmix_parameters_search(parameters, &protocol, "PROTOCOL")) {
      if (ESRCH == errno) {
        errno = EINVAL;
        RMIX_LOG((RMIX_WARN(no protocol specified)))
        errno = EINVAL;
        return -1;
      } else {
        int errno2 = errno;
        RMIX_LOG((RMIX_WARN(unable to search for protocol parameter)))
        errno = errno2;
        return -1;
      }
    }
  }
  /* Check binding parameter. */
  if (((NULL == binding)||('\0' == binding[0]))&&(NULL != parameters)) {
    /* Search for additional binding parameter. */
    if ((0     != rmix_parameters_search(parameters, &binding, "BINDING"))&&
        (ESRCH != errno)) {
      int errno2 = errno;
      RMIX_LOG((RMIX_WARN(unable to search for binding parameter)))
      errno = errno2;
      return -1;
    }
  }
  /* Add object to object registry. */
  if (0 != rmix_objects_add2(objectid, object, interface)) {
    int errno2 = errno;
    RMIX_LOG((RMIX_WARN(unable to add object to object registry)))
    errno = errno2;
    return -1;
  }
  /* Resolve provider plug-in export function. */
  if (0 != rmix_providers_symbol2(&export, provider, "export")) {
    int errno2 = errno;
    RMIX_LOG((RMIX_WARN(unable to resolve provider plug-in export function)))
    /* Remove object from object registry. */
    if (0 != rmix_objects_remove(objectid)) {
      RMIX_LOG((RMIX_WARN(unable to remove object from object registry)))
    }
    errno = errno2;
    return -1;
  }
  /* Check provider plug-in export function. */
  if (NULL == export) {
    errno = ENOTSUP;
    RMIX_LOG((RMIX_WARN(provider plug-in export function does not exist)))
    /* Remove object from object registry. */
    if (0 != rmix_objects_remove(objectid)) {
      RMIX_LOG((RMIX_WARN(unable to remove object from object registry)))
    }
    errno = ENOTSUP;
    return -1;
  }
  /* Export local object. */
  if (0 != ((rmix_providers_export_t)export)
            (localref, protocol, binding, parameters,
             provider, objectid, object, interface)) {
    int errno2 = errno;
    RMIX_LOG((RMIX_WARN(unable to export local object)))
    /* Remove object from object registry. */
    if (0 != rmix_objects_remove(objectid)) {
      RMIX_LOG((RMIX_WARN(unable to remove object from object registry)))
    }
    errno = errno2;
    return -1;
  }
  return 0;
}

/*
 * Exports a local object while forcing a specific object id. This is very
 * useful for standard object exports.
 *
 * localref   = The local object reference of the exported object (return).
 * parameters = The parameters must include PROTOCOL (for protocol) and
 *              OBJECTID (for objectid), and may include BINDING and additional
 *              protocol specific parameters (may not be NULL).
 * object     = The local object (may be NULL).
 * interface  = The local object interface.
 * return     = 0 on success or -1 on any error with errno set appropriately.
 */
#undef  __FUNC__
#define __FUNC__ "rmix_export3"
int rmix_export3 (rmix_localref_t         **localref,
                  const rmix_parameters_t  *parameters,
                  void                     *object,
                  const rmix_interface_t   *interface) {
  const char   *protocol, *binding, *objectstr;
  void         *export;
  unsigned int  objectid, provider;
#ifdef DEBUG
  /* Check localref parameter. */
  if (NULL == localref) {
    errno = EINVAL;
    RMIX_LOG((RMIX_WARN(localref parameter is null)))
    errno = EINVAL;
    return -1;
  }
  /* Check parameters parameter. */
  if (NULL == parameters) {
    errno = EINVAL;
    RMIX_LOG((RMIX_WARN(parameters parameter is null)))
    errno = EINVAL;
    return -1;
  }
  /* Check interface parameter. */
  if (NULL == interface) {
    errno = EINVAL;
    RMIX_LOG((RMIX_WARN(interface parameter is null)))
    errno = EINVAL;
    return -1;
  }
#endif /* DEBUG */
  /* Search for protocol parameter. */
  if (0 != rmix_parameters_search(parameters, &protocol, "PROTOCOL")) {
    if (ESRCH == errno) {
      errno = EINVAL;
      RMIX_LOG((RMIX_WARN(no protocol specified)))
      errno = EINVAL;
      return -1;
    } else {
      int errno2 = errno;
      RMIX_LOG((RMIX_WARN(unable to search for protocol parameter)))
      errno = errno2;
      return -1;
    }
  }
  /* Search for object id parameter. */
  if (0 != rmix_parameters_search(parameters, &objectstr, "OBJECTID")) {
    if (ESRCH == errno) {
      errno = EINVAL;
      RMIX_LOG((RMIX_WARN(no object id specified)))
      errno = EINVAL;
      return -1;
    } else {
      int errno2 = errno;
      RMIX_LOG((RMIX_WARN(unable to search for object id  parameter)))
      errno = errno2;
      return -1;
    }
  }
  /* Convert object id parameter to object id. */
  objectid = (unsigned int)atoi(objectstr);
  /* Search for additional binding parameter. */
  if (0 != rmix_parameters_search(parameters, &binding, "BINDING")) {
    if (ESRCH != errno) {
      int errno2 = errno;
      RMIX_LOG((RMIX_WARN(unable to search for binding parameter)))
      errno = errno2;
      return -1;
    }
    binding = NULL;
  }
  /* Add object to object registry. */
  if (0 != rmix_objects_add2(objectid, object, interface)) {
    int errno2 = errno;
    RMIX_LOG((RMIX_WARN(unable to add object to object registry)))
    errno = errno2;
    return -1;
  }
  /* Load protocol provider plug-in. */
  if (0 != rmix_providers_load(&provider, protocol)) {
    int errno2 = errno;
    RMIX_LOG((RMIX_WARN(unable to load protocol provider plug-in)))
    /* Remove object from object registry. */
    if (0 != rmix_objects_remove(objectid)) {
      RMIX_LOG((RMIX_WARN(unable to remove object from object registry)))
    }
    errno = errno2;
    return -1;
  }
  /* Resolve provider plug-in export function. */
  if (0 != rmix_providers_symbol2(&export, provider, "export")) {
    int errno2 = errno;
    RMIX_LOG((RMIX_WARN(unable to resolve provider plug-in export function)))
    /* Unload protocol provider plug-in. */
    if (0 != rmix_providers_unload(provider)) {
      RMIX_LOG((RMIX_WARN(unable to unoad protocol provider plug-in)))
    }
    /* Remove object from object registry. */
    if (0 != rmix_objects_remove(objectid)) {
      RMIX_LOG((RMIX_WARN(unable to remove object from object registry)))
    }
    errno = errno2;
    return -1;
  }
  /* Check provider plug-in export function. */
  if (NULL == export) {
    errno = ENOTSUP;
    RMIX_LOG((RMIX_WARN(provider plug-in export function does not exist)))
    /* Unload protocol provider plug-in. */
    if (0 != rmix_providers_unload(provider)) {
      RMIX_LOG((RMIX_WARN(unable to unoad protocol provider plug-in)))
    }
    /* Remove object from object registry. */
    if (0 != rmix_objects_remove(objectid)) {
      RMIX_LOG((RMIX_WARN(unable to remove object from object registry)))
    }
    errno = ENOTSUP;
    return -1;
  }
  /* Export local object. */
  if (0 != ((rmix_providers_export_t)export)
            (localref, protocol, binding, parameters,
             provider, objectid, object, interface)) {
    int errno2 = errno;
    RMIX_LOG((RMIX_WARN(unable to export local object)))
    /* Unload protocol provider plug-in. */
    if (0 != rmix_providers_unload(provider)) {
      RMIX_LOG((RMIX_WARN(unable to unoad protocol provider plug-in)))
    }
    /* Remove object from object registry. */
    if (0 != rmix_objects_remove(objectid)) {
      RMIX_LOG((RMIX_WARN(unable to remove object from object registry)))
    }
    errno = errno2;
    return -1;
  }
  return 0;
}

/*
 * Exports a local object while forcing a specific object id and without
 * loading the provider plug-in. This is very useful for standard object
 * exports.
 *
 * localref   = The local object reference of the exported object (return).
 * provider   = The preloaded provider plug-in handle.
 * parameters = The parameters must include PROTOCOL (for protocol) and
 *              OBJECTID (for objectid), and may include BINDING and additional
 *              protocol specific parameters (may not be NULL).
 * object     = The local object (may be NULL).
 * interface  = The local object interface.
 * return     = 0 on success or -1 on any error with errno set appropriately.
 */
#undef  __FUNC__
#define __FUNC__ "rmix_export3_p"
int rmix_export3_p (rmix_localref_t         **localref,
                    unsigned int              provider,
                    const rmix_parameters_t  *parameters,
                    void                     *object,
                    const rmix_interface_t   *interface) {
  const char   *protocol, *binding, *objectstr;
  void         *export;
  unsigned int  objectid;
#ifdef DEBUG
  /* Check localref parameter. */
  if (NULL == localref) {
    errno = EINVAL;
    RMIX_LOG((RMIX_WARN(localref parameter is null)))
    errno = EINVAL;
    return -1;
  }
  /* Check parameters parameter. */
  if (NULL == parameters) {
    errno = EINVAL;
    RMIX_LOG((RMIX_WARN(parameters parameter is null)))
    errno = EINVAL;
    return -1;
  }
  /* Check interface parameter. */
  if (NULL == interface) {
    errno = EINVAL;
    RMIX_LOG((RMIX_WARN(interface parameter is null)))
    errno = EINVAL;
    return -1;
  }
#endif /* DEBUG */
  /* Search for protocol parameter. */
  if (0 != rmix_parameters_search(parameters, &protocol, "PROTOCOL")) {
    if (ESRCH == errno) {
      errno = EINVAL;
      RMIX_LOG((RMIX_WARN(no protocol specified)))
      errno = EINVAL;
      return -1;
    } else {
      int errno2 = errno;
      RMIX_LOG((RMIX_WARN(unable to search for protocol parameter)))
      errno = errno2;
      return -1;
    }
  }
  /* Search for object id parameter. */
  if (0 != rmix_parameters_search(parameters, &objectstr, "OBJECTID")) {
    if (ESRCH == errno) {
      errno = EINVAL;
      RMIX_LOG((RMIX_WARN(no object id specified)))
      errno = EINVAL;
      return -1;
    } else {
      int errno2 = errno;
      RMIX_LOG((RMIX_WARN(unable to search for object id  parameter)))
      errno = errno2;
      return -1;
    }
  }
  /* Convert object id parameter to object id. */
  objectid = (unsigned int)atoi(objectstr);
  /* Search for additional binding parameter. */
  if (0 != rmix_parameters_search(parameters, &binding, "BINDING")) {
    if (ESRCH != errno) {
      int errno2 = errno;
      RMIX_LOG((RMIX_WARN(unable to search for binding parameter)))
      errno = errno2;
      return -1;
    }
    binding = NULL;
  }
  /* Add object to object registry. */
  if (0 != rmix_objects_add2(objectid, object, interface)) {
    int errno2 = errno;
    RMIX_LOG((RMIX_WARN(unable to add object to object registry)))
    errno = errno2;
    return -1;
  }
  /* Resolve provider plug-in export function. */
  if (0 != rmix_providers_symbol2(&export, provider, "export")) {
    int errno2 = errno;
    RMIX_LOG((RMIX_WARN(unable to resolve provider plug-in export function)))
    /* Remove object from object registry. */
    if (0 != rmix_objects_remove(objectid)) {
      RMIX_LOG((RMIX_WARN(unable to remove object from object registry)))
    }
    errno = errno2;
    return -1;
  }
  /* Check provider plug-in export function. */
  if (NULL == export) {
    errno = ENOTSUP;
    RMIX_LOG((RMIX_WARN(provider plug-in export function does not exist)))
    /* Remove object from object registry. */
    if (0 != rmix_objects_remove(objectid)) {
      RMIX_LOG((RMIX_WARN(unable to remove object from object registry)))
    }
    errno = ENOTSUP;
    return -1;
  }
  /* Export local object. */
  if (0 != ((rmix_providers_export_t)export)
            (localref, protocol, binding, parameters,
             provider, objectid, object, interface)) {
    int errno2 = errno;
    RMIX_LOG((RMIX_WARN(unable to export local object)))
    /* Remove object from object registry. */
    if (0 != rmix_objects_remove(objectid)) {
      RMIX_LOG((RMIX_WARN(unable to remove object from object registry)))
    }
    errno = errno2;
    return -1;
  }
  return 0;
}

/*
 * Exports a local object while forcing a specific object id. This is very
 * useful for standard object exports.
 *
 * localref   = The local object reference of the exported object (return).
 * parameters = The parameter pairs must include PROTOCOL (for protocol) and
 *              OBJECTID (for objectid), and may include BINDING and additional
 *              protocol specific parameters (may not be NULL).
 * object     = The local object (may be NULL).
 * interface  = The local object interface.
 * return     = 0 on success or -1 on any error with errno set appropriately.
 */
#undef  __FUNC__
#define __FUNC__ "rmix_export4"
int rmix_export4 (rmix_localref_t         **localref,
                  const char               *parameters,
                  void                     *object,
                  const rmix_interface_t   *interface) {
  rmix_parameters_t *parameters2;
#ifdef DEBUG
  /* Check localref parameter. */
  if (NULL == localref) {
    errno = EINVAL;
    RMIX_LOG((RMIX_WARN(localref parameter is null)))
    errno = EINVAL;
    return -1;
  }
  /* Check parameters parameter. */
  if (NULL == parameters) {
    errno = EINVAL;
    RMIX_LOG((RMIX_WARN(parameters parameter is null)))
    errno = EINVAL;
    return -1;
  }
  /* Check interface parameter. */
  if (NULL == interface) {
    errno = EINVAL;
    RMIX_LOG((RMIX_WARN(interface parameter is null)))
    errno = EINVAL;
    return -1;
  }
#endif /* DEBUG */
  /* Create papameters. */
  if (0 != rmix_parameters_create4(&parameters2, parameters)) {
    int errno2 = errno;
    RMIX_LOG((RMIX_WARN(unable to create parameters)))
    errno = errno2;
    return -1;
  }
  /* Export object. */
  if (0 != rmix_export3(localref, parameters2, object, interface)) {
    int errno2 = errno;
    RMIX_LOG((RMIX_WARN(unable to export object)))
    /* Destroy papameters. */
    if (0 != rmix_parameters_destroy(&parameters2)) {
      RMIX_LOG((RMIX_WARN(unable to destroy parameters)))
    }
    errno = errno2;
    return -1;
  }
  /* Destroy papameters. */
  if (0 != rmix_parameters_destroy(&parameters2)) {
    int errno2 = errno;
    RMIX_LOG((RMIX_WARN(unable to destroy parameters)))
    /* Unexport object. */
    if (0 != rmix_unexport(localref)) {
      RMIX_LOG((RMIX_WARN(unable to unexport object)))
    }
    errno = errno2;
    return -1;
  }
  return 0;
}

/*
 * Exports a local object while forcing a specific object id and without
 * loading the provider plug-in. This is very useful for standard object
 * exports.
 *
 * localref   = The local object reference of the exported object (return).
 * provider   = The preloaded provider plug-in handle.
 * parameters = The parameter pairs must include PROTOCOL (for protocol) and
 *              OBJECTID (for objectid), and may include BINDING and additional
 *              protocol specific parameters (may not be NULL).
 * object     = The local object (may be NULL).
 * interface  = The local object interface.
 * return     = 0 on success or -1 on any error with errno set appropriately.
 */
#undef  __FUNC__
#define __FUNC__ "rmix_export4_p"
int rmix_export4_p (rmix_localref_t         **localref,
                    unsigned int              provider,
                    const char               *parameters,
                    void                     *object,
                    const rmix_interface_t   *interface) {
  rmix_parameters_t *parameters2;
#ifdef DEBUG
  /* Check localref parameter. */
  if (NULL == localref) {
    errno = EINVAL;
    RMIX_LOG((RMIX_WARN(localref parameter is null)))
    errno = EINVAL;
    return -1;
  }
  /* Check parameters parameter. */
  if (NULL == parameters) {
    errno = EINVAL;
    RMIX_LOG((RMIX_WARN(parameters parameter is null)))
    errno = EINVAL;
    return -1;
  }
  /* Check interface parameter. */
  if (NULL == interface) {
    errno = EINVAL;
    RMIX_LOG((RMIX_WARN(interface parameter is null)))
    errno = EINVAL;
    return -1;
  }
#endif /* DEBUG */
  /* Create papameters. */
  if (0 != rmix_parameters_create4(&parameters2, parameters)) {
    int errno2 = errno;
    RMIX_LOG((RMIX_WARN(unable to create parameters)))
    errno = errno2;
    return -1;
  }
  /* Export object. */
  if (0 != rmix_export3_p(localref, provider, parameters2, object, interface)) {
    int errno2 = errno;
    RMIX_LOG((RMIX_WARN(unable to export object)))
    /* Destroy papameters. */
    if (0 != rmix_parameters_destroy(&parameters2)) {
      RMIX_LOG((RMIX_WARN(unable to destroy parameters)))
    }
    errno = errno2;
    return -1;
  }
  /* Destroy papameters. */
  if (0 != rmix_parameters_destroy(&parameters2)) {
    int errno2 = errno;
    RMIX_LOG((RMIX_WARN(unable to destroy parameters)))
    /* Unexport object. */
    if (0 != rmix_unexport_p(localref)) {
      RMIX_LOG((RMIX_WARN(unable to unexport object)))
    }
    errno = errno2;
    return -1;
  }
  return 0;
}

/*
 * Unexports a local object.
 *
 * localref = The local object reference of the exported object (also return).
 * return   = 0 on success or -1 on any error with errno set appropriately.
 */
#undef  __FUNC__
#define __FUNC__ "rmix_unexport"
int rmix_unexport (rmix_localref_t **localref) {
  int           result, errno2;
  void         *unexport;
  unsigned int  objectid, provider;
#ifdef DEBUG
  /* Check localref parameter. */
  if (NULL == localref) {
    errno = EINVAL;
    RMIX_LOG((RMIX_WARN(localref parameter is null)))
    errno = EINVAL;
    return -1;
  }
  /* Check *localref parameter. */
  if (NULL == *localref) {
    errno = EINVAL;
    RMIX_LOG((RMIX_WARN(*localref parameter is null)))
    errno = EINVAL;
    return -1;
  }
#endif /* DEBUG */
  result = 0;
  errno2 = 0;
  /* Save object and provider id. */
  objectid = (*localref)->objectid;
  provider = (*localref)->provider;
  /* Resolve provider plug-in unexport function. */
  if (0 != rmix_providers_symbol2(&unexport, (*localref)->provider,
                                  "unexport")) {
    errno2 = errno;
    RMIX_LOG((RMIX_WARN(unable to resolve provider plug-in unexport function)))
    result = -1;
  } else {
    /* Check provider plug-in unexport function. */
    if (NULL == unexport) {
      errno2 = ENOTSUP;
      errno = ENOTSUP;
      RMIX_LOG((RMIX_WARN(provider plug-in unexport function does not exist)))
      result = -1;
    } else {
      /* Unexport local object. */
      if (0 != ((rmix_providers_unexport_t)unexport)(localref)) {
        errno2 = errno;
        RMIX_LOG((RMIX_WARN(unable to export local object)))
        result = -1;
      }
    }
  }
  /* Unload protocol provider plug-in. */
  if (0 != rmix_providers_unload(provider)) {
    RMIX_LOG((RMIX_WARN(unable to unoad protocol provider plug-in)))
    errno2 = errno;
    result = -1;
  }
  /* Remove object from object registry. */
  if (0 != rmix_objects_remove(objectid)) {
    RMIX_LOG((RMIX_WARN(unable to remove object from object registry)))
    errno2 = errno;
    result = -1;
  }
  if (0 != errno2) {
    errno = errno2;
  }
  return result;
}

/*
 * Unexports a local object without unloading the provider plug-in.
 *
 * localref = The local object reference of the exported object (also return).
 * return   = 0 on success or -1 on any error with errno set appropriately.
 */
#undef  __FUNC__
#define __FUNC__ "rmix_unexport_p"
int rmix_unexport_p (rmix_localref_t **localref) {
  int           result, errno2;
  void         *unexport;
  unsigned int  objectid;
#ifdef DEBUG
  /* Check localref parameter. */
  if (NULL == localref) {
    errno = EINVAL;
    RMIX_LOG((RMIX_WARN(localref parameter is null)))
    errno = EINVAL;
    return -1;
  }
  /* Check *localref parameter. */
  if (NULL == *localref) {
    errno = EINVAL;
    RMIX_LOG((RMIX_WARN(*localref parameter is null)))
    errno = EINVAL;
    return -1;
  }
#endif /* DEBUG */
  result = 0;
  errno2 = 0;
  /* Save object id. */
  objectid = (*localref)->objectid;
  /* Resolve provider plug-in unexport function. */
  if (0 != rmix_providers_symbol2(&unexport, (*localref)->provider,
                                  "unexport")) {
    errno2 = errno;
    RMIX_LOG((RMIX_WARN(unable to resolve provider plug-in unexport function)))
    result = -1;
  } else {
    /* Check provider plug-in unexport function. */
    if (NULL == unexport) {
      errno2 = ENOTSUP;
      errno = ENOTSUP;
      RMIX_LOG((RMIX_WARN(provider plug-in unexport function does not exist)))
      result = -1;
    } else {
      /* Unexport local object. */
      if (0 != ((rmix_providers_unexport_t)unexport)(localref)) {
        errno2 = errno;
        RMIX_LOG((RMIX_WARN(unable to export local object)))
        result = -1;
      }
    }
  }
  /* Remove object from object registry. */
  if (0 != rmix_objects_remove((*localref)->objectid)) {
    errno2 = errno;
    RMIX_LOG((RMIX_WARN(unable to remove object from object registry)))
    result = -1;
  }
  if (0 != errno2) {
    errno = errno2;
  }
  return result;
}

/******************************************************************************
 *
 * END OF FILE
 *
 ******************************************************************************/
