/******************************************************************************
 *
 * Source file for rmix tcpip4 (TCP/IPv4 pool) module.
 * Copyright (c) 2002-2006 Oak Ridge National Laboratory.
 *
 * For more information see the following files in the source distribution top-
 * level directory or package data directory (usually /usr/local/share/package):
 *
 * - README    for general package information.
 * - INSTALL   for package install information.
 * - COPYING   for package license information and copying conditions.
 * - AUTHORS   for package authors information.
 * - ChangeLog for package changes information.
 *
 ******************************************************************************/

/******************************************************************************
 *
 * Macros
 *
 ******************************************************************************/

#define RMIX_TCPIP4_BUFFER_SIZE 64

/******************************************************************************
 *
 * Includes
 *
 ******************************************************************************/

/* Main module header file. */
#include "rmix.h"

/******************************************************************************
 *
 * Data Types
 *
 ******************************************************************************/

/* TCP/IPv4 pool server data type. */
typedef struct {
  rmix_tcpip4_server_t    *server;    /* server      */
  rmix_tcpip4_serverjob_t  serverjob; /* server job  */
  unsigned int             busy;      /* busy flag   */
} rmix_tcpip4_pool_server_t;          /* pool server */

/* TCP/IPv4 pool data type. */
typedef struct {
  pthread_mutex_t               mutex;     /* mutex                      */
  unsigned int                  ready;     /* initialized flag           */
  int                           wakeup[2]; /* wakeup pipe                */
  unsigned int                  count;     /* listener list count        */
  struct {
    rmix_tcpip4_listener_t     *listener;  /* TCP/IPv4 listener          */
    unsigned int                handle;    /* handle                     */
    unsigned int                close;     /* close flag                 */
    rmix_tcpip4_serverjob_t     serverjob; /* server job for new servers */
    unsigned int                count;     /* server list count          */
    rmix_tcpip4_pool_server_t **servers;   /* server list                */
  }                            *listeners; /* listener list              */
  char                          buffer[RMIX_TCPIP4_BUFFER_SIZE];
                                          /* wakeup pipe buffer         */
} rmix_tcpip4_data_t;

/******************************************************************************
 *
 * Data
 *
 ******************************************************************************/

/* TCP/IPv4 pool data. */
rmix_tcpip4_data_t rmix_tcpip4_data = {
  PTHREAD_MUTEX_INITIALIZER, /* mutex               */
  0,                         /* initialized flag    */
  { -1, -1},                 /* wakeup pipe         */
  0,                         /* listener list count */
  NULL                       /* listener list       */
};

/******************************************************************************
 *
 * Function Prototypes
 *
 ******************************************************************************/

/*
 * Initializes the TCP/IPv4 pool.
 *
 * return = 0 on success or -1 on any error with errno set appropriately.
 */
int rmix_tcpip4_init ();

/*
 * Finalizes the TCP/IPv4 pool.
 *
 * return = 0 on success or -1 on any error with errno set appropriately.
 */
int rmix_tcpip4_fini ();

/*
 * Serves all listener and server connections.
 *
 * argument = NULL
 * return   = 0 on success or -1 on any system error with errno set
 *            appropriately.
 */
RMIX_THREADS_JOB(rmix_tcpip4_job);

/*
 * The TCP/IPv4 pool module mutex unlock cleanup.
 *
 * argument = NULL.
 */
void rmix_tcpip4_unlock (void *argument);

/*
 * Serves server connections.
 *
 * argument = The pool server.
 * return   = 0 on success or -1 on any system error with errno set
 *            appropriately.
 */
RMIX_THREADS_JOB(rmix_tcpip4_server_job);

/*
 * TCP/IPv4 server job cleanup.
 *
 * server = The TCP/IPv4 server.
 * return = 0 on success or -1 on any error with errno set appropriately.
 */
void rmix_tcpip4_server_job_cleanup (void *pool_server);

/*
 * Reads a buffer from a file descriptor (non-/blocking).
 *
 * fd     = The file descriptor to read from.
 * length = The maximum buffer length to read.
 * buffer = The buffer to read into.
 * block  = The (non-)blocking flag (0 = non-blocking, 1 = blocking).
 * return = number of bytes read on success, -2 on closed file descriptor or -1
 *          on any other error with errno set appropriately.
 */
int rmix_tcpip4_read (int           fd,
                      unsigned int  length,
                      void         *buffer,
                      unsigned int  block);

/******************************************************************************
 *
 * Functions
 *
 ******************************************************************************/

/*
 * Returns the first local non-loopback address, if any, or the loopback
 * address. This function may be called before initializing.
 *
 * localaddr = The first local non-loopback address, if any, or the loopback
 *             address (return).
 * return    = 0 on success or -1 on any error with errno set appropriately.
 */
#undef  __FUNC__
#define __FUNC__ "rmix_tcpip4_address"
int rmix_tcpip4_address (struct sockaddr_in *localaddr) {
  int                  socketfd;
  unsigned int         index;
  struct ifreq         interface;
  struct if_nameindex *interfaces;
#ifdef DEBUG
  /* Check localaddr parameter. */
  if (NULL == localaddr) {
    errno = EINVAL;
    RMIX_LOG((RMIX_WARN(localaddr parameter is null)))
    errno = EINVAL;
    return -1;
  }
#endif /* DEBUG */
  /* Allocate interface names. */
  if (NULL == (interfaces = if_nameindex())) {
    int errno2 = errno;
    RMIX_LOG((RMIX_WARN(unable to allocate interface names)))
    errno = errno2;
    return -1;
  }
  /* Open socket. */
  if (-1 == (socketfd = socket(AF_INET, SOCK_DGRAM, IPPROTO_IP))) {
    int errno2 = errno;
    RMIX_LOG((RMIX_WARN(unable to open socket)))
    /* Deallocate interface names. */
    if_freenameindex(interfaces);
    errno = errno2;
    return -1;
  }
  /* Obtain interface addresses. */
  for (index = 0; NULL != interfaces[index].if_name; index++) {
    /* Duplicate interface name. */
    strcpy(interface.ifr_name, interfaces[index].if_name);
    /* Obtain interface address. */
    if (0 == ioctl(socketfd, SIOCGIFADDR, &interface)) {
      /* Save interface address. */
      localaddr->sin_addr =
        ((struct sockaddr_in*)&interface.ifr_addr)->sin_addr;
      /* Check for non-loopback address. */
      if ((((char*)&(localaddr->sin_addr))[0] != 127)&&
          (((char*)&(localaddr->sin_addr))[1] !=   0)&&
          (((char*)&(localaddr->sin_addr))[2] !=   0)&&
          (((char*)&(localaddr->sin_addr))[3] !=   1)) {
        /* Close socket. */
        close(socketfd);
        /* Deallocate interface names. */
        if_freenameindex(interfaces);
        return 0;
      }
    }
  }
  /* Close socket. */
  close(socketfd);
  /* Deallocate interface names. */
  if_freenameindex(interfaces);
  localaddr->sin_addr.s_addr = htonl(INADDR_LOOPBACK);
  return 0;
}

/*
 * Resolves a name/port pair to address/port. If name is NULL, the first local
 * non-loopback address, if any, or the loopback address is assigned.
 * This function may be called before initializing.
 *
 * localaddr = The local address/port to resolve (return).
 * localname = The local name to resolve (may be NULL).
 * localport = The local port to resolve.
 * return    = 0 on success or -1 on any error with errno set appropriately.
 */
#undef  __FUNC__
#define __FUNC__ "rmix_tcpip4_resolve"
int rmix_tcpip4_resolve (struct sockaddr_in *localaddr,
                         const char         *localname,
                         unsigned int        localport) {
  int              error;
  struct addrinfo *info;
#ifdef DEBUG
  /* Check localaddr parameter. */
  if (NULL == localaddr) {
    errno = EINVAL;
    RMIX_LOG((RMIX_WARN(localaddr parameter is null)))
    errno = EINVAL;
    return -1;
  }
#endif /* DEBUG */
  /* Check for missing name. */
  if (NULL == localname) {
    /* Find first address. */
    if (0 != rmix_tcpip4_address(localaddr)) {
      int errno2 = errno;
      RMIX_LOG((RMIX_WARN(unable to find first address)))
      errno = errno2;
      return -1;
    }
    localaddr->sin_port = htons(localport);
  } else {
    /* Lookup name. */
    if (0 != (error = getaddrinfo(localname, NULL, NULL, &info))) {
      errno = error;
      RMIX_LOG((RMIX_WARN(unable to lookup local name '%s'), localname))
      errno = error;
      return -1;
    }
    localaddr->sin_family = ((struct sockaddr_in*)info->ai_addr)->sin_family;
    localaddr->sin_addr   = ((struct sockaddr_in*)info->ai_addr)->sin_addr;
    localaddr->sin_port   = htons(localport);
    freeaddrinfo(info);
  }
  return 0;
}

/*
 * Opens a TCP/IPv4 listener and binds it to all local interfaces.
 * This function may be called before initializing.
 *
 * listener = The TCP/IPv4 listener (return).
 * port     = The port to bind to.
 * backlog  = The listen backlog.
 * return   = 0 on success or -1 on any error with errno set appropriately.
 */
#undef  __FUNC__
#define __FUNC__ "rmix_tcpip4_listener_open"
int rmix_tcpip4_listener_open (rmix_tcpip4_listener_t **listener,
                               unsigned int             localport,
                               unsigned int             backlog) {
  struct sockaddr_in localaddr;
#ifdef DEBUG
  /* Check listener parameter. */
  if (NULL == listener) {
    errno = EINVAL;
    RMIX_LOG((RMIX_WARN(listener parameter is null)))
    errno = EINVAL;
    return -1;
  }
#endif /* DEBUG */
  localaddr.sin_family      = AF_INET;
  localaddr.sin_addr.s_addr = htonl(INADDR_ANY);
  localaddr.sin_port        = htons(localport);
  /* Open listener. */
  if (0 != rmix_tcpip4_listener_open2(listener, &localaddr, backlog)) {
    int errno2 = errno;
    RMIX_LOG((RMIX_WARN(unable to open listener)))
    errno = errno2;
    return -1;
  }
  return 0;
}

/*
 * Opens a TCP/IPv4 listener and binds it to the addressed local interface.
 * This function may be called before initializing.
 *
 * listener  = The TCP/IPv4 listener (return).
 * localaddr = The local address/port to bind to.
 * backlog   = The listen backlog.
 * return    = 0 on success or -1 on any error with errno set appropriately.
 */
#undef  __FUNC__
#define __FUNC__ "rmix_tcpip4_listener_open2"
int rmix_tcpip4_listener_open2 (rmix_tcpip4_listener_t   **listener,
                                const struct sockaddr_in  *localaddr,
                                unsigned int               backlog) {
#ifdef DEBUG
  /* Check listener parameter. */
  if (NULL == listener) {
    errno = EINVAL;
    RMIX_LOG((RMIX_WARN(listener parameter is null)))
    errno = EINVAL;
    return -1;
  }
#endif /* DEBUG */
  /* Allocate listener. */
  if (NULL == (*listener = (rmix_tcpip4_listener_t*)
                           malloc(sizeof(rmix_tcpip4_listener_t)))) {
    int errno2 = errno;
    RMIX_LOG((RMIX_WARN(unable to allocate listener)))
    errno = errno2;
    return -1;
  }
  /* Create listener socket. */
  if (-1 == ((*listener)->socket = socket(AF_INET, SOCK_STREAM, 0))) {
    int errno2 = errno;
    RMIX_LOG((RMIX_WARN(unable to create listener socket)))
    /* Deallocate listener. */
    free(*listener);
    *listener = NULL;
    errno = errno2;
    return -1;
  }
  /* Bind listener socket. */
  if (0 != bind((*listener)->socket, localaddr, sizeof(*localaddr))) {
    int errno2 = errno;
    RMIX_LOG((RMIX_WARN(unable to bind listener socket)))
    /* Destroy listener socket. */
    close((*listener)->socket);
    /* Deallocate listener. */
    free(*listener);
    *listener = NULL;
    errno = errno2;
    return -1;
  }
  /* Listen on listener socket. */
  if (0 != listen((*listener)->socket, backlog)) {
    int errno2 = errno;
    RMIX_LOG((RMIX_WARN(unable to listen on listener socket)))
    /* Destroy listener socket. */
    close((*listener)->socket);
    /* Deallocate listener. */
    free(*listener);
    *listener = NULL;
    errno = errno2;
    return -1;
  }
  /* Set local address. */
  (*listener)->localaddr.sin_family = localaddr->sin_family;
  (*listener)->localaddr.sin_addr   = localaddr->sin_addr;
  (*listener)->localaddr.sin_port   = localaddr->sin_port;
  return 0;
}

/*
 * Opens a TCP/IPv4 listener and binds it to the addressed local interface.
 * This function may be called before initializing.
 *
 * listener   = The TCP/IPv4 listener (return).
 * localname  = The local name to bind to.
 * localport  = The local port to bind to.
 * backlog    = The listen backlog.
 * return     = 0 on success or -1 on any error with errno set appropriately.
 */
#undef  __FUNC__
#define __FUNC__ "rmix_tcpip4_listener_open3"
int rmix_tcpip4_listener_open3 (rmix_tcpip4_listener_t **listener,
                                const char              *localname,
                                unsigned int             localport,
                                unsigned int             backlog) {
  int                 error;
  struct addrinfo    *info;
  struct sockaddr_in  localaddr;
#ifdef DEBUG
  /* Check listener parameter. */
  if (NULL == listener) {
    errno = EINVAL;
    RMIX_LOG((RMIX_WARN(listener parameter is null)))
    errno = EINVAL;
    return -1;
  }
  /* Check localname parameter. */
  if (NULL == localname) {
    errno = EINVAL;
    RMIX_LOG((RMIX_WARN(localname parameter is null)))
    errno = EINVAL;
    return -1;
  }
#endif /* DEBUG */
  /* Lookup name. */
  if (0 != (error = getaddrinfo(localname, NULL, NULL, &info))) {
    errno = error;
    RMIX_LOG((RMIX_WARN(unable to lookup local name '%s'), localname))
    errno = error;
    return -1;
  }
  localaddr.sin_addr = ((struct sockaddr_in*)info->ai_addr)->sin_addr;
  localaddr.sin_port = htons((unsigned short)localport);
  freeaddrinfo(info);
  /* Open listener. */
  if (0 != rmix_tcpip4_listener_open2(listener, &localaddr, backlog)) {
    int errno2 = errno;
    RMIX_LOG((RMIX_WARN(unable to open listener)))
    errno = errno2;
    return -1;
  }
  return 0;
}

/*
 * Opens a TCP/IPv4 listener and binds it to the addressed local interface.
 * This function may be called before initializing.
 *
 * listener = The TCP/IPv4 listener (return).
 * localurl = The local URL to bind to.
 * backlog  = The listen backlog.
 * return   = 0 on success or -1 on any error with errno set appropriately.
 */
#undef  __FUNC__
#define __FUNC__ "rmix_tcpip4_listener_open4"
int rmix_tcpip4_listener_open4 (rmix_tcpip4_listener_t **listener,
                                const char              *localurl,
                                unsigned int             backlog) {
  char *localurl2, *localname, *rest, *localport;
#ifdef DEBUG
  /* Check listener parameter. */
  if (NULL == listener) {
    errno = EINVAL;
    RMIX_LOG((RMIX_WARN(listener parameter is null)))
    errno = EINVAL;
    return -1;
  }
  /* Check localurl parameter. */
  if (NULL == localurl) {
    errno = EINVAL;
    RMIX_LOG((RMIX_WARN(localurl parameter is null)))
    errno = EINVAL;
    return -1;
  }
#endif /* DEBUG */
  /* Duplicate URL. */
  if (NULL == (localurl2 = strdup(localurl))) {
    int errno2 = errno;
    RMIX_LOG((RMIX_WARN(unable to duplicate local URL)))
    errno = errno2;
    return -1;
  }
  if (NULL != (localname = strstr(localurl2, "://"))) {
    localname += 3;
  } else {
    localname = localurl2;
  }
  if (NULL != (rest = strstr(localname, "/"))) {
    rest[0] = '\0';
  }
  if (NULL != (localport = strstr(localname, ":"))) {
    localport += 1;
  } else {
    localport = "0";
  }
  /* Open listener. */
  if (0 != rmix_tcpip4_listener_open3(listener, localname,
                                      atoi(localport), backlog)) {
    int errno2 = errno;
    RMIX_LOG((RMIX_WARN(unable to open listener)))
    /* Deallocate local URL duplicate. */
    free(localurl2);
    errno = errno2;
    return -1;
  }
  /* Deallocate local URL duplicate. */
  free(localurl2);
  return 0;
}

/*
 * Closes a TCP/IPv4 listener.
 * This function may be called before initializing.
 *
 * listener = The TCP/IPv4 listener.
 * return   = 0 on success or -1 on any error with errno set appropriately.
 */
#undef  __FUNC__
#define __FUNC__ "rmix_tcpip4_listener_close"
int rmix_tcpip4_listener_close (rmix_tcpip4_listener_t **listener) {
#ifdef DEBUG
  /* Check listener parameter. */
  if (NULL == listener) {
    errno = EINVAL;
    RMIX_LOG((RMIX_WARN(listener parameter is null)))
    errno = EINVAL;
    return -1;
  }
  if (NULL == *listener) {
    errno = EINVAL;
    RMIX_LOG((RMIX_WARN(listener is null)))
    errno = EINVAL;
    return -1;
  }
#endif /* DEBUG */
  /* Destroy listener socket. */
  close((*listener)->socket);
  /* Deallocate listener. */
  free(*listener);
  *listener = NULL;
  return 0;
}

/*
 * TCP/IPv4 listener cleanup.
 * This function may be called before initializing.
 *
 * listener = The TCP/IPv4 listener.
 * return   = 0 on success or -1 on any error with errno set appropriately.
 */
#undef  __FUNC__
#define __FUNC__ "rmix_tcpip4_listener_close2"
void rmix_tcpip4_listener_close2 (rmix_tcpip4_listener_t *listener) {
#ifdef DEBUG
  /* Check listener parameter. */
  if (NULL == listener) {
    errno = EINVAL;
    RMIX_LOG((RMIX_WARN(listener parameter is null)))
    errno = EINVAL;
    return;
  }
#endif /* DEBUG */
  /* Destroy listener socket. */
  close(listener->socket);
  /* Deallocate listener. */
  free(listener);
}

/*
 * Opens a TCP/IPv4 server by accepting a connection to a TCP/IPv4 listener.
 * This function may be called before initializing.
 *
 * server   = The TCP/IPv4 server (return).
 * listener = The TCP/IPv4 listener.
 * return   = 0 on success or -1 on any error with errno set appropriately.
 */
#undef  __FUNC__
#define __FUNC__ "rmix_tcpip4_sever_open"
int rmix_tcpip4_sever_open (rmix_tcpip4_server_t   **server,
                            rmix_tcpip4_listener_t  *listener) {
  socklen_t remoteaddrlen;
#ifdef DEBUG
  /* Check server parameter. */
  if (NULL == server) {
    errno = EINVAL;
    RMIX_LOG((RMIX_WARN(server parameter is null)))
    errno = EINVAL;
    return -1;
  }
  /* Check listener parameter. */
  if (NULL == listener) {
    errno = EINVAL;
    RMIX_LOG((RMIX_WARN(listener parameter is null)))
    errno = EINVAL;
    return -1;
  }
#endif /* DEBUG */
  /* Allocate server. */
  if (NULL == (*server = (rmix_tcpip4_server_t*)
                           malloc(sizeof(rmix_tcpip4_server_t)))) {
    int errno2 = errno;
    RMIX_LOG((RMIX_WARN(unable to allocate server)))
    errno = errno2;
    return -1;
  }
  /* Push deallocate server to pthread cleanup stack. */
  pthread_cleanup_push(free, *server);
  /* Accept connection. */
  remoteaddrlen = sizeof((*server)->remoteaddr);
  do {
    (*server)->socket = accept(listener->socket,
                               (struct sockaddr *)&(*server)->remoteaddr,
                               &remoteaddrlen);
    if ((-1 == (*server)->socket)&&(EINTR == errno)) {
      pthread_testcancel();
    }
  } while ((-1 == (*server)->socket)&&(EINTR == errno));
  /* Pop deallocate server from pthread cleanup stack. */
  pthread_cleanup_pop(0);
  if (-1 == (*server)->socket) {
    int errno2 = errno;
    RMIX_LOG((RMIX_WARN(unable to accept connection)))
    /* Deallocate server. */
    free(*server);
    *server = NULL;
    errno = errno2;
    return -1;
  }
  /* Set local address. */
  (*server)->localaddr.sin_family = listener->localaddr.sin_family;
  (*server)->localaddr.sin_addr   = listener->localaddr.sin_addr;
  (*server)->localaddr.sin_port   = listener->localaddr.sin_port;
  return 0;
}

/*
 * Closes a TCP/IPv4 server previously accepted by a TCP/IPv4 listener.
 * This function may be called before initializing.
 *
 * server = The TCP/IPv4 server.
 * return = 0 on success or -1 on any error with errno set appropriately.
 */
#undef  __FUNC__
#define __FUNC__ "rmix_tcpip4_server_close"
int rmix_tcpip4_server_close (rmix_tcpip4_server_t **server) {
#ifdef DEBUG
  /* Check server parameter. */
  if (NULL == server) {
    errno = EINVAL;
    RMIX_LOG((RMIX_WARN(server parameter is null)))
    errno = EINVAL;
    return -1;
  }
  if (NULL == *server) {
    errno = EINVAL;
    RMIX_LOG((RMIX_WARN(server is null)))
    errno = EINVAL;
    return -1;
  }
#endif /* DEBUG */
  /* Destroy server socket. */
  close((*server)->socket);
  /* Deallocate server. */
  free(*server);
  *server = NULL;
  return 0;
}

/*
 * TCP/IPv4 server cleanup.
 * This function may be called before initializing.
 *
 * server = The TCP/IPv4 server.
 * return = 0 on success or -1 on any error with errno set appropriately.
 */
#undef  __FUNC__
#define __FUNC__ "rmix_tcpip4_server_close2"
void rmix_tcpip4_server_close2 (void *server) {
#ifdef DEBUG
  /* Check server parameter. */
  if (NULL == server) {
    errno = EINVAL;
    RMIX_LOG((RMIX_WARN(server parameter is null)))
    errno = EINVAL;
    return;
  }
#endif /* DEBUG */
  /* Destroy server socket. */
  close(((rmix_tcpip4_server_t*)server)->socket);
  /* Deallocate server. */
  free(server);
}

/*
 * Reads from a TCP/IPv4 server.
 * This function may be called before initializing.
 *
 * server = The TCP/IPv4 server.
 * size   = The number of bytes to read.
 * data   = The byte buffer to read into.
 * return = 0 on success or -1 on any error with errno set appropriately.
 *          ECONNABORTED = Connection closed before data transfer.
 *          EPIPE        = Connection closed during data transfer.
 */
#undef  __FUNC__
#define __FUNC__ "rmix_tcpip4_server_read"
int rmix_tcpip4_server_read (rmix_tcpip4_server_t *server,
                             unsigned int          size,
                             void                 *data) {
  int          bytes;
  unsigned int index;
#ifdef DEBUG
  /* Check server parameter. */
  if (NULL == server) {
    errno = EINVAL;
    RMIX_LOG((RMIX_WARN(server parameter is null)))
    errno = EINVAL;
    return -1;
  }
  /* Check data parameter. */
  if ((NULL == data)&&(0 != size)) {
    errno = EINVAL;
    RMIX_LOG((RMIX_WARN(data parameter is null)))
    errno = EINVAL;
    return -1;
  }
  /* Check size parameter. */
  if ((0 == size)&&(NULL != data)) {
    errno = EINVAL;
    RMIX_LOG((RMIX_WARN(size parameter is zero)))
    errno = EINVAL;
    return -1;
  }
#endif /* DEBUG */
  /* Read entire buffer to client socket. */
  for (index = 0; index < size; index += bytes) {
    /* Read entire or part of buffer to client socket. */
    switch (bytes = read(server->socket, data + index, size - index)) {
      /* Check for closed socket. */
      case 0: {
        /* Check if no data was written. */
        if (0 == index) {
          errno = ECONNABORTED;
          return -1;
        }
      }
      /* Check for error. */
      case -1: {
        /* Check for interruption. */
        if (EINTR == errno) {
          /* Reset the number of bytes written. */
          bytes = 0;
        } else {
          /* Check for connection reset and if no data was written. */
          if ((ECONNRESET == errno)&&(0 == index)) {
            errno = ECONNABORTED;
          } else {
            int errno2 = errno;
            RMIX_LOG((RMIX_WARN(unable to read from client)))
            errno = errno2;
          }
          return -1;
        }
        break;
      }
      /* Check for closed socket. */
      default: {
        break;
      }
    }
    pthread_testcancel();
  }
  return 0;
}

/*
 * Reads from a TCP/IPv4 server (stream read function).
 * This function may be called before initializing.
 *
 * incoming = The TCP/IPv4 server.
 * size     = The number of bytes to read.
 * data     = The byte buffer to read into.
 * return   = 0 on success or -1 on any error with errno set appropriately.
 *            ECONNABORTED = Connection closed before data transfer.
 *            EPIPE        = Connection closed during data transfer.
 */
#undef  __FUNC__
#define __FUNC__ "rmix_tcpip4_server_istream_read"
RMIX_STREAM_READ(rmix_tcpip4_server_istream_read) {
#ifdef DEBUG
  /* Check incoming parameter. */
  if (NULL == incoming) {
    errno = EINVAL;
    RMIX_LOG((RMIX_WARN(incoming parameter is null)))
    errno = EINVAL;
    return -1;
  }
  /* Check data parameter. */
  if ((NULL == data)&&(0 != size)) {
    errno = EINVAL;
    RMIX_LOG((RMIX_WARN(data parameter is null)))
    errno = EINVAL;
    return -1;
  }
  /* Check size parameter. */
  if ((0 == size)&&(NULL != data)) {
    errno = EINVAL;
    RMIX_LOG((RMIX_WARN(size parameter is zero)))
    errno = EINVAL;
    return -1;
  }
#endif /* DEBUG */
  /* Read from server, */
  if (0 != rmix_tcpip4_server_read((rmix_tcpip4_server_t*)incoming,
                                   size, data)) {
    /* Check for other errors than connection closed before data transfer. */
    if (ECONNABORTED != errno) {
      int errno2 = errno;
      RMIX_LOG((RMIX_WARN(unable to read from server)))
      errno = errno2;
    }
    return -1;
  }
  return 0;
}

/*
 * Writes to a TCP/IPv4 server.
 * This function may be called before initializing.
 *
 * server = The TCP/IPv4 server.
 * size   = The number of bytes to write.
 * data   = The byte buffer to write from.
 * return = 0 on success or -1 on any error with errno set appropriately.
 *          ECONNABORTED = Connection closed before data transfer.
 *          EPIPE        = Connection closed during data transfer.
 */
#undef  __FUNC__
#define __FUNC__ "rmix_tcpip4_server_write"
int rmix_tcpip4_server_write (rmix_tcpip4_server_t *server,
                              unsigned int          size,
                              const void           *data) {
  int          bytes;
  unsigned int index;
#ifdef DEBUG
  /* Check server parameter. */
  if (NULL == server) {
    errno = EINVAL;
    RMIX_LOG((RMIX_WARN(server parameter is null)))
    errno = EINVAL;
    return -1;
  }
  /* Check data parameter. */
  if ((NULL == data)&&(0 != size)) {
    errno = EINVAL;
    RMIX_LOG((RMIX_WARN(data parameter is null)))
    errno = EINVAL;
    return -1;
  }
  /* Check size parameter. */
  if ((0 == size)&&(NULL != data)) {
    errno = EINVAL;
    RMIX_LOG((RMIX_WARN(size parameter is zero)))
    errno = EINVAL;
    return -1;
  }
#endif /* DEBUG */
  /* Write entire buffer to client socket. */
  for (index = 0; index < size; index += bytes) {
    /* Write entire or part of buffer to server socket. */
    if (-1 == (bytes = write(server->socket, data + index, size - index))) {
      /* Check error. */
      switch (errno) {
        /* Check for interruption. */
        case EINTR: {
          /* Reset the number of bytes written. */
          bytes = 0;
          break;
        }
        /* Check for closed socket. */
        case EPIPE: {
          /* Check if no data was written. */
          if (0 == index) {
            errno = ECONNABORTED;
            return -1;
          }
        }
        /* Catch any other error. */
        default: {
          int errno2 = errno;
          RMIX_LOG((RMIX_WARN(unable to write to server)))
          errno = errno2;
          return -1;
        }
      }
    }
    pthread_testcancel();
  }
  return 0;
}

/*
 * Writes to a TCP/IPv4 server (stream write function).
 * This function may be called before initializing.
 *
 * outgoing = The TCP/IPv4 server.
 * size     = The number of bytes to write.
 * data     = The byte buffer to write from.
 * return   = 0 on success or -1 on any error with errno set appropriately.
 *            ECONNABORTED = Connection closed before data transfer.
 *            EPIPE        = Connection closed during data transfer.
 */
#undef  __FUNC__
#define __FUNC__ "rmix_tcpip4_server_ostream_write"
RMIX_STREAM_WRITE(rmix_tcpip4_server_ostream_write) {
#ifdef DEBUG
  /* Check outgoing parameter. */
  if (NULL == outgoing) {
    errno = EINVAL;
    RMIX_LOG((RMIX_WARN(outgoing parameter is null)))
    errno = EINVAL;
    return -1;
  }
  /* Check data parameter. */
  if ((NULL == data)&&(0 != size)) {
    errno = EINVAL;
    RMIX_LOG((RMIX_WARN(data parameter is null)))
    errno = EINVAL;
    return -1;
  }
  /* Check size parameter. */
  if ((0 == size)&&(NULL != data)) {
    errno = EINVAL;
    RMIX_LOG((RMIX_WARN(size parameter is zero)))
    errno = EINVAL;
    return -1;
  }
#endif /* DEBUG */
  /* Write to server, */
  if (0 != rmix_tcpip4_server_write((rmix_tcpip4_server_t*)outgoing,
                                    size, data)) {
    /* Check for other errors than connection closed before data transfer. */
    if (ECONNABORTED != errno) {
      int errno2 = errno;
      RMIX_LOG((RMIX_WARN(unable to write to server)))
      errno = errno2;
    }
    return -1;
  }
  return 0;
}

/*
 * Closes a TCP/IPv4 servers incoming stream.
 * This function may be called before initializing.
 *
 * stream = The TCP/IPv4 server.
 * return = 0 on success or -1 on any error with errno set appropriately.
 */
#undef  __FUNC__
#define __FUNC__ "rmix_tcpip4_server_istream_close"
RMIX_STREAM_CLOSE(rmix_tcpip4_server_istream_close) {
#ifdef DEBUG
  /* Check stream parameter. */
  if (NULL == stream) {
    errno = EINVAL;
    RMIX_LOG((RMIX_WARN(stream parameter is null)))
    errno = EINVAL;
    return -1;
  }
#endif /* DEBUG */
  /* Close the incoming server socket. */
  if (0 != shutdown(((rmix_tcpip4_server_t*)stream)->socket, SHUT_RD)) {
    int errno2 = errno;
    RMIX_LOG((RMIX_WARN(unable to close the incoming server socket)))
    errno = errno2;
    return -1;
  }
  return 0;
}

/*
 * Closes a TCP/IPv4 servers outgoing stream.
 * This function may be called before initializing.
 *
 * stream = The TCP/IPv4 server.
 * return = 0 on success or -1 on any error with errno set appropriately.
 */
#undef  __FUNC__
#define __FUNC__ "rmix_tcpip4_server_ostream_close"
RMIX_STREAM_CLOSE(rmix_tcpip4_server_ostream_close) {
#ifdef DEBUG
  /* Check stream parameter. */
  if (NULL == stream) {
    errno = EINVAL;
    RMIX_LOG((RMIX_WARN(stream parameter is null)))
    errno = EINVAL;
    return -1;
  }
#endif /* DEBUG */
  /* Close the outgoing server socket. */
  if (0 != shutdown(((rmix_tcpip4_server_t*)stream)->socket, SHUT_WR)) {
    int errno2 = errno;
    RMIX_LOG((RMIX_WARN(unable to close the outgoing server socket)))
    errno = errno2;
    return -1;
  }
  return 0;
}

/*
 * Opens a TCP/IPv4 client and connects to the addressed remote TCP/IPv4 server.
 * This function may be called before initializing.
 *
 * client     = The TCP/IPv4 client (return).
 * remoteaddr = The remote address/port to connect to.
 * return     = 0 on success or -1 on any error with errno set appropriately.
 */
#undef  __FUNC__
#define __FUNC__ "rmix_tcpip4_client_open"
int rmix_tcpip4_client_open (rmix_tcpip4_client_t     **client,
                             const struct sockaddr_in  *remoteaddr) {
  socklen_t remoteaddrlen;
#ifdef DEBUG
  /* Check client parameter. */
  if (NULL == client) {
    errno = EINVAL;
    RMIX_LOG((RMIX_WARN(client parameter is null)))
    errno = EINVAL;
    return -1;
  }
  /* Check remoteaddr parameter. */
  if (NULL == remoteaddr) {
    errno = EINVAL;
    RMIX_LOG((RMIX_WARN(remoteaddr parameter is null)))
    errno = EINVAL;
    return -1;
  }
#endif /* DEBUG */
  /* Allocate client. */
  if (NULL == (*client = (rmix_tcpip4_client_t*)
                         malloc(sizeof(rmix_tcpip4_client_t)))) {
    int errno2 = errno;
    RMIX_LOG((RMIX_WARN(unable to allocate server)))
    errno = errno2;
    return -1;
  }
  /* Create client socket. */
  if (-1 == ((*client)->socket = socket(AF_INET, SOCK_STREAM, 0))) {
    int errno2 = errno;
    RMIX_LOG((RMIX_WARN(unable to create client socket)))
    /* Deallocate client. */
    free(*client);
    *client = NULL;
    errno = errno2;
    return -1;
  }
  /* Push deallocate client to pthread cleanup stack. */
  pthread_cleanup_push(free, *client);
  /* Connect client socket. */
  while (-1 == connect((*client)->socket, (struct sockaddr *)remoteaddr,
                       sizeof(*remoteaddr))) {
    if (EINTR == errno) {
      /* Destroy client socket. */
      close((*client)->socket);
      pthread_testcancel();
      /* Create client socket. */
      if (-1 == ((*client)->socket = socket(AF_INET, SOCK_STREAM, 0))) {
        int errno2 = errno;
        RMIX_LOG((RMIX_WARN(unable to create client socket)))
        /* Deallocate client. */
        free(*client);
        *client = NULL;
        errno = errno2;
        return -1;
      }
    } else {
      int errno2 = errno;
      RMIX_LOG((RMIX_WARN(unable to connect client socket)))
      errno = errno2;
      return -1;
    }
  }
  /* Pop deallocate client from pthread cleanup stack. */
  pthread_cleanup_pop(0);
  /* Set remote address. */
  (*client)->remoteaddr.sin_family = remoteaddr->sin_family;
  (*client)->remoteaddr.sin_addr   = remoteaddr->sin_addr;
  (*client)->remoteaddr.sin_port   = remoteaddr->sin_port;
  /* Get local address. */
  remoteaddrlen = sizeof((*client)->localaddr);
  if (-1 == getsockname((*client)->socket, &(*client)->localaddr,
                        &remoteaddrlen)) {
    int errno2 = errno;
    RMIX_LOG((RMIX_WARN(unable to get local address)))
    /* Destroy client socket. */
    close((*client)->socket);
    /* Deallocate client. */
    free(*client);
    *client = NULL;
    errno = errno2;
    return -1;
  }
  return 0;
}

/*
 * Opens a TCP/IPv4 client and connects to the addressed remote TCP/IPv4 server.
 * This function may be called before initializing.
 *
 * client     = The TCP/IPv4 client (return).
 * remotename = The remote name to connect to.
 * remoteport = The remote port to connect to.
 * return     = 0 on success or -1 on any error with errno set appropriately.
 */
#undef  __FUNC__
#define __FUNC__ "rmix_tcpip4_client_open2"
int rmix_tcpip4_client_open2 (rmix_tcpip4_client_t **client,
                              const char            *remotename,
                              unsigned int           remoteport) {
  int                 error;
  struct addrinfo    *info;
  struct sockaddr_in  remoteaddr;
#ifdef DEBUG
  /* Check client parameter. */
  if (NULL == client) {
    errno = EINVAL;
    RMIX_LOG((RMIX_WARN(client parameter is null)))
    errno = EINVAL;
    return -1;
  }
  /* Check remotename parameter. */
  if (NULL == remotename) {
    errno = EINVAL;
    RMIX_LOG((RMIX_WARN(remotename parameter is null)))
    errno = EINVAL;
    return -1;
  }
#endif /* DEBUG */
  /* Lookup name. */
  if (0 != (error = getaddrinfo(remotename, NULL, NULL, &info))) {
    errno = error;
    RMIX_LOG((RMIX_WARN(unable to lookup remote name '%s'), remotename))
    errno = error;
    return -1;
  }
  remoteaddr.sin_family = AF_INET;
  remoteaddr.sin_addr   = ((struct sockaddr_in*)info->ai_addr)->sin_addr;
  remoteaddr.sin_port   = htons((unsigned short)remoteport);
  freeaddrinfo(info);
  /* Open listener. */
  if (0 != rmix_tcpip4_client_open(client, &remoteaddr)) {
    int errno2 = errno;
    RMIX_LOG((RMIX_WARN(unable to open client)))
    errno = errno2;
    return -1;
  }
  return 0;
}

/*
 * Opens a TCP/IPv4 client and connects to the addressed remote TCP/IPv4 server.
 * This function may be called before initializing.
 *
 * client    = The TCP/IPv4 client (return).
 * remoteurl = The remote URL to connect to.
 * return    = 0 on success or -1 on any error with errno set appropriately.
 */
#undef  __FUNC__
#define __FUNC__ "rmix_tcpip4_client_open3"
int rmix_tcpip4_client_open3 (rmix_tcpip4_client_t **client,
                              const char            *remoteurl) {
  char *remoteurl2, *remotename, *rest, *remoteport;
#ifdef DEBUG
  /* Check client parameter. */
  if (NULL == client) {
    errno = EINVAL;
    RMIX_LOG((RMIX_WARN(client parameter is null)))
    errno = EINVAL;
    return -1;
  }
  /* Check remoteurl parameter. */
  if (NULL == remoteurl) {
    errno = EINVAL;
    RMIX_LOG((RMIX_WARN(remoteurl parameter is null)))
    errno = EINVAL;
    return -1;
  }
#endif /* DEBUG */
  /* Duplicate remote URL. */
  if (NULL == (remoteurl2 = strdup(remoteurl))) {
    int errno2 = errno;
    RMIX_LOG((RMIX_WARN(unable to duplicate remote URL)))
    errno = errno2;
    return -1;
  }
  if (NULL != (remotename = strstr(remoteurl2, "://"))) {
    remotename += 3;
  } else {
    remotename = remoteurl2;
  }
  if (NULL != (rest = strstr(remotename, "/"))) {
    rest[0] = '\0';
  }
  if (NULL != (remoteport = strstr(remotename, ":"))) {
    remoteport[0] = '\0';
    remoteport   += 1;
  } else {
    remoteport = "0";
  }
  /* Open listener. */
  if (0 != rmix_tcpip4_client_open2(client, remotename, atoi(remoteport))) {
    int errno2 = errno;
    RMIX_LOG((RMIX_WARN(unable to open client)))
    /* Deallocate remote URL duplicate. */
    free(remoteurl2);
    errno = errno2;
    return -1;
  }
  /* Deallocate remote URL duplicate. */
  free(remoteurl2);
  return 0;
}

/*
 * Disconnects and closes a TCP/IPv4 client.
 * This function may be called before initializing.
 *
 * client = The TCP/IPv4 client.
 * return = 0 on success or -1 on any error with errno set appropriately.
 */
#undef  __FUNC__
#define __FUNC__ "rmix_tcpip4_client_close"
int rmix_tcpip4_client_close (rmix_tcpip4_client_t **client) {
#ifdef DEBUG
  /* Check client parameter. */
  if (NULL == client) {
    errno = EINVAL;
    RMIX_LOG((RMIX_WARN(client parameter is null)))
    errno = EINVAL;
    return -1;
  }
  if (NULL == *client) {
    errno = EINVAL;
    RMIX_LOG((RMIX_WARN(client is null)))
    errno = EINVAL;
    return -1;
  }
#endif /* DEBUG */
  /* Destroy client socket. */
  close((*client)->socket);
  /* Deallocate client. */
  free(*client);
  *client = NULL;
  return 0;
}

/*
 * TCP/IPv4 client cleanup.
 * This function may be called before initializing.
 *
 * client = The TCP/IPv4 client.
 * return = 0 on success or -1 on any error with errno set appropriately.
 */
#undef  __FUNC__
#define __FUNC__ "rmix_tcpip4_client_close2"
void rmix_tcpip4_client_close2 (void *client) {
#ifdef DEBUG
  /* Check client parameter. */
  if (NULL == client) {
    errno = EINVAL;
    RMIX_LOG((RMIX_WARN(client parameter is null)))
    errno = EINVAL;
    return;
  }
#endif /* DEBUG */
  /* Destroy client socket. */
  close(((rmix_tcpip4_client_t*)client)->socket);
  /* Deallocate client. */
  free(client);
}

/*
 * Reads from a TCP/IPv4 client.
 * This function may be called before initializing.
 *
 * client = The TCP/IPv4 client.
 * size   = The number of bytes to read.
 * data   = The byte buffer to read into.
 * return = 0 on success or -1 on any error with errno set appropriately.
 *          ECONNABORTED = Connection closed before data transfer.
 *          EPIPE        = Connection closed during data transfer.
 */
#undef  __FUNC__
#define __FUNC__ "rmix_tcpip4_client_read"
int rmix_tcpip4_client_read (rmix_tcpip4_client_t *client,
                             unsigned int          size,
                             void                 *data) {
  int          bytes;
  unsigned int index;
#ifdef DEBUG
  /* Check client parameter. */
  if (NULL == client) {
    errno = EINVAL;
    RMIX_LOG((RMIX_WARN(client parameter is null)))
    errno = EINVAL;
    return -1;
  }
  /* Check data parameter. */
  if ((NULL == data)&&(0 != size)) {
    errno = EINVAL;
    RMIX_LOG((RMIX_WARN(data parameter is null)))
    errno = EINVAL;
    return -1;
  }
  /* Check size parameter. */
  if ((0 == size)&&(NULL != data)) {
    errno = EINVAL;
    RMIX_LOG((RMIX_WARN(size parameter is zero)))
    errno = EINVAL;
    return -1;
  }
#endif /* DEBUG */
  /* Read entire buffer to client socket. */
  for (index = 0; index < size; index += bytes) {
    /* Read entire or part of buffer to client socket. */
    switch (bytes = read(client->socket, data + index, size - index)) {
      /* Check for closed socket. */
      case 0: {
        /* Check if no data was written. */
        if (0 == index) {
          errno = ECONNABORTED;
          return -1;
        }
      }
      /* Check for error. */
      case -1: {
        /* Check for interruption. */
        if (EINTR == errno) {
          /* Reset the number of bytes written. */
          bytes = 0;
        } else {
          /* Check for connection reset and if no data was written. */
          if ((ECONNRESET == errno)&&(0 == index)) {
            errno = ECONNABORTED;
          } else {
            int errno2 = errno;
            RMIX_LOG((RMIX_WARN(unable to read from client)))
            errno = errno2;
          }
          return -1;
        }
        break;
      }
      /* Check for closed socket. */
      default: {
        break;
      }
    }
    pthread_testcancel();
  }
  return 0;
}

/*
 * Reads from a TCP/IPv4 client (stream read function).
 * This function may be called before initializing.
 *
 * incoming = The TCP/IPv4 client.
 * size     = The number of bytes to read.
 * data     = The byte buffer to read into.
 * return   = 0 on success or -1 on any error with errno set appropriately.
 *            ECONNABORTED = Connection closed before data transfer.
 *            EPIPE        = Connection closed during data transfer.
 */
#undef  __FUNC__
#define __FUNC__ "rmix_tcpip4_client_istream_read"
RMIX_STREAM_READ(rmix_tcpip4_client_istream_read) {
#ifdef DEBUG
  /* Check incoming parameter. */
  if (NULL == incoming) {
    errno = EINVAL;
    RMIX_LOG((RMIX_WARN(incoming parameter is null)))
    errno = EINVAL;
    return -1;
  }
  /* Check data parameter. */
  if ((NULL == data)&&(0 != size)) {
    errno = EINVAL;
    RMIX_LOG((RMIX_WARN(data parameter is null)))
    errno = EINVAL;
    return -1;
  }
  /* Check size parameter. */
  if ((0 == size)&&(NULL != data)) {
    errno = EINVAL;
    RMIX_LOG((RMIX_WARN(size parameter is zero)))
    errno = EINVAL;
    return -1;
  }
#endif /* DEBUG */
  /* Read from client, */
  if (0 != rmix_tcpip4_client_read((rmix_tcpip4_client_t*)incoming,
                                    size, data)) {
    /* Check for other errors than connection closed before data transfer. */
    if (ECONNABORTED != errno) {
      int errno2 = errno;
      RMIX_LOG((RMIX_WARN(unable to read from client)))
      errno = errno2;
    }
    return -1;
  }
  return 0;
}

/*
 * Writes to a TCP/IPv4 client.
 * This function may be called before initializing.
 *
 * client = The TCP/IPv4 client.
 * size   = The number of bytes to write.
 * data   = The byte buffer to write from.
 * return = 0 on success or -1 on any error with errno set appropriately.
 *          ECONNABORTED = Connection closed before data transfer.
 *          EPIPE        = Connection closed during data transfer.
 */
#undef  __FUNC__
#define __FUNC__ "rmix_tcpip4_client_write"
int rmix_tcpip4_client_write (rmix_tcpip4_client_t *client,
                              unsigned int          size,
                              const void           *data) {
  int          bytes;
  unsigned int index;
#ifdef DEBUG
  /* Check client parameter. */
  if (NULL == client) {
    errno = EINVAL;
    RMIX_LOG((RMIX_WARN(client parameter is null)))
    errno = EINVAL;
    return -1;
  }
  /* Check data parameter. */
  if ((NULL == data)&&(0 != size)) {
    errno = EINVAL;
    RMIX_LOG((RMIX_WARN(data parameter is null)))
    errno = EINVAL;
    return -1;
  }
  /* Check size parameter. */
  if ((0 == size)&&(NULL != data)) {
    errno = EINVAL;
    RMIX_LOG((RMIX_WARN(size parameter is zero)))
    errno = EINVAL;
    return -1;
  }
#endif /* DEBUG */
  /* Write entire buffer to client socket. */
  for (index = 0; index < size; index += bytes) {
    /* Write entire or part of buffer to client socket. */
    if (-1 == (bytes = write(client->socket, data + index, size - index))) {
      /* Check error. */
      switch (errno) {
        /* Check for interruption. */
        case EINTR: {
          /* Reset the number of bytes written. */
          bytes = 0;
          break;
        }
        /* Check for closed socket. */
        case EPIPE: {
          /* Check if no data was written. */
          if (0 == index) {
            errno = ECONNABORTED;
            return -1;
          }
        }
        /* Catch any other error. */
        default: {
          int errno2 = errno;
          RMIX_LOG((RMIX_WARN(unable to write to client)))
          errno = errno2;
          return -1;
        }
      }
    }
    pthread_testcancel();
  }
  return 0;
}

/*
 * Writes to a TCP/IPv4 client (stream write function).
 * This function may be called before initializing.
 *
 * outgoing = The TCP/IPv4 client.
 * size     = The number of bytes to write.
 * data     = The byte buffer to write from.
 * return   = 0 on success or -1 on any error with errno set appropriately.
 *            ECONNABORTED = Connection closed before data transfer.
 *            EPIPE        = Connection closed during data transfer.
 */
#undef  __FUNC__
#define __FUNC__ "rmix_tcpip4_client_ostream_write"
RMIX_STREAM_WRITE(rmix_tcpip4_client_ostream_write) {
#ifdef DEBUG
  /* Check outgoing parameter. */
  if (NULL == outgoing) {
    errno = EINVAL;
    RMIX_LOG((RMIX_WARN(outgoing parameter is null)))
    errno = EINVAL;
    return -1;
  }
  /* Check data parameter. */
  if ((NULL == data)&&(0 != size)) {
    errno = EINVAL;
    RMIX_LOG((RMIX_WARN(data parameter is null)))
    errno = EINVAL;
    return -1;
  }
  /* Check size parameter. */
  if ((0 == size)&&(NULL != data)) {
    errno = EINVAL;
    RMIX_LOG((RMIX_WARN(size parameter is zero)))
    errno = EINVAL;
    return -1;
  }
#endif /* DEBUG */
  /* Write to client, */
  if (0 != rmix_tcpip4_client_write((rmix_tcpip4_client_t*)outgoing,
                                    size, data)) {
    /* Check for other errors than connection closed before data transfer. */
    if (ECONNABORTED != errno) {
      int errno2 = errno;
      RMIX_LOG((RMIX_WARN(unable to write to client)))
      errno = errno2;
    }
    return -1;
  }
  return 0;
}

/*
 * Closes a TCP/IPv4 clients incoming stream.
 * This function may be called before initializing.
 *
 * stream = The TCP/IPv4 client.
 * return = 0 on success or -1 on any error with errno set appropriately.
 */
#undef  __FUNC__
#define __FUNC__ "rmix_tcpip4_client_istream_close"
RMIX_STREAM_CLOSE(rmix_tcpip4_client_istream_close) {
#ifdef DEBUG
  /* Check stream parameter. */
  if (NULL == stream) {
    errno = EINVAL;
    RMIX_LOG((RMIX_WARN(stream parameter is null)))
    errno = EINVAL;
    return -1;
  }
#endif /* DEBUG */
  /* Close the incoming client socket. */
  if (0 != shutdown(((rmix_tcpip4_client_t*)stream)->socket, SHUT_RD)) {
    int errno2 = errno;
    RMIX_LOG((RMIX_WARN(unable to close the incoming client socket)))
    errno = errno2;
    return -1;
  }
  return 0;
}

/*
 * Closes a TCP/IPv4 clients outgoing stream.
 * This function may be called before initializing.
 *
 * stream = The TCP/IPv4 client.
 * return = 0 on success or -1 on any error with errno set appropriately.
 */
#undef  __FUNC__
#define __FUNC__ "rmix_tcpip4_client_ostream_close"
RMIX_STREAM_CLOSE(rmix_tcpip4_client_ostream_close) {
#ifdef DEBUG
  /* Check stream parameter. */
  if (NULL == stream) {
    errno = EINVAL;
    RMIX_LOG((RMIX_WARN(stream parameter is null)))
    errno = EINVAL;
    return -1;
  }
#endif /* DEBUG */
  /* Close the outgoing client socket. */
  if (0 != shutdown(((rmix_tcpip4_client_t*)stream)->socket, SHUT_WR)) {
    int errno2 = errno;
    RMIX_LOG((RMIX_WARN(unable to close the outgoing client socket)))
    errno = errno2;
    return -1;
  }
  return 0;
}

/*
 * Initializes the TCP/IPv4 pool.
 *
 * return = 0 on success or -1 on any error with errno set appropriately.
 */
#undef  __FUNC__
#define __FUNC__ "rmix_tcpip4_init"
int rmix_tcpip4_init () {
  int error;
  /* Lock TCP/IPv4 pool mutex. */
  if (0 != (error = pthread_mutex_lock(&rmix_tcpip4_data.mutex))) {
    errno = error;
    RMIX_LOG((RMIX_WARN(unable to lock TCP/IPv4 pool mutex)))
    errno = error;
    rmix_syserr();
    errno = error;
    return -1;
  }
#ifdef DEBUG
  /* Check for initialized TCP/IPv4 pool. */
  if (1 == rmix_tcpip4_data.ready) {
    errno = EACCES;
    RMIX_LOG((RMIX_WARN(TCP/IPv4 pool already initialized)))
    /* Unlock TCP/IPv4 pool mutex. */
    if (0 != (error = pthread_mutex_unlock(&rmix_tcpip4_data.mutex))) {
      errno = error;
      RMIX_LOG((RMIX_WARN(unable to unlock TCP/IPv4 pool mutex)))
      errno = error;
      rmix_syserr();
    }
    errno = EACCES;
    return -1;
  }
#endif /* DEBUG */
  /* Initialize wake-up pipe. */
  if (0 != pipe(rmix_tcpip4_data.wakeup)) {
    int errno2 = errno;
    RMIX_LOG((RMIX_WARN(unable to initialize wake-up pipe)))
    errno = errno2;
    rmix_syserr();
    /* Unlock TCP/IPv4 pool mutex. */
    if (0 != (error = pthread_mutex_unlock(&rmix_tcpip4_data.mutex))) {
      errno = error;
      RMIX_LOG((RMIX_WARN(unable to unlock TCP/IPv4 pool mutex)))
      errno = error;
      rmix_syserr();
    }
    errno = errno2;
    return -1;
  }
  /* Initialize listener list. */
  rmix_tcpip4_data.count     = 0;
  rmix_tcpip4_data.listeners = NULL;
  /* Increase the maximum thread count. */
  if (0 != rmix_threads_pool_maximum_add(1)) {
    int errno2 = errno;
    RMIX_LOG((RMIX_WARN(unable to increase maximum thread count)))
    /* Unlock TCP/IPv4 pool mutex. */
    if (0 != (error = pthread_mutex_unlock(&rmix_tcpip4_data.mutex))) {
      errno = error;
      RMIX_LOG((RMIX_WARN(unable to unlock TCP/IPv4 pool mutex)))
      errno = error;
      rmix_syserr();
    }
    errno = errno2;
    return -1;
  }
  /* Submit job to serve incoming connections. */
  if (0 != rmix_threads_submit(rmix_tcpip4_job, NULL, NULL)) {
    int errno2 = errno;
    RMIX_LOG((RMIX_WARN(unable to submit job to serve incoming connections)))
    /* Unlock TCP/IPv4 pool mutex. */
    if (0 != (error = pthread_mutex_unlock(&rmix_tcpip4_data.mutex))) {
      errno = error;
      RMIX_LOG((RMIX_WARN(unable to unlock TCP/IPv4 pool mutex)))
      errno = error;
      rmix_syserr();
    }
    errno = errno2;
    return -1;
  }
  /* Mark TCP/IPv4 pool initialized. */
  rmix_tcpip4_data.ready = 1; 
  /* Unlock TCP/IPv4 pool mutex. */
  if (0 != (error = pthread_mutex_unlock(&rmix_tcpip4_data.mutex))) {
    errno = error;
    RMIX_LOG((RMIX_WARN(unable to unlock TCP/IPv4 pool mutex)))
    errno = error;
    rmix_syserr();
    /* Mark TCP/IPv4 pool finalized. */
    rmix_tcpip4_data.ready = 0;
    errno = error;
    return -1;
  }
  return 0;
}

/*
 * Finalizes the TCP/IPv4 pool.
 *
 * return = 0 on success or -1 on any error with errno set appropriately.
 */
#undef  __FUNC__
#define __FUNC__ "rmix_tcpip4_fini"
int rmix_tcpip4_fini () {
  int          error, result, errno2;
  unsigned int index1, index2;
  result = 0;
  errno2 = 0;
  /* Lock TCP/IPv4 pool mutex. */
  if (0 != (error = pthread_mutex_lock(&rmix_tcpip4_data.mutex))) {
    errno = error;
    RMIX_LOG((RMIX_WARN(unable to lock TCP/IPv4 pool mutex)))
    errno = error;
    rmix_syserr();
    errno = error;
    return -1;
  }
#ifdef DEBUG
  /* Check for finalized TCP/IPv4 pool. */
  if (0 == rmix_tcpip4_data.ready) {
    errno = EACCES;
    RMIX_LOG((RMIX_WARN(TCP/IPv4 pool already finalized)))
    /* Unlock TCP/IPv4 pool mutex. */
    if (0 != (error = pthread_mutex_unlock(&rmix_tcpip4_data.mutex))) {
      errno = error;
      RMIX_LOG((RMIX_WARN(unable to unlock TCP/IPv4 pool mutex)))
      errno = error;
      rmix_syserr();
    }
    errno = EACCES;
    return -1;
  }
#endif /* DEBUG */
  /* Finalize wake-up pipe. */
  if ((0 != close(rmix_tcpip4_data.wakeup[0]))||
      (0 != close(rmix_tcpip4_data.wakeup[1]))) {
    errno2 = errno;
    RMIX_LOG((RMIX_WARN(unable to finalize wake-up pipe)))
    errno = errno2;
    rmix_syserr();
    result = -1;
  }
  /* Close TCP/IPv4 pool listeners and servers. */
  for (index1 = 0; index1 < rmix_tcpip4_data.count; index1++) {
    /* Close TCP/IPv4 pool listener. */
    if (0 != rmix_tcpip4_listener_close(
               &rmix_tcpip4_data.listeners[index1].listener)) {
      errno2 = errno;
      RMIX_LOG((RMIX_WARN(unable to close TCP/IPv4 listener)))
      result = -1;
    }
    /* Close related TCP/IPv4 pool servers. */
    for (index2 = 0;
         index2 < rmix_tcpip4_data.listeners[index1].count; index2++) {
      /* Close TCP/IPv4 server. */
      if (0 != rmix_tcpip4_server_close(
                 &rmix_tcpip4_data.listeners[index1].servers[index2]->server)) {
        errno2 = errno;
        RMIX_LOG((RMIX_WARN(unable to close TCP/IPv4 pool server)))
        result = -1;
      }
      /* Deallocate TCP/IPv4 pool server. */
      free(rmix_tcpip4_data.listeners[index1].servers[index2]);
    }
    /* Deallocate related TCP/IPv4 pool server list. */
    if (NULL != rmix_tcpip4_data.listeners[index1].servers) {
      free(rmix_tcpip4_data.listeners[index1].servers);
      rmix_tcpip4_data.listeners[index1].servers = NULL;
      rmix_tcpip4_data.listeners[index1].count   = 0;
    }
  }
  /* Deallocate listener list. */
  if (NULL != rmix_tcpip4_data.listeners) {
    free(rmix_tcpip4_data.listeners);
    rmix_tcpip4_data.listeners = NULL;
    rmix_tcpip4_data.count     = 0;
  }
  /* Mark TCP/IPv4 pool finalized. */
  rmix_tcpip4_data.ready = 0; 
  /* Unlock TCP/IPv4 pool mutex. */
  if (0 != (error = pthread_mutex_unlock(&rmix_tcpip4_data.mutex))) {
    errno = error;
    RMIX_LOG((RMIX_WARN(unable to unlock TCP/IPv4 pool mutex)))
    errno2 = errno;
    rmix_syserr();
    errno2 = errno;
    return -1;
  }
  if (0 != errno2) {
    errno = errno2;
  }
  return result;
}

/*
 * Opens a TCP/IPv4 server and binds it to all local interfaces.
 *
 * handle    = The TCP/IPv4 server handle (return).
 * port      = The port to bind to.
 * backlog   = The listen backlog.
 * serverjob = The server job.
 * return    = 0 on success or -1 on any error with errno set appropriately.
 */
#undef  __FUNC__
#define __FUNC__ "rmix_tcpip4_open"
int rmix_tcpip4_open (unsigned int            *handle,
                      unsigned int             localport,
                      unsigned int             backlog,
                      rmix_tcpip4_serverjob_t  serverjob) {
  int                     error;
  void                   *listeners;
  unsigned int            index1, index2;
  rmix_tcpip4_listener_t *listener;
#ifdef DEBUG
  /* Check handle parameter. */
  if (NULL == handle) {
    errno = EINVAL;
    RMIX_LOG((RMIX_WARN(handle parameter is null)))
    errno = EINVAL;
    return -1;
  }
  /* Check serverjob parameter. */
  if (NULL == serverjob) {
    errno = EINVAL;
    RMIX_LOG((RMIX_WARN(serverjob parameter is null)))
    errno = EINVAL;
    return -1;
  }
#endif /* DEBUG */
  /* Open TCP/IPv4 listener. */
  if (0 != rmix_tcpip4_listener_open(&listener, localport, backlog)) {
    int errno2 = errno;
    RMIX_LOG((RMIX_WARN(unable to open TCP/IPv4 listener)))
    errno = errno2;
    return -1;
  }
  /* Lock TCP/IPv4 pool mutex. */
  if (0 != (error = pthread_mutex_lock(&rmix_tcpip4_data.mutex))) {
    errno = error;
    RMIX_LOG((RMIX_WARN(unable to lock TCP/IPv4 pool mutex)))
    errno = error;
    rmix_syserr();
    /* Close TCP/IPv4 listener. */
    if (0 != rmix_tcpip4_listener_close(&listener)) {
      RMIX_LOG((RMIX_WARN(unable to close TCP/IPv4 listener)))
    }
    errno = error;
    return -1;
  }
#ifdef DEBUG
  /* Check for finalized TCP/IPv4 pool. */
  if (0 == rmix_tcpip4_data.ready) {
    errno = EACCES;
    RMIX_LOG((RMIX_WARN(TCP/IPv4 pool finalized)))
    /* Unlock TCP/IPv4 pool mutex. */
    if (0 != (error = pthread_mutex_unlock(&rmix_tcpip4_data.mutex))) {
      errno = error;
      RMIX_LOG((RMIX_WARN(unable to unlock TCP/IPv4 pool mutex)))
      errno = error;
      rmix_syserr();
    }
    /* Close TCP/IPv4 listener. */
    if (0 != rmix_tcpip4_listener_close(&listener)) {
      RMIX_LOG((RMIX_WARN(unable to close TCP/IPv4 listener)))
    }
    errno = EACCES;
    return -1;
  }
#endif /* DEBUG */
  /* Increase TCP/IPv4 pool listener list. */
  index1 = rmix_tcpip4_data.count;
  rmix_tcpip4_data.count++;
  /* Rellocate TCP/IPv4 pool listener list. */
  if (NULL == (listeners = realloc(rmix_tcpip4_data.listeners,
                                   rmix_tcpip4_data.count * 
                            sizeof(rmix_tcpip4_data.listeners[0])))) {
    int errno2 = errno;
    RMIX_LOG((RMIX_WARN(unable to reallocate TCP/IPv4 pool listener list)))
    /* Decrease TCP/IPv4 pool listener list. */
    rmix_tcpip4_data.count = index1;
    /* Unlock TCP/IPv4 pool mutex. */
    if (0 != (error = pthread_mutex_unlock(&rmix_tcpip4_data.mutex))) {
      errno = error;
      RMIX_LOG((RMIX_WARN(unable to unlock TCP/IPv4 pool mutex)))
      errno = error;
      rmix_syserr();
    }
    /* Close TCP/IPv4 listener. */
    if (0 != rmix_tcpip4_listener_close(&listener)) {
      RMIX_LOG((RMIX_WARN(unable to close TCP/IPv4 listener)))
    }
    errno = errno2;
    return -1;
  }
  rmix_tcpip4_data.listeners = listeners;
  /* Set up TCP/IPv4 pool listener list entry. */
  rmix_tcpip4_data.listeners[index1].listener  = listener;
  rmix_tcpip4_data.listeners[index1].handle    = 0;
  rmix_tcpip4_data.listeners[index1].close     = 0;
  rmix_tcpip4_data.listeners[index1].serverjob = serverjob;
  rmix_tcpip4_data.listeners[index1].count     = 0;
  rmix_tcpip4_data.listeners[index1].servers   = NULL;
  /* Randomly generate a unique non-zero 32-bit handle. */
  do {
    *handle = random() & 0xFFFFFFFF;
    for (index2 = 0; index2 < rmix_tcpip4_data.count; index2++) {
      if (*handle == rmix_tcpip4_data.listeners[index2].handle) {
        break;
      }
    }
  } while (index2 < rmix_tcpip4_data.count);
  rmix_tcpip4_data.listeners[index1].handle = *handle;
  /* Send byte to wake-up pipe. */
  write(rmix_tcpip4_data.wakeup[1], "", 1);
  /* Unlock TCP/IPv4 pool mutex. */
  if (0 != (error = pthread_mutex_unlock(&rmix_tcpip4_data.mutex))) {
    errno = error;
    RMIX_LOG((RMIX_WARN(unable to unlock TCP/IPv4 pool mutex)))
    errno = error;
    rmix_syserr();
    errno = error;
    return -1;
  }
  return 0;
}

/*
 * Opens a TCP/IPv4 server and binds it to the addressed local interface.
 *
 * handle    = The TCP/IPv4 server handle (return).
 * localaddr = The local address/port to bind to.
 * backlog   = The listen backlog.
 * serverjob = The server job.
 * return    = 0 on success or -1 on any error with errno set appropriately.
 */
#undef  __FUNC__
#define __FUNC__ "rmix_tcpip4_open2"
int rmix_tcpip4_open2 (unsigned int             *handle,
                       const struct sockaddr_in *localaddr,
                       unsigned int              backlog,
                       rmix_tcpip4_serverjob_t   serverjob) {
  int                     error;
  void                   *listeners;
  unsigned int            index1, index2;
  rmix_tcpip4_listener_t *listener;
#ifdef DEBUG
  /* Check handle parameter. */
  if (NULL == handle) {
    errno = EINVAL;
    RMIX_LOG((RMIX_WARN(handle parameter is null)))
    errno = EINVAL;
    return -1;
  }
  /* Check localaddr parameter. */
  if (NULL == localaddr) {
    errno = EINVAL;
    RMIX_LOG((RMIX_WARN(localaddr parameter is null)))
    errno = EINVAL;
    return -1;
  }
  /* Check serverjob parameter. */
  if (NULL == serverjob) {
    errno = EINVAL;
    RMIX_LOG((RMIX_WARN(serverjob parameter is null)))
    errno = EINVAL;
    return -1;
  }
#endif /* DEBUG */
  /* Open TCP/IPv4 listener. */
  if (0 != rmix_tcpip4_listener_open2(&listener, localaddr, backlog)) {
    int errno2 = errno;
    RMIX_LOG((RMIX_WARN(unable to open TCP/IPv4 listener)))
    errno = errno2;
    return -1;
  }
  /* Lock TCP/IPv4 pool mutex. */
  if (0 != (error = pthread_mutex_lock(&rmix_tcpip4_data.mutex))) {
    errno = error;
    RMIX_LOG((RMIX_WARN(unable to lock TCP/IPv4 pool mutex)))
    errno = error;
    rmix_syserr();
    /* Close TCP/IPv4 listener. */
    if (0 != rmix_tcpip4_listener_close(&listener)) {
      RMIX_LOG((RMIX_WARN(unable to close TCP/IPv4 listener)))
    }
    return -1;
  }
#ifdef DEBUG
  /* Check for finalized TCP/IPv4 pool. */
  if (0 == rmix_tcpip4_data.ready) {
    errno = EACCES;
    RMIX_LOG((RMIX_WARN(TCP/IPv4 pool finalized)))
    /* Unlock TCP/IPv4 pool mutex. */
    if (0 != (error = pthread_mutex_unlock(&rmix_tcpip4_data.mutex))) {
      errno = error;
      RMIX_LOG((RMIX_WARN(unable to unlock TCP/IPv4 pool mutex)))
      errno = error;
      rmix_syserr();
    }
    /* Close TCP/IPv4 listener. */
    if (0 != rmix_tcpip4_listener_close(&listener)) {
      RMIX_LOG((RMIX_WARN(unable to close TCP/IPv4 listener)))
    }
    errno = EACCES;
    return -1;
  }
#endif /* DEBUG */
  /* Increase TCP/IPv4 pool listener list. */
  index1 = rmix_tcpip4_data.count;
  rmix_tcpip4_data.count++;
  /* Rellocate TCP/IPv4 pool listener list. */
  if (NULL == (listeners = realloc(rmix_tcpip4_data.listeners,
                                   rmix_tcpip4_data.count * 
                            sizeof(rmix_tcpip4_data.listeners[0])))) {
    int errno2 = errno;
    RMIX_LOG((RMIX_WARN(unable to reallocate TCP/IPv4 pool listener list)))
    /* Decrease TCP/IPv4 pool listener list. */
    rmix_tcpip4_data.count = index1;
    /* Unlock TCP/IPv4 pool mutex. */
    if (0 != (error = pthread_mutex_unlock(&rmix_tcpip4_data.mutex))) {
      errno = error;
      RMIX_LOG((RMIX_WARN(unable to unlock TCP/IPv4 pool mutex)))
      errno = error;
      rmix_syserr();
    }
    /* Close TCP/IPv4 listener. */
    if (0 != rmix_tcpip4_listener_close(&listener)) {
      RMIX_LOG((RMIX_WARN(unable to close TCP/IPv4 listener)))
    }
    errno = errno2;
    return -1;
  }
  rmix_tcpip4_data.listeners = listeners;
  /* Set up TCP/IPv4 pool listener list entry. */
  rmix_tcpip4_data.listeners[index1].listener  = listener;
  rmix_tcpip4_data.listeners[index1].handle    = 0;
  rmix_tcpip4_data.listeners[index1].close     = 0;
  rmix_tcpip4_data.listeners[index1].serverjob = serverjob;
  rmix_tcpip4_data.listeners[index1].count     = 0;
  rmix_tcpip4_data.listeners[index1].servers   = NULL;
  /* Randomly generate a unique non-zero 32-bit handle. */
  do {
    *handle = random() & 0xFFFFFFFF;
    for (index2 = 0; index2 < rmix_tcpip4_data.count; index2++) {
      if (*handle == rmix_tcpip4_data.listeners[index2].handle) {
        break;
      }
    }
  } while (index2 < rmix_tcpip4_data.count);
  rmix_tcpip4_data.listeners[index1].handle = *handle;
  /* Send byte to wake-up pipe. */
  write(rmix_tcpip4_data.wakeup[1], "", 1);
  /* Unlock TCP/IPv4 pool mutex. */
  if (0 != (error = pthread_mutex_unlock(&rmix_tcpip4_data.mutex))) {
    errno = error;
    RMIX_LOG((RMIX_WARN(unable to unlock TCP/IPv4 pool mutex)))
    errno = error;
    rmix_syserr();
    errno = error;
    return -1;
  }
  return 0;
}

/*
 * Opens a TCP/IPv4 server and binds it to the addressed local interface.
 *
 * handle    = The TCP/IPv4 server handle (return).
 * localname = The local name to bind to.
 * localport = The local port to bind to.
 * backlog   = The listen backlog.
 * serverjob = The server job.
 * return    = 0 on success or -1 on any error with errno set appropriately.
 */
#undef  __FUNC__
#define __FUNC__ "rmix_tcpip4_open3"
int rmix_tcpip4_open3 (unsigned int            *handle,
                       char                    *localname,
                       unsigned int             localport,
                       unsigned int             backlog,
                       rmix_tcpip4_serverjob_t  serverjob) {
  int                     error;
  void                   *listeners;
  unsigned int            index1, index2;
  rmix_tcpip4_listener_t *listener;
#ifdef DEBUG
  /* Check handle parameter. */
  if (NULL == handle) {
    errno = EINVAL;
    RMIX_LOG((RMIX_WARN(handle parameter is null)))
    errno = EINVAL;
    return -1;
  }
  /* Check localname parameter. */
  if (NULL == localname) {
    errno = EINVAL;
    RMIX_LOG((RMIX_WARN(localname parameter is null)))
    errno = EINVAL;
    return -1;
  }
  /* Check serverjob parameter. */
  if (NULL == serverjob) {
    errno = EINVAL;
    RMIX_LOG((RMIX_WARN(serverjob parameter is null)))
    errno = EINVAL;
    return -1;
  }
#endif /* DEBUG */
  /* Open TCP/IPv4 listener. */
  if (0 != rmix_tcpip4_listener_open3(&listener, localname,
                                      localport, backlog)) {
    int errno2 = errno;
    RMIX_LOG((RMIX_WARN(unable to open TCP/IPv4 listener)))
    errno = errno2;
    return -1;
  }
  /* Lock TCP/IPv4 pool mutex. */
  if (0 != (error = pthread_mutex_lock(&rmix_tcpip4_data.mutex))) {
    errno = error;
    RMIX_LOG((RMIX_WARN(unable to lock TCP/IPv4 pool mutex)))
    errno = error;
    rmix_syserr();
    /* Close TCP/IPv4 listener. */
    if (0 != rmix_tcpip4_listener_close(&listener)) {
      RMIX_LOG((RMIX_WARN(unable to close TCP/IPv4 listener)))
    }
    errno = error;
    return -1;
  }
#ifdef DEBUG
  /* Check for finalized TCP/IPv4 pool. */
  if (0 == rmix_tcpip4_data.ready) {
    errno = EACCES;
    RMIX_LOG((RMIX_WARN(TCP/IPv4 pool finalized)))
    /* Unlock TCP/IPv4 pool mutex. */
    if (0 != (error = pthread_mutex_unlock(&rmix_tcpip4_data.mutex))) {
      errno = error;
      RMIX_LOG((RMIX_WARN(unable to unlock TCP/IPv4 pool mutex)))
      errno = error;
      rmix_syserr();
    }
    /* Close TCP/IPv4 listener. */
    if (0 != rmix_tcpip4_listener_close(&listener)) {
      RMIX_LOG((RMIX_WARN(unable to close TCP/IPv4 listener)))
    }
    errno = EACCES;
    return -1;
  }
#endif /* DEBUG */
  /* Increase TCP/IPv4 pool listener list. */
  index1 = rmix_tcpip4_data.count;
  rmix_tcpip4_data.count++;
  /* Rellocate TCP/IPv4 pool listener list. */
  if (NULL == (listeners = realloc(rmix_tcpip4_data.listeners,
                                   rmix_tcpip4_data.count * 
                            sizeof(rmix_tcpip4_data.listeners[0])))) {
    int errno2 = errno;
    RMIX_LOG((RMIX_WARN(unable to reallocate TCP/IPv4 pool listener list)))
    /* Decrease TCP/IPv4 pool listener list. */
    rmix_tcpip4_data.count = index1;
    /* Unlock TCP/IPv4 pool mutex. */
    if (0 != (error = pthread_mutex_unlock(&rmix_tcpip4_data.mutex))) {
      errno = error;
      RMIX_LOG((RMIX_WARN(unable to unlock TCP/IPv4 pool mutex)))
      errno = error;
      rmix_syserr();
    }
    /* Close TCP/IPv4 listener. */
    if (0 != rmix_tcpip4_listener_close(&listener)) {
      RMIX_LOG((RMIX_WARN(unable to close TCP/IPv4 listener)))
    }
    errno = errno2;
    return -1;
  }
  rmix_tcpip4_data.listeners = listeners;
  /* Set up TCP/IPv4 pool listener list entry. */
  rmix_tcpip4_data.listeners[index1].listener  = listener;
  rmix_tcpip4_data.listeners[index1].handle    = 0;
  rmix_tcpip4_data.listeners[index1].close     = 0;
  rmix_tcpip4_data.listeners[index1].serverjob = serverjob;
  rmix_tcpip4_data.listeners[index1].count     = 0;
  rmix_tcpip4_data.listeners[index1].servers   = NULL;
  /* Randomly generate a unique non-zero 32-bit handle. */
  do {
    *handle = random() & 0xFFFFFFFF;
    for (index2 = 0; index2 < rmix_tcpip4_data.count; index2++) {
      if (*handle == rmix_tcpip4_data.listeners[index2].handle) {
        break;
      }
    }
  } while (index2 < rmix_tcpip4_data.count);
  rmix_tcpip4_data.listeners[index1].handle = *handle;
  /* Send byte to wake-up pipe. */
  write(rmix_tcpip4_data.wakeup[1], "", 1);
  /* Unlock TCP/IPv4 pool mutex. */
  if (0 != (error = pthread_mutex_unlock(&rmix_tcpip4_data.mutex))) {
    errno = error;
    RMIX_LOG((RMIX_WARN(unable to unlock TCP/IPv4 pool mutex)))
    errno = error;
    rmix_syserr();
    errno = error;
    return -1;
  }
  return 0;
}

/*
 * Opens a TCP/IPv4 server and binds it to the addressed local interface.
 *
 * handle    = The TCP/IPv4 server handle (return).
 * localurl  = The local URL to bind to.
 * backlog   = The listen backlog.
 * serverjob = The server job.
 * return    = 0 on success or -1 on any error with errno set appropriately.
 */
#undef  __FUNC__
#define __FUNC__ "rmix_tcpip4_open4"
int rmix_tcpip4_open4 (unsigned int            *handle,
                       char                    *localurl,
                       unsigned int             backlog,
                       rmix_tcpip4_serverjob_t  serverjob) {
  int                     error;
  void                   *listeners;
  unsigned int            index1, index2;
  rmix_tcpip4_listener_t *listener;
#ifdef DEBUG
  /* Check handle parameter. */
  if (NULL == handle) {
    errno = EINVAL;
    RMIX_LOG((RMIX_WARN(handle parameter is null)))
    errno = EINVAL;
    return -1;
  }
  /* Check localurl parameter. */
  if (NULL == localurl) {
    errno = EINVAL;
    RMIX_LOG((RMIX_WARN(localurl parameter is null)))
    errno = EINVAL;
    return -1;
  }
  /* Check serverjob parameter. */
  if (NULL == serverjob) {
    errno = EINVAL;
    RMIX_LOG((RMIX_WARN(serverjob parameter is null)))
    errno = EINVAL;
    return -1;
  }
#endif /* DEBUG */
  /* Open TCP/IPv4 listener. */
  if (0 != rmix_tcpip4_listener_open4(&listener, localurl, backlog)) {
    int errno2 = errno;
    RMIX_LOG((RMIX_WARN(unable to open TCP/IPv4 listener)))
    errno = errno2;
    return -1;
  }
  /* Lock TCP/IPv4 pool mutex. */
  if (0 != (error = pthread_mutex_lock(&rmix_tcpip4_data.mutex))) {
    errno = error;
    RMIX_LOG((RMIX_WARN(unable to lock TCP/IPv4 pool mutex)))
    errno = error;
    rmix_syserr();
    /* Close TCP/IPv4 listener. */
    if (0 != rmix_tcpip4_listener_close(&listener)) {
      RMIX_LOG((RMIX_WARN(unable to close TCP/IPv4 listener)))
    }
    errno = error;
    return -1;
  }
#ifdef DEBUG
  /* Check for finalized TCP/IPv4 pool. */
  if (0 == rmix_tcpip4_data.ready) {
    errno = EACCES;
    RMIX_LOG((RMIX_WARN(TCP/IPv4 pool finalized)))
    /* Unlock TCP/IPv4 pool mutex. */
    if (0 != (error = pthread_mutex_unlock(&rmix_tcpip4_data.mutex))) {
      errno = error;
      RMIX_LOG((RMIX_WARN(unable to unlock TCP/IPv4 pool mutex)))
      errno = error;
      rmix_syserr();
    }
    /* Close TCP/IPv4 listener. */
    if (0 != rmix_tcpip4_listener_close(&listener)) {
      RMIX_LOG((RMIX_WARN(unable to close TCP/IPv4 listener)))
    }
    errno = error;
    return -1;
  }
#endif /* DEBUG */
  /* Increase TCP/IPv4 pool listener list. */
  index1 = rmix_tcpip4_data.count;
  rmix_tcpip4_data.count++;
  /* Rellocate TCP/IPv4 pool listener list. */
  if (NULL == (listeners = realloc(rmix_tcpip4_data.listeners,
                                   rmix_tcpip4_data.count * 
                            sizeof(rmix_tcpip4_data.listeners[0])))) {
    int errno2 = errno;
    RMIX_LOG((RMIX_WARN(unable to reallocate TCP/IPv4 pool listener list)))
    /* Decrease TCP/IPv4 pool listener list. */
    rmix_tcpip4_data.count = index1;
    /* Unlock TCP/IPv4 pool mutex. */
    if (0 != (error = pthread_mutex_unlock(&rmix_tcpip4_data.mutex))) {
      errno = error;
      RMIX_LOG((RMIX_WARN(unable to unlock TCP/IPv4 pool mutex)))
      errno = error;
      rmix_syserr();
    }
    /* Close TCP/IPv4 listener. */
    if (0 != rmix_tcpip4_listener_close(&listener)) {
      RMIX_LOG((RMIX_WARN(unable to close TCP/IPv4 listener)))
    }
    errno = errno2;
    return -1;
  }
  rmix_tcpip4_data.listeners = listeners;
  /* Set up TCP/IPv4 pool listener list entry. */
  rmix_tcpip4_data.listeners[index1].listener  = listener;
  rmix_tcpip4_data.listeners[index1].handle    = 0;
  rmix_tcpip4_data.listeners[index1].close     = 0;
  rmix_tcpip4_data.listeners[index1].serverjob = serverjob;
  rmix_tcpip4_data.listeners[index1].count     = 0;
  rmix_tcpip4_data.listeners[index1].servers   = NULL;
  /* Randomly generate a unique non-zero 32-bit handle. */
  do {
    *handle = random() & 0xFFFFFFFF;
    for (index2 = 0; index2 < rmix_tcpip4_data.count; index2++) {
      if (*handle == rmix_tcpip4_data.listeners[index2].handle) {
        break;
      }
    }
  } while (index2 < rmix_tcpip4_data.count);
  rmix_tcpip4_data.listeners[index1].handle = *handle;
  /* Send byte to wake-up pipe. */
  write(rmix_tcpip4_data.wakeup[1], "", 1);
  /* Unlock TCP/IPv4 pool mutex. */
  if (0 != (error = pthread_mutex_unlock(&rmix_tcpip4_data.mutex))) {
    errno = error;
    RMIX_LOG((RMIX_WARN(unable to unlock TCP/IPv4 pool mutex)))
    errno = error;
    rmix_syserr();
    errno = error;
    return -1;
  }
  return 0;
}

/*
 * Closes a TCP/IPv4 server.
 *
 * handle   = The TCP/IPv4 server handle.
 * return   = 0 on success or -1 on any error with errno set appropriately.
 */
#undef  __FUNC__
#define __FUNC__ "rmix_tcpip4_close"
int rmix_tcpip4_close (unsigned int handle) {
  int          error;
  unsigned int index1;
  /* Lock TCP/IPv4 pool mutex. */
  if (0 != (error = pthread_mutex_lock(&rmix_tcpip4_data.mutex))) {
    errno = error;
    RMIX_LOG((RMIX_WARN(unable to lock TCP/IPv4 pool mutex)))
    errno = error;
    rmix_syserr();
    errno = error;
    return -1;
  }
#ifdef DEBUG
  /* Check for finalized TCP/IPv4 pool. */
  if (0 == rmix_tcpip4_data.ready) {
    errno = EACCES;
    RMIX_LOG((RMIX_WARN(TCP/IPv4 pool finalized)))
    /* Unlock TCP/IPv4 pool mutex. */
    if (0 != (error = pthread_mutex_unlock(&rmix_tcpip4_data.mutex))) {
      errno = error;
      RMIX_LOG((RMIX_WARN(unable to unlock TCP/IPv4 pool mutex)))
      errno = error;
      rmix_syserr();
    }
    errno = EACCES;
    return -1;
  }
#endif /* DEBUG */
  /* Search for TCP/IPv4 pool listener in list. */
  for (index1 = 0; index1 < rmix_tcpip4_data.count; index1++) {
    if (handle == rmix_tcpip4_data.listeners[index1].handle) {
      break;
    }
  }
  if (index1 == rmix_tcpip4_data.count) {
    errno = ESRCH;
    RMIX_LOG((RMIX_WARN(unable to find TCP/IPv4 pool listener in list)))
    /* Unlock TCP/IPv4 pool mutex. */
    if (0 != (error = pthread_mutex_unlock(&rmix_tcpip4_data.mutex))) {
      errno = error;
      RMIX_LOG((RMIX_WARN(unable to unlock TCP/IPv4 pool mutex)))
      errno = error;
      rmix_syserr();
    }
    errno = ESRCH;
    return -1;
  }
  /* Check if TCP/IPv4 pool listener is already closed. */
  if (1 == rmix_tcpip4_data.listeners[index1].close) {
    errno = EALREADY;
    RMIX_LOG((RMIX_WARN(unable to find TCP/IPv4 pool listener in list)))
    /* Unlock TCP/IPv4 pool mutex. */
    if (0 != (error = pthread_mutex_unlock(&rmix_tcpip4_data.mutex))) {
      errno = error;
      RMIX_LOG((RMIX_WARN(unable to unlock TCP/IPv4 pool mutex)))
      errno = error;
      rmix_syserr();
    }
    errno = EALREADY;
    return -1;
  }
  /* Mark TCP/IPv4 pool listener for closing. */
  rmix_tcpip4_data.listeners[index1].close = 1;
  /* Send byte to wake-up pipe. */
  write(rmix_tcpip4_data.wakeup[1], "", 1);
  /* Unlock TCP/IPv4 pool mutex. */
  if (0 != (error = pthread_mutex_unlock(&rmix_tcpip4_data.mutex))) {
    errno = error;
    RMIX_LOG((RMIX_WARN(unable to unlock TCP/IPv4 pool mutex)))
    errno = error;
    rmix_syserr();
    return -1;
  }
  return 0;
}

/*
 * Serves all listener and server connections.
 *
 * argument = NULL
 * return   = 0 on success or -1 on any system error with errno set
 *            appropriately.
 */
#undef  __FUNC__
#define __FUNC__ "rmix_tcpip4_job"
RMIX_THREADS_JOB(rmix_tcpip4_job) {
  int                        maxfd, fds, error;
  fd_set                     fdset;
  void                      *servers, *listeners;
  unsigned int               index1, index2;
  rmix_tcpip4_server_t      *server;
  rmix_tcpip4_pool_server_t *pool_server;
  /* Lock TCP/IPv4 pool mutex. */
  if (0 != (error = pthread_mutex_lock(&rmix_tcpip4_data.mutex))) {
    errno = error;
    RMIX_LOG((RMIX_WARN(unable to lock TCP/IPv4 pool mutex)))
    errno = error;
    rmix_syserr();
    /* End job due to syserr. */
    errno = error;
    return -1;
  }
  /* Push unlock TCP/IPv4 pool mutex to pthread cleanup stack. */
  pthread_cleanup_push(rmix_tcpip4_unlock, NULL);
  /* Empty file descriptor set. */
  FD_ZERO(&fdset);
  /* Add wake-up pipe to file descriptor set. */
  FD_SET(rmix_tcpip4_data.wakeup[0], &fdset);
  /* Calculate the greatest file descriptor in the set. */
  maxfd = rmix_tcpip4_data.wakeup[0];
  /* Add sockets from all TCP/IPv4 pool listeners
     and non-busy TCP/IPv4 pool servers to file descriptor set. */
  for (index1 = 0; index1 < rmix_tcpip4_data.count; index1++) {
    /* Add socket from TCP/IPv4 pool listener to file descriptor set. */
    FD_SET(rmix_tcpip4_data.listeners[index1].listener->socket,  &fdset);
    /* Calculate the greatest file descriptor in the set. */
    if (maxfd < rmix_tcpip4_data.listeners[index1].listener->socket) {
      maxfd = rmix_tcpip4_data.listeners[index1].listener->socket;
    }
    /* Add sockets from non-busy TCP/IPv4 pool
       servers to file descriptor set. */
    for (index2 = 0;
         index2 < rmix_tcpip4_data.listeners[index1].count; index2++) {
      /* Check if TCP/IPv4 pool server is not busy. */
      if (0 == rmix_tcpip4_data.listeners[index1].servers[index2]->busy) {
        /* Add socket from TCP/IPv4 pool server to file descriptor set. */
        FD_SET(rmix_tcpip4_data.listeners[index1].
                                  servers[index2]->server->socket, &fdset);
        /* Calculate the greatest file descriptor in the set. */
        if (maxfd < rmix_tcpip4_data.listeners[index1].
                                       servers[index2]->server->socket) {
          maxfd = rmix_tcpip4_data.listeners[index1].
                                     servers[index2]->server->socket;
        }
      }
    }
  }
  /* Pop unlock TCP/IPv4 pool mutex from pthread cleanup stack. */
  pthread_cleanup_pop(0);
  /* Unlock TCP/IPv4 pool mutex. */
  if (0 != (error = pthread_mutex_unlock(&rmix_tcpip4_data.mutex))) {
    errno = error;
    RMIX_LOG((RMIX_WARN(unable to unlock TCP/IPv4 pool mutex)))
    errno = error;
    rmix_syserr();
    /* End job due to syserr. */
    errno = error;
    return -1;
  }
  while (1) {
    /* Wait on file descriptor set. */
    fds = select(maxfd + 1, &fdset, NULL, NULL, NULL);
    /* Lock TCP/IPv4 pool mutex. */
    if (0 != (error = pthread_mutex_lock(&rmix_tcpip4_data.mutex))) {
      errno = error;
      RMIX_LOG((RMIX_WARN(unable to lock TCP/IPv4 pool mutex)))
      errno = error;
      rmix_syserr();
      /* End job due to syserr. */
      errno = error;
      return -1;
    }
    /* Push unlock TCP/IPv4 pool mutex to pthread cleanup stack. */
    pthread_cleanup_push(rmix_tcpip4_unlock, NULL);
    /* Ckeck for error. */
    if ((-1 == fds)&&(EINTR != errno)) {
      int errno2 = errno;
      RMIX_LOG((RMIX_WARN(unable to wait on file descriptor set)))
      errno = errno2;
      rmix_syserr();
      /* Unlock TCP/IPv4 pool mutex. */
      if (0 != (error = pthread_mutex_unlock(&rmix_tcpip4_data.mutex))) {
        errno = error;
        RMIX_LOG((RMIX_WARN(unable to unlock TCP/IPv4 pool mutex)))
        errno = error;
        rmix_syserr();
      }
      /* End job due to syserr. */
      errno = errno2;
      return -1;
    }
    /* Check for finalized TCP/IPv4 pool. */
    if (0 == rmix_tcpip4_data.ready) {
      /* Unlock TCP/IPv4 pool mutex. */
      if (0 != (error = pthread_mutex_unlock(&rmix_tcpip4_data.mutex))) {
        errno = error;
        RMIX_LOG((RMIX_WARN(unable to unlock TCP/IPv4 pool mutex)))
        errno = error;
        rmix_syserr();
        /* End job due to syserr. */
        errno = error;
        return -1;
      }
      /* End job. */
      return 0;
    }
    /* Ckeck ready wake-up pipe. */
    if ((0 < fds)&&(0 != FD_ISSET(rmix_tcpip4_data.wakeup[0], &fdset))) {
      /* Reduce the number of ready file descriptors. */
      fds--;
      /* Try to empty wake-up pipe. */
      if (0 > rmix_tcpip4_read(rmix_tcpip4_data.wakeup[0],
                               RMIX_TCPIP4_BUFFER_SIZE,
                               rmix_tcpip4_data.buffer, 0)) {
        int errno2 = errno;
        RMIX_LOG((RMIX_WARN(unable to empty wake-up pipe)))
        errno = errno2;
        rmix_syserr();
        /* Unlock TCP/IPv4 pool mutex. */
        if (0 != (error = pthread_mutex_unlock(&rmix_tcpip4_data.mutex))) {
          errno = error;
          RMIX_LOG((RMIX_WARN(unable to unlock TCP/IPv4 pool mutex)))
          errno = error;
          rmix_syserr();
        }
        /* End job due to syserr. */
        errno = errno2;
        return -1;
      }
    }
    /* Search for TCP/IPv4 pool listeners to close. */
    for (index1 = 0; index1 < rmix_tcpip4_data.count; index1++) {
      if (1 == rmix_tcpip4_data.listeners[index1].close) {
        /* Close TCP/IPv4 listener. */
        if (0 != rmix_tcpip4_listener_close(
                   &rmix_tcpip4_data.listeners[index1].listener)) {
          int errno2 = errno;
          RMIX_LOG((RMIX_WARN(unable to close TCP/IPv4 listener)))
          errno = errno2;
          rmix_syserr();
          /* Unlock TCP/IPv4 pool mutex. */
          if (0 != (error = pthread_mutex_unlock(&rmix_tcpip4_data.mutex))) {
            errno = error;
            RMIX_LOG((RMIX_WARN(unable to unlock TCP/IPv4 pool mutex)))
            errno = error;
            rmix_syserr();
          }
          /* End job due to syserr. */
          errno = errno2;
          return -1;
        }
        /* Close related non-busy TCP/IPv4 pool servers. */
        for (index2 = 0;
             index2 < rmix_tcpip4_data.listeners[index1].count; index2++) {
          /* Check if TCP/IPv4 pool server is not busy. */
          if (0 == rmix_tcpip4_data.listeners[index1].servers[index2]->busy) {
            /* Close TCP/IPv4 server. */
            if (0 != rmix_tcpip4_server_close(
                       &rmix_tcpip4_data.listeners[index1].
                                           servers[index2]->server)) {
              int errno2 = errno;
              RMIX_LOG((RMIX_WARN(unable to close TCP/IPv4 pool server)))
              errno = errno2;
              rmix_syserr();
              /* Unlock TCP/IPv4 pool mutex. */
              if (0 != (error =
                        pthread_mutex_unlock(&rmix_tcpip4_data.mutex))) {
                errno = error;
                RMIX_LOG((RMIX_WARN(unable to unlock TCP/IPv4 pool mutex)))
                errno = error;
                rmix_syserr();
              }
              /* End job due to syserr. */
              errno = errno2;
              return -1;
            }
            /* Deallocate TCP/IPv4 pool server. */
            free(rmix_tcpip4_data.listeners[index1].servers[index2]);
            /* Remove from TCP/IPv4 pool server list. */
            rmix_tcpip4_data.listeners[index1].count--;
            memmove(rmix_tcpip4_data.listeners[index1].servers + index2,
                    rmix_tcpip4_data.listeners[index1].servers + index2 + 1,
                   (rmix_tcpip4_data.listeners[index1].count   - index2) *
             sizeof(rmix_tcpip4_data.listeners[index1].servers[0]));
            /* Reallocate related TCP/IPv4 pool server list. */
            if (0 == rmix_tcpip4_data.listeners[index1].count) {
              free(rmix_tcpip4_data.listeners[index1].servers);
              rmix_tcpip4_data.listeners[index1].servers = NULL;
            } else  if (NULL == (servers =
                realloc(rmix_tcpip4_data.listeners[index1].servers,
                        rmix_tcpip4_data.listeners[index1].count *
                 sizeof(rmix_tcpip4_data.listeners[index1].servers[0])))) {
              RMIX_LOG((
                RMIX_WARN(unable to reallocate TCP/IPv4 pool server list)))
              errno = 0;
            } else {
              rmix_tcpip4_data.listeners[index1].servers = servers;
            }
          }
        }
        /* Check for empty server list. */
        if (0 == rmix_tcpip4_data.listeners[index1].count) {
          /* Remove listener from TCP/IPv4 pool listener list. */
          rmix_tcpip4_data.count--;
          memmove(rmix_tcpip4_data.listeners + index1,
                  rmix_tcpip4_data.listeners + index1 + 1,
                 (rmix_tcpip4_data.count     - index1) *
                  sizeof(rmix_tcpip4_data.listeners[0]));
          /* Reallocate TCP/IPv4 pool listener list. */
          if (0 == rmix_tcpip4_data.count) {
            free(rmix_tcpip4_data.listeners);
            rmix_tcpip4_data.listeners = NULL;
          } else if (NULL == (listeners =
                              realloc(rmix_tcpip4_data.listeners,
                                      rmix_tcpip4_data.count *
                               sizeof(rmix_tcpip4_data.listeners[0])))) {
            RMIX_LOG((
              RMIX_WARN(unable to reallocate TCP/IPv4 pool listener list)))
            errno = 0;
          } else {
            rmix_tcpip4_data.listeners = listeners;
          }
        }
      }
    }
    /* Check for TCP/IPv4 pool servers to close. */
    for (index1 = 0; index1 < rmix_tcpip4_data.count; index1++) {
      for (index2 = 0;
           index2 < rmix_tcpip4_data.listeners[index1].count; index2++) {
        /* Check for TCP/IPv4 pool server to close. */
        if (NULL == rmix_tcpip4_data.listeners[index1].
                                       servers[index2]->serverjob) {
          /* Close TCP/IPv4 server. */
          if (0 != rmix_tcpip4_server_close(&rmix_tcpip4_data.listeners[index1].
                                             servers[index2]->server)) {
            RMIX_LOG((RMIX_WARN(unable to close server)))
          }
          /* Deallocate TCP/IPv4 pool server. */
          free(rmix_tcpip4_data.listeners[index1].servers[index2]);
          /* Remove server from TCP/IPv4 pool listener server list. */
          rmix_tcpip4_data.listeners[index1].count--;
          memmove(rmix_tcpip4_data.listeners[index1].servers + index2,
                  rmix_tcpip4_data.listeners[index1].servers + index2 + 1,
                 (rmix_tcpip4_data.listeners[index1].count   - index2) *
           sizeof(rmix_tcpip4_data.listeners[index1].servers[0]));
          /* Reallocate listener server list. */
          if (0 == rmix_tcpip4_data.listeners[index1].count) {
            free(rmix_tcpip4_data.listeners[index1].servers);
            rmix_tcpip4_data.listeners[index1].servers = NULL;
          } else if (NULL == (servers =
            realloc(rmix_tcpip4_data.listeners[index1].servers,
                    rmix_tcpip4_data.listeners[index1].count *
             sizeof(rmix_tcpip4_data.listeners[index1].servers[0])))) {
            RMIX_LOG((RMIX_WARN(unable to reallocate listener server list)))
            errno = 0;
          } else {
            rmix_tcpip4_data.listeners[index1].servers = servers;
          }
        }
      }
    }
    /* Check for ready sockets from TCP/IPv4 pool listeners and servers. */
    for (index1 = 0; 0 < fds && index1 < rmix_tcpip4_data.count; index1++) {
      /* Check for ready sockets from TCP/IPv4 pool listener servers. */
      for (index2 = 0; 0 < fds &&
           index2 < rmix_tcpip4_data.listeners[index1].count; index2++) {
        /* Check for ready socket from TCP/IPv4 pool listener server. */
        if (0  != FD_ISSET(rmix_tcpip4_data.listeners[index1].
                                              servers[index2]->server->socket,
                           &fdset)) {
          /* Reduce the number of ready file descriptors. */
          fds--;
          /* Set TCP/IPv4 pool server busy. */
          rmix_tcpip4_data.listeners[index1].servers[index2]->busy = 1;
          /* Submit TCP/IPv4 pool server job. */
          argument = rmix_tcpip4_data.listeners[index1].servers[index2];
          if (0 != rmix_threads_submit(rmix_tcpip4_server_job,
                                       NULL, &argument)) {
            int errno2 = errno;
            RMIX_LOG((RMIX_WARN(unable to submit server job)))
            errno = errno2;
            rmix_syserr();
            /* Check if job was not enqueued. */
            if (NULL != argument) {
              /* Reset TCP/IPv4 pool listener server busy flag. */
              rmix_tcpip4_data.listeners[index1].
                                 servers[index2]->busy = 0;
            }
            /* Unlock TCP/IPv4 pool mutex. */
            if (0 != (error =
                      pthread_mutex_unlock(&rmix_tcpip4_data.mutex))) {
              errno = error;
              RMIX_LOG((RMIX_WARN(unable to unlock TCP/IPv4 pool mutex)))
              errno = error;
              rmix_syserr();
            }
            /* End job due to syserr. */
            errno = errno2;
            return -1;
          }
        }
      }
      /* Check for ready socket from TCP/IPv4 pool listener. */
      if (0  != FD_ISSET(rmix_tcpip4_data.listeners[index1].listener->socket,
                         &fdset)) {
        /* Reduce the number of ready file descriptors. */
        fds--;
        /* Allocate TCP/IPv4 pool server. */
        if (NULL == (pool_server = malloc(sizeof(rmix_tcpip4_pool_server_t)))) {
          RMIX_LOG((RMIX_WARN(unable to allocate TCP/IPv4 pool server)))
        } else {
          /* Open TCP/IPv4 server. */
          if (0 != rmix_tcpip4_sever_open(&server,
                     rmix_tcpip4_data.listeners[index1].listener)) {
            /* Check if connection was not aborted. */
            if (ECONNABORTED != errno) {
              int errno2 = errno;
              RMIX_LOG((RMIX_WARN(unable to open TCP/IPv4 server)))
              errno = errno2;
              rmix_syserr();
              /* Deallocate TCP/IPv4 pool server. */
              free(pool_server);
              /* Unlock TCP/IPv4 pool mutex. */
              if (0 != (error =
                        pthread_mutex_unlock(&rmix_tcpip4_data.mutex))) {
                errno = error;
                RMIX_LOG((RMIX_WARN(unable to unlock TCP/IPv4 pool mutex)))
                errno = error;
                rmix_syserr();
              }
              /* End job due to syserr. */
              errno = errno2;
              return -1;
            }
          } else {
            /* Increase TCP/IPv4 pool listeners server list. */
            index2 = rmix_tcpip4_data.listeners[index1].count;
            rmix_tcpip4_data.listeners[index1].count++;
            /* Rellocate TCP/IPv4 pool listeners server list. */
            if (NULL == (servers =
                realloc(rmix_tcpip4_data.listeners[index1].servers,
                        rmix_tcpip4_data.listeners[index1].count *
                 sizeof(rmix_tcpip4_data.listeners[index1].servers[0])))) {
              RMIX_LOG((
                RMIX_WARN(unable to reallocate TCP/IPv4 pool listener list)))
              errno = 0;
              /* Decrease TCP/IPv4 pool listeners server list. */
              rmix_tcpip4_data.listeners[index1].count = index1;
              /* Deallocate TCP/IPv4 pool server. */
              free(pool_server);
              /* Close TCP/IPv4 server. */
              if (0 != rmix_tcpip4_server_close(&server)) {
                int errno2 = errno;
                RMIX_LOG((RMIX_WARN(unable to close TCP/IPv4 server)))
                rmix_syserr();
                /* Unlock TCP/IPv4 pool mutex. */
                if (0 != (error =
                          pthread_mutex_unlock(&rmix_tcpip4_data.mutex))) {
                  errno = error;
                  RMIX_LOG((RMIX_WARN(unable to unlock TCP/IPv4 pool mutex)))
                  errno = error;
                  rmix_syserr();
                }
                /* End job due to syserr. */
                errno = errno2;
                return -1;
              }
            }
            rmix_tcpip4_data.listeners[index1].servers = servers;
            /* Initialize TCP/IPv4 pool server. */
            rmix_tcpip4_data.listeners[index1].servers[index2] = pool_server;
            pool_server->server    = server;
            pool_server->busy      = 0;
            pool_server->serverjob =
              rmix_tcpip4_data.listeners[index1].serverjob;
          }
        }
      }
    }
    /* Empty file descriptor set. */
    FD_ZERO(&fdset);
    /* Add wake-up pipe to file descriptor set. */
    FD_SET(rmix_tcpip4_data.wakeup[0], &fdset);
    /* Calculate the greatest file descriptor in the set. */
    maxfd = rmix_tcpip4_data.wakeup[0];
    /* Add sockets from all TCP/IPv4 pool listeners
       and non-busy listeners servers to file descriptor set. */
    for (index1 = 0; index1 < rmix_tcpip4_data.count; index1++) {
      /* Add socket from TCP/IPv4 pool listener to file descriptor set. */
      FD_SET(rmix_tcpip4_data.listeners[index1].listener->socket,  &fdset);
      /* Calculate the greatest file descriptor in the set. */
      if (maxfd < rmix_tcpip4_data.listeners[index1].listener->socket) {
        maxfd = rmix_tcpip4_data.listeners[index1].listener->socket;
      }
      /* Add sockets from non-busy TCP/IPv4 pool
         listener servers to file descriptor set. */
      for (index2 = 0;
           index2 < rmix_tcpip4_data.listeners[index1].count; index2++) {
        /* Check if TCP/IPv4 pool server is not busy. */
        if (0 == rmix_tcpip4_data.listeners[index1].servers[index2]->busy) {
          /* Add socket from TCP/IPv4 pool server to file descriptor set. */
          FD_SET(rmix_tcpip4_data.listeners[index1].
                                    servers[index2]->server->socket, &fdset);
          /* Calculate the greatest file descriptor in the set. */
          if (maxfd < rmix_tcpip4_data.listeners[index1].
                                         servers[index2]->server->socket) {
            maxfd = rmix_tcpip4_data.listeners[index1].
                                       servers[index2]->server->socket;
          }
        }
      }
    }
    /* Pop unlock TCP/IPv4 pool mutex from pthread cleanup stack. */
    pthread_cleanup_pop(0);
    /* Unlock TCP/IPv4 pool mutex. */
    if (0 != (error = pthread_mutex_unlock(&rmix_tcpip4_data.mutex))) {
      errno = error;
      RMIX_LOG((RMIX_WARN(unable to unlock TCP/IPv4 pool mutex)))
      errno = error;
      rmix_syserr();
      /* End job due to syserr. */
      errno = error;
      return -1;
    }
  }
  return 0;
}

/*
 * The TCP/IPv4 pool module mutex unlock cleanup.
 *
 * argument = NULL.
 */
#undef  __FUNC__
#define __FUNC__ "rmix_tcpip4_unlock"
void rmix_tcpip4_unlock (void *argument) {
  int error;
  /* Unlock TCP/IPv4 pool mutex. */
  if (0 != (error = pthread_mutex_unlock(&rmix_tcpip4_data.mutex))) {
    errno = error;
    RMIX_LOG((RMIX_WARN(unable to unlock TCP/IPv4 pool mutex)))
    errno = error;
    rmix_syserr();
    errno = error;
  }
}

/*
 * Serves server connections.
 *
 * argument = The pool server.
 * return   = 0 on success or -1 on any system error with errno set
 *            appropriately.
 */
#undef  __FUNC__
#define __FUNC__ "rmix_tcpip4_server_job"
RMIX_THREADS_JOB(rmix_tcpip4_server_job) {
#ifdef DEBUG
  /* Check argument parameter. */
  if (NULL == argument) {
    errno = EINVAL;
    RMIX_LOG((RMIX_WARN(argument parameter is null)))
    errno = EINVAL;
    return -1;
  }
#endif /* DEBUG */
  rmix_tcpip4_pool_server_t *pool_server;
  /* Get the pool server. */
  pool_server = argument;
  /* Push release and reset TCP/IPv4 server job to pthread cleanup stack. */
  pthread_cleanup_push(rmix_tcpip4_server_job_cleanup, argument);
  /* Call and set the server job. */
  pool_server->serverjob = pool_server->serverjob(pool_server->server);
  /* Pop release and reset TCP/IPv4 server job from pthread cleanup stack. */
  pthread_cleanup_pop(0);
  /* Release the TCP/IPv4 pool server. */
  pool_server->busy = 0;
  /* Send byte to wake-up pipe. */
  write(rmix_tcpip4_data.wakeup[1], "", 1);
  return 0;
}

/*
 * TCP/IPv4 server job cleanup.
 *
 * server = The TCP/IPv4 server.
 * return = 0 on success or -1 on any error with errno set appropriately.
 */
#undef  __FUNC__
#define __FUNC__ "rmix_tcpip4_server_job_cleanup"
void rmix_tcpip4_server_job_cleanup (void *pool_server) {
#ifdef DEBUG
  /* Check server parameter. */
  if (NULL == pool_server) {
    errno = EINVAL;
    RMIX_LOG((RMIX_WARN(pool_server parameter is null)))
    errno = EINVAL;
    return;
  }
#endif /* DEBUG */
  /* Reset the TCP/IPv4 pool server job. */
  ((rmix_tcpip4_pool_server_t*)pool_server)->serverjob = NULL;
  /* Release the TCP/IPv4 pool server. */
  ((rmix_tcpip4_pool_server_t*)pool_server)->busy = 0;
  /* Send byte to wake-up pipe. */
  write(rmix_tcpip4_data.wakeup[1], "", 1);
}

/*
 * Reads a buffer from a file descriptor (non-/blocking).
 *
 * fd     = The file descriptor to read from.
 * length = The maximum buffer length to read.
 * buffer = The buffer to read into.
 * block  = The (non-)blocking flag (0 = non-blocking, 1 = blocking).
 * return = number of bytes read on success, -2 on closed file descriptor or -1
 *          on any other error with errno set appropriately.
 */
#undef  __FUNC__
#define __FUNC__ "rmix_tcpip4_read"
int rmix_tcpip4_read (int           fd,
                      unsigned int  length,
                      void         *buffer,
                      unsigned int  block) {
  int          bytes;
  unsigned int index;
#ifdef DEBUG
  /* Check buffer parameter. */
  if ((NULL == buffer)&&(0 != length)) {
    errno = EINVAL;
    RMIX_LOG((RMIX_WARN(buffer parameter is null)))
    errno = EINVAL;
    return -1;
  }
  /* Check length parameter. */
  if ((0 == length)&&(NULL != buffer)) {
    errno = EINVAL;
    RMIX_LOG((RMIX_WARN(length parameter is zero)))
    errno = EINVAL;
    return -1;
  }
#endif /* DEBUG */
  for (index = 0; index < length;) {
    /* Check for thread cancellation. */
    pthread_testcancel();
    /* Read some data. */
    switch (bytes = read(fd, buffer + index, length - index)) {
      case -1: {
        switch (errno) {
          case EINTR: {
            break;
          }
          case EAGAIN: {
            if (0 == block) {
              return index;
            }
            break;
          }
          default: {
            int errno2 = errno;
            RMIX_LOG((RMIX_WARN(unable to read from file descriptor)))
            errno = errno2;
            return -1;
          }
        }
        break;
      }
      case 0: {
        errno = EPIPE;
        if (0 != index) {
          RMIX_LOG((RMIX_WARN(unable to read from closed file descriptor)))
        }
        return -2;
      }
      default: {
        index += bytes;
        if (0 == block) {
          return index;
        }
      }
    }
  }
  return index;
}

/******************************************************************************
 *
 * END OF FILE
 *
 ******************************************************************************/
