/******************************************************************************
 *
 * Header file for rmix threads (thread pool) module.
 * Copyright (c) 2002-2006 Oak Ridge National Laboratory.
 *
 * For more information see the following files in the source distribution top-
 * level directory or package data directory (usually /usr/local/share/package):
 *
 * - README    for general package information.
 * - INSTALL   for package install information.
 * - COPYING   for package license information and copying conditions.
 * - AUTHORS   for package authors information.
 * - ChangeLog for package changes information.
 *
 ******************************************************************************/

/* Avoid to include the content of this header file twice. */
#ifndef RMIX_THREADS_H
#define RMIX_THREADS_H

/******************************************************************************
 *
 * Macros
 *
 ******************************************************************************/

/*
 * Thread pool module defaults.
 */
#define RMIX_THREADS_POOL_MINIMUM     5 /* minimum thread count   */
#define RMIX_THREADS_POOL_MAXIMUM    25 /* maximum thread count   */
#define RMIX_THREADS_POOL_TIMEOUT     2 /* exiting thread timeout */
#define RMIX_THREADS_JOBS_MAXIMUM  1024 /* maximum job count      */
#define RMIX_THREADS_JOBS_TIMEOUT   120 /* waiting thread timeout */

/* Thread pool job argument destructor function signature. */
/*
 * Cleans up the job argument if the job was cancelled.
 *
 * argument = The pointer to the job argument (if any).
 * return   = 0 on success or -1 on any error with errno set appropriately.
 */
#define RMIX_THREADS_JOB_CLEANUP(name) void name (void *argument)

/* Thread pool job function signature. */
/*
 * Executes something in a separate thread.
 *
 * cleanup  = The pointer to the job cleanup function (if any).
 * argument = The pointer to the job argument (if any).
 * return   = 0 on success or -1 on any error with errno set appropriately.
 */
#define RMIX_THREADS_JOB(name)\
  int name (rmix_threads_job_cleanup_t cleanup, void *argument)

/******************************************************************************
 *
 * Includes
 *
 ******************************************************************************/

/* None at this time. */

/******************************************************************************
 *
 * Data Types
 *
 ******************************************************************************/

/* Thread pool job argument destructor function data type. */
typedef RMIX_THREADS_JOB_CLEANUP((*rmix_threads_job_cleanup_t));

/* Thread pool job function data type. */
typedef RMIX_THREADS_JOB((*rmix_threads_job_t));

/******************************************************************************
 *
 * Function Prototypes
 *
 ******************************************************************************/

/*
 * Sets the minimum thread count of the thread pool.
 * This function may be called before initializing.
 *
 * minimum = The new minimum thread count.
 * return  = 0 on success or -1 on any error with errno set appropriately.
 */
int rmix_threads_pool_minimum_set (unsigned int minimum);

/*
 * Sets the maximum thread count of the thread pool.
 * This function may be called before initializing.
 *
 * maximum = The new maximum thread count.
 * return  = 0 on success or -1 on any error with errno set appropriately.
 */
int rmix_threads_pool_maximum_set (unsigned int maximum);

/*
 * Sets the exiting thread timeout of the thread pool.
 * This function may be called before initializing.
 *
 * timeout = The new exiting thread timeout.
 * return  = 0 on success or -1 on any error with errno set appropriately.
 */
int rmix_threads_pool_timeout_set (unsigned int timeout);

/*
 * Sets the maximum job count of the thread pools job queue.
 * This function may be called before initializing.
 *
 * maximum = The new maximum job count.
 * return  = 0 on success or -1 on any error with errno set appropriately.
 */
int rmix_threads_jobs_maximum_set (unsigned int maximum);

/*
 * Sets the waiting thread timeout of the thread pools job queue.
 * This function may be called before initializing.
 *
 * timeout = The new waiting thread timeout.
 * return  = 0 on success or -1 on any error with errno set appropriately.
 */
int rmix_threads_jobs_timeout_set (unsigned int timeout);

/*
 * Increases/Decreases the minimum thread count of the thread pool.
 * This function may be called before initializing.
 *
 * minimum = The minimum thread count increment/decrement.
 * return  = 0 on success or -1 on any error with errno set appropriately.
 */
int rmix_threads_pool_minimum_add (int minimum);

/*
 * Increases/Decreases the maximum thread count of the thread pool.
 * This function may be called before initializing.
 *
 * maximum = The maximum thread count increment/decrement.
 * return  = 0 on success or -1 on any error with errno set appropriately.
 */
int rmix_threads_pool_maximum_add (int maximum);

/*
 * Increases/Decreases the exiting thread timeout of the thread pool.
 * This function may be called before initializing.
 *
 * timeout = The exiting thread timeout increment/decrement.
 * return  = 0 on success or -1 on any error with errno set appropriately.
 */
int rmix_threads_pool_timeout_add (int timeout);

/*
 * Increases/Decreases the maximum job count of the thread pools job queue.
 * This function may be called before initializing.
 *
 * maxjobs = The maximum job count increment/decrement.
 * return  = 0 on success or -1 on any error with errno set appropriately.
 */
int rmix_threads_jobs_maximum_add (int maximum);

/*
 * Increases/Decreases the waiting thread timeout of the thread pools job queue.
 * This function may be called before initializing.
 *
 * timeout = The waiting thread timeout increment/decrement.
 * return  = 0 on success or -1 on any error with errno set appropriately.
 */
int rmix_threads_jobs_timeout_add (int timeout);

/*
 * Gets the minimum thread count of the thread pool.
 * This function may be called before initializing.
 *
 * minimum = The minimum thread count (return).
 * return  = 0 on success or -1 on any error with errno set appropriately.
 */
int rmix_threads_pool_minimum_get (unsigned int *minimum);

/*
 * Gets the maximum thread count of the thread pool.
 * This function may be called before initializing.
 *
 * maximum = The maximum thread count (return).
 * return  = 0 on success or -1 on any error with errno set appropriately.
 */
int rmix_threads_pool_maximum_get (unsigned int *maximum);

/*
 * Gets the exiting thread timeout of the thread pool.
 * This function may be called before initializing.
 *
 * timeout = The exiting thread timeout (return).
 * return  = 0 on success or -1 on any error with errno set appropriately.
 */
int rmix_threads_pool_timeout_get (unsigned int *timeout);

/*
 * Gets the maximum job count of the thread pools job queue.
 * This function may be called before initializing.
 *
 * maximum = The maximum job count (return).
 * return  = 0 on success or -1 on any error with errno set appropriately.
 */
int rmix_threads_jobs_maximum_get (unsigned int *maximum);

/*
 * Gets the waiting thread timeout of the thread pools job queue.
 * This function may be called before initializing.
 *
 * timeout = The waiting thread timeout (return).
 * return  = 0 on success or -1 on any error with errno set appropriately.
 */
int rmix_threads_jobs_timeout_get (unsigned int *timeout);

/*
 * Submits a job to the thread pool by enqueueing it into the job queue.
 *
 * function = The job function.
 * cleanup  = The job cleanup function.
 * argument = The job argument (returns NULL if job was enqueued).
 * return   = 0 on success or -1 on any error with errno set appropriately.
 */
int rmix_threads_submit (rmix_threads_job_t           function,
                         rmix_threads_job_cleanup_t   cleanup,
                         void                       **argument);

/******************************************************************************
 *
 * Data
 *
 ******************************************************************************/

/* None at this time. */

#endif /* RMIX_THREADS_H */

/******************************************************************************
 *
 * END OF FILE
 *
 ******************************************************************************/
