/******************************************************************************
 *
 * Header file for rmix stream (I/O valary) module.
 * Copyright (c) 2002-2006 Oak Ridge National Laboratory.
 *
 * For more information see the following files in the source distribution top-
 * level directory or package data directory (usually /usr/local/share/package):
 *
 * - README    for general package information.
 * - INSTALL   for package install information.
 * - COPYING   for package license information and copying conditions.
 * - AUTHORS   for package authors information.
 * - ChangeLog for package changes information.
 *
 ******************************************************************************/

/* Avoid to include the content of this header file twice. */
#ifndef RMIX_VALARY_H
#define RMIX_VALARY_H

/******************************************************************************
 *
 * Macros
 *
 ******************************************************************************/

/* Block size for block-wise value array allocation (cannot be 0). */
#define RMIX_VALARY_BLOCK 8

/******************************************************************************
 *
 * Includes
 *
 ******************************************************************************/

/* None at this time. */

/******************************************************************************
 *
 * Data Types
 *
 ******************************************************************************/

/* The values array cleanup data type. */
typedef struct {
  void         **valary; /* values array   */
  unsigned int   valcnt; /* values count   */
  const char    *sigstr; /* type signature */
} rmix_valary_cleanup_t;

/******************************************************************************
 *
 * Function Prototypes
 *
 ******************************************************************************/

/*
 * Creates a values array and all of its values. Strings and variable arrays
 * are initialized with NULL and 0-NULL respectively 
 *
 * valary = The values array to create (return).
 * valcnt = The values array size (return).
 * sigstr = The object method input/output signature.
 * return = 0 on success or -1 on any error with errno set appropriately.
 */
int rmix_valary_create (void         ***valary,
                        unsigned int   *valcnt,
                        const char     *sigstr);

/*
 * Destroys a values array and all of its values. The values array may be
 * incomplete, i.e. values may be missing at the end or set to NULL anywhere
 * within the values array with the exception of a variable array length.
 *
 * valary = The values array to destroy (also return).
 * valcnt = The values array size (also return).
 * sigstr = The object method input/output signature.
 * return = 0 on success or -1 on any error with errno set appropriately.
 */
int rmix_valary_destroy (void         ***valary,
                         unsigned int   *valcnt,
                         const char     *sigstr);

/*
 * Destroys all variable length values of a values array. The values array may
 * be incomplete, i.e. values may be missing at the end or set to NULL anywhere
 * within the values array with the exception of a variable array length.
 *
 * valary = The values array to destroy (also return).
 * valcnt = The values array size (also return).
 * sigstr = The object method input/output signature.
 * return = 0 on success or -1 on any error with errno set appropriately.
 */
int rmix_valary_destroy2 (void         **valary,
                          unsigned int   valcnt,
                          const char    *sigstr);

/*
 * Destroys a values array and all of its values. The values array may be
 * incomplete, i.e. values may be missing at the end or set to NULL anywhere
 * within the values array with the exception of a variable array length.
 * (Pthread cleanup function.)
 *
 * cleanup = The values array cleanup data to cleanup.
 */
void rmix_valary_cleanup (void *cleanup);

/*
 * Destroys all variable length values of a values array. The values array may
 * be incomplete, i.e. values may be missing at the end or set to NULL anywhere
 * within the values array with the exception of a variable array length.
 * (Pthread cleanup function.)
 *
 * cleanup = The values array cleanup data to cleanup.
 */
void rmix_valary_cleanup2 (void *cleanup);

/*
 * Moves all values from one values array to another. Both values arrays must
 * already exist and must be equal with the exception of strings and variable
 * arrays.
 *
 * valary1 = The values array to move values into.
 * valary1 = The values array to move values from.
 * valcnt  = The values array size.
 * sigstr  = The object method input/output signature.
 * return  = 0 on success or -1 on any error with errno set appropriately.
 */
int rmix_valary_move (void               **valary1,
                      void               **valary2,
                      const unsigned int   valcnt,
                      const char          *sigstr);

/*
 * Creates a value based on the type signature, reallocates the values array,
 * stores the allocated value in it, and advances the signature index. Strings
 * and variable arrays are initialized with NULL and 0-NULL respectively 
 *
 * valary = The values array (also return).
 * valcnt = The values array size (also return).
 * sigstr = The type signature.
 * sigidx = The type signature index (also return).
 * return = 0 on success or -1 on any error with errno set appropriately.
 */
int rmix_value_create (void         ***valary,
                       unsigned int   *valcnt,
                       const char     *sigstr,
                       unsigned int   *sigidx);

/*
 * Deallocates a value and advances the values array and signature indices.
 *
 * valary = The values array.
 * validx = The values array index (also return).
 * sigstr = The type signature.
 * sigidx = The type signature index (also return).
 * return = 0 on success or -1 on any error with errno set appropriately.
 */
int rmix_value_destroy (void         **valary,
                        unsigned int  *validx,
                        const char    *sigstr,
                        unsigned int  *sigidx);

/*
 * Deallocates a value if it is a string or variable array and advances the
 * values array and signature indices.
 *
 * valary = The values array.
 * validx = The values array index (also return).
 * sigstr = The type signature.
 * sigidx = The type signature index (also return).
 * return = 0 on success or -1 on any error with errno set appropriately.
 */
int rmix_value_destroy2 (void         **valary,
                         unsigned int  *validx,
                         const char    *sigstr,
                         unsigned int  *sigidx);

/*
 * Moves a value and advances the values array and signature indices.
 *
 * valary1 = The values array to move the value into.
 * valary2 = The values array to move the value from.
 * validx  = The values array index (also return).
 * sigstr  = The type signature.
 * sigidx  = The type signature index (also return).
 * return = 0 on success or -1 on any error with errno set appropriately.
 */
int rmix_value_move (void         **valary1,
                     void         **valary2,
                     unsigned int  *validx,
                     const char    *sigstr,
                     unsigned int  *sigidx);

/*
 * Allocates a value, reallocates the values array and stores the allocated
 * value in it.
 *
 * valary = The values array (also return).
 * valcnt = The values array size (also return).
 * length = The value length in bytes (may be 0).
 * return = 0 on success or -1 on any error with errno set appropriately.
 */
int rmix_value_malloc (void         ***valary,
                       unsigned int   *valcnt,
                       unsigned int    length);

/******************************************************************************
 *
 * Data
 *
 ******************************************************************************/

/* None at this time. */

#endif /* RMIX_VALARY_H */

/******************************************************************************
 *
 * END OF FILE
 *
 ******************************************************************************/
