/******************************************************************************
 *
 * Source file for rmixregistryclient (main) module.
 * Copyright (c) 2002-2006 Oak Ridge National Laboratory.
 *
 * For more information see the following files in the source distribution top-
 * level directory or package data directory (usually /usr/local/share/package):
 *
 * - README    for general package information.
 * - INSTALL   for package install information.
 * - COPYING   for package license information and copying conditions.
 * - AUTHORS   for package authors information.
 * - ChangeLog for package changes information.
 *
 ******************************************************************************/

/******************************************************************************
 *
 * Macros
 *
 ******************************************************************************/

/* None at this time. */

/******************************************************************************
 *
 * Includes
 *
 ******************************************************************************/

/* Main module header file. */
#include <rmixregistryclient.h>

/******************************************************************************
 *
 * Data Types
 *
 ******************************************************************************/

/* None at this time. */

/******************************************************************************
 *
 * Function Prototypes
 *
 ******************************************************************************/

/* None at this time. */

/******************************************************************************
 *
 * Data
 *
 ******************************************************************************/

/* Server-side method descriptors for rmix registry. */
const rmix_method_t rmixregistryclient_methods[RMIXREGISTRY_METHODS_COUNT] =
  RMIXREGISTRY_METHODS_CLIENT;

/* Server-side interface for rmix registry. */
const rmix_interface_t rmixregistryclient_interface = {
  RMIXREGISTRY_METHODS_COUNT,                 /* method descriptor count */
  (rmix_method_t*)rmixregistryclient_methods  /* method descriptor array */
};

/******************************************************************************
 *
 * Functions
 *
 ******************************************************************************/

/*
 * Invokes rmixregistry_bind (client-side method stub).
 * Binds a name with a remote object reference at a remote RMIX registry.
 *
 * registry  = The RMIX remote object reference of the RMIX registry.
 * name      = The name to bind.
 * remoteref = The RMIX remote object reference to bind.
 * return    = 0 on success or -1 on any error with errno set appropriately.
 */
#undef  __FUNC__
#define __FUNC__ "rmixregistryclient_bind"
int rmixregistryclient_bind (const rmix_remoteref_t *registry,
                             const char             *name,
                             const rmix_remoteref_t *remoteref) {
  int         result;
  void       *outary[1];
  const void *inary[2], *inary_remoteref[6];
#ifdef DEBUG
  /* Check registry parameter. */
  if (NULL == registry) {
    errno = EINVAL;
    RMIX_LOG((RMIX_WARN(registry parameter is null)))
    errno = EINVAL;
    return -1;
  }
  /* Check name parameter. */
  if (NULL == name) {
    errno = EINVAL;
    RMIX_LOG((RMIX_WARN(name parameter is null)))
    errno = EINVAL;
    return -1;
  }
  /* Check remoteref parameter. */
  if (NULL == remoteref) {
    errno = EINVAL;
    RMIX_LOG((RMIX_WARN(remoteref parameter is null)))
    errno = EINVAL;
    return -1;
  }
#endif /* DEBUG */
  /* Prepare input. */
  inary[0]           =  name;
  inary[1]           =  inary_remoteref;
  inary_remoteref[0] =  remoteref->protocol;
  inary_remoteref[1] =  remoteref->binding;
  inary_remoteref[2] = &remoteref->interfaces.count;
  inary_remoteref[3] =  remoteref->interfaces.array;
  inary_remoteref[4] =  remoteref->location;
  inary_remoteref[5] =  remoteref->object;
  /* Prepare output. */
  outary[0] = &result;
  /* Invoke remote object method. */
  if (0 != rmix_invoke(registry, &rmixregistryclient_interface,
                       RMIXREGISTRY_METHODS_BIND_INDEX, outary, 1, inary, 2)) {
    int errno2 = errno;
    RMIX_LOG((RMIX_WARN(unable to invoke remote object method)))
    errno = errno2;
    return -1;
  }
  return result;
}

/*
 * Invokes rmixregistry_bind (client-side method stub).
 * Binds a name with a remote object reference at a remote RMIX registry
 * without loading the provider plug-in.
 *
 * registry  = The RMIX remote object reference of the RMIX registry.
 * provider  = The preloaded RMIX provider plug-in handle.
 * name      = The name to bind.
 * remoteref = The RMIX remote object reference to bind.
 * return    = 0 on success or -1 on any error with errno set appropriately.
 */
#undef  __FUNC__
#define __FUNC__ "rmixregistryclient_bind_p"
int rmixregistryclient_bind_p (const rmix_remoteref_t *registry,
                               unsigned int            provider,
                               const char             *name,
                               const rmix_remoteref_t *remoteref) {
  int         result;
  void       *outary[1];
  const void *inary[2], *inary_remoteref[6];
#ifdef DEBUG
  /* Check registry parameter. */
  if (NULL == registry) {
    errno = EINVAL;
    RMIX_LOG((RMIX_WARN(registry parameter is null)))
    errno = EINVAL;
    return -1;
  }
  /* Check name parameter. */
  if (NULL == name) {
    errno = EINVAL;
    RMIX_LOG((RMIX_WARN(name parameter is null)))
    errno = EINVAL;
    return -1;
  }
  /* Check remoteref parameter. */
  if (NULL == remoteref) {
    errno = EINVAL;
    RMIX_LOG((RMIX_WARN(remoteref parameter is null)))
    errno = EINVAL;
    return -1;
  }
#endif /* DEBUG */
  /* Prepare input. */
  inary[0]           =  name;
  inary[1]           =  inary_remoteref;
  inary_remoteref[0] =  remoteref->protocol;
  inary_remoteref[1] =  remoteref->binding;
  inary_remoteref[2] = &remoteref->interfaces.count;
  inary_remoteref[3] =  remoteref->interfaces.array;
  inary_remoteref[4] =  remoteref->location;
  inary_remoteref[5] =  remoteref->object;
  /* Prepare output. */
  outary[0] = &result;
  /* Invoke remote object method. */
  if (0 != rmix_invoke_p(registry, provider,
                         &rmixregistryclient_interface,
                         RMIXREGISTRY_METHODS_BIND_INDEX,
                         outary, 1, inary, 2)) {
    int errno2 = errno;
    RMIX_LOG((RMIX_WARN(unable to invoke remote object method)))
    errno = errno2;
    return -1;
  }
  return result;
}

/*
 * Invokes rmixregistry_bind (client-side method stub).
 * Binds a name with a remote object reference at a remote RMIX registry using
 * a one-way invocation without loading the provider plug-in.
 *
 * registry  = The RMIX remote object reference of the RMIX registry.
 * name      = The name to bind.
 * remoteref = The RMIX remote object reference to bind.
 * return    = 0 on success or -1 on any error with errno set appropriately.
 *             (-1 on local errors only due to one-way invocation)
 */
#undef  __FUNC__
#define __FUNC__ "rmixregistryclient_bind_oneway"
int rmixregistryclient_bind_oneway (const rmix_remoteref_t *registry,
                                    const char             *name,
                                    const rmix_remoteref_t *remoteref) {
  const void *inary[2], *inary_remoteref[6];
#ifdef DEBUG
  /* Check registry parameter. */
  if (NULL == registry) {
    errno = EINVAL;
    RMIX_LOG((RMIX_WARN(registry parameter is null)))
    errno = EINVAL;
    return -1;
  }
  /* Check name parameter. */
  if (NULL == name) {
    errno = EINVAL;
    RMIX_LOG((RMIX_WARN(name parameter is null)))
    errno = EINVAL;
    return -1;
  }
  /* Check remoteref parameter. */
  if (NULL == remoteref) {
    errno = EINVAL;
    RMIX_LOG((RMIX_WARN(remoteref parameter is null)))
    errno = EINVAL;
    return -1;
  }
#endif /* DEBUG */
  /* Prepare parameters. */
  inary[0]           =  name;
  inary[1]           =  inary_remoteref;
  inary_remoteref[0] =  remoteref->protocol;
  inary_remoteref[1] =  remoteref->binding;
  inary_remoteref[2] = &remoteref->interfaces.count;
  inary_remoteref[3] =  remoteref->interfaces.array;
  inary_remoteref[4] =  remoteref->location;
  inary_remoteref[5] =  remoteref->object;
  /* Invoke remote object method. */
  if (0 != rmix_oneway(registry, &rmixregistryclient_interface,
                       RMIXREGISTRY_METHODS_BIND_INDEX, inary, 2)) {
    int errno2 = errno;
    RMIX_LOG((RMIX_WARN(unable to invoke remote object method)))
    errno = errno2;
    return -1;
  }
  return 0;
}

/*
 * Invokes rmixregistry_bind (client-side method stub).
 * Binds a name with a remote object reference at a remote RMIX registry using
 * a one-way invocation without loading the provider plug-in.
 *
 * registry  = The RMIX remote object reference of the RMIX registry.
 * provider  = The preloaded RMIX provider plug-in handle.
 * name      = The name to bind.
 * remoteref = The RMIX remote object reference to bind.
 * return    = 0 on success or -1 on any error with errno set appropriately.
 *             (-1 on local errors only due to one-way invocation)
 */
#undef  __FUNC__
#define __FUNC__ "rmixregistryclient_bind_oneway_p"
int rmixregistryclient_bind_oneway_p (const rmix_remoteref_t *registry,
                                      unsigned int            provider,
                                      const char             *name,
                                      const rmix_remoteref_t *remoteref) {
  const void *inary[2], *inary_remoteref[6];
#ifdef DEBUG
  /* Check registry parameter. */
  if (NULL == registry) {
    errno = EINVAL;
    RMIX_LOG((RMIX_WARN(registry parameter is null)))
    errno = EINVAL;
    return -1;
  }
  /* Check name parameter. */
  if (NULL == name) {
    errno = EINVAL;
    RMIX_LOG((RMIX_WARN(name parameter is null)))
    errno = EINVAL;
    return -1;
  }
  /* Check remoteref parameter. */
  if (NULL == remoteref) {
    errno = EINVAL;
    RMIX_LOG((RMIX_WARN(remoteref parameter is null)))
    errno = EINVAL;
    return -1;
  }
#endif /* DEBUG */
  /* Prepare parameters. */
  inary[0]           =  name;
  inary[1]           =  inary_remoteref;
  inary_remoteref[0] =  remoteref->protocol;
  inary_remoteref[1] =  remoteref->binding;
  inary_remoteref[2] = &remoteref->interfaces.count;
  inary_remoteref[3] =  remoteref->interfaces.array;
  inary_remoteref[4] =  remoteref->location;
  inary_remoteref[5] =  remoteref->object;
  /* Invoke remote object method. */
  if (0 != rmix_oneway_p(registry, provider,
                         &rmixregistryclient_interface,
                         RMIXREGISTRY_METHODS_BIND_INDEX, inary, 2)) {
    int errno2 = errno;
    RMIX_LOG((RMIX_WARN(unable to invoke remote object method)))
    errno = errno2;
    return -1;
  }
  return 0;
}

/*
 * Invokes rmixregistry_bind (client-side method stub).
 * Binds a name with a remote object reference at a remote RMIX registry using
 * an asynchronous invocation.
 *
 * invokeref = The invocation reference (return).
 * registry  = The RMIX remote object reference of the RMIX registry.
 * name      = The name to bind.
 * remoteref = The RMIX remote object reference to bind.
 * return    = 0 on success or -1 on any error with errno set appropriately.
 */
#undef  __FUNC__
#define __FUNC__ "rmixregistryclient_bind_send"
int rmixregistryclient_bind_send (rmix_invokeref_t       **invokeref,
                                  const rmix_remoteref_t  *registry,
                                  const char              *name,
                                  const rmix_remoteref_t  *remoteref) {
  const void *inary[2], *inary_remoteref[6];
#ifdef DEBUG
  /* Check invokeref parameter. */
  if (NULL == invokeref) {
    errno = EINVAL;
    RMIX_LOG((RMIX_WARN(invokeref parameter is null)))
    errno = EINVAL;
    return -1;
  }
  /* Check registry parameter. */
  if (NULL == registry) {
    errno = EINVAL;
    RMIX_LOG((RMIX_WARN(registry parameter is null)))
    errno = EINVAL;
    return -1;
  }
  /* Check name parameter. */
  if (NULL == name) {
    errno = EINVAL;
    RMIX_LOG((RMIX_WARN(name parameter is null)))
    errno = EINVAL;
    return -1;
  }
  /* Check remoteref parameter. */
  if (NULL == remoteref) {
    errno = EINVAL;
    RMIX_LOG((RMIX_WARN(remoteref parameter is null)))
    errno = EINVAL;
    return -1;
  }
#endif /* DEBUG */
  /* Prepare parameters. */
  inary[0]           =  name;
  inary[1]           =  inary_remoteref;
  inary_remoteref[0] =  remoteref->protocol;
  inary_remoteref[1] =  remoteref->binding;
  inary_remoteref[2] = &remoteref->interfaces.count;
  inary_remoteref[3] =  remoteref->interfaces.array;
  inary_remoteref[4] =  remoteref->location;
  inary_remoteref[5] =  remoteref->object;
  /* Invoke remote object method. */
  if (0 != rmix_send(invokeref, registry,
                     &rmixregistryclient_interface,
                     RMIXREGISTRY_METHODS_BIND_INDEX, inary, 2)) {
    int errno2 = errno;
    RMIX_LOG((RMIX_WARN(unable to invoke remote object method)))
    errno = errno2;
    return -1;
  }
  return 0;
}

/*
 * Invokes rmixregistry_bind (client-side method stub).
 * Binds a name with a remote object reference at a remote RMIX registry using
 * an asynchronous invocation without loading the provider plug-in.
 *
 * invokeref = The invocation reference (return).
 * registry  = The RMIX remote object reference of the RMIX registry.
 * provider  = The preloaded RMIX provider plug-in handle.
 * name      = The name to bind.
 * remoteref = The RMIX remote object reference to bind.
 * return    = 0 on success or -1 on any error with errno set appropriately.
 */
#undef  __FUNC__
#define __FUNC__ "rmixregistryclient_bind_send_p"
int rmixregistryclient_bind_send_p (rmix_invokeref_t       **invokeref,
                                    const rmix_remoteref_t  *registry,
                                    unsigned int             provider,
                                    const char              *name,
                                    const rmix_remoteref_t  *remoteref) {
  const void *inary[2], *inary_remoteref[6];
#ifdef DEBUG
  /* Check invokeref parameter. */
  if (NULL == invokeref) {
    errno = EINVAL;
    RMIX_LOG((RMIX_WARN(invokeref parameter is null)))
    errno = EINVAL;
    return -1;
  }
  /* Check registry parameter. */
  if (NULL == registry) {
    errno = EINVAL;
    RMIX_LOG((RMIX_WARN(registry parameter is null)))
    errno = EINVAL;
    return -1;
  }
  /* Check name parameter. */
  if (NULL == name) {
    errno = EINVAL;
    RMIX_LOG((RMIX_WARN(name parameter is null)))
    errno = EINVAL;
    return -1;
  }
  /* Check remoteref parameter. */
  if (NULL == remoteref) {
    errno = EINVAL;
    RMIX_LOG((RMIX_WARN(remoteref parameter is null)))
    errno = EINVAL;
    return -1;
  }
#endif /* DEBUG */
  /* Prepare parameters. */
  inary[0]           =  name;
  inary[1]           =  inary_remoteref;
  inary_remoteref[0] =  remoteref->protocol;
  inary_remoteref[1] =  remoteref->binding;
  inary_remoteref[2] = &remoteref->interfaces.count;
  inary_remoteref[3] =  remoteref->interfaces.array;
  inary_remoteref[4] =  remoteref->location;
  inary_remoteref[5] =  remoteref->object;
  /* Invoke remote object method. */
  if (0 != rmix_send_p(invokeref, registry, provider,
                       &rmixregistryclient_interface,
                       RMIXREGISTRY_METHODS_BIND_INDEX, inary, 2)) {
    int errno2 = errno;
    RMIX_LOG((RMIX_WARN(unable to invoke remote object method)))
    errno = errno2;
    return -1;
  }
  return 0;
}

/*
 * Retrieves rmixregistry_bind invocation output (client-side method stub).
 * Returns the output of a previous asynchronous invocation that binds a name
 * with a remote object reference at a remote RMIX registry.
 *
 * invokeref = The invocation reference (also return).
 * return    = 0 on success or -1 on any error with errno set appropriately.
 */
#undef  __FUNC__
#define __FUNC__ "rmixregistryclient_bind_retrieve"
int rmixregistryclient_bind_retrieve (rmix_invokeref_t **invokeref) {
  int   result;
  void *outary[1];
#ifdef DEBUG
  /* Check invokeref parameter. */
  if (NULL == invokeref) {
    errno = EINVAL;
    RMIX_LOG((RMIX_WARN(invokeref parameter is null)))
    errno = EINVAL;
    return -1;
  }
#endif /* DEBUG */
  /* Prepare output. */
  outary[0] = &result;
  /* Retrieve invocation result. */
  if (0 != rmix_retrieve(invokeref, outary, 1)) {
    int errno2 = errno;
    RMIX_LOG((RMIX_WARN(unable to invoke remote object method)))
    errno = errno2;
    return -1;
  }
  return result;
}

/*
 * Retrieves rmixregistry_bind invocation output (client-side method stub).
 * Returns the output of a previous asynchronous invocation that binds a name
 * with a remote object reference at a remote RMIX registry without unloading
 * the provider plug-in.
 *
 * invokeref = The invocation reference (also return).
 * return    = 0 on success or -1 on any error with errno set appropriately.
 */
#undef  __FUNC__
#define __FUNC__ "rmixregistryclient_bind_retrieve_p"
int rmixregistryclient_bind_retrieve_p (rmix_invokeref_t **invokeref) {
  int   result;
  void *outary[1];
#ifdef DEBUG
  /* Check invokeref parameter. */
  if (NULL == invokeref) {
    errno = EINVAL;
    RMIX_LOG((RMIX_WARN(invokeref parameter is null)))
    errno = EINVAL;
    return -1;
  }
#endif /* DEBUG */
  /* Prepare output. */
  outary[0] = &result;
  /* Retrieve invocation result. */
  if (0 != rmix_retrieve_p(invokeref, outary, 1)) {
    int errno2 = errno;
    RMIX_LOG((RMIX_WARN(unable to invoke remote object method)))
    errno = errno2;
    return -1;
  }
  return result;
}

/*
 * Invokes rmixregistry_list (client-side method stub).
 * Returns an array of the names in a remote RMIX registry
 *
 * registry = The RMIX remote object reference of the RMIX registry.
 * count    = The number of names in the registry (return).
 * names    = The NULL terminated array of names (return).
 * return   = 0 on success or -1 on any error with errno set appropriately.
 */
#undef  __FUNC__
#define __FUNC__ "rmixregistryclient_list"
int rmixregistryclient_list (const rmix_remoteref_t   *registry,
                             unsigned int             *count,
                             char                   ***names) {
  int   result;
  void *outary[3];
#ifdef DEBUG
  /* Check registry parameter. */
  if (NULL == registry) {
    errno = EINVAL;
    RMIX_LOG((RMIX_WARN(registry parameter is null)))
    errno = EINVAL;
    return -1;
  }
  /* Check count parameter. */
  if (NULL == count) {
    errno = EINVAL;
    RMIX_LOG((RMIX_WARN(count parameter is null)))
    errno = EINVAL;
    return -1;
  }
  /* Check names parameter. */
  if (NULL == names) {
    errno = EINVAL;
    RMIX_LOG((RMIX_WARN(names parameter is null)))
    errno = EINVAL;
    return -1;
  }
#endif /* DEBUG */
  /* Prepare output. */
  *count    =  0;
  outary[0] = &result;
  outary[1] =  count;
  outary[2] =  NULL;
  /* Invoke remote object method. */
  if (0 != rmix_invoke(registry, &rmixregistryclient_interface,
                       RMIXREGISTRY_METHODS_LIST_INDEX, outary, 3, NULL, 0)) {
    int errno2 = errno;
    RMIX_LOG((RMIX_WARN(unable to invoke remote object method)))
    errno = errno2;
    return -1;
  }
  /* Save output. */
  *names = (char**)outary[2];
  return result;
}

/*
 * Invokes rmixregistry_list (client-side method stub).
 * Returns an array of the names in a remote RMIX registry without loading the
 * provider plug-in.
 *
 * registry = The RMIX remote object reference of the RMIX registry.
 * provider = The preloaded RMIX provider plug-in handle.
 * count    = The number of names in the registry (return).
 * names    = The NULL terminated array of names (return).
 * return   = 0 on success or -1 on any error with errno set appropriately.
 */
#undef  __FUNC__
#define __FUNC__ "rmixregistryclient_list_p"
int rmixregistryclient_list_p (const rmix_remoteref_t   *registry,
                               unsigned int              provider,
                               unsigned int             *count,
                               char                   ***names){
  int   result;
  void *outary[3];
#ifdef DEBUG
  /* Check registry parameter. */
  if (NULL == registry) {
    errno = EINVAL;
    RMIX_LOG((RMIX_WARN(registry parameter is null)))
    errno = EINVAL;
    return -1;
  }
  /* Check count parameter. */
  if (NULL == count) {
    errno = EINVAL;
    RMIX_LOG((RMIX_WARN(count parameter is null)))
    errno = EINVAL;
    return -1;
  }
  /* Check names parameter. */
  if (NULL == names) {
    errno = EINVAL;
    RMIX_LOG((RMIX_WARN(names parameter is null)))
    errno = EINVAL;
    return -1;
  }
#endif /* DEBUG */
  /* Prepare output. */
  *count    =  0;
  outary[0] = &result;
  outary[1] =  count;
  outary[2] =  NULL;
  /* Invoke remote object method. */
  if (0 != rmix_invoke_p(registry, provider,
                         &rmixregistryclient_interface,
                         RMIXREGISTRY_METHODS_LIST_INDEX,
                         outary, 3, NULL, 0)) {
    int errno2 = errno;
    RMIX_LOG((RMIX_WARN(unable to invoke remote object method)))
    errno = errno2;
    return -1;
  }
  /* Save output. */
  *names = (char**)outary[2];
  return result;
}

/*
 * Invokes rmixregistry_list (client-side method stub).
 * Returns an array of the names in a remote RMIX registry using an asynchronous
 * invocation.
 *
 * invokeref = The invocation reference (return).
 * registry  = The RMIX remote object reference of the RMIX registry.
 * return    = 0 on success or -1 on any error with errno set appropriately.
 */
#undef  __FUNC__
#define __FUNC__ "rmixregistryclient_list_send"
int rmixregistryclient_list_send (rmix_invokeref_t       **invokeref,
                                  const rmix_remoteref_t  *registry) {
#ifdef DEBUG
  /* Check invokeref parameter. */
  if (NULL == invokeref) {
    errno = EINVAL;
    RMIX_LOG((RMIX_WARN(invokeref parameter is null)))
    errno = EINVAL;
    return -1;
  }
  /* Check registry parameter. */
  if (NULL == registry) {
    errno = EINVAL;
    RMIX_LOG((RMIX_WARN(registry parameter is null)))
    errno = EINVAL;
    return -1;
  }
#endif /* DEBUG */
  /* Invoke remote object method. */
  if (0 != rmix_send(invokeref, registry,
                     &rmixregistryclient_interface,
                     RMIXREGISTRY_METHODS_LIST_INDEX, NULL, 0)) {
    int errno2 = errno;
    RMIX_LOG((RMIX_WARN(unable to invoke remote object method)))
    errno = errno2;
    return -1;
  }
  return 0;
}

/*
 * Invokes rmixregistry_list (client-side method stub).
 * Returns an array of the names in in a remote RMIX registry using an
 * asynchronous invocation without loading the provider plug-in.
 *
 * invokeref = The invocation reference (return).
 * registry  = The RMIX remote object reference of the RMIX registry.
 * provider  = The preloaded RMIX provider plug-in handle.
 * return    = 0 on success or -1 on any error with errno set appropriately.
 */
#undef  __FUNC__
#define __FUNC__ "rmixregistryclient_list_send_p"
int rmixregistryclient_list_send_p (rmix_invokeref_t       **invokeref,
                                    const rmix_remoteref_t  *registry,
                                    unsigned int             provider) {
#ifdef DEBUG
  /* Check invokeref parameter. */
  if (NULL == invokeref) {
    errno = EINVAL;
    RMIX_LOG((RMIX_WARN(invokeref parameter is null)))
    errno = EINVAL;
    return -1;
  }
  /* Check registry parameter. */
  if (NULL == registry) {
    errno = EINVAL;
    RMIX_LOG((RMIX_WARN(registry parameter is null)))
    errno = EINVAL;
    return -1;
  }
#endif /* DEBUG */
  /* Invoke remote object method. */
  if (0 != rmix_send_p(invokeref, registry, provider,
                       &rmixregistryclient_interface,
                       RMIXREGISTRY_METHODS_LIST_INDEX, NULL, 0)) {
    int errno2 = errno;
    RMIX_LOG((RMIX_WARN(unable to invoke remote object method)))
    errno = errno2;
    return -1;
  }
  return 0;
}

/*
 * Retrieves rmixregistry_list invocation output (client-side method stub).
 * Returns the output of a previous asynchronous invocation that returns an
 * array of the names in a remote RMIX registry.
 *
 * invokeref = The invocation reference (also return).
 * count     = The number of names in the registry (return).
 * names     = The NULL terminated array of names (return).
 * return    = 0 on success or -1 on any error with errno set appropriately.
 */
#undef  __FUNC__
#define __FUNC__ "rmixregistryclient_list_retrieve"
int rmixregistryclient_list_retrieve (rmix_invokeref_t  **invokeref,
                                      unsigned int       *count,
                                      char             ***names) {
  int   result;
  void *outary[3];
#ifdef DEBUG
  /* Check invokeref parameter. */
  if (NULL == invokeref) {
    errno = EINVAL;
    RMIX_LOG((RMIX_WARN(invokeref parameter is null)))
    errno = EINVAL;
    return -1;
  }
  /* Check count parameter. */
  if (NULL == count) {
    errno = EINVAL;
    RMIX_LOG((RMIX_WARN(count parameter is null)))
    errno = EINVAL;
    return -1;
  }
  /* Check names parameter. */
  if (NULL == names) {
    errno = EINVAL;
    RMIX_LOG((RMIX_WARN(names parameter is null)))
    errno = EINVAL;
    return -1;
  }
#endif /* DEBUG */
  /* Prepare output. */
  *count    =  0;
  outary[0] = &result;
  outary[1] =  count;
  outary[2] =  NULL;
  /* Retrieve invocation result. */
  if (0 != rmix_retrieve(invokeref, outary, 3)) {
    int errno2 = errno;
    RMIX_LOG((RMIX_WARN(unable to invoke remote object method)))
    errno = errno2;
    return -1;
  }
  /* Save output. */
  *names = (char**)outary[2];
  return result;
}

/*
 * Retrieves rmixregistry_list invocation output (client-side method stub).
 * Returns the output of a previous asynchronous invocation that returns an
 * array of the names in a remote RMIX registry without unloading the provider
 * plug-in.
 *
 * invokeref = The invocation reference (also return).
 * count     = The number of names in the registry (return).
 * names     = The NULL terminated array of names (return).
 * return    = 0 on success or -1 on any error with errno set appropriately.
 */
#undef  __FUNC__
#define __FUNC__ "rmixregistryclient_list_retrieve_p"
int rmixregistryclient_list_retrieve_p (rmix_invokeref_t  **invokeref,
                                        unsigned int       *count,
                                        char             ***names) {
  int   result;
  void *outary[3];
#ifdef DEBUG
  /* Check invokeref parameter. */
  if (NULL == invokeref) {
    errno = EINVAL;
    RMIX_LOG((RMIX_WARN(invokeref parameter is null)))
    errno = EINVAL;
    return -1;
  }
  /* Check count parameter. */
  if (NULL == count) {
    errno = EINVAL;
    RMIX_LOG((RMIX_WARN(count parameter is null)))
    errno = EINVAL;
    return -1;
  }
  /* Check names parameter. */
  if (NULL == names) {
    errno = EINVAL;
    RMIX_LOG((RMIX_WARN(names parameter is null)))
    errno = EINVAL;
    return -1;
  }
#endif /* DEBUG */
  /* Prepare output. */
  *count    =  0;
  outary[0] = &result;
  outary[1] =  count;
  outary[2] =  NULL;
  /* Retrieve invocation result. */
  if (0 != rmix_retrieve_p(invokeref, outary, 3)) {
    int errno2 = errno;
    RMIX_LOG((RMIX_WARN(unable to invoke remote object method)))
    errno = errno2;
    return -1;
  }
  /* Save output. */
  *names = (char**)outary[2];
  return result;
}

/*
 * Invokes rmixregistry_lookup (client-side method stub).
 * Returns the remote object reference associated with a name  at a remote RMIX
 * registry.
 *
 * registry  = The RMIX remote object reference of the RMIX registry.
 * name      = The name to lookup.
 * remoteref = The RMIX remote object reference (return).
 * return    = 0 on success or -1 on any error with errno set appropriately.
 */
#undef  __FUNC__
#define __FUNC__ "rmixregistryclient_lookup"
int rmixregistryclient_lookup (const rmix_remoteref_t  *registry,
                               const char              *name,
                               rmix_remoteref_t       **remoteref) {
  int         result;
  void       *outary[2], *outary_remoteref[6];
  const void *inary[1];
#ifdef DEBUG
  /* Check registry parameter. */
  if (NULL == registry) {
    errno = EINVAL;
    RMIX_LOG((RMIX_WARN(registry parameter is null)))
    errno = EINVAL;
    return -1;
  }
  /* Check name parameter. */
  if (NULL == name) {
    errno = EINVAL;
    RMIX_LOG((RMIX_WARN(name parameter is null)))
    errno = EINVAL;
    return -1;
  }
  /* Check remoteref parameter. */
  if (NULL == remoteref) {
    errno = EINVAL;
    RMIX_LOG((RMIX_WARN(remoteref parameter is null)))
    errno = EINVAL;
    return -1;
  }
#endif /* DEBUG */
  /* Allocate remote object reference. */
  if (NULL == (*remoteref = malloc(sizeof(rmix_remoteref_t)))) {
    int errno2 = errno;
    RMIX_LOG((RMIX_WARN(unable to allocate remote object reference)))
    errno = errno2;
    return -1;
  }
  /* Prepare input. */
  inary[0]            =  name;
  /* Prepare output. */
  outary[0]           = &result;
  outary[1]           =  outary_remoteref;
  outary_remoteref[0] =  NULL;
  outary_remoteref[1] =  NULL;
  outary_remoteref[2] = &(*remoteref)->interfaces.count;
  outary_remoteref[3] =  NULL;
  outary_remoteref[4] =  NULL;
  outary_remoteref[5] =  NULL;
  /* Invoke remote object method. */
  if (0 != rmix_invoke(registry, &rmixregistryclient_interface,
                       RMIXREGISTRY_METHODS_LOOKUP_INDEX,
                       outary, 2, inary, 1)) {
    int errno2 = errno;
    RMIX_LOG((RMIX_WARN(unable to invoke remote object method)))
    /* Deallocate remote object reference. */
    free(*remoteref);
    errno = errno2;
    return -1;
  }
  /* Save output. */
  (*remoteref)->protocol         = (char* )outary_remoteref[0];
  (*remoteref)->binding          = (char* )outary_remoteref[1];
  (*remoteref)->interfaces.array = (char**)outary_remoteref[3];
  (*remoteref)->location         = (char* )outary_remoteref[4];
  (*remoteref)->object           = (char* )outary_remoteref[5];
  return result;
}

/*
 * Invokes rmixregistry_lookup (client-side method stub).
 * Returns the remote object reference associated with a name at a remote RMIX
 * registry without loading the provider plug-in.
 *
 * registry  = The RMIX remote object reference of the RMIX registry.
 * provider  = The preloaded RMIX provider plug-in handle.
 * name      = The name to lookup.
 * remoteref = The RMIX remote object reference (return).
 * return    = 0 on success or -1 on any error with errno set appropriately.
 */
#undef  __FUNC__
#define __FUNC__ "rmixregistryclient_lookup_p"
int rmixregistryclient_lookup_p (const rmix_remoteref_t  *registry,
                                 unsigned int             provider,
                                 const char              *name,
                                 rmix_remoteref_t       **remoteref) {
  int         result;
  void       *outary[2], *outary_remoteref[6];
  const void *inary[1];
#ifdef DEBUG
  /* Check registry parameter. */
  if (NULL == registry) {
    errno = EINVAL;
    RMIX_LOG((RMIX_WARN(registry parameter is null)))
    errno = EINVAL;
    return -1;
  }
  /* Check name parameter. */
  if (NULL == name) {
    errno = EINVAL;
    RMIX_LOG((RMIX_WARN(name parameter is null)))
    errno = EINVAL;
    return -1;
  }
  /* Check remoteref parameter. */
  if (NULL == remoteref) {
    errno = EINVAL;
    RMIX_LOG((RMIX_WARN(remoteref parameter is null)))
    errno = EINVAL;
    return -1;
  }
#endif /* DEBUG */
  /* Allocate remote object reference. */
  if (NULL == (*remoteref = malloc(sizeof(rmix_remoteref_t)))) {
    int errno2 = errno;
    RMIX_LOG((RMIX_WARN(unable to allocate remote object reference)))
    errno = errno2;
    return -1;
  }
  /* Prepare input. */
  inary[0]            =  name;
  /* Prepare output. */
  outary[0]           = &result;
  outary[1]           =  outary_remoteref;
  outary_remoteref[0] =  NULL;
  outary_remoteref[1] =  NULL;
  outary_remoteref[2] = &(*remoteref)->interfaces.count;
  outary_remoteref[3] =  NULL;
  outary_remoteref[4] =  NULL;
  outary_remoteref[5] =  NULL;
  /* Invoke remote object method. */
  if (0 != rmix_invoke_p(registry, provider,
                         &rmixregistryclient_interface,
                         RMIXREGISTRY_METHODS_LOOKUP_INDEX,
                         outary, 2, inary, 1)) {
    int errno2 = errno;
    RMIX_LOG((RMIX_WARN(unable to invoke remote object method)))
    /* Deallocate remote object reference. */
    free(*remoteref);
    errno = errno2;
    return -1;
  }
  /* Save output. */
  (*remoteref)->protocol         = (char* )outary_remoteref[0];
  (*remoteref)->binding          = (char* )outary_remoteref[1];
  (*remoteref)->interfaces.array = (char**)outary_remoteref[3];
  (*remoteref)->location         = (char* )outary_remoteref[4];
  (*remoteref)->object           = (char* )outary_remoteref[5];
  return result;
}

/*
 * Invokes rmixregistry_lookup (client-side method stub).
 * Returns the remote object reference associated with a name at a remote RMIX
 * registry using an asynchronous invocation.
 *
 * invokeref = The invocation reference (return).
 * registry  = The RMIX remote object reference of the RMIX registry.
 * name      = The name to lookup.
 * return    = 0 on success or -1 on any error with errno set appropriately.
 */
#undef  __FUNC__
#define __FUNC__ "rmixregistryclient_lookup_send"
int rmixregistryclient_lookup_send (rmix_invokeref_t       **invokeref,
                                    const rmix_remoteref_t  *registry,
                                    const char              *name) {
  const void *inary[1];
#ifdef DEBUG
  /* Check invokeref parameter. */
  if (NULL == invokeref) {
    errno = EINVAL;
    RMIX_LOG((RMIX_WARN(invokeref parameter is null)))
    errno = EINVAL;
    return -1;
  }
  /* Check registry parameter. */
  if (NULL == registry) {
    errno = EINVAL;
    RMIX_LOG((RMIX_WARN(registry parameter is null)))
    errno = EINVAL;
    return -1;
  }
  /* Check name parameter. */
  if (NULL == name) {
    errno = EINVAL;
    RMIX_LOG((RMIX_WARN(name parameter is null)))
    errno = EINVAL;
    return -1;
  }
#endif /* DEBUG */
  /* Prepare input. */
  inary[0] = name;
  /* Invoke remote object method. */
  if (0 != rmix_send(invokeref, registry,
                     &rmixregistryclient_interface,
                     RMIXREGISTRY_METHODS_LOOKUP_INDEX, inary, 1)) {
    int errno2 = errno;
    RMIX_LOG((RMIX_WARN(unable to invoke remote object method)))
    errno = errno2;
    return -1;
  }
  return 0;
}

/*
 * Invokes rmixregistry_lookup (client-side method stub).
 * Returns the remote object reference associated with a name at a remote RMIX
 * registry using an asynchronous invocation without loading the provider
 * plug-in.
 *
 * invokeref = The invocation reference (return).
 * registry  = The RMIX remote object reference of the RMIX registry.
 * provider  = The preloaded RMIX provider plug-in handle.
 * name      = The name to lookup.
 * return    = 0 on success or -1 on any error with errno set appropriately.
 */
#undef  __FUNC__
#define __FUNC__ "rmixregistryclient_lookup_send_p"
int rmixregistryclient_lookup_send_p (rmix_invokeref_t       **invokeref,
                                      const rmix_remoteref_t  *registry,
                                      unsigned int             provider,
                                      const char              *name) {
  const void *inary[1];
#ifdef DEBUG
  /* Check invokeref parameter. */
  if (NULL == invokeref) {
    errno = EINVAL;
    RMIX_LOG((RMIX_WARN(invokeref parameter is null)))
    errno = EINVAL;
    return -1;
  }
  /* Check registry parameter. */
  if (NULL == registry) {
    errno = EINVAL;
    RMIX_LOG((RMIX_WARN(registry parameter is null)))
    errno = EINVAL;
    return -1;
  }
  /* Check name parameter. */
  if (NULL == name) {
    errno = EINVAL;
    RMIX_LOG((RMIX_WARN(name parameter is null)))
    errno = EINVAL;
    return -1;
  }
#endif /* DEBUG */
  /* Prepare input. */
  inary[0] = name;
  /* Invoke remote object method. */
  if (0 != rmix_send_p(invokeref, registry, provider,
                      &rmixregistryclient_interface,
                      RMIXREGISTRY_METHODS_LOOKUP_INDEX, inary, 1)) {
    int errno2 = errno;
    RMIX_LOG((RMIX_WARN(unable to invoke remote object method)))
    errno = errno2;
    return -1;
  }
  return 0;
}

/*
 * Retrieves rmixregistry_lookup invocation output (client-side method stub).
 * Returns the output of a previous asynchronous invocation that returns the
 * remote object reference associated with a name at a remote RMIX registry.
 *
 * invokeref = The invocation reference (also return).
 * remoteref = The RMIX remote object reference (return).
 * return    = 0 on success or -1 on any error with errno set appropriately.
 */
#undef  __FUNC__
#define __FUNC__ "rmixregistryclient_lookup_retrieve"
int rmixregistryclient_lookup_retrieve (rmix_invokeref_t **invokeref,
                                        rmix_remoteref_t **remoteref) {
  int   result;
  void *outary[2], *outary_remoteref[6];
#ifdef DEBUG
  /* Check invokeref parameter. */
  if (NULL == invokeref) {
    errno = EINVAL;
    RMIX_LOG((RMIX_WARN(invokeref parameter is null)))
    errno = EINVAL;
    return -1;
  }
  /* Check remoteref parameter. */
  if (NULL == remoteref) {
    errno = EINVAL;
    RMIX_LOG((RMIX_WARN(remoteref parameter is null)))
    errno = EINVAL;
    return -1;
  }
#endif /* DEBUG */
  /* Allocate remote object reference. */
  if (NULL == (*remoteref = malloc(sizeof(rmix_remoteref_t)))) {
    int errno2 = errno;
    RMIX_LOG((RMIX_WARN(unable to allocate remote object reference)))
    errno = errno2;
    return -1;
  }
  /* Prepare output. */
  outary[0]           = &result;
  outary[1]           =  outary_remoteref;
  outary_remoteref[0] =  NULL;
  outary_remoteref[1] =  NULL;
  outary_remoteref[2] = &(*remoteref)->interfaces.count;
  outary_remoteref[3] =  NULL;
  outary_remoteref[4] =  NULL;
  outary_remoteref[5] =  NULL;
  /* Retrieve invocation result. */
  if (0 != rmix_retrieve(invokeref, outary, 2)) {
    int errno2 = errno;
    RMIX_LOG((RMIX_WARN(unable to invoke remote object method)))
    /* Deallocate remote object reference. */
    free(*remoteref);
    errno = errno2;
    return -1;
  }
  /* Save output. */
  (*remoteref)->protocol         = (char* )outary_remoteref[0];
  (*remoteref)->binding          = (char* )outary_remoteref[1];
  (*remoteref)->interfaces.array = (char**)outary_remoteref[3];
  (*remoteref)->location         = (char* )outary_remoteref[4];
  (*remoteref)->object           = (char* )outary_remoteref[5];
  return result;
}

/*
 * Retrieves rmixregistry_lookup invocation output (client-side method stub).
 * Returns the output of a previous asynchronous invocation that returns the
 * remote object reference associated with a name at a remote RMIX registry
 * without unloading the provider plug-in.
 *
 * invokeref = The invocation reference (also return).
 * remoteref = The RMIX remote object reference (return).
 * return    = 0 on success or -1 on any error with errno set appropriately.
 */
#undef  __FUNC__
#define __FUNC__ "rmixregistryclient_lookup_retrieve_p"
int rmixregistryclient_lookup_retrieve_p (rmix_invokeref_t **invokeref,
                                          rmix_remoteref_t **remoteref) {
  int   result;
  void *outary[2], *outary_remoteref[6];
#ifdef DEBUG
  /* Check invokeref parameter. */
  if (NULL == invokeref) {
    errno = EINVAL;
    RMIX_LOG((RMIX_WARN(invokeref parameter is null)))
    errno = EINVAL;
    return -1;
  }
  /* Check remoteref parameter. */
  if (NULL == remoteref) {
    errno = EINVAL;
    RMIX_LOG((RMIX_WARN(remoteref parameter is null)))
    errno = EINVAL;
    return -1;
  }
#endif /* DEBUG */
  /* Allocate remote object reference. */
  if (NULL == (*remoteref = malloc(sizeof(rmix_remoteref_t)))) {
    int errno2 = errno;
    RMIX_LOG((RMIX_WARN(unable to allocate remote object reference)))
    errno = errno2;
    return -1;
  }
  /* Prepare output. */
  outary[0]           = &result;
  outary[1]           =  outary_remoteref;
  outary_remoteref[0] =  NULL;
  outary_remoteref[1] =  NULL;
  outary_remoteref[2] = &(*remoteref)->interfaces.count;
  outary_remoteref[3] =  NULL;
  outary_remoteref[4] =  NULL;
  outary_remoteref[5] =  NULL;
  /* Retrieve invocation result. */
  if (0 != rmix_retrieve_p(invokeref, outary, 2)) {
    int errno2 = errno;
    RMIX_LOG((RMIX_WARN(unable to invoke remote object method)))
    /* Deallocate remote object reference. */
    free(*remoteref);
    errno = errno2;
    return -1;
  }
  /* Save output. */
  (*remoteref)->protocol         = (char* )outary_remoteref[0];
  (*remoteref)->binding          = (char* )outary_remoteref[1];
  (*remoteref)->interfaces.array = (char**)outary_remoteref[3];
  (*remoteref)->location         = (char* )outary_remoteref[4];
  (*remoteref)->object           = (char* )outary_remoteref[5];
  return result;
}

/*
 * Invokes rmixregistry_rebind (client-side method stub).
 * Rebinds (creates a new or replaces an existing binding) a name with a remote
 * object reference at a remote RMIX registry.
 *
 * registry  = The RMIX remote object reference of the RMIX registry.
 * name      = The name to bind.
 * remoteref = The RMIX remote object reference to bind.
 * return    = 0 on success or -1 on any error with errno set appropriately.
 */
#undef  __FUNC__
#define __FUNC__ "rmixregistryclient_rebind"
int rmixregistryclient_rebind (const rmix_remoteref_t *registry,
                               const char             *name,
                               const rmix_remoteref_t *remoteref) {
  int         result;
  void       *outary[1];
  const void *inary[2], *inary_remoteref[6];
#ifdef DEBUG
  /* Check registry parameter. */
  if (NULL == registry) {
    errno = EINVAL;
    RMIX_LOG((RMIX_WARN(registry parameter is null)))
    errno = EINVAL;
    return -1;
  }
  /* Check name parameter. */
  if (NULL == name) {
    errno = EINVAL;
    RMIX_LOG((RMIX_WARN(name parameter is null)))
    errno = EINVAL;
    return -1;
  }
  /* Check remoteref parameter. */
  if (NULL == remoteref) {
    errno = EINVAL;
    RMIX_LOG((RMIX_WARN(remoteref parameter is null)))
    errno = EINVAL;
    return -1;
  }
#endif /* DEBUG */
  /* Prepare input. */
  inary[0]           =  name;
  inary[1]           =  inary_remoteref;
  inary_remoteref[0] =  remoteref->protocol;
  inary_remoteref[1] =  remoteref->binding;
  inary_remoteref[2] = &remoteref->interfaces.count;
  inary_remoteref[3] =  remoteref->interfaces.array;
  inary_remoteref[4] =  remoteref->location;
  inary_remoteref[5] =  remoteref->object;
  /* Prepare output. */
  outary[0] = &result;
  /* Invoke remote object method. */
  if (0 != rmix_invoke(registry, &rmixregistryclient_interface,
                       RMIXREGISTRY_METHODS_REBIND_INDEX,
                       outary, 1, inary, 2)) {
    int errno2 = errno;
    RMIX_LOG((RMIX_WARN(unable to invoke remote object method)))
    errno = errno2;
    return -1;
  }
  return result;
}

/*
 * Invokes rmixregistry_rebind (client-side method stub).
 * Rebinds (creates a new or replaces an existing binding) a name with a remote
 * object reference at a remote RMIX registry without loading the provider
 * plug-in.
 *
 * registry  = The RMIX remote object reference of the RMIX registry.
 * provider  = The preloaded RMIX provider plug-in handle.
 * name      = The name to bind.
 * remoteref = The RMIX remote object reference to bind.
 * return    = 0 on success or -1 on any error with errno set appropriately.
 */
#undef  __FUNC__
#define __FUNC__ "rmixregistryclient_rebind_p"
int rmixregistryclient_rebind_p (const rmix_remoteref_t *registry,
                                 unsigned int            provider,
                                 const char             *name,
                                 const rmix_remoteref_t *remoteref) {
  int         result;
  void       *outary[1];
  const void *inary[2], *inary_remoteref[6];
#ifdef DEBUG
  /* Check registry parameter. */
  if (NULL == registry) {
    errno = EINVAL;
    RMIX_LOG((RMIX_WARN(registry parameter is null)))
    errno = EINVAL;
    return -1;
  }
  /* Check name parameter. */
  if (NULL == name) {
    errno = EINVAL;
    RMIX_LOG((RMIX_WARN(name parameter is null)))
    errno = EINVAL;
    return -1;
  }
  /* Check remoteref parameter. */
  if (NULL == remoteref) {
    errno = EINVAL;
    RMIX_LOG((RMIX_WARN(remoteref parameter is null)))
    errno = EINVAL;
    return -1;
  }
#endif /* DEBUG */
  /* Prepare input. */
  inary[0]           =  name;
  inary[1]           =  inary_remoteref;
  inary_remoteref[0] =  remoteref->protocol;
  inary_remoteref[1] =  remoteref->binding;
  inary_remoteref[2] = &remoteref->interfaces.count;
  inary_remoteref[3] =  remoteref->interfaces.array;
  inary_remoteref[4] =  remoteref->location;
  inary_remoteref[5] =  remoteref->object;
  /* Prepare output. */
  outary[0] = &result;
  /* Invoke remote object method. */
  if (0 != rmix_invoke_p(registry, provider,
                         &rmixregistryclient_interface,
                         RMIXREGISTRY_METHODS_REBIND_INDEX,
                         outary, 1, inary, 2)) {
    int errno2 = errno;
    RMIX_LOG((RMIX_WARN(unable to invoke remote object method)))
    errno = errno2;
    return -1;
  }
  return result;
}

/*
 * Invokes rmixregistry_rebind (client-side method stub).
 * Rebinds (creates a new or replaces an existing binding) a name with a remote
 * object reference at a remote RMIX registry using a one-way invocation.
 *
 * registry  = The RMIX remote object reference of the RMIX registry.
 * name      = The n ame to bind.
 * remoteref = The RMIX remote object reference to bind.
 * return    = 0 on success or -1 on any error with errno set appropriately.
 *             (-1 on local errors only due to one-way invocation)
 */
#undef  __FUNC__
#define __FUNC__ "rmixregistryclient_rebind_oneway"
int rmixregistryclient_rebind_oneway (const rmix_remoteref_t *registry,
                                      const char             *name,
                                      const rmix_remoteref_t *remoteref) {
  const void *inary[2], *inary_remoteref[6];
#ifdef DEBUG
  /* Check registry parameter. */
  if (NULL == registry) {
    errno = EINVAL;
    RMIX_LOG((RMIX_WARN(registry parameter is null)))
    errno = EINVAL;
    return -1;
  }
  /* Check name parameter. */
  if (NULL == name) {
    errno = EINVAL;
    RMIX_LOG((RMIX_WARN(name parameter is null)))
    errno = EINVAL;
    return -1;
  }
  /* Check remoteref parameter. */
  if (NULL == remoteref) {
    errno = EINVAL;
    RMIX_LOG((RMIX_WARN(remoteref parameter is null)))
    errno = EINVAL;
    return -1;
  }
#endif /* DEBUG */
  /* Prepare input. */
  inary[0]           =  name;
  inary[1]           =  inary_remoteref;
  inary_remoteref[0] =  remoteref->protocol;
  inary_remoteref[1] =  remoteref->binding;
  inary_remoteref[2] = &remoteref->interfaces.count;
  inary_remoteref[3] =  remoteref->interfaces.array;
  inary_remoteref[4] =  remoteref->location;
  inary_remoteref[5] =  remoteref->object;
  /* Invoke remote object method. */
  if (0 != rmix_oneway(registry,
                       &rmixregistryclient_interface,
                       RMIXREGISTRY_METHODS_REBIND_INDEX, inary, 2)) {
    int errno2 = errno;
    RMIX_LOG((RMIX_WARN(unable to invoke remote object method)))
    errno = errno2;
    return -1;
  }
  return 0;
}

/*
 * Invokes rmixregistry_rebind (client-side method stub).
 * Rebinds (creates a new or replaces an existing binding) a name with a remote
 * object reference at a remote RMIX registry using a one-way invocation
 * without loading the provider plug-in.
 *
 * registry  = The RMIX remote object reference of the RMIX registry.
 * provider  = The preloaded RMIX provider plug-in handle.
 * name      = The n ame to bind.
 * remoteref = The RMIX remote object reference to bind.
 * return    = 0 on success or -1 on any error with errno set appropriately.
 *             (-1 on local errors only due to one-way invocation)
 */
#undef  __FUNC__
#define __FUNC__ "rmixregistryclient_rebind_oneway_p"
int rmixregistryclient_rebind_oneway_p (const rmix_remoteref_t *registry,
                                        unsigned int            provider,
                                        const char             *name,
                                        const rmix_remoteref_t *remoteref) {
  const void *inary[2], *inary_remoteref[6];
#ifdef DEBUG
  /* Check registry parameter. */
  if (NULL == registry) {
    errno = EINVAL;
    RMIX_LOG((RMIX_WARN(registry parameter is null)))
    errno = EINVAL;
    return -1;
  }
  /* Check name parameter. */
  if (NULL == name) {
    errno = EINVAL;
    RMIX_LOG((RMIX_WARN(name parameter is null)))
    errno = EINVAL;
    return -1;
  }
  /* Check remoteref parameter. */
  if (NULL == remoteref) {
    errno = EINVAL;
    RMIX_LOG((RMIX_WARN(remoteref parameter is null)))
    errno = EINVAL;
    return -1;
  }
#endif /* DEBUG */
  /* Prepare input. */
  inary[0]           =  name;
  inary[1]           =  inary_remoteref;
  inary_remoteref[0] =  remoteref->protocol;
  inary_remoteref[1] =  remoteref->binding;
  inary_remoteref[2] = &remoteref->interfaces.count;
  inary_remoteref[3] =  remoteref->interfaces.array;
  inary_remoteref[4] =  remoteref->location;
  inary_remoteref[5] =  remoteref->object;
  /* Invoke remote object method. */
  if (0 != rmix_oneway_p(registry, provider,
                         &rmixregistryclient_interface,
                         RMIXREGISTRY_METHODS_REBIND_INDEX, inary, 2)) {
    int errno2 = errno;
    RMIX_LOG((RMIX_WARN(unable to invoke remote object method)))
    errno = errno2;
    return -1;
  }
  return 0;
}

/*
 * Invokes rmixregistry_rebind (client-side method stub).
 * Rebinds (creates a new or replaces an existing binding) a name with a remote
 * object reference at a remote RMIX registry using an asynchronous invocation.
 *
 * invokeref = The invocation reference (return).
 * registry  = The RMIX remote object reference of the RMIX registry.
 * name      = The name to bind.
 * remoteref = The RMIX remote object reference to bind.
 * return    = 0 on success or -1 on any error with errno set appropriately.
 */
#undef  __FUNC__
#define __FUNC__ "rmixregistryclient_rebind_send"
int rmixregistryclient_rebind_send (rmix_invokeref_t       **invokeref,
                                    const rmix_remoteref_t  *registry,
                                    const char              *name,
                                    const rmix_remoteref_t  *remoteref) {
  const void *inary[2], *inary_remoteref[6];
#ifdef DEBUG
  /* Check invokeref parameter. */
  if (NULL == invokeref) {
    errno = EINVAL;
    RMIX_LOG((RMIX_WARN(invokeref parameter is null)))
    errno = EINVAL;
    return -1;
  }
  /* Check registry parameter. */
  if (NULL == registry) {
    errno = EINVAL;
    RMIX_LOG((RMIX_WARN(registry parameter is null)))
    errno = EINVAL;
    return -1;
  }
  /* Check name parameter. */
  if (NULL == name) {
    errno = EINVAL;
    RMIX_LOG((RMIX_WARN(name parameter is null)))
    errno = EINVAL;
    return -1;
  }
  /* Check remoteref parameter. */
  if (NULL == remoteref) {
    errno = EINVAL;
    RMIX_LOG((RMIX_WARN(remoteref parameter is null)))
    errno = EINVAL;
    return -1;
  }
#endif /* DEBUG */
  /* Prepare parameters. */
  inary[0]           =  name;
  inary[1]           =  inary_remoteref;
  inary_remoteref[0] =  remoteref->protocol;
  inary_remoteref[1] =  remoteref->binding;
  inary_remoteref[2] = &remoteref->interfaces.count;
  inary_remoteref[3] =  remoteref->interfaces.array;
  inary_remoteref[4] =  remoteref->location;
  inary_remoteref[5] =  remoteref->object;
  /* Invoke remote object method. */
  if (0 != rmix_send(invokeref, registry,
                     &rmixregistryclient_interface,
                     RMIXREGISTRY_METHODS_REBIND_INDEX, inary, 2)) {
    int errno2 = errno;
    RMIX_LOG((RMIX_WARN(unable to invoke remote object method)))
    errno = errno2;
    return -1;
  }
  return 0;
}

/*
 * Invokes rmixregistry_rebind (client-side method stub).
 * Rebinds (creates a new or replaces an existing binding) a name with a remote
 * object reference at a remote RMIX registry using an asynchronous invocation
 * without loading the provider plug-in.
 *
 * invokeref = The invocation reference (return).
 * registry  = The RMIX remote object reference of the RMIX registry.
 * provider  = The preloaded RMIX provider plug-in handle.
 * name      = The name to bind.
 * remoteref = The RMIX remote object reference to bind.
 * return    = 0 on success or -1 on any error with errno set appropriately.
 */
#undef  __FUNC__
#define __FUNC__ "rmixregistryclient_rebind_send_p"
int rmixregistryclient_rebind_send_p (rmix_invokeref_t       **invokeref,
                                      const rmix_remoteref_t  *registry,
                                      unsigned int             provider,
                                      const char              *name,
                                      const rmix_remoteref_t  *remoteref) {
  const void *inary[2], *inary_remoteref[6];
#ifdef DEBUG
  /* Check invokeref parameter. */
  if (NULL == invokeref) {
    errno = EINVAL;
    RMIX_LOG((RMIX_WARN(invokeref parameter is null)))
    errno = EINVAL;
    return -1;
  }
  /* Check registry parameter. */
  if (NULL == registry) {
    errno = EINVAL;
    RMIX_LOG((RMIX_WARN(registry parameter is null)))
    errno = EINVAL;
    return -1;
  }
  /* Check name parameter. */
  if (NULL == name) {
    errno = EINVAL;
    RMIX_LOG((RMIX_WARN(name parameter is null)))
    errno = EINVAL;
    return -1;
  }
  /* Check remoteref parameter. */
  if (NULL == remoteref) {
    errno = EINVAL;
    RMIX_LOG((RMIX_WARN(remoteref parameter is null)))
    errno = EINVAL;
    return -1;
  }
#endif /* DEBUG */
  /* Prepare parameters. */
  inary[0]           =  name;
  inary[1]           =  inary_remoteref;
  inary_remoteref[0] =  remoteref->protocol;
  inary_remoteref[1] =  remoteref->binding;
  inary_remoteref[2] = &remoteref->interfaces.count;
  inary_remoteref[3] =  remoteref->interfaces.array;
  inary_remoteref[4] =  remoteref->location;
  inary_remoteref[5] =  remoteref->object;
  /* Invoke remote object method. */
  if (0 != rmix_send_p(invokeref, registry, provider,
                       &rmixregistryclient_interface,
                       RMIXREGISTRY_METHODS_REBIND_INDEX, inary, 2)) {
    int errno2 = errno;
    RMIX_LOG((RMIX_WARN(unable to invoke remote object method)))
    errno = errno2;
    return -1;
  }
  return 0;
}

/*
 * Retrieves rmixregistry_rebind invocation output (client-side method stub).
 * Returns the output of a previous asynchronous invocation that rebinds
 * (creates a new or replaces an existing binding) a name with a remote
 * object reference at a remote RMIX registry.
 *
 * invokeref = The invocation reference (also return).
 * return    = 0 on success or -1 on any error with errno set appropriately.
 */
#undef  __FUNC__
#define __FUNC__ "rmixregistryclient_rebind_retrieve"
int rmixregistryclient_rebind_retrieve (rmix_invokeref_t **invokeref) {
  int   result;
  void *outary[1];
#ifdef DEBUG
  /* Check invokeref parameter. */
  if (NULL == invokeref) {
    errno = EINVAL;
    RMIX_LOG((RMIX_WARN(invokeref parameter is null)))
    errno = EINVAL;
    return -1;
  }
#endif /* DEBUG */
  /* Prepare output. */
  outary[0] = &result;
  /* Retrieve invocation result. */
  if (0 != rmix_retrieve(invokeref, outary, 1)) {
    int errno2 = errno;
    RMIX_LOG((RMIX_WARN(unable to invoke remote object method)))
    errno = errno2;
    return -1;
  }
  return result;
}

/*
 * Retrieves rmixregistry_rebind invocation output (client-side method stub).
 * Returns the output of a previous asynchronous invocation that rebinds
 * (creates a new or replaces an existing binding) a name with a remote
 * object reference at a remote RMIX registry without unloading the provider
 * plug-in.
 *
 * invokeref = The invocation reference (also return).
 * return    = 0 on success or -1 on any error with errno set appropriately.
 */
#undef  __FUNC__
#define __FUNC__ "rmixregistryclient_rebind_retrieve_p"
int rmixregistryclient_rebind_retrieve_p (rmix_invokeref_t **invokeref) {
  int   result;
  void *outary[1];
#ifdef DEBUG
  /* Check invokeref parameter. */
  if (NULL == invokeref) {
    errno = EINVAL;
    RMIX_LOG((RMIX_WARN(invokeref parameter is null)))
    errno = EINVAL;
    return -1;
  }
#endif /* DEBUG */
  /* Prepare output. */
  outary[0] = &result;
  /* Retrieve invocation result. */
  if (0 != rmix_retrieve_p(invokeref, outary, 1)) {
    int errno2 = errno;
    RMIX_LOG((RMIX_WARN(unable to invoke remote object method)))
    errno = errno2;
    return -1;
  }
  return result;
}

/*
 * Invokes rmixregistry_unbind (client-side method stub).
 * Unbinds a name from a remote object reference at a remote RMIX registry.
 *
 * registry = The RMIX remote object reference of the RMIX registry.
 * name     = The name to unbind.
 * return   = 0 on success or -1 on any error with errno set appropriately.
 */
#undef  __FUNC__
#define __FUNC__ "rmixregistryclient_unbind"
int rmixregistryclient_unbind (const rmix_remoteref_t *registry,
                               const char             *name) {
  int         result;
  void       *outary[1];
  const void *inary[1];
#ifdef DEBUG
  /* Check registry parameter. */
  if (NULL == registry) {
    errno = EINVAL;
    RMIX_LOG((RMIX_WARN(registry parameter is null)))
    errno = EINVAL;
    return -1;
  }
  /* Check name parameter. */
  if (NULL == name) {
    errno = EINVAL;
    RMIX_LOG((RMIX_WARN(name parameter is null)))
    errno = EINVAL;
    return -1;
  }
#endif /* DEBUG */
  /* Prepare input. */
  inary[0]  = name;
  /* Prepare output. */
  outary[0] = &result;
  /* Invoke remote object method. */
  if (0 != rmix_invoke(registry, &rmixregistryclient_interface,
                       RMIXREGISTRY_METHODS_UNBIND_INDEX,
                       outary, 1, inary, 1)) {
    int errno2 = errno;
    RMIX_LOG((RMIX_WARN(unable to invoke remote object method)))
    errno = errno2;
    return -1;
  }
  return result;
}

/*
 * Invokes rmixregistry_unbind (client-side method stub).
 * Unbinds a name from a remote object reference at a remote RMIX registry
 * without loading the provider plug-in.
 *
 * registry = The RMIX remote object reference of the RMIX registry.
 * provider = The preloaded RMIX provider plug-in handle.
 * name     = The name to unbind.
 * return   = 0 on success or -1 on any error with errno set appropriately.
 */
#undef  __FUNC__
#define __FUNC__ "rmixregistryclient_unbind_p"
int rmixregistryclient_unbind_p (const rmix_remoteref_t *registry,
                                 unsigned int            provider,
                                 const char             *name) {
  int         result;
  void       *outary[1];
  const void *inary[1];
#ifdef DEBUG
  /* Check registry parameter. */
  if (NULL == registry) {
    errno = EINVAL;
    RMIX_LOG((RMIX_WARN(registry parameter is null)))
    errno = EINVAL;
    return -1;
  }
  /* Check name parameter. */
  if (NULL == name) {
    errno = EINVAL;
    RMIX_LOG((RMIX_WARN(name parameter is null)))
    errno = EINVAL;
    return -1;
  }
#endif /* DEBUG */
  /* Prepare input. */
  inary[0]  = name;
  /* Prepare output. */
  outary[0] = &result;
  /* Invoke remote object method. */
  if (0 != rmix_invoke_p(registry, provider,
                         &rmixregistryclient_interface,
                         RMIXREGISTRY_METHODS_UNBIND_INDEX,
                         outary, 1, inary, 1)) {
    int errno2 = errno;
    RMIX_LOG((RMIX_WARN(unable to invoke remote object method)))
    errno = errno2;
    return -1;
  }
  return result;
}

/*
 * Invokes rmixregistry_unbind (client-side method stub).
 * Unbinds a name from a remote object reference at a remote RMIX registry
 * using a one-way invocation.
 *
 * registry = The RMIX remote object reference of the RMIX registry.
 * name     = The name to unbind.
 * return   = 0 on success or -1 on any error with errno set appropriately.
 *            (-1 on local errors only due to one-way invocation)
 */
#undef  __FUNC__
#define __FUNC__ "rmixregistryclient_unbind_oneway"
int rmixregistryclient_unbind_oneway (const rmix_remoteref_t *registry,
                                      const char             *name) {
  const void *inary[1];
#ifdef DEBUG
  /* Check registry parameter. */
  if (NULL == registry) {
    errno = EINVAL;
    RMIX_LOG((RMIX_WARN(registry parameter is null)))
    errno = EINVAL;
    return -1;
  }
  /* Check name parameter. */
  if (NULL == name) {
    errno = EINVAL;
    RMIX_LOG((RMIX_WARN(name parameter is null)))
    errno = EINVAL;
    return -1;
  }
#endif /* DEBUG */
  /* Prepare input. */
  inary[0] = name;
  /* Invoke remote object method. */
  if (0 != rmix_oneway(registry,
                       &rmixregistryclient_interface,
                       RMIXREGISTRY_METHODS_UNBIND_INDEX, inary, 1)) {
    int errno2 = errno;
    RMIX_LOG((RMIX_WARN(unable to invoke remote object method)))
    errno = errno2;
    return -1;
  }
  return 0;
}

/*
 * Invokes rmixregistry_unbind (client-side method stub).
 * Unbinds a name from a remote object reference at a remote RMIX registry
 * using a one-way invocation without loading the provider plug-in.
 *
 * registry = The RMIX remote object reference of the RMIX registry.
 * provider = The preloaded RMIX provider plug-in handle.
 * name     = The name to unbind.
 * return   = 0 on success or -1 on any error with errno set appropriately.
 *            (-1 on local errors only due to one-way invocation)
 */
#undef  __FUNC__
#define __FUNC__ "rmixregistryclient_unbind_oneway_p"
int rmixregistryclient_unbind_oneway_p (const rmix_remoteref_t *registry,
                                        unsigned int            provider,
                                        const char             *name) {
  const void *inary[1];
#ifdef DEBUG
  /* Check registry parameter. */
  if (NULL == registry) {
    errno = EINVAL;
    RMIX_LOG((RMIX_WARN(registry parameter is null)))
    errno = EINVAL;
    return -1;
  }
  /* Check name parameter. */
  if (NULL == name) {
    errno = EINVAL;
    RMIX_LOG((RMIX_WARN(name parameter is null)))
    errno = EINVAL;
    return -1;
  }
#endif /* DEBUG */
  /* Prepare input. */
  inary[0] = name;
  /* Invoke remote object method. */
  if (0 != rmix_oneway_p(registry, provider,
                         &rmixregistryclient_interface,
                         RMIXREGISTRY_METHODS_UNBIND_INDEX, inary, 1)) {
    int errno2 = errno;
    RMIX_LOG((RMIX_WARN(unable to invoke remote object method)))
    errno = errno2;
    return -1;
  }
  return 0;
}

/*
 * Invokes rmixregistry_unbind (client-side method stub).
 * Unbinds a name from a remote object reference at a remote RMIX registry
 * using an asynchronous invocation.
 *
 * invokeref = The invocation reference (return).
 * registry  = The RMIX remote object reference of the RMIX registry.
 * name      = The name to unbind.
 * return    = 0 on success or -1 on any error with errno set appropriately.
 *             (-1 on local errors only due to one-way invocation)
 */
#undef  __FUNC__
#define __FUNC__ "rmixregistryclient_unbind_send"
int rmixregistryclient_unbind_send (rmix_invokeref_t       **invokeref,
                                    const rmix_remoteref_t  *registry,
                                    const char              *name) {
  const void *inary[1];
#ifdef DEBUG
  /* Check invokeref parameter. */
  if (NULL == invokeref) {
    errno = EINVAL;
    RMIX_LOG((RMIX_WARN(invokeref parameter is null)))
    errno = EINVAL;
    return -1;
  }
  /* Check registry parameter. */
  if (NULL == registry) {
    errno = EINVAL;
    RMIX_LOG((RMIX_WARN(registry parameter is null)))
    errno = EINVAL;
    return -1;
  }
  /* Check name parameter. */
  if (NULL == name) {
    errno = EINVAL;
    RMIX_LOG((RMIX_WARN(name parameter is null)))
    errno = EINVAL;
    return -1;
  }
#endif /* DEBUG */
  /* Prepare input. */
  inary[0] = name;
  /* Invoke remote object method. */
  if (0 != rmix_send(invokeref, registry,
                     &rmixregistryclient_interface,
                     RMIXREGISTRY_METHODS_UNBIND_INDEX, inary, 1)) {
    int errno2 = errno;
    RMIX_LOG((RMIX_WARN(unable to invoke remote object method)))
    errno = errno2;
    return -1;
  }
  return 0;
}

/*
 * Invokes rmixregistry_unbind (client-side method stub).
 * Unbinds a name from a remote object reference at a remote RMIX registry
 * using an asynchronous invocation without loading the provider plug-in.
 *
 * invokeref = The invocation reference (return).
 * registry = The RMIX remote object reference of the RMIX registry.
 * provider = The preloaded RMIX provider plug-in handle.
 * name     = The name to unbind.
 * return   = 0 on success or -1 on any error with errno set appropriately.
 *            (-1 on local errors only due to one-way invocation)
 */
#undef  __FUNC__
#define __FUNC__ "rmixregistryclient_unbind_send_p"
int rmixregistryclient_unbind_send_p (rmix_invokeref_t       **invokeref,
                                      const rmix_remoteref_t  *registry,
                                      unsigned int             provider,
                                      const char              *name) {
  const void *inary[1];
#ifdef DEBUG
  /* Check invokeref parameter. */
  if (NULL == invokeref) {
    errno = EINVAL;
    RMIX_LOG((RMIX_WARN(invokeref parameter is null)))
    errno = EINVAL;
    return -1;
  }
  /* Check registry parameter. */
  if (NULL == registry) {
    errno = EINVAL;
    RMIX_LOG((RMIX_WARN(registry parameter is null)))
    errno = EINVAL;
    return -1;
  }
  /* Check name parameter. */
  if (NULL == name) {
    errno = EINVAL;
    RMIX_LOG((RMIX_WARN(name parameter is null)))
    errno = EINVAL;
    return -1;
  }
#endif /* DEBUG */
  /* Prepare input. */
  inary[0] = name;
  /* Invoke remote object method. */
  if (0 != rmix_send_p(invokeref, registry, provider,
                       &rmixregistryclient_interface,
                       RMIXREGISTRY_METHODS_UNBIND_INDEX, inary, 1)) {
    int errno2 = errno;
    RMIX_LOG((RMIX_WARN(unable to invoke remote object method)))
    errno = errno2;
    return -1;
  }
  return 0;
}

/*
 * Retrieves rmixregistry_unbind invocation output (client-side method stub).
 * Returns the output of a previous asynchronous invocation that unbinds a name
 * from a remote object reference at a remote RMIX registry.
 *
 * invokeref = The invocation reference (also return).
 * return    = 0 on success or -1 on any error with errno set appropriately.
 */
#undef  __FUNC__
#define __FUNC__ "rmixregistryclient_unbind_retrieve"
int rmixregistryclient_unbind_retrieve (rmix_invokeref_t **invokeref) {
  int   result;
  void *outary[1];
#ifdef DEBUG
  /* Check invokeref parameter. */
  if (NULL == invokeref) {
    errno = EINVAL;
    RMIX_LOG((RMIX_WARN(invokeref parameter is null)))
    errno = EINVAL;
    return -1;
  }
#endif /* DEBUG */
  /* Prepare output. */
  outary[0] = &result;
  /* Retrieve invocation result. */
  if (0 != rmix_retrieve(invokeref, outary, 1)) {
    int errno2 = errno;
    RMIX_LOG((RMIX_WARN(unable to invoke remote object method)))
    errno = errno2;
    return -1;
  }
  return result;
}

/*
 * Retrieves rmixregistry_unbind invocation output (client-side method stub).
 * Returns the output of a previous asynchronous invocation that unbinds a name
 * from a remote object reference at a remote RMIX registry without unloading
 * the provider plug-in.
 *
 * invokeref = The invocation reference (also return).
 * return    = 0 on success or -1 on any error with errno set appropriately.
 */
#undef  __FUNC__
#define __FUNC__ "rmixregistryclient_unbind_retrieve_p"
int rmixregistryclient_unbind_retrieve_p (rmix_invokeref_t **invokeref) {
  int   result;
  void *outary[1];
#ifdef DEBUG
  /* Check invokeref parameter. */
  if (NULL == invokeref) {
    errno = EINVAL;
    RMIX_LOG((RMIX_WARN(invokeref parameter is null)))
    errno = EINVAL;
    return -1;
  }
#endif /* DEBUG */
  /* Prepare output. */
  outary[0] = &result;
  /* Retrieve invocation result. */
  if (0 != rmix_retrieve_p(invokeref, outary, 1)) {
    int errno2 = errno;
    RMIX_LOG((RMIX_WARN(unable to invoke remote object method)))
    errno = errno2;
    return -1;
  }
  return result;
}

/******************************************************************************
 *
 * END OF FILE
 *
 ******************************************************************************/
