/******************************************************************************
 *
 * Source file for harnessclient (main) module.
 * Copyright (c) 2002-2006 Oak Ridge National Laboratory.
 *
 * For more information see the following files in the source distribution top-
 * level directory or package data directory (usually /usr/local/share/package):
 *
 * - README    for general package information.
 * - INSTALL   for package install information.
 * - COPYING   for package license information and copying conditions.
 * - AUTHORS   for package authors information.
 * - ChangeLog for package changes information.
 *
 ******************************************************************************/

/******************************************************************************
 *
 * Macros
 *
 ******************************************************************************/

/* None at this time. */

/******************************************************************************
 *
 * Includes
 *
 ******************************************************************************/

/* Main module header file. */
#include <harnessclient.h>

/******************************************************************************
 *
 * Data Types
 *
 ******************************************************************************/

/* None at this time. */

/******************************************************************************
 *
 * Function Prototypes
 *
 ******************************************************************************/

/* None at this time. */

/******************************************************************************
 *
 * Data
 *
 ******************************************************************************/

/* Server-side method descriptors for harness. */
const rmix_method_t harnessclient_methods[HARNESS_RMIX_METHODS_COUNT] =
  HARNESS_RMIX_METHODS_CLIENT;

/* Server-side interface for harness. */
const rmix_interface_t harnessclient_interface = {
  HARNESS_RMIX_METHODS_COUNT,            /* method descriptor count */
  (rmix_method_t*)harnessclient_methods  /* method descriptor array */
};

/******************************************************************************
 *
 * Functions
 *
 ******************************************************************************/

/*
 * Invokes harness_plugins_load (client-side method stub).
 * Loads a plug-in at a remote harness instance.
 *
 * harness = The RMIX remote object reference of harness.
 * handle  = The plug-in handle (return). Plug-in handles are unique.
 * name    = The plug-in name. Plug-in names are unique and do
 *           not contain '/..' and do not start with '/'.
 * flags   = The plug-in loading flags, e.g. HARNESS_PLUGINS_UNIQUE.
 * return  = 0 on success or -1 on any error with errno set appropriately.
 */
#undef  __FUNC__
#define __FUNC__ "harnessclient_plugins_load"
int harnessclient_plugins_load (const rmix_remoteref_t *harness,
                                unsigned int           *handle,
                                const char             *name,
                                unsigned int            flags) {
  int         result;
  void       *outary[2];
  const void *inary[2];
#ifdef DEBUG
  /* Check harness parameter. */
  if (NULL == harness) {
    errno = EINVAL;
    HARNESS_LOG((HARNESS_WARN(harness parameter is null)))
    errno = EINVAL;
  }
  /* Check handle parameter. */
  if (NULL == handle) {
    errno = EINVAL;
    HARNESS_LOG((HARNESS_WARN(handle parameter is null)))
    errno = EINVAL;
    return -1;
  }
  /* Check name parameter. */
  if (NULL == name) {
    errno = EINVAL;
    HARNESS_LOG((HARNESS_WARN(name parameter is null)))
    errno = EINVAL;
    return -1;
  }
#endif /* DEBUG */
  /* Prepare input. */
  inary[0] =  name;
  inary[1] = &flags;
  /* Prepare output. */
  outary[0] = &result;
  outary[1] =  handle;
  /* Invoke remote object method. */
  if (0 != rmix_invoke(harness, &harnessclient_interface,
                       HARNESS_RMIX_METHODS_PLUGINS_LOAD_INDEX,
                       outary, 2, inary, 2)) {
    int errno2 = errno;
    RMIX_LOG((RMIX_WARN(unable to invoke remote object method)))
    errno = errno2;
    return -1;
  }
  return result;
}

/*
 * Invokes harness_plugins_load (client-side method stub).
 * Loads a plug-in at a remote harness instance.
 *
 * harness  = The RMIX remote object reference of harness.
 * provider = The preloaded RMIX provider plug-in handle.
 * handle   = The plug-in handle (return). Plug-in handles are unique.
 * name     = The plug-in name. Plug-in names are unique and do
 *            not contain '/..' and do not start with '/'.
 * flags    = The plug-in loading flags, e.g. HARNESS_PLUGINS_UNIQUE.
 * return   = 0 on success or -1 on any error with errno set appropriately.
 */
#undef  __FUNC__
#define __FUNC__ "harnessclient_plugins_load_p"
int harnessclient_plugins_load_p (const rmix_remoteref_t *harness,
                                  unsigned int            provider,
                                  unsigned int           *handle,
                                  const char             *name,
                                  unsigned int            flags) {
  int         result;
  void       *outary[2];
  const void *inary[2];
#ifdef DEBUG
  /* Check harness parameter. */
  if (NULL == harness) {
    errno = EINVAL;
    HARNESS_LOG((HARNESS_WARN(harness parameter is null)))
    errno = EINVAL;
    return -1;
  }
  /* Check handle parameter. */
  if (NULL == handle) {
    errno = EINVAL;
    HARNESS_LOG((HARNESS_WARN(handle parameter is null)))
    errno = EINVAL;
    return -1;
  }
  /* Check name parameter. */
  if (NULL == name) {
    errno = EINVAL;
    HARNESS_LOG((HARNESS_WARN(name parameter is null)))
    errno = EINVAL;
    return -1;
  }
#endif /* DEBUG */
  /* Prepare input. */
  inary[0] =  name;
  inary[1] = &flags;
  /* Prepare output. */
  outary[0] = &result;
  outary[1] =  handle;
  /* Invoke remote object method. */
  if (0 != rmix_invoke_p(harness, provider, &harnessclient_interface,
                         HARNESS_RMIX_METHODS_PLUGINS_LOAD_INDEX,
                         outary, 2, inary, 2)) {
    int errno2 = errno;
    RMIX_LOG((RMIX_WARN(unable to invoke remote object method)))
    errno = errno2;
    return -1;
  }
  return result;
}

/*
 * Invokes harness_plugins_load (client-side method stub).
 * Loads a plug-in at a remote harness instance using a oneway invocation.
 *
 * harness = The RMIX remote object reference of harness.
 * name    = The plug-in name. Plug-in names are unique and do
 *           not contain '/..' and do not start with '/'.
 * flags   = The plug-in loading flags, e.g. HARNESS_PLUGINS_UNIQUE.
 * return  = 0 on success or -1 on any error with errno set appropriately.
 */
#undef  __FUNC__
#define __FUNC__ "harnessclient_plugins_load_oneway"
int harnessclient_plugins_load_oneway (const rmix_remoteref_t *harness,
                                       const char             *name,
                                       unsigned int            flags) {
  const void *inary[2];
#ifdef DEBUG
  /* Check harness parameter. */
  if (NULL == harness) {
    errno = EINVAL;
    HARNESS_LOG((HARNESS_WARN(harness parameter is null)))
    errno = EINVAL;
    return -1;
  }
  /* Check name parameter. */
  if (NULL == name) {
    errno = EINVAL;
    HARNESS_LOG((HARNESS_WARN(name parameter is null)))
    errno = EINVAL;
    return -1;
  }
#endif /* DEBUG */
  /* Prepare input. */
  inary[0] =  name;
  inary[1] = &flags;
  /* Invoke remote object method. */
  if (0 != rmix_oneway(harness, &harnessclient_interface,
                       HARNESS_RMIX_METHODS_PLUGINS_LOAD_INDEX,
                       inary, 2)) {
    int errno2 = errno;
    RMIX_LOG((RMIX_WARN(unable to invoke remote object method)))
    errno = errno2;
    return -1;
  }
  return 0;
}

/*
 * Invokes harness_plugins_load (client-side method stub).
 * Loads a plug-in at a remote harness instance using a oneway invocation
 * without loading the provider plug-in.
 *
 * harness  = The RMIX remote object reference of harness.
 * provider = The preloaded RMIX provider plug-in handle.
 * name     = The plug-in name. Plug-in names are unique and do
 *            not contain '/..' and do not start with '/'.
 * flags    = The plug-in loading flags, e.g. HARNESS_PLUGINS_UNIQUE.
 * return   = 0 on success or -1 on any error with errno set appropriately.
 */
#undef  __FUNC__
#define __FUNC__ "harnessclient_plugins_load_oneway_p"
int harnessclient_plugins_load_oneway_p (const rmix_remoteref_t *harness,
                                         unsigned int            provider,
                                         const char             *name,
                                         unsigned int            flags) {
  const void *inary[2];
#ifdef DEBUG
  /* Check harness parameter. */
  if (NULL == harness) {
    errno = EINVAL;
    HARNESS_LOG((HARNESS_WARN(harness parameter is null)))
    errno = EINVAL;
    return -1;
  }
  /* Check name parameter. */
  if (NULL == name) {
    errno = EINVAL;
    HARNESS_LOG((HARNESS_WARN(name parameter is null)))
    errno = EINVAL;
    return -1;
  }
#endif /* DEBUG */
  /* Prepare input. */
  inary[0] =  name;
  inary[1] = &flags;
  /* Invoke remote object method. */
  if (0 != rmix_oneway_p(harness, provider, &harnessclient_interface,
                         HARNESS_RMIX_METHODS_PLUGINS_LOAD_INDEX,
                         inary, 2)) {
    int errno2 = errno;
    RMIX_LOG((RMIX_WARN(unable to invoke remote object method)))
    errno = errno2;
    return -1;
  }
  return 0;
}

/*
 * Invokes harness_plugins_load (client-side method stub).
 * Loads a plug-in at a remote harness instance using an asynchronous
 * invocation.
 *
 * invokeref = The invocation reference (return).
 * harness   = The RMIX remote object reference of harness.
 * name      = The plug-in name. Plug-in names are unique and do
 *             not contain '/..' and do not start with '/'.
 * flags     = The plug-in loading flags, e.g. HARNESS_PLUGINS_UNIQUE.
 * return    = 0 on success or -1 on any error with errno set appropriately.
 */
#undef  __FUNC__
#define __FUNC__ "harnessclient_plugins_load_send"
int harnessclient_plugins_load_send (rmix_invokeref_t       **invokeref,
                                     const rmix_remoteref_t  *harness,
                                     const char              *name,
                                     unsigned int             flags) {
  const void *inary[2];
#ifdef DEBUG
  /* Check invokeref parameter. */
  if (NULL == invokeref) {
    errno = EINVAL;
    HARNESS_LOG((HARNESS_WARN(invokeref parameter is null)))
    errno = EINVAL;
    return -1;
  }
  /* Check harness parameter. */
  if (NULL == harness) {
    errno = EINVAL;
    HARNESS_LOG((HARNESS_WARN(harness parameter is null)))
    errno = EINVAL;
    return -1;
  }
  /* Check name parameter. */
  if (NULL == name) {
    errno = EINVAL;
    HARNESS_LOG((HARNESS_WARN(name parameter is null)))
    errno = EINVAL;
    return -1;
  }
#endif /* DEBUG */
  /* Prepare input. */
  inary[0] =  name;
  inary[1] = &flags;
  /* Invoke remote object method. */
  if (0 != rmix_send(invokeref, harness, &harnessclient_interface,
                     HARNESS_RMIX_METHODS_PLUGINS_LOAD_INDEX,
                     inary, 2)) {
    int errno2 = errno;
    RMIX_LOG((RMIX_WARN(unable to invoke remote object method)))
    errno = errno2;
    return -1;
  }
  return 0;
}

/*
 * Invokes harness_plugins_load (client-side method stub).
 * Loads a plug-in at a remote harness instance using an asynchronous invocation
 * without loading the provider plug-in.
 *
 * invokeref = The invocation reference (return).
 * harness   = The RMIX remote object reference of harness.
 * provider  = The preloaded RMIX provider plug-in handle.
 * name      = The plug-in name. Plug-in names are unique and do
 *             not contain '/..' and do not start with '/'.
 * flags     = The plug-in loading flags, e.g. HARNESS_PLUGINS_UNIQUE.
 * return    = 0 on success or -1 on any error with errno set appropriately.
 */
#undef  __FUNC__
#define __FUNC__ "harnessclient_plugins_load_send_p"
int harnessclient_plugins_load_send_p (rmix_invokeref_t       **invokeref,
                                       const rmix_remoteref_t  *harness,
                                       unsigned int             provider,
                                       const char              *name,
                                       unsigned int             flags) {
  const void *inary[2];
#ifdef DEBUG
  /* Check invokeref parameter. */
  if (NULL == invokeref) {
    errno = EINVAL;
    HARNESS_LOG((HARNESS_WARN(invokeref parameter is null)))
    errno = EINVAL;
    return -1;
  }
  /* Check harness parameter. */
  if (NULL == harness) {
    errno = EINVAL;
    HARNESS_LOG((HARNESS_WARN(harness parameter is null)))
    errno = EINVAL;
    return -1;
  }
  /* Check name parameter. */
  if (NULL == name) {
    errno = EINVAL;
    HARNESS_LOG((HARNESS_WARN(name parameter is null)))
    errno = EINVAL;
    return -1;
  }
#endif /* DEBUG */
  /* Prepare input. */
  inary[0] =  name;
  inary[1] = &flags;
  /* Invoke remote object method. */
  if (0 != rmix_send_p(invokeref, harness, provider, &harnessclient_interface,
                       HARNESS_RMIX_METHODS_PLUGINS_LOAD_INDEX,
                       inary, 2)) {
    int errno2 = errno;
    RMIX_LOG((RMIX_WARN(unable to invoke remote object method)))
    errno = errno2;
    return -1;
  }
  return 0;
}

/*
 * Retrieves harness_plugins_load output (client-side method stub).
 * Returns the output of a previous asynchronous invocation that loads a plug-in
 * at a remote harness instance.
 *
 * invokeref = The invocation reference (also return).
 * handle    = The plug-in handle (return). Plug-in handles are unique.
 * return    = 0 on success or -1 on any error with errno set appropriately.
 */
#undef  __FUNC__
#define __FUNC__ "harnessclient_plugins_load_retrieve"
int harnessclient_plugins_load_retrieve (rmix_invokeref_t **invokeref,
                                         unsigned int      *handle) {
  int   result;
  void *outary[2];
#ifdef DEBUG
  /* Check invokeref parameter. */
  if (NULL == invokeref) {
    errno = EINVAL;
    HARNESS_LOG((HARNESS_WARN(invokeref parameter is null)))
    errno = EINVAL;
    return -1;
  }
  /* Check *invokeref parameter. */
  if (NULL == *invokeref) {
    errno = EINVAL;
    HARNESS_LOG((HARNESS_WARN(*invokeref parameter is null)))
    errno = EINVAL;
    return -1;
  }
  /* Check handle parameter. */
  if (NULL == handle) {
    errno = EINVAL;
    HARNESS_LOG((HARNESS_WARN(handle parameter is null)))
    errno = EINVAL;
    return -1;
  }
#endif /* DEBUG */
  /* Prepare output. */
  outary[0] = &result;
  outary[1] =  handle;
  /* Invoke remote object method. */
  if (0 != rmix_retrieve(invokeref, outary, 2)) {
    int errno2 = errno;
    RMIX_LOG((RMIX_WARN(unable to invoke remote object method)))
    errno = errno2;
    return -1;
  }
  return result;
}

/*
 * Retrieves harness_plugins_load output (client-side method stub).
 * Returns the output of a previous asynchronous invocation that loads a plug-in
 * at a remote harness instance without unloading the provider plug-in.
 *
 * invokeref = The invocation reference (also return).
 * handle    = The plug-in handle (return). Plug-in handles are unique.
 * return    = 0 on success or -1 on any error with errno set appropriately.
 */
#undef  __FUNC__
#define __FUNC__ "harnessclient_plugins_load_retrieve_p"
int harnessclient_plugins_load_retrieve_p (rmix_invokeref_t **invokeref,
                                           unsigned int      *handle) {
  int   result;
  void *outary[2];
#ifdef DEBUG
  /* Check invokeref parameter. */
  if (NULL == invokeref) {
    errno = EINVAL;
    HARNESS_LOG((HARNESS_WARN(invokeref parameter is null)))
    errno = EINVAL;
    return -1;
  }
  /* Check *invokeref parameter. */
  if (NULL == *invokeref) {
    errno = EINVAL;
    HARNESS_LOG((HARNESS_WARN(*invokeref parameter is null)))
    errno = EINVAL;
    return -1;
  }
  /* Check handle parameter. */
  if (NULL == handle) {
    errno = EINVAL;
    HARNESS_LOG((HARNESS_WARN(handle parameter is null)))
    errno = EINVAL;
    return -1;
  }
#endif /* DEBUG */
  /* Prepare output. */
  outary[0] = &result;
  outary[1] =  handle;
  /* Invoke remote object method. */
  if (0 != rmix_retrieve_p(invokeref, outary, 2)) {
    int errno2 = errno;
    RMIX_LOG((RMIX_WARN(unable to invoke remote object method)))
    errno = errno2;
    return -1;
  }
  return result;
}

/*
 * Invokes harness_plugins_load (client-side method stub).
 * Loads a plug-in at multiple remote harness instances in sequence.
 * The plug-in may be different for every remote harness instance.
 *
 * count     = The number of remote harness instances.
 * harnesses = The array of RMIX remote object references of harness instances.
 * handles   = The plug-in handle array (also return). Plug-in handles are
 *             unique, and 0 for unsuccessful loading attempts.
 * names     = The plug-in name array. Plug-in names are unique and do
 *             not contain '/..' and do not start with '/'.
 * flags     = The plug-in loading flags array, e.g. HARNESS_PLUGINS_UNIQUE.
 * order     = The order flag. 0 for forward and 1 for reverse.
 * fail      = The fail flag. 0 to continue and 1 to fail on any error.
 * return    = The number of loaded plug-ins with errno set appropriately.
 */
#undef  __FUNC__
#define __FUNC__ "harnessclient_plugins_sload_mp"
unsigned int harnessclient_plugins_sload_mp (unsigned int             count,
                                             const rmix_remoteref_t **harnesses,
                                             unsigned int            *handles,
                                             const char             **names,
                                             const unsigned int      *flags,
                                             unsigned int             order,
                                             unsigned int             fail) {
  int          errno2;
  unsigned int loaded, index;
#ifdef DEBUG
  /* Check harnesses parameter. */
  if ((NULL == harnesses)&&(0 != count)) {
    errno = EINVAL;
    RMIX_LOG((RMIX_WARN(harnesses parameter is null)))
    errno = EINVAL;
    return -1;
  }
  /* Check count parameter. */
  if ((0 == count)&&(NULL != harnesses)) {
    errno = EINVAL;
    RMIX_LOG((RMIX_WARN(count parameter is zero)))
    errno = EINVAL;
    return -1;
  }
  /* Check handles parameter. */
  if ((NULL == handles)&&(0 != count)) {
    errno = EINVAL;
    RMIX_LOG((RMIX_WARN(handles parameter is null)))
    errno = EINVAL;
    return -1;
  }
  /* Check count parameter. */
  if ((0 == count)&&(NULL != handles)) {
    errno = EINVAL;
    RMIX_LOG((RMIX_WARN(count parameter is zero)))
    errno = EINVAL;
    return -1;
  }
  /* Check names parameter. */
  if ((NULL == names)&&(0 != count)) {
    errno = EINVAL;
    RMIX_LOG((RMIX_WARN(names parameter is null)))
    errno = EINVAL;
    return -1;
  }
  /* Check count parameter. */
  if ((0 == count)&&(NULL != names)) {
    errno = EINVAL;
    RMIX_LOG((RMIX_WARN(count parameter is zero)))
    errno = EINVAL;
    return -1;
  }
  /* Check flags parameter. */
  if ((NULL == flags)&&(0 != count)) {
    errno = EINVAL;
    RMIX_LOG((RMIX_WARN(flags parameter is null)))
    errno = EINVAL;
    return -1;
  }
  /* Check count parameter. */
  if ((0 == count)&&(NULL != flags)) {
    errno = EINVAL;
    RMIX_LOG((RMIX_WARN(count parameter is zero)))
    errno = EINVAL;
    return -1;
  }
#endif /* DEBUG */
  /* Initialize errno and loaded counter.*/
  errno2 = 0;
  loaded = 0;
  /* Initialize all handles. */
  memset(handles, 0, count * sizeof(*handles));
  /* Check order flag. */
  if (0 == order) {
    /* Load all plug-ins in forward order. */
    for (index = 0; index < count; index++) {
      /* Load plug-in. */
      if (0 != harnessclient_plugins_load(harnesses[index], &handles[index],
                                          names[index], flags[index])) {
        errno2 = (0 == errno2)?errno:errno2;
        RMIX_LOG((RMIX_WARN(unable to load plug-in)))
        /* Check fail flag. */
        if (1 == fail) {
          errno = errno2;
          return loaded;
        }
      } else {
        loaded++;
      }
    }
  } else {
    /* Load all plug-ins in reverse order. */
    for (index = count; 0 < index--;) {
      /* Load plug-in. */
      if (0 != harnessclient_plugins_load(harnesses[index], &handles[index],
                                          names[index], flags[index])) {
        errno2 = (0 == errno2)?errno:errno2;
        RMIX_LOG((RMIX_WARN(unable to load plug-in)))
        /* Check fail flag. */
        if (1 == fail) {
          errno = errno2;
          return loaded;
        }
      } else {
        loaded++;
      }
    }
  }
  errno = errno2;
  return loaded;
}

/*
 * Invokes harness_plugins_load (client-side method stub).
 * Loads a plug-in at multiple remote harness instances in sequence.
 * The plug-in is the same for every remote harness instance.
 *
 * count     = The number of remote harness instances.
 * harnesses = The array of RMIX remote object references of harness instances.
 * handles   = The plug-in handle array (also return). Plug-in handles are
 *             unique, and 0 for unsuccessful loading attempts.
 * name      = The plug-in name. Plug-in names are unique and do
 *             not contain '/..' and do not start with '/'.
 * flags     = The plug-in loading flags, e.g. HARNESS_PLUGINS_UNIQUE.
 * order     = The order flag. 0 for forward and 1 for reverse.
 * fail      = The fail flag. 0 to continue and 1 to fail on any error.
 * return    = The number of loaded plug-ins with errno set appropriately.
 */
#undef  __FUNC__
#define __FUNC__ "harnessclient_plugins_sload_sp"
unsigned int harnessclient_plugins_sload_sp (unsigned int             count,
                                             const rmix_remoteref_t **harnesses,
                                             unsigned int            *handles,
                                             const char              *name,
                                             unsigned int             flags,
                                             unsigned int             order,
                                             unsigned int             fail) {
  int          errno2;
  unsigned int loaded, index;
#ifdef DEBUG
  /* Check harnesses parameter. */
  if ((NULL == harnesses)&&(0 != count)) {
    errno = EINVAL;
    RMIX_LOG((RMIX_WARN(harnesses parameter is null)))
    errno = EINVAL;
    return -1;
  }
  /* Check count parameter. */
  if ((0 == count)&&(NULL != harnesses)) {
    errno = EINVAL;
    RMIX_LOG((RMIX_WARN(count parameter is zero)))
    errno = EINVAL;
    return -1;
  }
  /* Check handles parameter. */
  if ((NULL == handles)&&(0 != count)) {
    errno = EINVAL;
    RMIX_LOG((RMIX_WARN(handles parameter is null)))
    errno = EINVAL;
    return -1;
  }
  /* Check count parameter. */
  if ((0 == count)&&(NULL != handles)) {
    errno = EINVAL;
    RMIX_LOG((RMIX_WARN(count parameter is zero)))
    errno = EINVAL;
    return -1;
  }
  /* Check name parameter. */
  if (NULL == name) {
    errno = EINVAL;
    HARNESS_LOG((HARNESS_WARN(name parameter is null)))
    errno = EINVAL;
    return -1;
  }
#endif /* DEBUG */
  /* Initialize errno and loaded counter.*/
  errno2 = 0;
  loaded = 0;
  /* Initialize all handles. */
  memset(handles, 0, count * sizeof(*handles));
  /* Check order flag. */
  if (0 == order) {
    /* Load all plug-ins in forward order. */
    for (index = 0; index < count; index++) {
      /* Load plug-in. */
      if (0 != harnessclient_plugins_load(harnesses[index], &handles[index],
                                          name, flags)) {
        errno2 = (0 == errno2)?errno:errno2;
        RMIX_LOG((RMIX_WARN(unable to load plug-in)))
        /* Check fail flag. */
        if (1 == fail) {
          errno = errno2;
          return loaded;
        }
      } else {
        loaded++;
      }
    }
  } else {
    /* Load all plug-ins in reverse order. */
    for (index = count; 0 < index--;) {
      /* Load plug-in. */
      if (0 != harnessclient_plugins_load(harnesses[index], &handles[index],
                                          name, flags)) {
        errno2 = (0 == errno2)?errno:errno2;
        RMIX_LOG((RMIX_WARN(unable to load plug-in)))
        /* Check fail flag. */
        if (1 == fail) {
          errno = errno2;
          return loaded;
        }
      } else {
        loaded++;
      }
    }
  }
  errno = errno2;
  return loaded;
}

/*
 * Invokes harness_plugins_load (client-side method stub).
 * Loads a plug-in at multiple remote harness instances in parallel using
 * asynchonous method invocations.
 * The plug-in may be different for every remote harness instance.
 *
 * count     = The number of remote harness instances.
 * harnesses = The array of RMIX remote object references of harness instances.
 * handles   = The plug-in handle array (also return). Plug-in handles are
 *             unique, and 0 for unsuccessful loading attempts.
 * names     = The plug-in name array. Plug-in names are unique and do
 *             not contain '/..' and do not start with '/'.
 * flags     = The plug-in loading flags array, e.g. HARNESS_PLUGINS_UNIQUE.
 * return    = The number of loaded plug-ins with errno set appropriately.
 */
#undef  __FUNC__
#define __FUNC__ "harnessclient_plugins_pload_mp"
unsigned int harnessclient_plugins_pload_mp (unsigned int             count,
                                             const rmix_remoteref_t **harnesses,
                                             unsigned int            *handles,
                                             const char             **names,
                                             const unsigned int      *flags) {
  int                errno2;
  unsigned int       loaded, index;
  rmix_invokeref_t **invocations;
#ifdef DEBUG
  /* Check harnesses parameter. */
  if ((NULL == harnesses)&&(0 != count)) {
    errno = EINVAL;
    RMIX_LOG((RMIX_WARN(harnesses parameter is null)))
    errno = EINVAL;
    return -1;
  }
  /* Check count parameter. */
  if ((0 == count)&&(NULL != harnesses)) {
    errno = EINVAL;
    RMIX_LOG((RMIX_WARN(count parameter is zero)))
    errno = EINVAL;
    return -1;
  }
  /* Check handles parameter. */
  if ((NULL == handles)&&(0 != count)) {
    errno = EINVAL;
    RMIX_LOG((RMIX_WARN(handles parameter is null)))
    errno = EINVAL;
    return -1;
  }
  /* Check count parameter. */
  if ((0 == count)&&(NULL != handles)) {
    errno = EINVAL;
    RMIX_LOG((RMIX_WARN(count parameter is zero)))
    errno = EINVAL;
    return -1;
  }
  /* Check names parameter. */
  if ((NULL == names)&&(0 != count)) {
    errno = EINVAL;
    RMIX_LOG((RMIX_WARN(names parameter is null)))
    errno = EINVAL;
    return -1;
  }
  /* Check count parameter. */
  if ((0 == count)&&(NULL != names)) {
    errno = EINVAL;
    RMIX_LOG((RMIX_WARN(count parameter is zero)))
    errno = EINVAL;
    return -1;
  }
  /* Check flags parameter. */
  if ((NULL == flags)&&(0 != count)) {
    errno = EINVAL;
    RMIX_LOG((RMIX_WARN(flags parameter is null)))
    errno = EINVAL;
    return -1;
  }
  /* Check count parameter. */
  if ((0 == count)&&(NULL != flags)) {
    errno = EINVAL;
    RMIX_LOG((RMIX_WARN(count parameter is zero)))
    errno = EINVAL;
    return -1;
  }
#endif /* DEBUG */
  /* Allocate invocation reference array. */
  if (NULL == (invocations = malloc(count * sizeof(rmix_invokeref_t*)))) {
    errno2 = errno;
    RMIX_LOG((RMIX_WARN(unable to allocate invocation reference array)))
    errno = errno2;
    return 0;
  }
  /* Initialize errno and loaded counter.*/
  errno2 = 0;
  loaded = 0;
  /* Initialize all handles. */
  memset(handles, 0, count * sizeof(*handles));
  /* Load all plug-ins in forward order (send part). */
  for (index = 0; index < count; index++) {
    /* Load plug-in (send part). */
    if (0 != harnessclient_plugins_load_send(&invocations[index],
                                             harnesses[index],
                                             names[index], flags[index])) {
      errno2 = (0 == errno2)?errno:errno2;
      RMIX_LOG((RMIX_WARN(unable to load plug-in (send part))))
      invocations[index] = NULL;
    }
  }
  /* Load all plug-ins in forward order (retrieve part). */
  for (index = 0; index < count; index++) {
    /* Check for successful send. */
    if (NULL != invocations[index]) {
      /* Load plug-in (retrieve part). */
      if (0 != harnessclient_plugins_load_retrieve(&invocations[index],
                                                   &handles[index])) {
        errno2 = (0 == errno2)?errno:errno2;
        RMIX_LOG((RMIX_WARN(unable to load plug-in (retrieve part))))
      } else {
        loaded++;
      }
    }
  }
  /* Deallocate invocation reference array. */
  free(invocations);
  errno = errno2;
  return loaded;
}

/*
 * Invokes harness_plugins_load (client-side method stub).
 * Loads a plug-in at multiple remote harness instances in parallel using
 * asynchonous method invocations.
 * The plug-in is the same for every remote harness instance.
 *
 * count     = The number of remote harness instances.
 * harnesses = The array of RMIX remote object references of harness instances.
 * handles   = The plug-in handle array (also return). Plug-in handles are
 *             unique, and 0 for unsuccessful loading attempts.
 * name      = The plug-in name. Plug-in names are unique and do
 *             not contain '/..' and do not start with '/'.
 * flags     = The plug-in loading flags, e.g. HARNESS_PLUGINS_UNIQUE.
 * return    = The number of loaded plug-ins with errno set appropriately.
 */
#undef  __FUNC__
#define __FUNC__ "harnessclient_plugins_pload_sp"
unsigned int harnessclient_plugins_pload_sp (unsigned int             count,
                                             const rmix_remoteref_t **harnesses,
                                             unsigned int            *handles,
                                             const char              *name,
                                             unsigned int             flags) {
  int                errno2;
  unsigned int       loaded, index;
  rmix_invokeref_t **invocations;
#ifdef DEBUG
  /* Check harnesses parameter. */
  if ((NULL == harnesses)&&(0 != count)) {
    errno = EINVAL;
    RMIX_LOG((RMIX_WARN(harnesses parameter is null)))
    errno = EINVAL;
    return -1;
  }
  /* Check count parameter. */
  if ((0 == count)&&(NULL != harnesses)) {
    errno = EINVAL;
    RMIX_LOG((RMIX_WARN(count parameter is zero)))
    errno = EINVAL;
    return -1;
  }
  /* Check handles parameter. */
  if ((NULL == handles)&&(0 != count)) {
    errno = EINVAL;
    RMIX_LOG((RMIX_WARN(handles parameter is null)))
    errno = EINVAL;
    return -1;
  }
  /* Check count parameter. */
  if ((0 == count)&&(NULL != handles)) {
    errno = EINVAL;
    RMIX_LOG((RMIX_WARN(count parameter is zero)))
    errno = EINVAL;
    return -1;
  }
  /* Check name parameter. */
  if (NULL == name) {
    errno = EINVAL;
    HARNESS_LOG((HARNESS_WARN(name parameter is null)))
    errno = EINVAL;
    return -1;
  }
#endif /* DEBUG */
  /* Allocate invocation reference array. */
  if (NULL == (invocations = malloc(count * sizeof(rmix_invokeref_t*)))) {
    errno2 = errno;
    RMIX_LOG((RMIX_WARN(unable to allocate invocation reference array)))
    errno = errno2;
    return 0;
  }
  /* Initialize errno and loaded counter.*/
  errno2 = 0;
  loaded = 0;
  /* Initialize all handles. */
  memset(handles, 0, count * sizeof(*handles));
  /* Load all plug-ins in forward order (send part). */
  for (index = 0; index < count; index++) {
    /* Load plug-in (send part). */
    if (0 != harnessclient_plugins_load_send(&invocations[index],
                                             harnesses[index],
                                             name, flags)) {
      errno2 = (0 == errno2)?errno:errno2;
      RMIX_LOG((RMIX_WARN(unable to load plug-in (send part))))
      invocations[index] = NULL;
    }
  }
  /* Load all plug-ins in forward order (retrieve part). */
  for (index = 0; index < count; index++) {
    /* Check for successful send. */
    if (NULL != invocations[index]) {
      /* Load plug-in (retrieve part). */
      if (0 != harnessclient_plugins_load_retrieve(&invocations[index],
                                                   &handles[index])) {
        errno2 = (0 == errno2)?errno:errno2;
        RMIX_LOG((RMIX_WARN(unable to load plug-in (retrieve part))))
      } else {
        loaded++;
      }
    }
  }
  /* Deallocate invocation reference array. */
  free(invocations);
  errno = errno2;
  return loaded;
}

/*
 * Invokes harness_plugins_load2 (client-side method stub).
 * Loads a plug-in at a remote harness instance.
 *
 * harness = The RMIX remote object reference of harness.
 * handle  = The plug-in handle (return). Plug-in handles are unique.
 * name    = The plug-in name. Plug-in names are unique and do
 *           not contain '/..' and do not start with '/'.
 * version = The version of the plug-in to load.
 * flags   = The plug-in loading flags, e.g. HARNESS_PLUGINS_UNIQUE.
 * return  = 0 on success or -1 on any error with errno set appropriately.
 */
#undef  __FUNC__
#define __FUNC__ "harnessclient_plugins_load2"
int harnessclient_plugins_load2 (const rmix_remoteref_t *harness,
                                 unsigned int           *handle,
                                 const char             *name,
                                 unsigned int            version,
                                 unsigned int            flags) {
  int         result;
  void       *outary[2];
  const void *inary[3];
#ifdef DEBUG
  /* Check harness parameter. */
  if (NULL == harness) {
    errno = EINVAL;
    HARNESS_LOG((HARNESS_WARN(harness parameter is null)))
    errno = EINVAL;
    return -1;
  }
  /* Check handle parameter. */
  if (NULL == handle) {
    errno = EINVAL;
    HARNESS_LOG((HARNESS_WARN(handle parameter is null)))
    errno = EINVAL;
    return -1;
  }
  /* Check name parameter. */
  if (NULL == name) {
    errno = EINVAL;
    HARNESS_LOG((HARNESS_WARN(name parameter is null)))
    errno = EINVAL;
    return -1;
  }
#endif /* DEBUG */
  /* Prepare input. */
  inary[0] =  name;
  inary[1] = &version;
  inary[2] = &flags;
  /* Prepare output. */
  outary[0] = &result;
  outary[1] =  handle;
  /* Invoke remote object method. */
  if (0 != rmix_invoke(harness, &harnessclient_interface,
                       HARNESS_RMIX_METHODS_PLUGINS_LOAD2_INDEX,
                       outary, 2, inary, 3)) {
    int errno2 = errno;
    RMIX_LOG((RMIX_WARN(unable to invoke remote object method)))
    errno = errno2;
    return -1;
  }
  return result;
}

/*
 * Invokes harness_plugins_load2 (client-side method stub).
 * Loads a plug-in at a remote harness instance.
 *
 * harness  = The RMIX remote object reference of harness.
 * provider = The preloaded RMIX provider plug-in handle.
 * handle   = The plug-in handle (return). Plug-in handles are unique.
 * name     = The plug-in name. Plug-in names are unique and do
 *            not contain '/..' and do not start with '/'.
 * version  = The version of the plug-in to load.
 * flags    = The plug-in loading flags, e.g. HARNESS_PLUGINS_UNIQUE.
 * return   = 0 on success or -1 on any error with errno set appropriately.
 */
#undef  __FUNC__
#define __FUNC__ "harnessclient_plugins_load2_p"
int harnessclient_plugins_load2_p (const rmix_remoteref_t *harness,
                                   unsigned int            provider,
                                   unsigned int           *handle,
                                   const char             *name,
                                   unsigned int            version,
                                   unsigned int            flags) {
  int         result;
  void       *outary[2];
  const void *inary[3];
#ifdef DEBUG
  /* Check harness parameter. */
  if (NULL == harness) {
    errno = EINVAL;
    HARNESS_LOG((HARNESS_WARN(harness parameter is null)))
    errno = EINVAL;
    return -1;
  }
  /* Check handle parameter. */
  if (NULL == handle) {
    errno = EINVAL;
    HARNESS_LOG((HARNESS_WARN(handle parameter is null)))
    errno = EINVAL;
    return -1;
  }
  /* Check name parameter. */
  if (NULL == name) {
    errno = EINVAL;
    HARNESS_LOG((HARNESS_WARN(name parameter is null)))
    errno = EINVAL;
    return -1;
  }
#endif /* DEBUG */
  /* Prepare input. */
  inary[0] =  name;
  inary[1] = &version;
  inary[2] = &flags;
  /* Prepare output. */
  outary[0] = &result;
  outary[1] =  handle;
  /* Invoke remote object method. */
  if (0 != rmix_invoke_p(harness, provider, &harnessclient_interface,
                         HARNESS_RMIX_METHODS_PLUGINS_LOAD2_INDEX,
                         outary, 2, inary, 3)) {
    int errno2 = errno;
    RMIX_LOG((RMIX_WARN(unable to invoke remote object method)))
    errno = errno2;
    return -1;
  }
  return result;
}

/*
 * Invokes harness_plugins_load2 (client-side method stub).
 * Loads a plug-in at a remote harness instance using a oneway invocation.
 *
 * harness = The RMIX remote object reference of harness.
 * name    = The plug-in name. Plug-in names are unique and do
 *           not contain '/..' and do not start with '/'.
 * version = The version of the plug-in to load.
 * flags   = The plug-in loading flags, e.g. HARNESS_PLUGINS_UNIQUE.
 * return  = 0 on success or -1 on any error with errno set appropriately.
 */
#undef  __FUNC__
#define __FUNC__ "harnessclient_plugins_load2_oneway"
int harnessclient_plugins_load2_oneway (const rmix_remoteref_t *harness,
                                        const char             *name,
                                        unsigned int            version,
                                        unsigned int            flags) {
  const void *inary[3];
#ifdef DEBUG
  /* Check harness parameter. */
  if (NULL == harness) {
    errno = EINVAL;
    HARNESS_LOG((HARNESS_WARN(harness parameter is null)))
    errno = EINVAL;
    return -1;
  }
  /* Check name parameter. */
  if (NULL == name) {
    errno = EINVAL;
    HARNESS_LOG((HARNESS_WARN(name parameter is null)))
    errno = EINVAL;
    return -1;
  }
#endif /* DEBUG */
  /* Prepare input. */
  inary[0] =  name;
  inary[1] = &version;
  inary[2] = &flags;
  /* Invoke remote object method. */
  if (0 != rmix_oneway(harness, &harnessclient_interface,
                       HARNESS_RMIX_METHODS_PLUGINS_LOAD2_INDEX,
                       inary, 3)) {
    int errno2 = errno;
    RMIX_LOG((RMIX_WARN(unable to invoke remote object method)))
    errno = errno2;
    return -1;
  }
  return 0;
}

/*
 * Invokes harness_plugins_load2 (client-side method stub).
 * Loads a plug-in at a remote harness instance using a oneway invocation
 * without loading the provider plug-in.
 *
 * harness  = The RMIX remote object reference of harness.
 * provider = The preloaded RMIX provider plug-in handle.
 * name     = The plug-in name. Plug-in names are unique and do
 *            not contain '/..' and do not start with '/'.
 * version  = The version of the plug-in to load.
 * flags    = The plug-in loading flags, e.g. HARNESS_PLUGINS_UNIQUE.
 * return   = 0 on success or -1 on any error with errno set appropriately.
 */
#undef  __FUNC__
#define __FUNC__ "harnessclient_plugins_load2_oneway_p"
int harnessclient_plugins_load2_oneway_p (const rmix_remoteref_t *harness,
                                          unsigned int            provider,
                                          const char             *name,
                                          unsigned int            version,
                                          unsigned int            flags) {
  const void *inary[3];
#ifdef DEBUG
  /* Check harness parameter. */
  if (NULL == harness) {
    errno = EINVAL;
    HARNESS_LOG((HARNESS_WARN(harness parameter is null)))
    errno = EINVAL;
    return -1;
  }
  /* Check name parameter. */
  if (NULL == name) {
    errno = EINVAL;
    HARNESS_LOG((HARNESS_WARN(name parameter is null)))
    errno = EINVAL;
    return -1;
  }
#endif /* DEBUG */
  /* Prepare input. */
  inary[0] =  name;
  inary[1] = &version;
  inary[2] = &flags;
  /* Invoke remote object method. */
  if (0 != rmix_oneway_p(harness, provider, &harnessclient_interface,
                         HARNESS_RMIX_METHODS_PLUGINS_LOAD2_INDEX,
                         inary, 3)) {
    int errno2 = errno;
    RMIX_LOG((RMIX_WARN(unable to invoke remote object method)))
    errno = errno2;
    return -1;
  }
  return 0;
}

/*
 * Invokes harness_plugins_load2 (client-side method stub).
 * Loads a plug-in at a remote harness instance using an asynchronous
 * invocation.
 *
 * invokeref = The invocation reference (return).
 * harness   = The RMIX remote object reference of harness.
 * name      = The plug-in name. Plug-in names are unique and do
 *             not contain '/..' and do not start with '/'.
 * version   = The version of the plug-in to load.
 * flags     = The plug-in loading flags, e.g. HARNESS_PLUGINS_UNIQUE.
 * return    = 0 on success or -1 on any error with errno set appropriately.
 */
#undef  __FUNC__
#define __FUNC__ "harnessclient_plugins_load2_send"
int harnessclient_plugins_load2_send (rmix_invokeref_t       **invokeref,
                                      const rmix_remoteref_t  *harness,
                                      const char              *name,
                                      unsigned int             version,
                                      unsigned int             flags) {
  const void *inary[3];
#ifdef DEBUG
  /* Check invokeref parameter. */
  if (NULL == invokeref) {
    errno = EINVAL;
    HARNESS_LOG((HARNESS_WARN(invokeref parameter is null)))
    errno = EINVAL;
    return -1;
  }
  /* Check harness parameter. */
  if (NULL == harness) {
    errno = EINVAL;
    HARNESS_LOG((HARNESS_WARN(harness parameter is null)))
    errno = EINVAL;
    return -1;
  }
  /* Check name parameter. */
  if (NULL == name) {
    errno = EINVAL;
    HARNESS_LOG((HARNESS_WARN(name parameter is null)))
    errno = EINVAL;
    return -1;
  }
#endif /* DEBUG */
  /* Prepare input. */
  inary[0] =  name;
  inary[1] = &version;
  inary[2] = &flags;
  /* Invoke remote object method. */
  if (0 != rmix_send(invokeref, harness, &harnessclient_interface,
                     HARNESS_RMIX_METHODS_PLUGINS_LOAD2_INDEX,
                     inary, 3)) {
    int errno2 = errno;
    RMIX_LOG((RMIX_WARN(unable to invoke remote object method)))
    errno = errno2;
    return -1;
  }
  return 0;
}

/*
 * Invokes harness_plugins_load2 (client-side method stub).
 * Loads a plug-in at a remote harness instance using an asynchronous invocation
 * without loading the provider plug-in.
 *
 * invokeref = The invocation reference (return).
 * harness   = The RMIX remote object reference of harness.
 * provider  = The preloaded RMIX provider plug-in handle.
 * name      = The plug-in name. Plug-in names are unique and do
 *             not contain '/..' and do not start with '/'.
 * version   = The version of the plug-in to load.
 * flags     = The plug-in loading flags, e.g. HARNESS_PLUGINS_UNIQUE.
 * return    = 0 on success or -1 on any error with errno set appropriately.
 */
#undef  __FUNC__
#define __FUNC__ "harnessclient_plugins_load2_send_p"
int harnessclient_plugins_load2_send_p (rmix_invokeref_t       **invokeref,
                                        const rmix_remoteref_t  *harness,
                                        unsigned int             provider,
                                        const char              *name,
                                        unsigned int             version,
                                        unsigned int             flags) {
  const void *inary[3];
#ifdef DEBUG
  /* Check invokeref parameter. */
  if (NULL == invokeref) {
    errno = EINVAL;
    HARNESS_LOG((HARNESS_WARN(invokeref parameter is null)))
    errno = EINVAL;
    return -1;
  }
  /* Check harness parameter. */
  if (NULL == harness) {
    errno = EINVAL;
    HARNESS_LOG((HARNESS_WARN(harness parameter is null)))
    errno = EINVAL;
    return -1;
  }
  /* Check name parameter. */
  if (NULL == name) {
    errno = EINVAL;
    HARNESS_LOG((HARNESS_WARN(name parameter is null)))
    errno = EINVAL;
    return -1;
  }
#endif /* DEBUG */
  /* Prepare input. */
  inary[0] =  name;
  inary[1] = &version;
  inary[2] = &flags;
  /* Invoke remote object method. */
  if (0 != rmix_send_p(invokeref, harness, provider, &harnessclient_interface,
                       HARNESS_RMIX_METHODS_PLUGINS_LOAD2_INDEX,
                       inary, 3)) {
    int errno2 = errno;
    RMIX_LOG((RMIX_WARN(unable to invoke remote object method)))
    errno = errno2;
    return -1;
  }
  return 0;
}

/*
 * Retrieves harness_plugins_load2 output (client-side method stub).
 * Returns the output of a previous asynchronous invocation that loads a plug-in
 * at a remote harness instance.
 *
 * invokeref = The invocation reference (also return).
 * handle    = The plug-in handle (return). Plug-in handles are unique.
 * return    = 0 on success or -1 on any error with errno set appropriately.
 */
#undef  __FUNC__
#define __FUNC__ "harnessclient_plugins_load2_retrieve"
int harnessclient_plugins_load2_retrieve (rmix_invokeref_t **invokeref,
                                          unsigned int      *handle) {
  int   result;
  void *outary[2];
#ifdef DEBUG
  /* Check invokeref parameter. */
  if (NULL == invokeref) {
    errno = EINVAL;
    HARNESS_LOG((HARNESS_WARN(invokeref parameter is null)))
    errno = EINVAL;
    return -1;
  }
  /* Check *invokeref parameter. */
  if (NULL == *invokeref) {
    errno = EINVAL;
    HARNESS_LOG((HARNESS_WARN(*invokeref parameter is null)))
    errno = EINVAL;
    return -1;
  }
  /* Check handle parameter. */
  if (NULL == handle) {
    errno = EINVAL;
    HARNESS_LOG((HARNESS_WARN(handle parameter is null)))
    errno = EINVAL;
    return -1;
  }
#endif /* DEBUG */
  /* Prepare output. */
  outary[0] = &result;
  outary[1] =  handle;
  /* Invoke remote object method. */
  if (0 != rmix_retrieve(invokeref, outary, 2)) {
    int errno2 = errno;
    RMIX_LOG((RMIX_WARN(unable to invoke remote object method)))
    errno = errno2;
    return -1;
  }
  return result;
}

/*
 * Retrieves harness_plugins_load2 output (client-side method stub).
 * Returns the output of a previous asynchronous invocation that loads a plug-in
 * at a remote harness instance without unloading the provider plug-in.
 *
 * invokeref = The invocation reference (also return).
 * handle    = The plug-in handle (return). Plug-in handles are unique.
 * return    = 0 on success or -1 on any error with errno set appropriately.
 */
#undef  __FUNC__
#define __FUNC__ "harnessclient_plugins_load2_retrieve_p"
int harnessclient_plugins_load2_retrieve_p (rmix_invokeref_t **invokeref,
                                            unsigned int      *handle) {
  int   result;
  void *outary[2];
#ifdef DEBUG
  /* Check invokeref parameter. */
  if (NULL == invokeref) {
    errno = EINVAL;
    HARNESS_LOG((HARNESS_WARN(invokeref parameter is null)))
    return -1;
  }
  /* Check *invokeref parameter. */
  if (NULL == *invokeref) {
    errno = EINVAL;
    HARNESS_LOG((HARNESS_WARN(*invokeref parameter is null)))
    return -1;
  }
  /* Check handle parameter. */
  if (NULL == handle) {
    errno = EINVAL;
    HARNESS_LOG((HARNESS_WARN(handle parameter is null)))
    return -1;
  }
#endif /* DEBUG */
  /* Prepare output. */
  outary[0] = &result;
  outary[1] =  handle;
  /* Invoke remote object method. */
  if (0 != rmix_retrieve_p(invokeref, outary, 2)) {
    int errno2 = errno;
    RMIX_LOG((RMIX_WARN(unable to invoke remote object method)))
    errno = errno2;
    return -1;
  }
  return result;
}

/*
 * Invokes harness_plugins_load2 (client-side method stub).
 * Loads a plug-in at multiple remote harness instances in sequence.
 * The plug-in may be different for every remote harness instance.
 *
 * count     = The number of remote harness instances.
 * harnesses = The array of RMIX remote object references of harness instances.
 * handles   = The plug-in handle array (also return). Plug-in handles are
 *             unique, and 0 for unsuccessful loading attempts.
 * names     = The plug-in name array. Plug-in names are unique and do
 *             not contain '/..' and do not start with '/'.
 * versions  = The version array of the plug-ins to load.
 * flags     = The plug-in loading flags array, e.g. HARNESS_PLUGINS_UNIQUE.
 * order     = The order flag. 0 for forward and 1 for reverse.
 * fail      = The fail flag. 0 to continue and 1 to fail on any error.
 * return    = The number of loaded plug-ins with errno set appropriately.
 */
#undef  __FUNC__
#define __FUNC__ "harnessclient_plugins_sload2_mp"
unsigned int harnessclient_plugins_sload2_mp
  (unsigned int             count,
   const rmix_remoteref_t **harnesses,
   unsigned int            *handles,
   const char             **names,
   const unsigned int      *versions,
   const unsigned int      *flags,
   unsigned int             order,
   unsigned int             fail) {
  int          errno2;
  unsigned int loaded, index;
#ifdef DEBUG
  /* Check harnesses parameter. */
  if ((NULL == harnesses)&&(0 != count)) {
    errno = EINVAL;
    RMIX_LOG((RMIX_WARN(harnesses parameter is null)))
    errno = EINVAL;
    return -1;
  }
  /* Check count parameter. */
  if ((0 == count)&&(NULL != harnesses)) {
    errno = EINVAL;
    RMIX_LOG((RMIX_WARN(count parameter is zero)))
    errno = EINVAL;
    return -1;
  }
  /* Check handles parameter. */
  if ((NULL == handles)&&(0 != count)) {
    errno = EINVAL;
    RMIX_LOG((RMIX_WARN(handles parameter is null)))
    errno = EINVAL;
    return -1;
  }
  /* Check count parameter. */
  if ((0 == count)&&(NULL != handles)) {
    errno = EINVAL;
    RMIX_LOG((RMIX_WARN(count parameter is zero)))
    errno = EINVAL;
    return -1;
  }
  /* Check names parameter. */
  if ((NULL == names)&&(0 != count)) {
    errno = EINVAL;
    RMIX_LOG((RMIX_WARN(names parameter is null)))
    errno = EINVAL;
    return -1;
  }
  /* Check count parameter. */
  if ((0 == count)&&(NULL != names)) {
    errno = EINVAL;
    RMIX_LOG((RMIX_WARN(count parameter is zero)))
    errno = EINVAL;
    return -1;
  }
  /* Check versions parameter. */
  if ((NULL == versions)&&(0 != count)) {
    errno = EINVAL;
    RMIX_LOG((RMIX_WARN(versions parameter is null)))
    errno = EINVAL;
    return -1;
  }
  /* Check count parameter. */
  if ((0 == count)&&(NULL != versions)) {
    errno = EINVAL;
    RMIX_LOG((RMIX_WARN(count parameter is zero)))
    errno = EINVAL;
    return -1;
  }
  /* Check flags parameter. */
  if ((NULL == flags)&&(0 != count)) {
    errno = EINVAL;
    RMIX_LOG((RMIX_WARN(flags parameter is null)))
    errno = EINVAL;
    return -1;
  }
  /* Check count parameter. */
  if ((0 == count)&&(NULL != flags)) {
    errno = EINVAL;
    RMIX_LOG((RMIX_WARN(count parameter is zero)))
    errno = EINVAL;
    return -1;
  }
#endif /* DEBUG */
  /* Initialize errno and loaded counter.*/
  errno2 = 0;
  loaded = 0;
  /* Initialize all handles. */
  memset(handles, 0, count * sizeof(*handles));
  /* Check order flag. */
  if (0 == order) {
    /* Load all plug-ins in forward order. */
    for (index = 0; index < count; index++) {
      /* Load plug-in. */
      if (0 != harnessclient_plugins_load2(harnesses[index], &handles[index],
                                           names[index], versions[index],
                                           flags[index])) {
        errno2 = (0 == errno2)?errno:errno2;
        RMIX_LOG((RMIX_WARN(unable to load plug-in)))
        /* Check fail flag. */
        if (1 == fail) {
          errno = errno2;
          return loaded;
        }
      } else {
        loaded++;
      }
    }
  } else {
    /* Load all plug-ins in reverse order. */
    for (index = count; 0 < index--;) {
      /* Load plug-in. */
      if (0 != harnessclient_plugins_load2(harnesses[index], &handles[index],
                                           names[index], versions[index],
                                           flags[index])) {
        errno2 = (0 == errno2)?errno:errno2;
        RMIX_LOG((RMIX_WARN(unable to load plug-in)))
        /* Check fail flag. */
        if (1 == fail) {
          errno = errno2;
          return loaded;
        }
      } else {
        loaded++;
      }
    }
  }
  errno = errno2;
  return loaded;
}

/*
 * Invokes harness_plugins_load2 (client-side method stub).
 * Loads a plug-in at multiple remote harness instances in sequence.
 * The plug-in is the same for every remote harness instance.
 *
 * count     = The number of remote harness instances.
 * harnesses = The array of RMIX remote object references of harness instances.
 * handles   = The plug-in handle array (also return). Plug-in handles are
 *             unique, and 0 for unsuccessful loading attempts.
 * name      = The plug-in name. Plug-in names are unique and do
 *             not contain '/..' and do not start with '/'.
 * version   = The version of the plug-in to load.
 * flags     = The plug-in loading flags, e.g. HARNESS_PLUGINS_UNIQUE.
 * order     = The order flag. 0 for forward and 1 for reverse.
 * fail      = The fail flag. 0 to continue and 1 to fail on any error.
 * return    = The number of loaded plug-ins with errno set appropriately.
 */
#undef  __FUNC__
#define __FUNC__ "harnessclient_plugins_sload2_sp"
unsigned int harnessclient_plugins_sload2_sp
  (unsigned int             count,
   const rmix_remoteref_t **harnesses,
   unsigned int            *handles,
   const char              *name,
   unsigned int             version,
   unsigned int             flags,
   unsigned int             order,
   unsigned int             fail) {
  int          errno2;
  unsigned int loaded, index;
#ifdef DEBUG
  /* Check harnesses parameter. */
  if ((NULL == harnesses)&&(0 != count)) {
    errno = EINVAL;
    RMIX_LOG((RMIX_WARN(harnesses parameter is null)))
    errno = EINVAL;
    return -1;
  }
  /* Check count parameter. */
  if ((0 == count)&&(NULL != harnesses)) {
    errno = EINVAL;
    RMIX_LOG((RMIX_WARN(count parameter is zero)))
    errno = EINVAL;
    return -1;
  }
  /* Check handles parameter. */
  if ((NULL == handles)&&(0 != count)) {
    errno = EINVAL;
    RMIX_LOG((RMIX_WARN(handles parameter is null)))
    errno = EINVAL;
    return -1;
  }
  /* Check count parameter. */
  if ((0 == count)&&(NULL != handles)) {
    errno = EINVAL;
    RMIX_LOG((RMIX_WARN(count parameter is zero)))
    errno = EINVAL;
    return -1;
  }
  /* Check name parameter. */
  if (NULL == name) {
    errno = EINVAL;
    HARNESS_LOG((HARNESS_WARN(name parameter is null)))
    return -1;
  }
#endif /* DEBUG */
  /* Initialize errno and loaded counter.*/
  errno2 = 0;
  loaded = 0;
  /* Initialize all handles. */
  memset(handles, 0, count * sizeof(*handles));
  /* Check order flag. */
  if (0 == order) {
    /* Load all plug-ins in forward order. */
    for (index = 0; index < count; index++) {
      /* Load plug-in. */
      if (0 != harnessclient_plugins_load2(harnesses[index], &handles[index],
                                           name, version, flags)) {
        errno2 = (0 == errno2)?errno:errno2;
        RMIX_LOG((RMIX_WARN(unable to load plug-in)))
        /* Check fail flag. */
        if (1 == fail) {
          errno = errno2;
          return loaded;
        }
      } else {
        loaded++;
      }
    }
  } else {
    /* Load all plug-ins in reverse order. */
    for (index = count; 0 < index--;) {
      /* Load plug-in. */
      if (0 != harnessclient_plugins_load2(harnesses[index], &handles[index],
                                           name, version, flags)) {
        errno2 = (0 == errno2)?errno:errno2;
        RMIX_LOG((RMIX_WARN(unable to load plug-in)))
        /* Check fail flag. */
        if (1 == fail) {
          errno = errno2;
          return loaded;
        }
      } else {
        loaded++;
      }
    }
  }
  errno = errno2;
  return loaded;
}

/*
 * Invokes harness_plugins_load2 (client-side method stub).
 * Loads a plug-in at multiple remote harness instances in parallel using
 * asynchonous method invocations.
 * The plug-in may be different for every remote harness instance.
 *
 * count     = The number of remote harness instances.
 * harnesses = The array of RMIX remote object references of harness instances.
 * handles   = The plug-in handle array (also return). Plug-in handles are
 *             unique, and 0 for unsuccessful loading attempts.
 * names     = The plug-in name array. Plug-in names are unique and do
 *             not contain '/..' and do not start with '/'.
 * versions  = The version array of the plug-ins to load.
 * flags     = The plug-in loading flags array, e.g. HARNESS_PLUGINS_UNIQUE.
 * return    = The number of loaded plug-ins with errno set appropriately.
 */
#undef  __FUNC__
#define __FUNC__ "harnessclient_plugins_pload2_mp"
unsigned int harnessclient_plugins_pload2_mp
  (unsigned int             count,
   const rmix_remoteref_t **harnesses,
   unsigned int            *handles,
   const char             **names,
   const unsigned int      *versions,
   const unsigned int      *flags) {
  int                errno2;
  unsigned int       loaded, index;
  rmix_invokeref_t **invocations;
#ifdef DEBUG
  /* Check harnesses parameter. */
  if ((NULL == harnesses)&&(0 != count)) {
    errno = EINVAL;
    RMIX_LOG((RMIX_WARN(harnesses parameter is null)))
    errno = EINVAL;
    return -1;
  }
  /* Check count parameter. */
  if ((0 == count)&&(NULL != harnesses)) {
    errno = EINVAL;
    RMIX_LOG((RMIX_WARN(count parameter is zero)))
    errno = EINVAL;
    return -1;
  }
  /* Check handles parameter. */
  if ((NULL == handles)&&(0 != count)) {
    errno = EINVAL;
    RMIX_LOG((RMIX_WARN(handles parameter is null)))
    errno = EINVAL;
    return -1;
  }
  /* Check count parameter. */
  if ((0 == count)&&(NULL != handles)) {
    errno = EINVAL;
    RMIX_LOG((RMIX_WARN(count parameter is zero)))
    errno = EINVAL;
    return -1;
  }
  /* Check names parameter. */
  if ((NULL == names)&&(0 != count)) {
    errno = EINVAL;
    RMIX_LOG((RMIX_WARN(names parameter is null)))
    errno = EINVAL;
    return -1;
  }
  /* Check count parameter. */
  if ((0 == count)&&(NULL != names)) {
    errno = EINVAL;
    RMIX_LOG((RMIX_WARN(count parameter is zero)))
    errno = EINVAL;
    return -1;
  }
  /* Check versions parameter. */
  if ((NULL == versions)&&(0 != count)) {
    errno = EINVAL;
    RMIX_LOG((RMIX_WARN(versions parameter is null)))
    errno = EINVAL;
    return -1;
  }
  /* Check count parameter. */
  if ((0 == count)&&(NULL != versions)) {
    errno = EINVAL;
    RMIX_LOG((RMIX_WARN(count parameter is zero)))
    errno = EINVAL;
    return -1;
  }
  /* Check flags parameter. */
  if ((NULL == flags)&&(0 != count)) {
    errno = EINVAL;
    RMIX_LOG((RMIX_WARN(flags parameter is null)))
    errno = EINVAL;
    return -1;
  }
  /* Check count parameter. */
  if ((0 == count)&&(NULL != flags)) {
    errno = EINVAL;
    RMIX_LOG((RMIX_WARN(count parameter is zero)))
    errno = EINVAL;
    return -1;
  }
#endif /* DEBUG */
  /* Allocate invocation reference array. */
  if (NULL == (invocations = malloc(count * sizeof(rmix_invokeref_t*)))) {
    errno2 = errno;
    RMIX_LOG((RMIX_WARN(unable to allocate invocation reference array)))
    errno = errno2;
    return 0;
  }
  /* Initialize errno and loaded counter.*/
  errno2 = 0;
  loaded = 0;
  /* Initialize all handles. */
  memset(handles, 0, count * sizeof(*handles));
  /* Load all plug-ins in forward order (send part). */
  for (index = 0; index < count; index++) {
    /* Load plug-in (send part). */
    if (0 != harnessclient_plugins_load2_send(&invocations[index],
                                              harnesses[index], names[index],
                                              versions[index], flags[index])) {
      errno2 = (0 == errno2)?errno:errno2;
      RMIX_LOG((RMIX_WARN(unable to load plug-in (send part))))
      invocations[index] = NULL;
    }
  }
  /* Load all plug-ins in forward order (retrieve part). */
  for (index = 0; index < count; index++) {
    /* Check for successful send. */
    if (NULL != invocations[index]) {
      /* Load plug-in (retrieve part). */
      if (0 != harnessclient_plugins_load2_retrieve(&invocations[index],
                                                    &handles[index])) {
        errno2 = (0 == errno2)?errno:errno2;
        RMIX_LOG((RMIX_WARN(unable to load plug-in (retrieve part))))
      } else {
        loaded++;
      }
    }
  }
  /* Deallocate invocation reference array. */
  free(invocations);
  errno = errno2;
  return loaded;
}

/*
 * Invokes harness_plugins_load2 (client-side method stub).
 * Loads a plug-in at multiple remote harness instances in parallel using
 * asynchonous method invocations.
 * The plug-in is the same for every remote harness instance.
 *
 * count     = The number of remote harness instances.
 * harnesses = The array of RMIX remote object references of harness instances.
 * handles   = The plug-in handle array (also return). Plug-in handles are
 *             unique, and 0 for unsuccessful loading attempts.
 * name      = The plug-in name. Plug-in names are unique and do
 *             not contain '/..' and do not start with '/'.
 * version   = The version of the plug-in to load.
 * flags     = The plug-in loading flags, e.g. HARNESS_PLUGINS_UNIQUE.
 * return    = The number of loaded plug-ins with errno set appropriately.
 */
#undef  __FUNC__
#define __FUNC__ "harnessclient_plugins_pload2_sp"
unsigned int harnessclient_plugins_pload2_sp
  (unsigned int             count,
   const rmix_remoteref_t **harnesses,
   unsigned int            *handles,
   const char              *name,
   unsigned int             version,
   unsigned int             flags) {
  int                errno2;
  unsigned int       loaded, index;
  rmix_invokeref_t **invocations;
#ifdef DEBUG
  /* Check harnesses parameter. */
  if ((NULL == harnesses)&&(0 != count)) {
    errno = EINVAL;
    RMIX_LOG((RMIX_WARN(harnesses parameter is null)))
    errno = EINVAL;
    return -1;
  }
  /* Check count parameter. */
  if ((0 == count)&&(NULL != harnesses)) {
    errno = EINVAL;
    RMIX_LOG((RMIX_WARN(count parameter is zero)))
    errno = EINVAL;
    return -1;
  }
  /* Check handles parameter. */
  if ((NULL == handles)&&(0 != count)) {
    errno = EINVAL;
    RMIX_LOG((RMIX_WARN(handles parameter is null)))
    errno = EINVAL;
    return -1;
  }
  /* Check count parameter. */
  if ((0 == count)&&(NULL != handles)) {
    errno = EINVAL;
    RMIX_LOG((RMIX_WARN(count parameter is zero)))
    errno = EINVAL;
    return -1;
  }
  /* Check name parameter. */
  if (NULL == name) {
    errno = EINVAL;
    HARNESS_LOG((HARNESS_WARN(name parameter is null)))
    return -1;
  }
#endif /* DEBUG */
  /* Allocate invocation reference array. */
  if (NULL == (invocations = malloc(count * sizeof(rmix_invokeref_t*)))) {
    errno2 = errno;
    RMIX_LOG((RMIX_WARN(unable to allocate invocation reference array)))
    errno = errno2;
    return 0;
  }
  /* Initialize errno and loaded counter.*/
  errno2 = 0;
  loaded = 0;
  /* Initialize all handles. */
  memset(handles, 0, count * sizeof(*handles));
  /* Load all plug-ins in forward order (send part). */
  for (index = 0; index < count; index++) {
    /* Load plug-in (send part). */
    if (0 != harnessclient_plugins_load2_send(&invocations[index],
                                              harnesses[index],
                                              name, version, flags)) {
      errno2 = (0 == errno2)?errno:errno2;
      RMIX_LOG((RMIX_WARN(unable to load plug-in (send part))))
      invocations[index] = NULL;
    }
  }
  /* Load all plug-ins in forward order (retrieve part). */
  for (index = 0; index < count; index++) {
    /* Check for successful send. */
    if (NULL != invocations[index]) {
      /* Load plug-in (retrieve part). */
      if (0 != harnessclient_plugins_load2_retrieve(&invocations[index],
                                                    &handles[index])) {
        errno2 = (0 == errno2)?errno:errno2;
        RMIX_LOG((RMIX_WARN(unable to load plug-in (retrieve part))))
      } else {
        loaded++;
      }
    }
  }
  /* Deallocate invocation reference array. */
  free(invocations);
  errno = errno2;
  return loaded;
}

/*
 * Invokes harness_plugins_load3 (client-side method stub).
 * Loads a plug-in at a remote harness instance.
 *
 * harness  = The RMIX remote object reference of harness.
 * handle   = The plug-in handle (return). Plug-in handles are unique.
 * name     = The plug-in name. Plug-in names are unique and do
 *            not contain '/..' and do not start with '/'.
 * first    = The first version supported of the plug-in to load.
 * age      = The number of followup versions supported of the plug-in to load.
 * revision = The revision of the plug-in to load.
 * flags    = The plug-in loading flags, e.g. HARNESS_PLUGINS_UNIQUE.
 * return   = 0 on success or -1 on any error with errno set appropriately.
 */
#undef  __FUNC__
#define __FUNC__ "harnessclient_plugins_load3"
int harnessclient_plugins_load3 (const rmix_remoteref_t *harness,
                                 unsigned int           *handle,
                                 const char             *name,
                                 unsigned int            first,
                                 unsigned int            age,
                                 unsigned int            revision,
                                 unsigned int            flags) {
  int         result;
  void       *outary[2];
  const void *inary[5];
#ifdef DEBUG
  /* Check harness parameter. */
  if (NULL == harness) {
    errno = EINVAL;
    HARNESS_LOG((HARNESS_WARN(harness parameter is null)))
    return -1;
  }
  /* Check handle parameter. */
  if (NULL == handle) {
    errno = EINVAL;
    HARNESS_LOG((HARNESS_WARN(handle parameter is null)))
    return -1;
  }
  /* Check name parameter. */
  if (NULL == name) {
    errno = EINVAL;
    HARNESS_LOG((HARNESS_WARN(name parameter is null)))
    return -1;
  }
#endif /* DEBUG */
  /* Prepare input. */
  inary[0] =  name;
  inary[1] = &first;
  inary[2] = &age;
  inary[3] = &revision;
  inary[4] = &flags;
  /* Prepare output. */
  outary[0] = &result;
  outary[1] =  handle;
  /* Invoke remote object method. */
  if (0 != rmix_invoke(harness, &harnessclient_interface,
                       HARNESS_RMIX_METHODS_PLUGINS_LOAD3_INDEX,
                       outary, 2, inary, 5)) {
    int errno2 = errno;
    RMIX_LOG((RMIX_WARN(unable to invoke remote object method)))
    errno = errno2;
    return -1;
  }
  return result;
}

/*
 * Invokes harness_plugins_load3 (client-side method stub).
 * Loads a plug-in at a remote harness instance.
 *
 * harness  = The RMIX remote object reference of harness.
 * provider = The preloaded RMIX provider plug-in handle.
 * handle   = The plug-in handle (return). Plug-in handles are unique.
 * name     = The plug-in name. Plug-in names are unique and do
 *            not contain '/..' and do not start with '/'.
 * first    = The first version supported of the plug-in to load.
 * age      = The number of followup versions supported of the plug-in to load.
 * revision = The revision of the plug-in to load.
 * flags    = The plug-in loading flags, e.g. HARNESS_PLUGINS_UNIQUE.
 * return   = 0 on success or -1 on any error with errno set appropriately.
 */
#undef  __FUNC__
#define __FUNC__ "harnessclient_plugins_load3_p"
int harnessclient_plugins_load3_p (const rmix_remoteref_t *harness,
                                   unsigned int            provider,
                                   unsigned int           *handle,
                                   const char             *name,
                                   unsigned int            first,
                                   unsigned int            age,
                                   unsigned int            revision,
                                   unsigned int            flags) {
  int         result;
  void       *outary[2];
  const void *inary[5];
#ifdef DEBUG
  /* Check harness parameter. */
  if (NULL == harness) {
    errno = EINVAL;
    HARNESS_LOG((HARNESS_WARN(harness parameter is null)))
    return -1;
  }
  /* Check handle parameter. */
  if (NULL == handle) {
    errno = EINVAL;
    HARNESS_LOG((HARNESS_WARN(handle parameter is null)))
    return -1;
  }
  /* Check name parameter. */
  if (NULL == name) {
    errno = EINVAL;
    HARNESS_LOG((HARNESS_WARN(name parameter is null)))
    return -1;
  }
#endif /* DEBUG */
  /* Prepare input. */
  inary[0] =  name;
  inary[1] = &first;
  inary[2] = &age;
  inary[3] = &revision;
  inary[4] = &flags;
  /* Prepare output. */
  outary[0] = &result;
  outary[1] =  handle;
  /* Invoke remote object method. */
  if (0 != rmix_invoke_p(harness, provider, &harnessclient_interface,
                         HARNESS_RMIX_METHODS_PLUGINS_LOAD3_INDEX,
                         outary, 2, inary, 5)) {
    int errno2 = errno;
    RMIX_LOG((RMIX_WARN(unable to invoke remote object method)))
    errno = errno2;
    return -1;
  }
  return result;
}

/*
 * Invokes harness_plugins_load3 (client-side method stub).
 * Loads a plug-in at a remote harness instance using a oneway invocation.
 *
 * harness  = The RMIX remote object reference of harness.
 * name     = The plug-in name. Plug-in names are unique and do
 *            not contain '/..' and do not start with '/'.
 * first    = The first version supported of the plug-in to load.
 * age      = The number of followup versions supported of the plug-in to load.
 * revision = The revision of the plug-in to load.
 * flags    = The plug-in loading flags, e.g. HARNESS_PLUGINS_UNIQUE.
 * return   = 0 on success or -1 on any error with errno set appropriately.
 */
#undef  __FUNC__
#define __FUNC__ "harnessclient_plugins_load3_oneway"
int harnessclient_plugins_load3_oneway (const rmix_remoteref_t *harness,
                                        const char             *name,
                                        unsigned int            first,
                                        unsigned int            age,
                                        unsigned int            revision,
                                        unsigned int            flags) {
  const void *inary[5];
#ifdef DEBUG
  /* Check harness parameter. */
  if (NULL == harness) {
    errno = EINVAL;
    HARNESS_LOG((HARNESS_WARN(harness parameter is null)))
    return -1;
  }
  /* Check name parameter. */
  if (NULL == name) {
    errno = EINVAL;
    HARNESS_LOG((HARNESS_WARN(name parameter is null)))
    return -1;
  }
#endif /* DEBUG */
  /* Prepare input. */
  inary[0] =  name;
  inary[1] = &first;
  inary[2] = &age;
  inary[3] = &revision;
  inary[4] = &flags;
  /* Invoke remote object method. */
  if (0 != rmix_oneway(harness, &harnessclient_interface,
                       HARNESS_RMIX_METHODS_PLUGINS_LOAD3_INDEX,
                     inary, 5)) {
    int errno2 = errno;
    RMIX_LOG((RMIX_WARN(unable to invoke remote object method)))
    errno = errno2;
    return -1;
  }
  return 0;
}

/*
 * Invokes harness_plugins_load3 (client-side method stub).
 * Loads a plug-in at a remote harness instance using a oneway invocation
 * without loading the provider plug-in.
 *
 * harness  = The RMIX remote object reference of harness.
 * provider = The preloaded RMIX provider plug-in handle.
 * name     = The plug-in name. Plug-in names are unique and do
 *            not contain '/..' and do not start with '/'.
 * first    = The first version supported of the plug-in to load.
 * age      = The number of followup versions supported of the plug-in to load.
 * revision = The revision of the plug-in to load.
 * flags    = The plug-in loading flags, e.g. HARNESS_PLUGINS_UNIQUE.
 * return   = 0 on success or -1 on any error with errno set appropriately.
 */
#undef  __FUNC__
#define __FUNC__ "harnessclient_plugins_load3_oneway_p"
int harnessclient_plugins_load3_oneway_p (const rmix_remoteref_t *harness,
                                          unsigned int            provider,
                                          const char             *name,
                                          unsigned int            first,
                                          unsigned int            age,
                                          unsigned int            revision,
                                          unsigned int            flags) {
  const void *inary[5];
#ifdef DEBUG
  /* Check harness parameter. */
  if (NULL == harness) {
    errno = EINVAL;
    HARNESS_LOG((HARNESS_WARN(harness parameter is null)))
    return -1;
  }
  /* Check name parameter. */
  if (NULL == name) {
    errno = EINVAL;
    HARNESS_LOG((HARNESS_WARN(name parameter is null)))
    return -1;
  }
#endif /* DEBUG */
  /* Prepare input. */
  inary[0] =  name;
  inary[1] = &first;
  inary[2] = &age;
  inary[3] = &revision;
  inary[4] = &flags;
  /* Invoke remote object method. */
  if (0 != rmix_oneway_p(harness, provider, &harnessclient_interface,
                         HARNESS_RMIX_METHODS_PLUGINS_LOAD3_INDEX,
                         inary, 5)) {
    int errno2 = errno;
    RMIX_LOG((RMIX_WARN(unable to invoke remote object method)))
    errno = errno2;
    return -1;
  }
  return 0;
}

/*
 * Invokes harness_plugins_load3 (client-side method stub).
 * Loads a plug-in at a remote harness instance using an asynchronous
 * invocation.
 *
 * invokeref = The invocation reference (return).
 * harness   = The RMIX remote object reference of harness.
 * name      = The plug-in name. Plug-in names are unique and do
 *             not contain '/..' and do not start with '/'.
 * first     = The first version supported of the plug-in to load.
 * age       = The number of followup versions supported of the plug-in to load.
 * revision  = The revision of the plug-in to load.
 * flags     = The plug-in loading flags, e.g. HARNESS_PLUGINS_UNIQUE.
 * return    = 0 on success or -1 on any error with errno set appropriately.
 */
#undef  __FUNC__
#define __FUNC__ "harnessclient_plugins_load3_send"
int harnessclient_plugins_load3_send (rmix_invokeref_t       **invokeref,
                                      const rmix_remoteref_t  *harness,
                                      const char              *name,
                                      unsigned int             first,
                                      unsigned int             age,
                                      unsigned int             revision,
                                      unsigned int             flags) {
  const void *inary[5];
#ifdef DEBUG
  /* Check invokeref parameter. */
  if (NULL == invokeref) {
    errno = EINVAL;
    HARNESS_LOG((HARNESS_WARN(invokeref parameter is null)))
    return -1;
  }
  /* Check harness parameter. */
  if (NULL == harness) {
    errno = EINVAL;
    HARNESS_LOG((HARNESS_WARN(harness parameter is null)))
    return -1;
  }
  /* Check name parameter. */
  if (NULL == name) {
    errno = EINVAL;
    HARNESS_LOG((HARNESS_WARN(name parameter is null)))
    errno = EINVAL;
    return -1;
  }
#endif /* DEBUG */
  /* Prepare input. */
  inary[0] =  name;
  inary[1] = &first;
  inary[2] = &age;
  inary[3] = &revision;
  inary[4] = &flags;
  /* Invoke remote object method. */
  if (0 != rmix_send(invokeref, harness, &harnessclient_interface,
                     HARNESS_RMIX_METHODS_PLUGINS_LOAD3_INDEX,
                     inary, 5)) {
    int errno2 = errno;
    RMIX_LOG((RMIX_WARN(unable to invoke remote object method)))
    errno = errno2;
    return -1;
  }
  return 0;
}

/*
 * Invokes harness_plugins_load3 (client-side method stub).
 * Loads a plug-in at a remote harness instance using an asynchronous invocation
 * without loading the provider plug-in.
 *
 * invokeref = The invocation reference (return).
 * harness   = The RMIX remote object reference of harness.
 * provider  = The preloaded RMIX provider plug-in handle.
 * name      = The plug-in name. Plug-in names are unique and do
 *             not contain '/..' and do not start with '/'.
 * first     = The first version supported of the plug-in to load.
 * age       = The number of followup versions supported of the plug-in to load.
 * revision  = The revision of the plug-in to load.
 * flags     = The plug-in loading flags, e.g. HARNESS_PLUGINS_UNIQUE.
 * return    = 0 on success or -1 on any error with errno set appropriately.
 */
#undef  __FUNC__
#define __FUNC__ "harnessclient_plugins_load3_send_p"
int harnessclient_plugins_load3_send_p (rmix_invokeref_t       **invokeref,
                                        const rmix_remoteref_t  *harness,
                                        unsigned int             provider,
                                        const char              *name,
                                        unsigned int             first,
                                        unsigned int             age,
                                        unsigned int             revision,
                                        unsigned int             flags) {
  const void *inary[5];
#ifdef DEBUG
  /* Check invokeref parameter. */
  if (NULL == invokeref) {
    errno = EINVAL;
    HARNESS_LOG((HARNESS_WARN(invokeref parameter is null)))
    errno = EINVAL;
    return -1;
  }
  /* Check harness parameter. */
  if (NULL == harness) {
    errno = EINVAL;
    HARNESS_LOG((HARNESS_WARN(harness parameter is null)))
    errno = EINVAL;
    return -1;
  }
  /* Check name parameter. */
  if (NULL == name) {
    errno = EINVAL;
    HARNESS_LOG((HARNESS_WARN(name parameter is null)))
    errno = EINVAL;
    return -1;
  }
#endif /* DEBUG */
  /* Prepare input. */
  inary[0] =  name;
  inary[1] = &first;
  inary[2] = &age;
  inary[3] = &revision;
  inary[4] = &flags;
  /* Invoke remote object method. */
  if (0 != rmix_send_p(invokeref, harness, provider, &harnessclient_interface,
                       HARNESS_RMIX_METHODS_PLUGINS_LOAD3_INDEX,
                       inary, 5)) {
    int errno2 = errno;
    RMIX_LOG((RMIX_WARN(unable to invoke remote object method)))
    errno = errno2;
    return -1;
  }
  return 0;
}

/*
 * Retrieves harness_plugins_load3 output (client-side method stub).
 * Returns the output of a previous asynchronous invocation that loads a plug-in
 * at a remote harness instance.
 *
 * invokeref = The invocation reference (also return).
 * handle    = The plug-in handle (return). Plug-in handles are unique.
 * return    = 0 on success or -1 on any error with errno set appropriately.
 */
#undef  __FUNC__
#define __FUNC__ "harnessclient_plugins_load3_retrieve"
int harnessclient_plugins_load3_retrieve (rmix_invokeref_t **invokeref,
                                          unsigned int      *handle) {
  int   result;
  void *outary[2];
#ifdef DEBUG
  /* Check invokeref parameter. */
  if (NULL == invokeref) {
    errno = EINVAL;
    HARNESS_LOG((HARNESS_WARN(invokeref parameter is null)))
    errno = EINVAL;
    return -1;
  }
  /* Check *invokeref parameter. */
  if (NULL == *invokeref) {
    errno = EINVAL;
    HARNESS_LOG((HARNESS_WARN(*invokeref parameter is null)))
    errno = EINVAL;
    return -1;
  }
  /* Check handle parameter. */
  if (NULL == handle) {
    errno = EINVAL;
    HARNESS_LOG((HARNESS_WARN(handle parameter is null)))
    errno = EINVAL;
    return -1;
  }
#endif /* DEBUG */
  /* Prepare output. */
  outary[0] = &result;
  outary[1] =  handle;
  /* Invoke remote object method. */
  if (0 != rmix_retrieve(invokeref, outary, 2)) {
    int errno2 = errno;
    RMIX_LOG((RMIX_WARN(unable to invoke remote object method)))
    errno = errno2;
    return -1;
  }
  return result;
}

/*
 * Retrieves harness_plugins_load3 output (client-side method stub).
 * Returns the output of a previous asynchronous invocation that loads a plug-in
 * at a remote harness instance without unloading the provider plug-in.
 *
 * invokeref = The invocation reference (also return).
 * handle    = The plug-in handle (return). Plug-in handles are unique.
 * return    = 0 on success or -1 on any error with errno set appropriately.
 */
#undef  __FUNC__
#define __FUNC__ "harnessclient_plugins_load3_retrieve_p"
int harnessclient_plugins_load3_retrieve_p (rmix_invokeref_t **invokeref,
                                            unsigned int      *handle) {
  int   result;
  void *outary[2];
#ifdef DEBUG
  /* Check invokeref parameter. */
  if (NULL == invokeref) {
    errno = EINVAL;
    HARNESS_LOG((HARNESS_WARN(invokeref parameter is null)))
    errno = EINVAL;
    return -1;
  }
  /* Check *invokeref parameter. */
  if (NULL == *invokeref) {
    errno = EINVAL;
    HARNESS_LOG((HARNESS_WARN(*invokeref parameter is null)))
    errno = EINVAL;
    return -1;
  }
  /* Check handle parameter. */
  if (NULL == handle) {
    errno = EINVAL;
    HARNESS_LOG((HARNESS_WARN(handle parameter is null)))
    errno = EINVAL;
    return -1;
  }
#endif /* DEBUG */
  /* Prepare output. */
  outary[0] = &result;
  outary[1] =  handle;
  /* Invoke remote object method. */
  if (0 != rmix_retrieve_p(invokeref, outary, 2)) {
    int errno2 = errno;
    RMIX_LOG((RMIX_WARN(unable to invoke remote object method)))
    errno = errno2;
    return -1;
  }
  return result;
}



/*
 * Invokes harness_plugins_load3 (client-side method stub).
 * Loads a plug-in at multiple remote harness instances in sequence.
 * The plug-in may be different for every remote harness instance.
 *
 * count     = The number of remote harness instances.
 * harnesses = The array of RMIX remote object references of harness instances.
 * handles   = The plug-in handle array (also return). Plug-in handles are
 *             unique, and 0 for unsuccessful loading attempts.
 * names     = The plug-in name array. Plug-in names are unique and do
 *             not contain '/..' and do not start with '/'.
 * firsts    = The array of first versions supported of the plug-ins to load.
 * ages      = The array number of followup versions supported of the plug-ins
 *             to load.
 * revisions = The array revisions of the plug-ins to load.
 * flags     = The plug-in loading flags array, e.g. HARNESS_PLUGINS_UNIQUE.
 * order     = The order flag. 0 for forward and 1 for reverse.
 * fail      = The fail flag. 0 to continue and 1 to fail on any error.
 * return    = The number of loaded plug-ins with errno set appropriately.
 */
#undef  __FUNC__
#define __FUNC__ "harnessclient_plugins_sload3_mp"
unsigned int harnessclient_plugins_sload3_mp
  (unsigned int             count,
   const rmix_remoteref_t **harnesses,
   unsigned int            *handles,
   const char             **names,
   const unsigned int      *firsts,
   const unsigned int      *ages,
   const unsigned int      *revisions,
   const unsigned int      *flags,
   unsigned int             order,
   unsigned int             fail) {
  int          errno2;
  unsigned int loaded, index;
#ifdef DEBUG
  /* Check harnesses parameter. */
  if ((NULL == harnesses)&&(0 != count)) {
    errno = EINVAL;
    RMIX_LOG((RMIX_WARN(harnesses parameter is null)))
    errno = EINVAL;
    return -1;
  }
  /* Check count parameter. */
  if ((0 == count)&&(NULL != harnesses)) {
    errno = EINVAL;
    RMIX_LOG((RMIX_WARN(count parameter is zero)))
    errno = EINVAL;
    return -1;
  }
  /* Check handles parameter. */
  if ((NULL == handles)&&(0 != count)) {
    errno = EINVAL;
    RMIX_LOG((RMIX_WARN(handles parameter is null)))
    errno = EINVAL;
    return -1;
  }
  /* Check count parameter. */
  if ((0 == count)&&(NULL != handles)) {
    errno = EINVAL;
    RMIX_LOG((RMIX_WARN(count parameter is zero)))
    errno = EINVAL;
    return -1;
  }
  /* Check names parameter. */
  if ((NULL == names)&&(0 != count)) {
    errno = EINVAL;
    RMIX_LOG((RMIX_WARN(names parameter is null)))
    errno = EINVAL;
    return -1;
  }
  /* Check count parameter. */
  if ((0 == count)&&(NULL != names)) {
    errno = EINVAL;
    RMIX_LOG((RMIX_WARN(count parameter is zero)))
    errno = EINVAL;
    return -1;
  }
  /* Check firsts parameter. */
  if ((NULL == firsts)&&(0 != count)) {
    errno = EINVAL;
    RMIX_LOG((RMIX_WARN(firsts parameter is null)))
    errno = EINVAL;
    return -1;
  }
  /* Check count parameter. */
  if ((0 == count)&&(NULL != firsts)) {
    errno = EINVAL;
    RMIX_LOG((RMIX_WARN(count parameter is zero)))
    errno = EINVAL;
    return -1;
  }
  /* Check ages parameter. */
  if ((NULL == ages)&&(0 != count)) {
    errno = EINVAL;
    RMIX_LOG((RMIX_WARN(ages parameter is null)))
    errno = EINVAL;
    return -1;
  }
  /* Check count parameter. */
  if ((0 == count)&&(NULL != ages)) {
    errno = EINVAL;
    RMIX_LOG((RMIX_WARN(count parameter is zero)))
    errno = EINVAL;
    return -1;
  }
  /* Check revisions parameter. */
  if ((NULL == revisions)&&(0 != count)) {
    errno = EINVAL;
    RMIX_LOG((RMIX_WARN(revisions parameter is null)))
    errno = EINVAL;
    return -1;
  }
  /* Check count parameter. */
  if ((0 == count)&&(NULL != revisions)) {
    errno = EINVAL;
    RMIX_LOG((RMIX_WARN(count parameter is zero)))
    errno = EINVAL;
    return -1;
  }
  /* Check flags parameter. */
  if ((NULL == flags)&&(0 != count)) {
    errno = EINVAL;
    RMIX_LOG((RMIX_WARN(flags parameter is null)))
    errno = EINVAL;
    return -1;
  }
  /* Check count parameter. */
  if ((0 == count)&&(NULL != flags)) {
    errno = EINVAL;
    RMIX_LOG((RMIX_WARN(count parameter is zero)))
    errno = EINVAL;
    return -1;
  }
#endif /* DEBUG */
  /* Initialize errno and loaded counter.*/
  errno2 = 0;
  loaded = 0;
  /* Initialize all handles. */
  memset(handles, 0, count * sizeof(*handles));
  /* Check order flag. */
  if (0 == order) {
    /* Load all plug-ins in forward order. */
    for (index = 0; index < count; index++) {
      /* Load plug-in. */
      if (0 != harnessclient_plugins_load3(harnesses[index], &handles[index],
                                           names[index], firsts[index],
                                           ages[index], revisions[index],
                                           flags[index])) {
        errno2 = (0 == errno2)?errno:errno2;
        RMIX_LOG((RMIX_WARN(unable to load plug-in)))
        /* Check fail flag. */
        if (1 == fail) {
          errno = errno2;
          return loaded;
        }
      } else {
        loaded++;
      }
    }
  } else {
    /* Load all plug-ins in reverse order. */
    for (index = count; 0 < index--;) {
      /* Load plug-in. */
      if (0 != harnessclient_plugins_load3(harnesses[index], &handles[index],
                                           names[index], firsts[index],
                                           ages[index], revisions[index],
                                           flags[index])) {
        errno2 = (0 == errno2)?errno:errno2;
        RMIX_LOG((RMIX_WARN(unable to load plug-in)))
        /* Check fail flag. */
        if (1 == fail) {
          errno = errno2;
          return loaded;
        }
      } else {
        loaded++;
      }
    }
  }
  errno = errno2;
  return loaded;
}

/*
 * Invokes harness_plugins_load3 (client-side method stub).
 * Loads a plug-in at multiple remote harness instances in sequence.
 * The plug-in is the same for every remote harness instance.
 *
 * count     = The number of remote harness instances.
 * harnesses = The array of RMIX remote object references of harness instances.
 * handles   = The plug-in handle array (also return). Plug-in handles are
 *             unique, and 0 for unsuccessful loading attempts.
 * name      = The plug-in name. Plug-in names are unique and do
 *             not contain '/..' and do not start with '/'.
 * first     = The first version supported of the plug-in to load.
 * age       = The number of followup versions supported of the plug-in to load.
 * revision  = The revision of the plug-in to load.
 * flags     = The plug-in loading flags, e.g. HARNESS_PLUGINS_UNIQUE.
 * order     = The order flag. 0 for forward and 1 for reverse.
 * fail      = The fail flag. 0 to continue and 1 to fail on any error.
 * return    = The number of loaded plug-ins with errno set appropriately.
 */
#undef  __FUNC__
#define __FUNC__ "harnessclient_plugins_sload3_sp"
unsigned int harnessclient_plugins_sload3_sp
  (unsigned int             count,
   const rmix_remoteref_t **harnesses,
   unsigned int            *handles,
   const char              *name,
   unsigned int             first,
   unsigned int             age,
   unsigned int             revision,
   unsigned int             flags,
   unsigned int             order,
   unsigned int             fail) {
  int          errno2;
  unsigned int loaded, index;
#ifdef DEBUG
  /* Check harnesses parameter. */
  if ((NULL == harnesses)&&(0 != count)) {
    errno = EINVAL;
    RMIX_LOG((RMIX_WARN(harnesses parameter is null)))
    errno = EINVAL;
    return -1;
  }
  /* Check count parameter. */
  if ((0 == count)&&(NULL != harnesses)) {
    errno = EINVAL;
    RMIX_LOG((RMIX_WARN(count parameter is zero)))
    errno = EINVAL;
    return -1;
  }
  /* Check handles parameter. */
  if ((NULL == handles)&&(0 != count)) {
    errno = EINVAL;
    RMIX_LOG((RMIX_WARN(handles parameter is null)))
    errno = EINVAL;
    return -1;
  }
  /* Check count parameter. */
  if ((0 == count)&&(NULL != handles)) {
    errno = EINVAL;
    RMIX_LOG((RMIX_WARN(count parameter is zero)))
    errno = EINVAL;
    return -1;
  }
  /* Check name parameter. */
  if (NULL == name) {
    errno = EINVAL;
    HARNESS_LOG((HARNESS_WARN(name parameter is null)))
    errno = EINVAL;
    return -1;
  }
#endif /* DEBUG */
  /* Initialize errno and loaded counter.*/
  errno2 = 0;
  loaded = 0;
  /* Initialize all handles. */
  memset(handles, 0, count * sizeof(*handles));
  /* Check order flag. */
  if (0 == order) {
    /* Load all plug-ins in forward order. */
    for (index = 0; index < count; index++) {
      /* Load plug-in. */
      if (0 != harnessclient_plugins_load3(harnesses[index], &handles[index],
                                           name, first, age, revision, flags)) {
        errno2 = (0 == errno2)?errno:errno2;
        RMIX_LOG((RMIX_WARN(unable to load plug-in)))
        /* Check fail flag. */
        if (1 == fail) {
          errno = errno2;
          return loaded;
        }
      } else {
        loaded++;
      }
    }
  } else {
    /* Load all plug-ins in reverse order. */
    for (index = count; 0 < index--;) {
      /* Load plug-in. */
      if (0 != harnessclient_plugins_load3(harnesses[index], &handles[index],
                                           name, first, age, revision, flags)) {
        errno2 = (0 == errno2)?errno:errno2;
        RMIX_LOG((RMIX_WARN(unable to load plug-in)))
        /* Check fail flag. */
        if (1 == fail) {
          errno = errno2;
          return loaded;
        }
      } else {
        loaded++;
      }
    }
  }
  errno = errno2;
  return loaded;
}

/*
 * Invokes harness_plugins_load3 (client-side method stub).
 * Loads a plug-in at multiple remote harness instances in parallel using
 * asynchonous method invocations.
 * The plug-in may be different for every remote harness instance.
 *
 * count     = The number of remote harness instances.
 * harnesses = The array of RMIX remote object references of harness instances.
 * handles   = The plug-in handle array (also return). Plug-in handles are
 *             unique, and 0 for unsuccessful loading attempts.
 * names     = The plug-in name array. Plug-in names are unique and do
 *             not contain '/..' and do not start with '/'.
 * firsts    = The array of first versions supported of the plug-ins to load.
 * ages      = The array number of followup versions supported of the plug-ins
 *             to load.
 * revisions = The array revisions of the plug-ins to load.
 * flags     = The plug-in loading flags array, e.g. HARNESS_PLUGINS_UNIQUE.
 * return    = The number of loaded plug-ins with errno set appropriately.
 */
#undef  __FUNC__
#define __FUNC__ "harnessclient_plugins_pload3_mp"
unsigned int harnessclient_plugins_pload3_mp
  (unsigned int             count,
   const rmix_remoteref_t **harnesses,
   unsigned int            *handles,
   const char             **names,
   const unsigned int      *firsts,
   const unsigned int      *ages,
   const unsigned int      *revisions,
   const unsigned int      *flags) {
  int                errno2;
  unsigned int       loaded, index;
  rmix_invokeref_t **invocations;
#ifdef DEBUG
  /* Check harnesses parameter. */
  if ((NULL == harnesses)&&(0 != count)) {
    errno = EINVAL;
    RMIX_LOG((RMIX_WARN(harnesses parameter is null)))
    errno = EINVAL;
    return -1;
  }
  /* Check count parameter. */
  if ((0 == count)&&(NULL != harnesses)) {
    errno = EINVAL;
    RMIX_LOG((RMIX_WARN(count parameter is zero)))
    errno = EINVAL;
    return -1;
  }
  /* Check handles parameter. */
  if ((NULL == handles)&&(0 != count)) {
    errno = EINVAL;
    RMIX_LOG((RMIX_WARN(handles parameter is null)))
    errno = EINVAL;
    return -1;
  }
  /* Check count parameter. */
  if ((0 == count)&&(NULL != handles)) {
    errno = EINVAL;
    RMIX_LOG((RMIX_WARN(count parameter is zero)))
    errno = EINVAL;
    return -1;
  }
  /* Check names parameter. */
  if ((NULL == names)&&(0 != count)) {
    errno = EINVAL;
    RMIX_LOG((RMIX_WARN(names parameter is null)))
    errno = EINVAL;
    return -1;
  }
  /* Check count parameter. */
  if ((0 == count)&&(NULL != names)) {
    errno = EINVAL;
    RMIX_LOG((RMIX_WARN(count parameter is zero)))
    errno = EINVAL;
    return -1;
  }
  /* Check firsts parameter. */
  if ((NULL == firsts)&&(0 != count)) {
    errno = EINVAL;
    RMIX_LOG((RMIX_WARN(firsts parameter is null)))
    errno = EINVAL;
    return -1;
  }
  /* Check count parameter. */
  if ((0 == count)&&(NULL != firsts)) {
    errno = EINVAL;
    RMIX_LOG((RMIX_WARN(count parameter is zero)))
    errno = EINVAL;
    return -1;
  }
  /* Check ages parameter. */
  if ((NULL == ages)&&(0 != count)) {
    errno = EINVAL;
    RMIX_LOG((RMIX_WARN(ages parameter is null)))
    errno = EINVAL;
    return -1;
  }
  /* Check count parameter. */
  if ((0 == count)&&(NULL != ages)) {
    errno = EINVAL;
    RMIX_LOG((RMIX_WARN(count parameter is zero)))
    errno = EINVAL;
    return -1;
  }
  /* Check revisions parameter. */
  if ((NULL == revisions)&&(0 != count)) {
    errno = EINVAL;
    RMIX_LOG((RMIX_WARN(revisions parameter is null)))
    errno = EINVAL;
    return -1;
  }
  /* Check count parameter. */
  if ((0 == count)&&(NULL != revisions)) {
    errno = EINVAL;
    RMIX_LOG((RMIX_WARN(count parameter is zero)))
    errno = EINVAL;
    return -1;
  }
  /* Check flags parameter. */
  if ((NULL == flags)&&(0 != count)) {
    errno = EINVAL;
    RMIX_LOG((RMIX_WARN(flags parameter is null)))
    errno = EINVAL;
    return -1;
  }
  /* Check count parameter. */
  if ((0 == count)&&(NULL != flags)) {
    errno = EINVAL;
    RMIX_LOG((RMIX_WARN(count parameter is zero)))
    errno = EINVAL;
    return -1;
  }
#endif /* DEBUG */
  /* Allocate invocation reference array. */
  if (NULL == (invocations = malloc(count * sizeof(rmix_invokeref_t*)))) {
    errno2 = errno;
    RMIX_LOG((RMIX_WARN(unable to allocate invocation reference array)))
    errno = errno2;
    return 0;
  }
  /* Initialize errno and loaded counter.*/
  errno2 = 0;
  loaded = 0;
  /* Initialize all handles. */
  memset(handles, 0, count * sizeof(*handles));
  /* Load all plug-ins in forward order (send part). */
  for (index = 0; index < count; index++) {
    /* Load plug-in (send part). */
    if (0 != harnessclient_plugins_load3_send(&invocations[index],
                                              harnesses[index], names[index],
                                              firsts[index], ages[index],
                                              revisions[index], flags[index])) {
      errno2 = (0 == errno2)?errno:errno2;
      RMIX_LOG((RMIX_WARN(unable to load plug-in (send part))))
      invocations[index] = NULL;
    }
  }
  /* Load all plug-ins in forward order (retrieve part). */
  for (index = 0; index < count; index++) {
    /* Check for successful send. */
    if (NULL != invocations[index]) {
      /* Load plug-in (retrieve part). */
      if (0 != harnessclient_plugins_load3_retrieve(&invocations[index],
                                                    &handles[index])) {
        errno2 = (0 == errno2)?errno:errno2;
        RMIX_LOG((RMIX_WARN(unable to load plug-in (retrieve part))))
      } else {
        loaded++;
      }
    }
  }
  /* Deallocate invocation reference array. */
  free(invocations);
  errno = errno2;
  return loaded;
}

/*
 * Invokes harness_plugins_load3 (client-side method stub).
 * Loads a plug-in at multiple remote harness instances in parallel using
 * asynchonous method invocations.
 * The plug-in is the same for every remote harness instance.
 *
 * count     = The number of remote harness instances.
 * harnesses = The array of RMIX remote object references of harness instances.
 * handles   = The plug-in handle array (also return). Plug-in handles are
 *             unique, and 0 for unsuccessful loading attempts.
 * name      = The plug-in name. Plug-in names are unique and do
 *             not contain '/..' and do not start with '/'.
 * first     = The first version supported of the plug-in to load.
 * age       = The number of followup versions supported of the plug-in to load.
 * revision  = The revision of the plug-in to load.
 * flags     = The plug-in loading flags, e.g. HARNESS_PLUGINS_UNIQUE.
 * return    = The number of loaded plug-ins with errno set appropriately.
 */
#undef  __FUNC__
#define __FUNC__ "harnessclient_plugins_pload3_sp"
unsigned int harnessclient_plugins_pload3_sp
  (unsigned int             count,
   const rmix_remoteref_t **harnesses,
   unsigned int            *handles,
   const char              *name,
   unsigned int             first,
   unsigned int             age,
   unsigned int             revision,
   unsigned int             flags) {
  int                errno2;
  unsigned int       loaded, index;
  rmix_invokeref_t **invocations;
#ifdef DEBUG
  /* Check harnesses parameter. */
  if ((NULL == harnesses)&&(0 != count)) {
    errno = EINVAL;
    RMIX_LOG((RMIX_WARN(harnesses parameter is null)))
    errno = EINVAL;
    return -1;
  }
  /* Check count parameter. */
  if ((0 == count)&&(NULL != harnesses)) {
    errno = EINVAL;
    RMIX_LOG((RMIX_WARN(count parameter is zero)))
    errno = EINVAL;
    return -1;
  }
  /* Check handles parameter. */
  if ((NULL == handles)&&(0 != count)) {
    errno = EINVAL;
    RMIX_LOG((RMIX_WARN(handles parameter is null)))
    errno = EINVAL;
    return -1;
  }
  /* Check count parameter. */
  if ((0 == count)&&(NULL != handles)) {
    errno = EINVAL;
    RMIX_LOG((RMIX_WARN(count parameter is zero)))
    errno = EINVAL;
    return -1;
  }
  /* Check name parameter. */
  if (NULL == name) {
    errno = EINVAL;
    HARNESS_LOG((HARNESS_WARN(name parameter is null)))
    errno = EINVAL;
    return -1;
  }
#endif /* DEBUG */
  /* Allocate invocation reference array. */
  if (NULL == (invocations = malloc(count * sizeof(rmix_invokeref_t*)))) {
    errno2 = errno;
    RMIX_LOG((RMIX_WARN(unable to allocate invocation reference array)))
    errno = errno2;
    return 0;
  }
  /* Initialize errno and loaded counter.*/
  errno2 = 0;
  loaded = 0;
  /* Initialize all handles. */
  memset(handles, 0, count * sizeof(*handles));
  /* Load all plug-ins in forward order (send part). */
  for (index = 0; index < count; index++) {
    /* Load plug-in (send part). */
    if (0 != harnessclient_plugins_load3_send(&invocations[index],
                                              harnesses[index], name,
                                              first, age, revision, flags)) {
      errno2 = (0 == errno2)?errno:errno2;
      RMIX_LOG((RMIX_WARN(unable to load plug-in (send part))))
      invocations[index] = NULL;
    }
  }
  /* Load all plug-ins in forward order (retrieve part). */
  for (index = 0; index < count; index++) {
    /* Check for successful send. */
    if (NULL != invocations[index]) {
      /* Load plug-in (retrieve part). */
      if (0 != harnessclient_plugins_load3_retrieve(&invocations[index],
                                                    &handles[index])) {
        errno2 = (0 == errno2)?errno:errno2;
        RMIX_LOG((RMIX_WARN(unable to load plug-in (retrieve part))))
      } else {
        loaded++;
      }
    }
  }
  /* Deallocate invocation reference array. */
  free(invocations);
  errno = errno2;
  return loaded;
}

/*
 * Invokes harness_plugins_load4 (client-side method stub).
 * Loads a plug-in at a remote harness instance.
 *
 * harness  = The RMIX remote object reference of harness.
 * handle   = The plug-in handle (return). Plug-in handles are unique.
 * name     = The plug-in name. Plug-in names are unique and do
 *            not contain '/..' and do not start with '/'.
 * current  = The current version supported of the plug-in to load.
 * age      = The number of previous versions supported of the plug-in to load.
 * revision = The revision of the plug-in to load.
 * flags    = The plug-in loading flags, e.g. HARNESS_PLUGINS_UNIQUE.
 * return   = 0 on success or -1 on any error with errno set appropriately.
 */
#undef  __FUNC__
#define __FUNC__ "harnessclient_plugins_load4"
int harnessclient_plugins_load4 (const rmix_remoteref_t *harness,
                                 unsigned int           *handle,
                                 const char             *name,
                                 unsigned int            current,
                                 unsigned int            age,
                                 unsigned int            revision,
                                 unsigned int            flags) {
  int         result;
  void       *outary[2];
  const void *inary[5];
#ifdef DEBUG
  /* Check harness parameter. */
  if (NULL == harness) {
    errno = EINVAL;
    HARNESS_LOG((HARNESS_WARN(harness parameter is null)))
    errno = EINVAL;
    return -1;
  }
  /* Check handle parameter. */
  if (NULL == handle) {
    errno = EINVAL;
    HARNESS_LOG((HARNESS_WARN(handle parameter is null)))
    errno = EINVAL;
    return -1;
  }
  /* Check name parameter. */
  if (NULL == name) {
    errno = EINVAL;
    HARNESS_LOG((HARNESS_WARN(name parameter is null)))
    errno = EINVAL;
    return -1;
  }
#endif /* DEBUG */
  /* Prepare input. */
  inary[0] =  name;
  inary[1] = &current;
  inary[2] = &age;
  inary[3] = &revision;
  inary[4] = &flags;
  /* Prepare output. */
  outary[0] = &result;
  outary[1] =  handle;
  /* Invoke remote object method. */
  if (0 != rmix_invoke(harness, &harnessclient_interface,
                       HARNESS_RMIX_METHODS_PLUGINS_LOAD4_INDEX,
                       outary, 2, inary, 5)) {
    int errno2 = errno;
    RMIX_LOG((RMIX_WARN(unable to invoke remote object method)))
    errno = errno2;
    return -1;
  }
  return result;
}

/*
 * Invokes harness_plugins_load4 (client-side method stub).
 * Loads a plug-in at a remote harness instance.
 *
 * harness  = The RMIX remote object reference of harness.
 * provider = The preloaded RMIX provider plug-in handle.
 * handle   = The plug-in handle (return). Plug-in handles are unique.
 * name     = The plug-in name. Plug-in names are unique and do
 *            not contain '/..' and do not start with '/'.
 * current  = The current version supported of the plug-in to load.
 * age      = The number of previous versions supported of the plug-in to load.
 * revision = The revision of the plug-in to load.
 * flags    = The plug-in loading flags, e.g. HARNESS_PLUGINS_UNIQUE.
 * return   = 0 on success or -1 on any error with errno set appropriately.
 */
#undef  __FUNC__
#define __FUNC__ "harnessclient_plugins_load4_p"
int harnessclient_plugins_load4_p (const rmix_remoteref_t *harness,
                                   unsigned int            provider,
                                   unsigned int           *handle,
                                   const char             *name,
                                   unsigned int            current,
                                   unsigned int            age,
                                   unsigned int            revision,
                                   unsigned int            flags) {
  int         result;
  void       *outary[2];
  const void *inary[5];
#ifdef DEBUG
  /* Check harness parameter. */
  if (NULL == harness) {
    errno = EINVAL;
    HARNESS_LOG((HARNESS_WARN(harness parameter is null)))
    errno = EINVAL;
    return -1;
  }
  /* Check handle parameter. */
  if (NULL == handle) {
    errno = EINVAL;
    HARNESS_LOG((HARNESS_WARN(handle parameter is null)))
    errno = EINVAL;
    return -1;
  }
  /* Check name parameter. */
  if (NULL == name) {
    errno = EINVAL;
    HARNESS_LOG((HARNESS_WARN(name parameter is null)))
    errno = EINVAL;
    return -1;
  }
#endif /* DEBUG */
  /* Prepare input. */
  inary[0] =  name;
  inary[1] = &current;
  inary[2] = &age;
  inary[3] = &revision;
  inary[4] = &flags;
  /* Prepare output. */
  outary[0] = &result;
  outary[1] =  handle;
  /* Invoke remote object method. */
  if (0 != rmix_invoke_p(harness, provider, &harnessclient_interface,
                         HARNESS_RMIX_METHODS_PLUGINS_LOAD4_INDEX,
                         outary, 2, inary, 5)) {
    int errno2 = errno;
    RMIX_LOG((RMIX_WARN(unable to invoke remote object method)))
    errno = errno2;
    return -1;
  }
  return result;
}

/*
 * Invokes harness_plugins_load4 (client-side method stub).
 * Loads a plug-in at a remote harness instance using a oneway invocation.
 *
 * harness  = The RMIX remote object reference of harness.
 * name     = The plug-in name. Plug-in names are unique and do
 *            not contain '/..' and do not start with '/'.
 * current  = The current version supported of the plug-in to load.
 * age      = The number of followup versions supported of the plug-in to load.
 * revision = The revision of the plug-in to load.
 * flags    = The plug-in loading flags, e.g. HARNESS_PLUGINS_UNIQUE.
 * return   = 0 on success or -1 on any error with errno set appropriately.
 */
#undef  __FUNC__
#define __FUNC__ "harnessclient_plugins_load4_oneway"
int harnessclient_plugins_load4_oneway (const rmix_remoteref_t *harness,
                                        const char             *name,
                                        unsigned int            current,
                                        unsigned int            age,
                                        unsigned int            revision,
                                        unsigned int            flags) {
  const void *inary[5];
#ifdef DEBUG
  /* Check harness parameter. */
  if (NULL == harness) {
    errno = EINVAL;
    HARNESS_LOG((HARNESS_WARN(harness parameter is null)))
    errno = EINVAL;
    return -1;
  }
  /* Check name parameter. */
  if (NULL == name) {
    errno = EINVAL;
    HARNESS_LOG((HARNESS_WARN(name parameter is null)))
    errno = EINVAL;
    return -1;
  }
#endif /* DEBUG */
  /* Prepare input. */
  inary[0] =  name;
  inary[1] = &current;
  inary[2] = &age;
  inary[3] = &revision;
  inary[4] = &flags;
  /* Invoke remote object method. */
  if (0 != rmix_oneway(harness, &harnessclient_interface,
                       HARNESS_RMIX_METHODS_PLUGINS_LOAD4_INDEX,
                       inary, 5)) {
    int errno2 = errno;
    RMIX_LOG((RMIX_WARN(unable to invoke remote object method)))
    errno = errno2;
    return -1;
  }
  return 0;
}

/*
 * Invokes harness_plugins_load4 (client-side method stub).
 * Loads a plug-in at a remote harness instance using a oneway invocation
 * without loading the provider plug-in.
 *
 * harness  = The RMIX remote object reference of harness.
 * provider = The preloaded RMIX provider plug-in handle.
 * name     = The plug-in name. Plug-in names are unique and do
 *            not contain '/..' and do not start with '/'.
 * current  = The current version supported of the plug-in to load.
 * age      = The number of followup versions supported of the plug-in to load.
 * revision = The revision of the plug-in to load.
 * flags    = The plug-in loading flags, e.g. HARNESS_PLUGINS_UNIQUE.
 * return   = 0 on success or -1 on any error with errno set appropriately.
 */
#undef  __FUNC__
#define __FUNC__ "harnessclient_plugins_load4_oneway_p"
int harnessclient_plugins_load4_oneway_p (const rmix_remoteref_t *harness,
                                          unsigned int            provider,
                                          const char             *name,
                                          unsigned int            current,
                                          unsigned int            age,
                                          unsigned int            revision,
                                          unsigned int            flags) {
  const void *inary[5];
#ifdef DEBUG
  /* Check harness parameter. */
  if (NULL == harness) {
    errno = EINVAL;
    HARNESS_LOG((HARNESS_WARN(harness parameter is null)))
    errno = EINVAL;
    return -1;
  }
  /* Check name parameter. */
  if (NULL == name) {
    errno = EINVAL;
    HARNESS_LOG((HARNESS_WARN(name parameter is null)))
    errno = EINVAL;
    return -1;
  }
#endif /* DEBUG */
  /* Prepare input. */
  inary[0] =  name;
  inary[1] = &current;
  inary[2] = &age;
  inary[3] = &revision;
  inary[4] = &flags;
  /* Invoke remote object method. */
  if (0 != rmix_oneway_p(harness, provider, &harnessclient_interface,
                         HARNESS_RMIX_METHODS_PLUGINS_LOAD4_INDEX,
                         inary, 5)) {
    int errno2 = errno;
    RMIX_LOG((RMIX_WARN(unable to invoke remote object method)))
    errno = errno2;
    return -1;
  }
  return 0;
}

/*
 * Invokes harness_plugins_load4 (client-side method stub).
 * Loads a plug-in at a remote harness instance using an asynchronous
 * invocation.
 *
 * invokeref = The invocation reference (return).
 * harness   = The RMIX remote object reference of harness.
 * name      = The plug-in name. Plug-in names are unique and do
 *             not contain '/..' and do not start with '/'.
 * current   = The current version supported of the plug-in to load.
 * age       = The number of followup versions supported of the plug-in to load.
 * revision  = The revision of the plug-in to load.
 * flags     = The plug-in loading flags, e.g. HARNESS_PLUGINS_UNIQUE.
 * return    = 0 on success or -1 on any error with errno set appropriately.
 */
#undef  __FUNC__
#define __FUNC__ "harnessclient_plugins_load4_send"
int harnessclient_plugins_load4_send (rmix_invokeref_t       **invokeref,
                                      const rmix_remoteref_t  *harness,
                                      const char              *name,
                                      unsigned int             current,
                                      unsigned int             age,
                                      unsigned int             revision,
                                      unsigned int             flags) {
  const void *inary[5];
#ifdef DEBUG
  /* Check invokeref parameter. */
  if (NULL == invokeref) {
    errno = EINVAL;
    HARNESS_LOG((HARNESS_WARN(invokeref parameter is null)))
    errno = EINVAL;
    return -1;
  }
  /* Check harness parameter. */
  if (NULL == harness) {
    errno = EINVAL;
    HARNESS_LOG((HARNESS_WARN(harness parameter is null)))
    errno = EINVAL;
    return -1;
  }
  /* Check name parameter. */
  if (NULL == name) {
    errno = EINVAL;
    HARNESS_LOG((HARNESS_WARN(name parameter is null)))
    errno = EINVAL;
    return -1;
  }
#endif /* DEBUG */
  /* Prepare input. */
  inary[0] =  name;
  inary[1] = &current;
  inary[2] = &age;
  inary[3] = &revision;
  inary[4] = &flags;
  /* Invoke remote object method. */
  if (0 != rmix_send(invokeref, harness, &harnessclient_interface,
                     HARNESS_RMIX_METHODS_PLUGINS_LOAD4_INDEX,
                     inary, 5)) {
    int errno2 = errno;
    RMIX_LOG((RMIX_WARN(unable to invoke remote object method)))
    errno = errno2;
    return -1;
  }
  return 0;
}

/*
 * Invokes harness_plugins_load4 (client-side method stub).
 * Loads a plug-in at a remote harness instance using an asynchronous invocation
 * without loading the provider plug-in.
 *
 * invokeref = The invocation reference (return).
 * harness   = The RMIX remote object reference of harness.
 * provider  = The preloaded RMIX provider plug-in handle.
 * name      = The plug-in name. Plug-in names are unique and do
 *             not contain '/..' and do not start with '/'.
 * current   = The current version supported of the plug-in to load.
 * age       = The number of followup versions supported of the plug-in to load.
 * revision  = The revision of the plug-in to load.
 * flags     = The plug-in loading flags, e.g. HARNESS_PLUGINS_UNIQUE.
 * return    = 0 on success or -1 on any error with errno set appropriately.
 */
#undef  __FUNC__
#define __FUNC__ "harnessclient_plugins_load4_send_p"
int harnessclient_plugins_load4_send_p (rmix_invokeref_t       **invokeref,
                                        const rmix_remoteref_t  *harness,
                                        unsigned int             provider,
                                        const char              *name,
                                        unsigned int             current,
                                        unsigned int             age,
                                        unsigned int             revision,
                                        unsigned int             flags) {
  const void *inary[5];
#ifdef DEBUG
  /* Check invokeref parameter. */
  if (NULL == invokeref) {
    errno = EINVAL;
    HARNESS_LOG((HARNESS_WARN(invokeref parameter is null)))
    errno = EINVAL;
    return -1;
  }
  /* Check harness parameter. */
  if (NULL == harness) {
    errno = EINVAL;
    HARNESS_LOG((HARNESS_WARN(harness parameter is null)))
    errno = EINVAL;
    return -1;
  }
  /* Check name parameter. */
  if (NULL == name) {
    errno = EINVAL;
    HARNESS_LOG((HARNESS_WARN(name parameter is null)))
    errno = EINVAL;
    return -1;
  }
#endif /* DEBUG */
  /* Prepare input. */
  inary[0] =  name;
  inary[1] = &current;
  inary[2] = &age;
  inary[3] = &revision;
  inary[4] = &flags;
  /* Invoke remote object method. */
  if (0 != rmix_send_p(invokeref, harness, provider, &harnessclient_interface,
                       HARNESS_RMIX_METHODS_PLUGINS_LOAD4_INDEX,
                       inary, 5)) {
    int errno2 = errno;
    RMIX_LOG((RMIX_WARN(unable to invoke remote object method)))
    errno = errno2;
    return -1;
  }
  return 0;
}

/*
 * Retrieves harness_plugins_load4 output (client-side method stub).
 * Returns the output of a previous asynchronous invocation that loads a plug-in
 * at a remote harness instance.
 *
 * invokeref = The invocation reference (also return).
 * handle    = The plug-in handle (return). Plug-in handles are unique.
 * return    = 0 on success or -1 on any error with errno set appropriately.
 */
#undef  __FUNC__
#define __FUNC__ "harnessclient_plugins_load4_retrieve"
int harnessclient_plugins_load4_retrieve (rmix_invokeref_t **invokeref,
                                          unsigned int      *handle) {
  int   result;
  void *outary[2];
#ifdef DEBUG
  /* Check invokeref parameter. */
  if (NULL == invokeref) {
    errno = EINVAL;
    HARNESS_LOG((HARNESS_WARN(invokeref parameter is null)))
    errno = EINVAL;
    return -1;
  }
  /* Check *invokeref parameter. */
  if (NULL == *invokeref) {
    errno = EINVAL;
    HARNESS_LOG((HARNESS_WARN(*invokeref parameter is null)))
    errno = EINVAL;
    return -1;
  }
  /* Check handle parameter. */
  if (NULL == handle) {
    errno = EINVAL;
    HARNESS_LOG((HARNESS_WARN(handle parameter is null)))
    errno = EINVAL;
    return -1;
  }
#endif /* DEBUG */
  /* Prepare output. */
  outary[0] = &result;
  outary[1] =  handle;
  /* Invoke remote object method. */
  if (0 != rmix_retrieve(invokeref, outary, 2)) {
    int errno2 = errno;
    RMIX_LOG((RMIX_WARN(unable to invoke remote object method)))
    errno = errno2;
    return -1;
  }
  return result;
}

/*
 * Retrieves harness_plugins_load4 output (client-side method stub).
 * Returns the output of a previous asynchronous invocation that loads a plug-in
 * at a remote harness instance without unloading the provider plug-in.
 *
 * invokeref = The invocation reference (also return).
 * handle    = The plug-in handle (return). Plug-in handles are unique.
 * return    = 0 on success or -1 on any error with errno set appropriately.
 */
#undef  __FUNC__
#define __FUNC__ "harnessclient_plugins_load4_retrieve_p"
int harnessclient_plugins_load4_retrieve_p (rmix_invokeref_t **invokeref,
                                            unsigned int      *handle) {
  int   result;
  void *outary[2];
#ifdef DEBUG
  /* Check invokeref parameter. */
  if (NULL == invokeref) {
    errno = EINVAL;
    HARNESS_LOG((HARNESS_WARN(invokeref parameter is null)))
    errno = EINVAL;
    return -1;
  }
  /* Check *invokeref parameter. */
  if (NULL == *invokeref) {
    errno = EINVAL;
    HARNESS_LOG((HARNESS_WARN(*invokeref parameter is null)))
    errno = EINVAL;
    return -1;
  }
  /* Check handle parameter. */
  if (NULL == handle) {
    errno = EINVAL;
    HARNESS_LOG((HARNESS_WARN(handle parameter is null)))
    errno = EINVAL;
    return -1;
  }
#endif /* DEBUG */
  /* Prepare output. */
  outary[0] = &result;
  outary[1] =  handle;
  /* Invoke remote object method. */
  if (0 != rmix_retrieve_p(invokeref, outary, 2)) {
    int errno2 = errno;
    RMIX_LOG((RMIX_WARN(unable to invoke remote object method)))
    errno = errno2;
    return -1;
  }
  return result;
}

/*
 * Invokes harness_plugins_load4 (client-side method stub).
 * Loads a plug-in at multiple remote harness instances in sequence.
 * The plug-in may be different for every remote harness instance.
 *
 * count     = The number of remote harness instances.
 * harnesses = The array of RMIX remote object references of harness instances.
 * handles   = The plug-in handle array (also return). Plug-in handles are
 *             unique, and 0 for unsuccessful loading attempts.
 * names     = The plug-in name array. Plug-in names are unique and do
 *             not contain '/..' and do not start with '/'.
 * currents  = The array of current versions supported of the plug-ins to load.
 * ages      = The array number of followup versions supported of the plug-ins
 *             to load.
 * revisions = The array revisions of the plug-ins to load.
 * flags     = The plug-in loading flags array, e.g. HARNESS_PLUGINS_UNIQUE.
 * order     = The order flag. 0 for forward and 1 for reverse.
 * fail      = The fail flag. 0 to continue and 1 to fail on any error.
 * return    = The number of loaded plug-ins with errno set appropriately.
 */
#undef  __FUNC__
#define __FUNC__ "harnessclient_plugins_sload4_mp"
unsigned int harnessclient_plugins_sload4_mp
  (unsigned int             count,
   const rmix_remoteref_t **harnesses,
   unsigned int            *handles,
   const char             **names,
   const unsigned int      *currents,
   const unsigned int      *ages,
   const unsigned int      *revisions,
   const unsigned int      *flags,
   unsigned int             order,
   unsigned int             fail) {
  int          errno2;
  unsigned int loaded, index;
#ifdef DEBUG
  /* Check harnesses parameter. */
  if ((NULL == harnesses)&&(0 != count)) {
    errno = EINVAL;
    RMIX_LOG((RMIX_WARN(harnesses parameter is null)))
    errno = EINVAL;
    return -1;
  }
  /* Check count parameter. */
  if ((0 == count)&&(NULL != harnesses)) {
    errno = EINVAL;
    RMIX_LOG((RMIX_WARN(count parameter is zero)))
    errno = EINVAL;
    return -1;
  }
  /* Check handles parameter. */
  if ((NULL == handles)&&(0 != count)) {
    errno = EINVAL;
    RMIX_LOG((RMIX_WARN(handles parameter is null)))
    errno = EINVAL;
    return -1;
  }
  /* Check count parameter. */
  if ((0 == count)&&(NULL != handles)) {
    errno = EINVAL;
    RMIX_LOG((RMIX_WARN(count parameter is zero)))
    errno = EINVAL;
    return -1;
  }
  /* Check names parameter. */
  if ((NULL == names)&&(0 != count)) {
    errno = EINVAL;
    RMIX_LOG((RMIX_WARN(names parameter is null)))
    errno = EINVAL;
    return -1;
  }
  /* Check count parameter. */
  if ((0 == count)&&(NULL != names)) {
    errno = EINVAL;
    RMIX_LOG((RMIX_WARN(count parameter is zero)))
    errno = EINVAL;
    return -1;
  }
  /* Check currents parameter. */
  if ((NULL == currents)&&(0 != count)) {
    errno = EINVAL;
    RMIX_LOG((RMIX_WARN(currents parameter is null)))
    errno = EINVAL;
    return -1;
  }
  /* Check count parameter. */
  if ((0 == count)&&(NULL != currents)) {
    errno = EINVAL;
    RMIX_LOG((RMIX_WARN(count parameter is zero)))
    errno = EINVAL;
    return -1;
  }
  /* Check ages parameter. */
  if ((NULL == ages)&&(0 != count)) {
    errno = EINVAL;
    RMIX_LOG((RMIX_WARN(ages parameter is null)))
    errno = EINVAL;
    return -1;
  }
  /* Check count parameter. */
  if ((0 == count)&&(NULL != ages)) {
    errno = EINVAL;
    RMIX_LOG((RMIX_WARN(count parameter is zero)))
    errno = EINVAL;
    return -1;
  }
  /* Check revisions parameter. */
  if ((NULL == revisions)&&(0 != count)) {
    errno = EINVAL;
    RMIX_LOG((RMIX_WARN(revisions parameter is null)))
    errno = EINVAL;
    return -1;
  }
  /* Check count parameter. */
  if ((0 == count)&&(NULL != revisions)) {
    errno = EINVAL;
    RMIX_LOG((RMIX_WARN(count parameter is zero)))
    errno = EINVAL;
    return -1;
  }
  /* Check flags parameter. */
  if ((NULL == flags)&&(0 != count)) {
    errno = EINVAL;
    RMIX_LOG((RMIX_WARN(flags parameter is null)))
    errno = EINVAL;
    return -1;
  }
  /* Check count parameter. */
  if ((0 == count)&&(NULL != flags)) {
    errno = EINVAL;
    RMIX_LOG((RMIX_WARN(count parameter is zero)))
    errno = EINVAL;
    return -1;
  }
#endif /* DEBUG */
  /* Initialize errno and loaded counter.*/
  errno2 = 0;
  loaded = 0;
  /* Initialize all handles. */
  memset(handles, 0, count * sizeof(*handles));
  /* Check order flag. */
  if (0 == order) {
    /* Load all plug-ins in forward order. */
    for (index = 0; index < count; index++) {
      /* Load plug-in. */
      if (0 != harnessclient_plugins_load4(harnesses[index], &handles[index],
                                           names[index], currents[index],
                                           ages[index], revisions[index],
                                           flags[index])) {
        errno2 = (0 == errno2)?errno:errno2;
        RMIX_LOG((RMIX_WARN(unable to load plug-in)))
        /* Check fail flag. */
        if (1 == fail) {
          errno = errno2;
          return loaded;
        }
      } else {
        loaded++;
      }
    }
  } else {
    /* Load all plug-ins in reverse order. */
    for (index = count; 0 < index--;) {
      /* Load plug-in. */
      if (0 != harnessclient_plugins_load4(harnesses[index], &handles[index],
                                           names[index], currents[index],
                                           ages[index], revisions[index],
                                           flags[index])) {
        errno2 = (0 == errno2)?errno:errno2;
        RMIX_LOG((RMIX_WARN(unable to load plug-in)))
        /* Check fail flag. */
        if (1 == fail) {
          errno = errno2;
          return loaded;
        }
      } else {
        loaded++;
      }
    }
  }
  errno = errno2;
  return loaded;
}

/*
 * Invokes harness_plugins_load4 (client-side method stub).
 * Loads a plug-in at multiple remote harness instances in sequence.
 * The plug-in is the same for every remote harness instance.
 *
 * count     = The number of remote harness instances.
 * harnesses = The array of RMIX remote object references of harness instances.
 * handles   = The plug-in handle array (also return). Plug-in handles are
 *             unique, and 0 for unsuccessful loading attempts.
 * name      = The plug-in name. Plug-in names are unique and do
 *             not contain '/..' and do not start with '/'.
 * current   = The current version supported of the plug-in to load.
 * age       = The number of followup versions supported of the plug-in to load.
 * revision  = The revision of the plug-in to load.
 * flags     = The plug-in loading flags, e.g. HARNESS_PLUGINS_UNIQUE.
 * order     = The order flag. 0 for forward and 1 for reverse.
 * fail      = The fail flag. 0 to continue and 1 to fail on any error.
 * return    = The number of loaded plug-ins with errno set appropriately.
 */
#undef  __FUNC__
#define __FUNC__ "harnessclient_plugins_sload4_sp"
unsigned int harnessclient_plugins_sload4_sp
  (unsigned int             count,
   const rmix_remoteref_t **harnesses,
   unsigned int            *handles,
   const char              *name,
   unsigned int             current,
   unsigned int             age,
   unsigned int             revision,
   unsigned int             flags,
   unsigned int             order,
   unsigned int             fail) {
  int          errno2;
  unsigned int loaded, index;
#ifdef DEBUG
  /* Check harnesses parameter. */
  if ((NULL == harnesses)&&(0 != count)) {
    errno = EINVAL;
    RMIX_LOG((RMIX_WARN(harnesses parameter is null)))
    errno = EINVAL;
    return -1;
  }
  /* Check count parameter. */
  if ((0 == count)&&(NULL != harnesses)) {
    errno = EINVAL;
    RMIX_LOG((RMIX_WARN(count parameter is zero)))
    errno = EINVAL;
    return -1;
  }
  /* Check handles parameter. */
  if ((NULL == handles)&&(0 != count)) {
    errno = EINVAL;
    RMIX_LOG((RMIX_WARN(handles parameter is null)))
    errno = EINVAL;
    return -1;
  }
  /* Check count parameter. */
  if ((0 == count)&&(NULL != handles)) {
    errno = EINVAL;
    RMIX_LOG((RMIX_WARN(count parameter is zero)))
    errno = EINVAL;
    return -1;
  }
  /* Check name parameter. */
  if (NULL == name) {
    errno = EINVAL;
    HARNESS_LOG((HARNESS_WARN(name parameter is null)))
    errno = EINVAL;
    return -1;
  }
#endif /* DEBUG */
  /* Initialize errno and loaded counter.*/
  errno2 = 0;
  loaded = 0;
  /* Initialize all handles. */
  memset(handles, 0, count * sizeof(*handles));
  /* Check order flag. */
  if (0 == order) {
    /* Load all plug-ins in forward order. */
    for (index = 0; index < count; index++) {
      /* Load plug-in. */
      if (0 != harnessclient_plugins_load4(harnesses[index], &handles[index],
                                           name, current, age, revision,
                                           flags)) {
        errno2 = (0 == errno2)?errno:errno2;
        RMIX_LOG((RMIX_WARN(unable to load plug-in)))
        /* Check fail flag. */
        if (1 == fail) {
          errno = errno2;
          return loaded;
        }
      } else {
        loaded++;
      }
    }
  } else {
    /* Load all plug-ins in reverse order. */
    for (index = count; 0 < index--;) {
      /* Load plug-in. */
      if (0 != harnessclient_plugins_load4(harnesses[index], &handles[index],
                                           name, current, age, revision,
                                           flags)) {
        errno2 = (0 == errno2)?errno:errno2;
        RMIX_LOG((RMIX_WARN(unable to load plug-in)))
        /* Check fail flag. */
        if (1 == fail) {
          errno = errno2;
          return loaded;
        }
      } else {
        loaded++;
      }
    }
  }
  errno = errno2;
  return loaded;
}

/*
 * Invokes harness_plugins_load4 (client-side method stub).
 * Loads a plug-in at multiple remote harness instances in parallel using
 * asynchonous method invocations.
 * The plug-in may be different for every remote harness instance.
 *
 * count     = The number of remote harness instances.
 * harnesses = The array of RMIX remote object references of harness instances.
 * handles   = The plug-in handle array (also return). Plug-in handles are
 *             unique, and 0 for unsuccessful loading attempts.
 * names     = The plug-in name array. Plug-in names are unique and do
 *             not contain '/..' and do not start with '/'.
 * currents  = The array of current versions supported of the plug-ins to load.
 * ages      = The array number of followup versions supported of the plug-ins
 *             to load.
 * revisions = The array revisions of the plug-ins to load.
 * flags     = The plug-in loading flags array, e.g. HARNESS_PLUGINS_UNIQUE.
 * return    = The number of loaded plug-ins with errno set appropriately.
 */
#undef  __FUNC__
#define __FUNC__ "harnessclient_plugins_pload4_mp"
unsigned int harnessclient_plugins_pload4_mp
  (unsigned int             count,
   const rmix_remoteref_t **harnesses,
   unsigned int            *handles,
   const char             **names,
   const unsigned int      *currents,
   const unsigned int      *ages,
   const unsigned int      *revisions,
   const unsigned int      *flags) {
  int                errno2;
  unsigned int       loaded, index;
  rmix_invokeref_t **invocations;
#ifdef DEBUG
  /* Check harnesses parameter. */
  if ((NULL == harnesses)&&(0 != count)) {
    errno = EINVAL;
    RMIX_LOG((RMIX_WARN(harnesses parameter is null)))
    errno = EINVAL;
    return -1;
  }
  /* Check count parameter. */
  if ((0 == count)&&(NULL != harnesses)) {
    errno = EINVAL;
    RMIX_LOG((RMIX_WARN(count parameter is zero)))
    errno = EINVAL;
    return -1;
  }
  /* Check handles parameter. */
  if ((NULL == handles)&&(0 != count)) {
    errno = EINVAL;
    RMIX_LOG((RMIX_WARN(handles parameter is null)))
    errno = EINVAL;
    return -1;
  }
  /* Check count parameter. */
  if ((0 == count)&&(NULL != handles)) {
    errno = EINVAL;
    RMIX_LOG((RMIX_WARN(count parameter is zero)))
    errno = EINVAL;
    return -1;
  }
  /* Check names parameter. */
  if ((NULL == names)&&(0 != count)) {
    errno = EINVAL;
    RMIX_LOG((RMIX_WARN(names parameter is null)))
    errno = EINVAL;
    return -1;
  }
  /* Check count parameter. */
  if ((0 == count)&&(NULL != names)) {
    errno = EINVAL;
    RMIX_LOG((RMIX_WARN(count parameter is zero)))
    errno = EINVAL;
    return -1;
  }
  /* Check currents parameter. */
  if ((NULL == currents)&&(0 != count)) {
    errno = EINVAL;
    RMIX_LOG((RMIX_WARN(currents parameter is null)))
    errno = EINVAL;
    return -1;
  }
  /* Check count parameter. */
  if ((0 == count)&&(NULL != currents)) {
    errno = EINVAL;
    RMIX_LOG((RMIX_WARN(count parameter is zero)))
    errno = EINVAL;
    return -1;
  }
  /* Check ages parameter. */
  if ((NULL == ages)&&(0 != count)) {
    errno = EINVAL;
    RMIX_LOG((RMIX_WARN(ages parameter is null)))
    errno = EINVAL;
    return -1;
  }
  /* Check count parameter. */
  if ((0 == count)&&(NULL != ages)) {
    errno = EINVAL;
    RMIX_LOG((RMIX_WARN(count parameter is zero)))
    errno = EINVAL;
    return -1;
  }
  /* Check revisions parameter. */
  if ((NULL == revisions)&&(0 != count)) {
    errno = EINVAL;
    RMIX_LOG((RMIX_WARN(revisions parameter is null)))
    errno = EINVAL;
    return -1;
  }
  /* Check count parameter. */
  if ((0 == count)&&(NULL != revisions)) {
    errno = EINVAL;
    RMIX_LOG((RMIX_WARN(count parameter is zero)))
    errno = EINVAL;
    return -1;
  }
  /* Check flags parameter. */
  if ((NULL == flags)&&(0 != count)) {
    errno = EINVAL;
    RMIX_LOG((RMIX_WARN(flags parameter is null)))
    errno = EINVAL;
    return -1;
  }
  /* Check count parameter. */
  if ((0 == count)&&(NULL != flags)) {
    errno = EINVAL;
    RMIX_LOG((RMIX_WARN(count parameter is zero)))
    errno = EINVAL;
    return -1;
  }
#endif /* DEBUG */
  /* Allocate invocation reference array. */
  if (NULL == (invocations = malloc(count * sizeof(rmix_invokeref_t*)))) {
    errno2 = errno;
    RMIX_LOG((RMIX_WARN(unable to allocate invocation reference array)))
    errno = errno2;
    return 0;
  }
  /* Initialize errno and loaded counter.*/
  errno2 = 0;
  loaded = 0;
  /* Initialize all handles. */
  memset(handles, 0, count * sizeof(*handles));
  /* Load all plug-ins in forward order (send part). */
  for (index = 0; index < count; index++) {
    /* Load plug-in (send part). */
    if (0 != harnessclient_plugins_load4_send(&invocations[index],
                                              harnesses[index], names[index],
                                              currents[index], ages[index],
                                              revisions[index], flags[index])) {
      errno2 = (0 == errno2)?errno:errno2;
      RMIX_LOG((RMIX_WARN(unable to load plug-in (send part))))
      invocations[index] = NULL;
    }
  }
  /* Load all plug-ins in forward order (retrieve part). */
  for (index = 0; index < count; index++) {
    /* Check for successful send. */
    if (NULL != invocations[index]) {
      /* Load plug-in (retrieve part). */
      if (0 != harnessclient_plugins_load4_retrieve(&invocations[index],
                                                    &handles[index])) {
        errno2 = (0 == errno2)?errno:errno2;
        RMIX_LOG((RMIX_WARN(unable to load plug-in (retrieve part))))
      } else {
        loaded++;
      }
    }
  }
  /* Deallocate invocation reference array. */
  free(invocations);
  errno = errno2;
  return loaded;
}

/*
 * Invokes harness_plugins_load4 (client-side method stub).
 * Loads a plug-in at multiple remote harness instances in parallel using
 * asynchonous method invocations.
 * The plug-in is the same for every remote harness instance.
 *
 * count     = The number of remote harness instances.
 * harnesses = The array of RMIX remote object references of harness instances.
 * handles   = The plug-in handle array (also return). Plug-in handles are
 *             unique, and 0 for unsuccessful loading attempts.
 * name      = The plug-in name. Plug-in names are unique and do
 *             not contain '/..' and do not start with '/'.
 * current   = The current version supported of the plug-in to load.
 * age       = The number of followup versions supported of the plug-in to load.
 * revision  = The revision of the plug-in to load.
 * flags     = The plug-in loading flags, e.g. HARNESS_PLUGINS_UNIQUE.
 * return    = The number of loaded plug-ins with errno set appropriately.
 */
#undef  __FUNC__
#define __FUNC__ "harnessclient_plugins_pload4_sp"
unsigned int harnessclient_plugins_pload4_sp
  (unsigned int             count,
   const rmix_remoteref_t **harnesses,
   unsigned int            *handles,
   const char              *name,
   unsigned int             current,
   unsigned int             age,
   unsigned int             revision,
   unsigned int             flags) {
  int                errno2;
  unsigned int       loaded, index;
  rmix_invokeref_t **invocations;
#ifdef DEBUG
  /* Check harnesses parameter. */
  if ((NULL == harnesses)&&(0 != count)) {
    errno = EINVAL;
    RMIX_LOG((RMIX_WARN(harnesses parameter is null)))
    errno = EINVAL;
    return -1;
  }
  /* Check count parameter. */
  if ((0 == count)&&(NULL != harnesses)) {
    errno = EINVAL;
    RMIX_LOG((RMIX_WARN(count parameter is zero)))
    errno = EINVAL;
    return -1;
  }
  /* Check handles parameter. */
  if ((NULL == handles)&&(0 != count)) {
    errno = EINVAL;
    RMIX_LOG((RMIX_WARN(handles parameter is null)))
    errno = EINVAL;
    return -1;
  }
  /* Check count parameter. */
  if ((0 == count)&&(NULL != handles)) {
    errno = EINVAL;
    RMIX_LOG((RMIX_WARN(count parameter is zero)))
    errno = EINVAL;
    return -1;
  }
  /* Check name parameter. */
  if (NULL == name) {
    errno = EINVAL;
    HARNESS_LOG((HARNESS_WARN(name parameter is null)))
    errno = EINVAL;
    return -1;
  }
#endif /* DEBUG */
  /* Allocate invocation reference array. */
  if (NULL == (invocations = malloc(count * sizeof(rmix_invokeref_t*)))) {
    errno2 = errno;
    RMIX_LOG((RMIX_WARN(unable to allocate invocation reference array)))
    errno = errno2;
    return 0;
  }
  /* Initialize errno and loaded counter.*/
  errno2 = 0;
  loaded = 0;
  /* Initialize all handles. */
  memset(handles, 0, count * sizeof(*handles));
  /* Load all plug-ins in forward order (send part). */
  for (index = 0; index < count; index++) {
    /* Load plug-in (send part). */
    if (0 != harnessclient_plugins_load4_send(&invocations[index],
                                              harnesses[index], name,
                                              current, age, revision, flags)) {
      errno2 = (0 == errno2)?errno:errno2;
      RMIX_LOG((RMIX_WARN(unable to load plug-in (send part))))
      invocations[index] = NULL;
    }
  }
  /* Load all plug-ins in forward order (retrieve part). */
  for (index = 0; index < count; index++) {
    /* Check for successful send. */
    if (NULL != invocations[index]) {
      /* Load plug-in (retrieve part). */
      if (0 != harnessclient_plugins_load4_retrieve(&invocations[index],
                                                    &handles[index])) {
        errno2 = (0 == errno2)?errno:errno2;
        RMIX_LOG((RMIX_WARN(unable to load plug-in (retrieve part))))
      } else {
        loaded++;
      }
    }
  }
  /* Deallocate invocation reference array. */
  free(invocations);
  errno = errno2;
  return loaded;
}

/*
 * Invokes harness_plugins_name (client-side method stub).
 * Gets the name of a plug-in at a remote harness instance.
 *
 * harness = The RMIX remote object reference of harness.
 * name    = The plug-in name (return).
 * handle  = The plug-in handle.
 * return  = 0 on success or -1 on any error with errno set appropriately.
 */
#undef  __FUNC__
#define __FUNC__ "harnessclient_plugins_name"
int harnessclient_plugins_name (const rmix_remoteref_t  *harness,
                                char                   **name,
                                unsigned int             handle) {
  int         result;
  void       *outary[2];
  const void *inary[1];
#ifdef DEBUG
  /* Check harness parameter. */
  if (NULL == harness) {
    errno = EINVAL;
    HARNESS_LOG((HARNESS_WARN(harness parameter is null)))
    errno = EINVAL;
    return -1;
  }
  /* Check name parameter. */
  if (NULL == name) {
    errno = EINVAL;
    HARNESS_LOG((HARNESS_WARN(names parameter is null)))
    errno = EINVAL;
    return -1;
  }
#endif /* DEBUG */
  /* Prepare input. */
  inary[0] = &handle;
  /* Prepare output. */
  outary[0] = &result;
  outary[1] = NULL;
  /* Invoke remote object method. */
  if (0 != rmix_invoke(harness, &harnessclient_interface,
                       HARNESS_RMIX_METHODS_PLUGINS_NAME_INDEX,
                       outary, 2, inary, 1)) {
    int errno2 = errno;
    RMIX_LOG((RMIX_WARN(unable to invoke remote object method)))
    errno = errno2;
    return -1;
  }
  /* Prepare output. */
  *name = outary[0];
  return result;
}

/*
 * Invokes harness_plugins_name (client-side method stub).
 * Gets the name of a plug-in at a remote harness instance.
 *
 * harness  = The RMIX remote object reference of harness.
 * provider = The preloaded RMIX provider plug-in handle.
 * name     = The plug-in name (return).
 * handle   = The plug-in handle.
 * return   = 0 on success or -1 on any error with errno set appropriately.
 */
#undef  __FUNC__
#define __FUNC__ "harnessclient_plugins_name_p"
int harnessclient_plugins_name_p (const rmix_remoteref_t  *harness,
                                  unsigned int             provider,
                                  char                   **name,
                                  unsigned int             handle) {
  int         result;
  void       *outary[2];
  const void *inary[1];
#ifdef DEBUG
  /* Check harness parameter. */
  if (NULL == harness) {
    errno = EINVAL;
    HARNESS_LOG((HARNESS_WARN(harness parameter is null)))
    errno = EINVAL;
    return -1;
  }
  /* Check names parameter. */
  if (NULL == name) {
    errno = EINVAL;
    HARNESS_LOG((HARNESS_WARN(names parameter is null)))
    errno = EINVAL;
    return -1;
  }
#endif /* DEBUG */
  /* Prepare input. */
  inary[0] = &handle;
  /* Prepare output. */
  outary[0] = &result;
  outary[1] = NULL;
  /* Invoke remote object method. */
  if (0 != rmix_invoke_p(harness, provider, &harnessclient_interface,
                         HARNESS_RMIX_METHODS_PLUGINS_NAME_INDEX,
                         outary, 2, inary, 1)) {
    int errno2 = errno;
    RMIX_LOG((RMIX_WARN(unable to invoke remote object method)))
    errno = errno2;
    return -1;
  }
  /* Prepare output. */
  *name = outary[0];
  return result;
}

/*
 * Invokes harness_plugins_name (client-side method stub).
 * Gets the name of a plug-in at a remote harness instance using an asynchronous
 * invocation.
 *
 * invokeref = The invocation reference (return).
 * harness   = The RMIX remote object reference of harness.
 * provider  = The preloaded RMIX provider plug-in handle.
 * handle    = The plug-in handle.
 * return    = 0 on success or -1 on any error with errno set appropriately.
 */
#undef  __FUNC__
#define __FUNC__ "harnessclient_plugins_name_send"
int harnessclient_plugins_name_send (rmix_invokeref_t       **invokeref,
                                     const rmix_remoteref_t  *harness,
                                     unsigned int             handle) {
  const void *inary[1];
#ifdef DEBUG
  /* Check invokeref parameter. */
  if (NULL == invokeref) {
    errno = EINVAL;
    HARNESS_LOG((HARNESS_WARN(invokeref parameter is null)))
    errno = EINVAL;
    return -1;
  }
  /* Check harness parameter. */
  if (NULL == harness) {
    errno = EINVAL;
    HARNESS_LOG((HARNESS_WARN(harness parameter is null)))
    errno = EINVAL;
    return -1;
  }
#endif /* DEBUG */
  /* Prepare input. */
  inary[0] = &handle;
  /* Invoke remote object method. */
  if (0 != rmix_send(invokeref, harness, &harnessclient_interface,
                     HARNESS_RMIX_METHODS_PLUGINS_NAME_INDEX,
                     inary, 1)) {
    int errno2 = errno;
    RMIX_LOG((RMIX_WARN(unable to invoke remote object method)))
    errno = errno2;
    return -1;
  }
  return 0;
}

/*
 * Invokes harness_plugins_name (client-side method stub).
 * Gets the name of a plug-in at a remote harness instance using an asynchronous
 * invocation without loading the provider plug-in.
 *
 * invokeref = The invocation reference (return).
 * harness   = The RMIX remote object reference of harness.
 * provider  = The preloaded RMIX provider plug-in handle.
 * handle    = The plug-in handle.
 * return    = 0 on success or -1 on any error with errno set appropriately.
 */
#undef  __FUNC__
#define __FUNC__ "harnessclient_plugins_name_send_p"
int harnessclient_plugins_name_send_p (rmix_invokeref_t       **invokeref,
                                       const rmix_remoteref_t  *harness,
                                       unsigned int             provider,
                                       unsigned int             handle) {
  const void *inary[1];
#ifdef DEBUG
  /* Check invokeref parameter. */
  if (NULL == invokeref) {
    errno = EINVAL;
    HARNESS_LOG((HARNESS_WARN(invokeref parameter is null)))
    errno = EINVAL;
    return -1;
  }
  /* Check harness parameter. */
  if (NULL == harness) {
    errno = EINVAL;
    HARNESS_LOG((HARNESS_WARN(harness parameter is null)))
    errno = EINVAL;
    return -1;
  }
#endif /* DEBUG */
  /* Prepare input. */
  inary[0] = &handle;
  /* Invoke remote object method. */
  if (0 != rmix_send_p(invokeref, harness, provider, &harnessclient_interface,
                       HARNESS_RMIX_METHODS_PLUGINS_NAME_INDEX,
                       inary, 1)) {
    int errno2 = errno;
    RMIX_LOG((RMIX_WARN(unable to invoke remote object method)))
    errno = errno2;
    return -1;
  }
  return 0;
}

/*
 * Retrieves harness_plugins_name output (client-side method stub).
 * Returns the output of a previous asynchronous invocation that gets the name
 * of a plug-in at a remote harness instance.
 *
 * invokeref = The invocation reference (also return).
 * name      = The plug-in name (return).
 * return    = 0 on success or -1 on any error with errno set appropriately.
 */
#undef  __FUNC__
#define __FUNC__ "harnessclient_plugins_name_retrieve"
int harnessclient_plugins_name_retrieve (rmix_invokeref_t **invokeref,
                                         char             **name) {
  int   result;
  void *outary[2];
#ifdef DEBUG
  /* Check invokeref parameter. */
  if (NULL == invokeref) {
    errno = EINVAL;
    HARNESS_LOG((HARNESS_WARN(invokeref parameter is null)))
    errno = EINVAL;
    return -1;
  }
  /* Check *invokeref parameter. */
  if (NULL == *invokeref) {
    errno = EINVAL;
    HARNESS_LOG((HARNESS_WARN(*invokeref parameter is null)))
    errno = EINVAL;
    return -1;
  }
  /* Check name parameter. */
  if (NULL == name) {
    errno = EINVAL;
    HARNESS_LOG((HARNESS_WARN(name parameter is null)))
    errno = EINVAL;
    return -1;
  }
#endif /* DEBUG */
  /* Prepare output. */
  outary[0] = &result;
  outary[1] = NULL;
  /* Invoke remote object method. */
  if (0 != rmix_retrieve(invokeref, outary, 2)) {
    int errno2 = errno;
    RMIX_LOG((RMIX_WARN(unable to invoke remote object method)))
    errno = errno2;
    return -1;
  }
  /* Prepare output. */
  *name = outary[0];
  return result;
}

/*
 * Retrieves harness_plugins_name output (client-side method stub).
 * Returns the output of a previous asynchronous invocation that gets the name
 * of a plug-in at a remote harness instance without unloading the provider
 * plug-in.
 *
 * invokeref = The invocation reference (also return).
 * name      = The plug-in name (return).
 * return    = 0 on success or -1 on any error with errno set appropriately.
 */
#undef  __FUNC__
#define __FUNC__ "harnessclient_plugins_name_retrieve_p"
int harnessclient_plugins_name_retrieve_p (rmix_invokeref_t **invokeref,
                                           char             **name) {
  int   result;
  void *outary[2];
#ifdef DEBUG
  /* Check invokeref parameter. */
  if (NULL == invokeref) {
    errno = EINVAL;
    HARNESS_LOG((HARNESS_WARN(invokeref parameter is null)))
    errno = EINVAL;
    return -1;
  }
  /* Check *invokeref parameter. */
  if (NULL == *invokeref) {
    errno = EINVAL;
    HARNESS_LOG((HARNESS_WARN(*invokeref parameter is null)))
    errno = EINVAL;
    return -1;
  }
  /* Check name parameter. */
  if (NULL == name) {
    errno = EINVAL;
    HARNESS_LOG((HARNESS_WARN(name parameter is null)))
    errno = EINVAL;
    return -1;
  }
#endif /* DEBUG */
  /* Prepare output. */
  outary[0] = &result;
  outary[1] = NULL;
  /* Invoke remote object method. */
  if (0 != rmix_retrieve_p(invokeref, outary, 2)) {
    int errno2 = errno;
    RMIX_LOG((RMIX_WARN(unable to invoke remote object method)))
    errno = errno2;
    return -1;
  }
  /* Prepare output. */
  *name = outary[0];
  return result;
}

/*
 * Invokes harness_plugins_name (client-side method stub).
 * Gets the name of a plug-in at multiple remote harness instances.
 * The plug-in may be different for every remote harness instance.
 *
 * count     = The number of remote harness instances.
 * harnesses = The array of RMIX remote object references of harness instances.
 * names     = The plug-in name array (also return). Names of failed invocations
 *             are set to NULL.
 * handles   = The plug-in handle array. 0 handles are ignored.
 * order     = The order flag. 0 for forward and 1 for reverse.
 * fail      = The fail flag. 0 to continue and 1 to fail on any error.
 * return    = The number of names retrieved with errno set appropriately.
 */
#undef  __FUNC__
#define __FUNC__ "harnessclient_plugins_sname"
int harnessclient_plugins_sname (unsigned int             count,
                                 const rmix_remoteref_t **harnesses,
                                 char                   **names,
                                 const unsigned int      *handles,
                                 unsigned int             order,
                                 unsigned int             fail) {
  int          errno2;
  unsigned int obtained, index;
#ifdef DEBUG
  /* Check harnesses parameter. */
  if ((NULL == harnesses)&&(0 != count)) {
    errno = EINVAL;
    RMIX_LOG((RMIX_WARN(harnesses parameter is null)))
    errno = EINVAL;
    return -1;
  }
  /* Check count parameter. */
  if ((0 == count)&&(NULL != harnesses)) {
    errno = EINVAL;
    RMIX_LOG((RMIX_WARN(count parameter is zero)))
    errno = EINVAL;
    return -1;
  }
  /* Check names parameter. */
  if ((NULL == names)&&(0 != count)) {
    errno = EINVAL;
    RMIX_LOG((RMIX_WARN(names parameter is null)))
    errno = EINVAL;
    return -1;
  }
  /* Check count parameter. */
  if ((0 == count)&&(NULL != names)) {
    errno = EINVAL;
    RMIX_LOG((RMIX_WARN(count parameter is zero)))
    errno = EINVAL;
    return -1;
  }
  /* Check handles parameter. */
  if ((NULL == handles)&&(0 != count)) {
    errno = EINVAL;
    RMIX_LOG((RMIX_WARN(handles parameter is null)))
    errno = EINVAL;
    return -1;
  }
  /* Check count parameter. */
  if ((0 == count)&&(NULL != handles)) {
    errno = EINVAL;
    RMIX_LOG((RMIX_WARN(count parameter is zero)))
    errno = EINVAL;
    return -1;
  }
#endif /* DEBUG */
  /* Initialize errno and obtained counter.*/
  errno2   = 0;
  obtained = 0;
  /* Initialize all names. */
  memset(names, 0, count * sizeof(char*));
  /* Check order flag. */
  if (0 == order) {
    /* Get all plug-in names in forward order. */
    for (index = 0; index < count; index++) {
      /* Check for non-zero plug-in handle. */
      if (0 != handles[index]) {
        /* Get plug-in name. */
        if (0 != harnessclient_plugins_name(harnesses[index],
                                            &names[index], handles[index])) {
          errno2 = (0 == errno2)?errno:errno2;
          RMIX_LOG((RMIX_WARN(unable to get plug-in name)))
          /* Check fail flag. */
          if (1 == fail) {
            errno = errno2;
            return obtained;
          }
        } else {
          obtained++;
        }
      }
    }
  } else {
    /* all plug-in names in reverse order. */
    for (index = count; 0 < index--;) {
      /* Check for non-zero plug-in handle. */
      if (0 != handles[index]) {
        /* Get plug-in name. */
        if (0 != harnessclient_plugins_name(harnesses[index],
                                            &names[index], handles[index])) {
          errno2 = (0 == errno2)?errno:errno2;
          RMIX_LOG((RMIX_WARN(unable to get plug-in name)))
          /* Check fail flag. */
          if (1 == fail) {
            errno = errno2;
            return obtained;
          }
        } else {
          obtained++;
        }
      }
    }
  }
  errno = errno2;
  return obtained;
}

/*
 * Invokes harness_plugins_name (client-side method stub).
 * Gets the name of a plug-in at multiple remote harness instances in parallel
 * using asynchonous method invocations.
 * The plug-in may be different for every remote harness instance.
 *
 * count     = The number of remote harness instances.
 * harnesses = The array of RMIX remote object references of harness instances.
 * names     = The plug-in name array (also return). Names of failed invocations
 *             are set to NULL.
 * handles   = The plug-in handle array. 0 handles are ignored.
 * return    = The number of names retrieved with errno set appropriately.
 */
#undef  __FUNC__
#define __FUNC__ "harnessclient_plugins_pname"
int harnessclient_plugins_pname (unsigned int             count,
                                 const rmix_remoteref_t **harnesses,
                                 char                   **names,
                                 const unsigned int      *handles) {
  int                errno2;
  unsigned int       obtained, index;
  rmix_invokeref_t **invocations;
#ifdef DEBUG
  /* Check harnesses parameter. */
  if ((NULL == harnesses)&&(0 != count)) {
    errno = EINVAL;
    RMIX_LOG((RMIX_WARN(harnesses parameter is null)))
    errno = EINVAL;
    return -1;
  }
  /* Check count parameter. */
  if ((0 == count)&&(NULL != harnesses)) {
    errno = EINVAL;
    RMIX_LOG((RMIX_WARN(count parameter is zero)))
    errno = EINVAL;
    return -1;
  }
  /* Check names parameter. */
  if ((NULL == names)&&(0 != count)) {
    errno = EINVAL;
    RMIX_LOG((RMIX_WARN(names parameter is null)))
    errno = EINVAL;
    return -1;
  }
  /* Check count parameter. */
  if ((0 == count)&&(NULL != names)) {
    errno = EINVAL;
    RMIX_LOG((RMIX_WARN(count parameter is zero)))
    errno = EINVAL;
    return -1;
  }
  /* Check handles parameter. */
  if ((NULL == handles)&&(0 != count)) {
    errno = EINVAL;
    RMIX_LOG((RMIX_WARN(handles parameter is null)))
    errno = EINVAL;
    return -1;
  }
  /* Check count parameter. */
  if ((0 == count)&&(NULL != handles)) {
    errno = EINVAL;
    RMIX_LOG((RMIX_WARN(count parameter is zero)))
    errno = EINVAL;
    return -1;
  }
#endif /* DEBUG */
  /* Allocate invocation reference array. */
  if (NULL == (invocations = malloc(count * sizeof(rmix_invokeref_t*)))) {
    errno2 = errno;
    RMIX_LOG((RMIX_WARN(unable to allocate invocation reference array)))
    errno = errno2;
    return 0;
  }
  /* Initialize errno and obtained counter.*/
  errno2   = 0;
  obtained = 0;
  /* Initialize all names. */
  memset(names, 0, count * sizeof(char*));
  /* Get all plug-in names in forward order (send part). */
  for (index = 0; index < count; index++) {
    /* Check for non-zero plug-in handle. */
    if (0 != handles[index]) {
      /* Get plug-in name (send part). */
      if (0 != harnessclient_plugins_name_send(&invocations[index],
                                               harnesses[index],
                                               handles[index])) {
        errno2 = (0 == errno2)?errno:errno2;
        RMIX_LOG((RMIX_WARN(unable to get plug-in name (send part))))
        invocations[index] = NULL;
      }
    }
  }
  /* Get all plug-in names in forward order (retrieve part). */
  for (index = 0; index < count; index++) {
    /* Check for non-zero plug-in handle. */
    if (0 != handles[index]) {
      /* Check for successful send. */
      if (NULL != invocations[index]) {
        /* Get plug-in name (retrieve part). */
        if (0 != harnessclient_plugins_name_retrieve(&invocations[index],
                                                     &names[index])) {
          errno2 = (0 == errno2)?errno:errno2;
          RMIX_LOG((RMIX_WARN(unable to get plug-in name (retrieve part))))
        } else {
          obtained++;
        }
      }
    }
  }
  /* Deallocate invocation reference array. */
  free(invocations);
  errno = errno2;
  return obtained;
}

/*
 * Invokes harness_plugins_unload (client-side method stub).
 * Unloads a plug-in at a remote harness instance.
 *
 * harness = The RMIX remote object reference of harness.
 * handle  = The plug-in handle.
 * return  = 0 on success or -1 on any error with errno set appropriately.
 */
#undef  __FUNC__
#define __FUNC__ "harnessclient_plugins_unload"
int harnessclient_plugins_unload (const rmix_remoteref_t *harness,
                                  unsigned int            handle) {
  int         result;
  void       *outary[1];
  const void *inary[1];
#ifdef DEBUG
  /* Check harness parameter. */
  if (NULL == harness) {
    errno = EINVAL;
    HARNESS_LOG((HARNESS_WARN(harness parameter is null)))
    errno = EINVAL;
    return -1;
  }
#endif /* DEBUG */
  /* Prepare input. */
  inary[0] = &handle;
  /* Prepare output. */
  outary[0] = &result;
  /* Invoke remote object method. */
  if (0 != rmix_invoke(harness, &harnessclient_interface,
                       HARNESS_RMIX_METHODS_PLUGINS_UNLOAD_INDEX,
                       outary, 1, inary, 1)) {
    int errno2 = errno;
    RMIX_LOG((RMIX_WARN(unable to invoke remote object method)))
    errno = errno2;
    return -1;
  }
  return result;
}

/*
 * Invokes harness_plugins_unload (client-side method stub).
 * Unloads a plug-in at a remote harness instance.
 *
 * harness  = The RMIX remote object reference of harness.
 * provider = The preloaded RMIX provider plug-in handle.
 * handle   = The plug-in handle.
 * return   = 0 on success or -1 on any error with errno set appropriately.
 */
#undef  __FUNC__
#define __FUNC__ "harnessclient_plugins_unload_p"
int harnessclient_plugins_unload_p (const rmix_remoteref_t *harness,
                                    unsigned int            provider,
                                    unsigned int            handle) {
  int         result;
  void       *outary[1];
  const void *inary[1];
#ifdef DEBUG
  /* Check harness parameter. */
  if (NULL == harness) {
    errno = EINVAL;
    HARNESS_LOG((HARNESS_WARN(harness parameter is null)))
    errno = EINVAL;
    return -1;
  }
#endif /* DEBUG */
  /* Prepare input. */
  inary[0] = &handle;
  /* Prepare output. */
  outary[0] = &result;
  /* Invoke remote object method. */
  if (0 != rmix_invoke_p(harness, provider, &harnessclient_interface,
                         HARNESS_RMIX_METHODS_PLUGINS_UNLOAD_INDEX,
                         outary, 1, inary, 1)) {
    int errno2 = errno;
    RMIX_LOG((RMIX_WARN(unable to invoke remote object method)))
    errno = errno2;
    return -1;
  }
  return result;
}

/*
 * Invokes harness_plugins_unload (client-side method stub).
 * Unloads a plug-in at a remote harness instance using a oneway invocation.
 *
 * harness = The RMIX remote object reference of harness.
 * handle  = The plug-in handle.
 * return  = 0 on success or -1 on any error with errno set appropriately.
 */
#undef  __FUNC__
#define __FUNC__ "harnessclient_plugins_unload_oneway"
int harnessclient_plugins_unload_oneway (const rmix_remoteref_t *harness,
                                         unsigned int            handle) {
  const void *inary[1];
#ifdef DEBUG
  /* Check harness parameter. */
  if (NULL == harness) {
    errno = EINVAL;
    HARNESS_LOG((HARNESS_WARN(harness parameter is null)))
    errno = EINVAL;
    return -1;
  }
#endif /* DEBUG */
  /* Prepare input. */
  inary[0] = &handle;
  /* Invoke remote object method. */
  if (0 != rmix_oneway(harness, &harnessclient_interface,
                       HARNESS_RMIX_METHODS_PLUGINS_UNLOAD_INDEX,
                       inary, 1)) {
    int errno2 = errno;
    RMIX_LOG((RMIX_WARN(unable to invoke remote object method)))
    errno = errno2;
    return -1;
  }
  return 0;
}

/*
 * Invokes harness_plugins_unload (client-side method stub).
 * Unloads a plug-in at a remote harness instance using a oneway invocation
 * without loading the provider plug-in.
 *
 * harness  = The RMIX remote object reference of harness.
 * provider = The preloaded RMIX provider plug-in handle.
 * handle   = The plug-in handle.
 * return   = 0 on success or -1 on any error with errno set appropriately.
 */
#undef  __FUNC__
#define __FUNC__ "harnessclient_plugins_unload_oneway_p"
int harnessclient_plugins_unload_oneway_p (const rmix_remoteref_t *harness,
                                           unsigned int            provider,
                                           unsigned int            handle) {
  const void *inary[1];
#ifdef DEBUG
  /* Check harness parameter. */
  if (NULL == harness) {
    errno = EINVAL;
    HARNESS_LOG((HARNESS_WARN(harness parameter is null)))
    errno = EINVAL;
    return -1;
  }
#endif /* DEBUG */
  /* Prepare input. */
  inary[0] = &handle;
  /* Invoke remote object method. */
  if (0 != rmix_oneway_p(harness, provider, &harnessclient_interface,
                         HARNESS_RMIX_METHODS_PLUGINS_UNLOAD_INDEX,
                         inary, 1)) {
    int errno2 = errno;
    RMIX_LOG((RMIX_WARN(unable to invoke remote object method)))
    errno = errno2;
    return -1;
  }
  return 0;
}

/*
 * Invokes harness_plugins_unload (client-side method stub).
 * Unloads a plug-in at a remote harness instance using an asynchronous
 * invocation.
 *
 * invokeref = The invocation reference (return).
 * harness   = The RMIX remote object reference of harness.
 * handle    = The plug-in handle.
 * return    = 0 on success or -1 on any error with errno set appropriately.
 */
#undef  __FUNC__
#define __FUNC__ "harnessclient_plugins_unload_send"
int harnessclient_plugins_unload_send (rmix_invokeref_t       **invokeref,
                                       const rmix_remoteref_t  *harness,
                                       unsigned int             handle) {
  const void *inary[1];
#ifdef DEBUG
  /* Check invokeref parameter. */
  if (NULL == invokeref) {
    errno = EINVAL;
    HARNESS_LOG((HARNESS_WARN(invokeref parameter is null)))
    errno = EINVAL;
    return -1;
  }
  /* Check harness parameter. */
  if (NULL == harness) {
    errno = EINVAL;
    HARNESS_LOG((HARNESS_WARN(harness parameter is null)))
    errno = EINVAL;
    return -1;
  }
#endif /* DEBUG */
  /* Prepare input. */
  inary[0] = &handle;
  /* Invoke remote object method. */
  if (0 != rmix_send(invokeref, harness, &harnessclient_interface,
                     HARNESS_RMIX_METHODS_PLUGINS_UNLOAD_INDEX,
                     inary, 1)) {
    int errno2 = errno;
    RMIX_LOG((RMIX_WARN(unable to invoke remote object method)))
    errno = errno2;
    return -1;
  }
  return 0;
}

/*
 * Invokes harness_plugins_unload (client-side method stub).
 * Unloads a plug-in at a remote harness instance using an asynchronous
 * invocation without loading the provider plug-in.
 *
 * invokeref = The invocation reference (return).
 * harness   = The RMIX remote object reference of harness.
 * provider  = The preloaded RMIX provider plug-in handle.
 * handle    = The plug-in handle.
 * return    = 0 on success or -1 on any error with errno set appropriately.
 */
#undef  __FUNC__
#define __FUNC__ "harnessclient_plugins_unload_send_p"
int harnessclient_plugins_unload_send_p (rmix_invokeref_t       **invokeref,
                                         const rmix_remoteref_t  *harness,
                                         unsigned int             provider,
                                         unsigned int             handle) {
  const void *inary[1];
#ifdef DEBUG
  /* Check invokeref parameter. */
  if (NULL == invokeref) {
    errno = EINVAL;
    HARNESS_LOG((HARNESS_WARN(invokeref parameter is null)))
    errno = EINVAL;
    return -1;
  }
  /* Check harness parameter. */
  if (NULL == harness) {
    errno = EINVAL;
    HARNESS_LOG((HARNESS_WARN(harness parameter is null)))
    errno = EINVAL;
    return -1;
  }
#endif /* DEBUG */
  /* Prepare input. */
  inary[0] = &handle;
  /* Invoke remote object method. */
  if (0 != rmix_send_p(invokeref, harness, provider, &harnessclient_interface,
                       HARNESS_RMIX_METHODS_PLUGINS_UNLOAD_INDEX,
                       inary, 1)) {
    int errno2 = errno;
    RMIX_LOG((RMIX_WARN(unable to invoke remote object method)))
    errno = errno2;
    return -1;
  }
  return 0;
}

/*
 * Retrieves harness_plugins_unload output (client-side method stub).
 * Returns the output of a previous asynchronous invocation that unloads a
 * plug-in at a remote harness instance.
 *
 * invokeref = The invocation reference (also return).
 * return    = 0 on success or -1 on any error with errno set appropriately.
 */
#undef  __FUNC__
#define __FUNC__ "harnessclient_plugins_unload_retrieve"
int harnessclient_plugins_unload_retrieve (rmix_invokeref_t **invokeref) {
  int   result;
  void *outary[1];
#ifdef DEBUG
  /* Check invokeref parameter. */
  if (NULL == invokeref) {
    errno = EINVAL;
    HARNESS_LOG((HARNESS_WARN(invokeref parameter is null)))
    errno = EINVAL;
    return -1;
  }
  /* Check *invokeref parameter. */
  if (NULL == *invokeref) {
    errno = EINVAL;
    HARNESS_LOG((HARNESS_WARN(*invokeref parameter is null)))
    errno = EINVAL;
    return -1;
  }
#endif /* DEBUG */
  /* Prepare output. */
  outary[0] = &result;
  /* Invoke remote object method. */
  if (0 != rmix_retrieve(invokeref, outary, 1)) {
    int errno2 = errno;
    RMIX_LOG((RMIX_WARN(unable to invoke remote object method)))
    errno = errno2;
    return -1;
  }
  return result;
}

/*
 * Retrieves harness_plugins_unload output (client-side method stub).
 * Returns the output of a previous asynchronous invocation that unloads a
 * plug-in at a remote harness instance without unloading the provider plug-in.
 *
 * invokeref = The invocation reference (also return).
 * return    = 0 on success or -1 on any error with errno set appropriately.
 */
#undef  __FUNC__
#define __FUNC__ "harnessclient_plugins_unload_retrieve_p"
int harnessclient_plugins_unload_retrieve_p (rmix_invokeref_t **invokeref) {
  int   result;
  void *outary[1];
#ifdef DEBUG
  /* Check invokeref parameter. */
  if (NULL == invokeref) {
    errno = EINVAL;
    HARNESS_LOG((HARNESS_WARN(invokeref parameter is null)))
    errno = EINVAL;
    return -1;
  }
  /* Check *invokeref parameter. */
  if (NULL == *invokeref) {
    errno = EINVAL;
    HARNESS_LOG((HARNESS_WARN(*invokeref parameter is null)))
    errno = EINVAL;
    return -1;
  }
#endif /* DEBUG */
  /* Prepare output. */
  outary[0] = &result;
  /* Invoke remote object method. */
  if (0 != rmix_retrieve_p(invokeref, outary, 1)) {
    int errno2 = errno;
    RMIX_LOG((RMIX_WARN(unable to invoke remote object method)))
    errno = errno2;
    return -1;
  }
  return result;
}

/*
 * Invokes harness_plugins_unload (client-side method stub).
 * Unloads a plug-in at multiple remote harness instances.
 * The plug-in may be different for every remote harness instance.
 *
 * count     = The number of remote harness instances.
 * harnesses = The array of RMIX remote object references of harness instances.
 * handles   = The plug-in handle array (also return). 0 handles are ignored.
 *             Plug-in handles for successful unloading attempts are set to 0.
 * order     = The order flag. 0 for forward and 1 for reverse.
 * fail      = The fail flag. 0 to continue and 1 to fail on any error.
 * return    = The number of unloaded plug-ins with errno set appropriately.
 */
#undef  __FUNC__
#define __FUNC__ "harnessclient_plugins_sunload"
int harnessclient_plugins_sunload (unsigned int             count,
                                   const rmix_remoteref_t **harnesses,
                                   unsigned int            *handles,
                                   unsigned int             order,
                                   unsigned int             fail) {
  int          errno2;
  unsigned int unloaded, index;
#ifdef DEBUG
  /* Check harnesses parameter. */
  if ((NULL == harnesses)&&(0 != count)) {
    errno = EINVAL;
    RMIX_LOG((RMIX_WARN(harnesses parameter is null)))
    errno = EINVAL;
    return -1;
  }
  /* Check count parameter. */
  if ((0 == count)&&(NULL != harnesses)) {
    errno = EINVAL;
    RMIX_LOG((RMIX_WARN(count parameter is zero)))
    errno = EINVAL;
    return -1;
  }
  /* Check handles parameter. */
  if ((NULL == handles)&&(0 != count)) {
    errno = EINVAL;
    RMIX_LOG((RMIX_WARN(handles parameter is null)))
    errno = EINVAL;
    return -1;
  }
  /* Check count parameter. */
  if ((0 == count)&&(NULL != handles)) {
    errno = EINVAL;
    RMIX_LOG((RMIX_WARN(count parameter is zero)))
    errno = EINVAL;
    return -1;
  }
#endif /* DEBUG */
  /* Initialize errno and unloaded counter.*/
  errno2   = 0;
  unloaded = 0;
  /* Check order flag. */
  if (0 == order) {
    /* Unload all plug-ins in forward order. */
    for (index = 0; index < count; index++) {
      /* Check for non-zero plug-in handle. */
      if (0 != handles[index]) {
        /* Unload plug-in. */
        if (0 != harnessclient_plugins_unload(harnesses[index],
                                              handles[index])) {
          errno2 = (0 == errno2)?errno:errno2;
          RMIX_LOG((RMIX_WARN(unable to unload plug-in)))
          /* Check fail flag. */
          if (1 == fail) {
            errno = errno2;
            return unloaded;
          }
        } else {
          unloaded++;
        }
        /* Set plug-in handle to zero. */
        handles[index] = 0;
      }
    }
  } else {
    /* Unload all plug-ins in reverse order. */
    for (index = count; 0 < index--;) {
      /* Check for non-zero plug-in handle. */
      if (0 != handles[index]) {
        /* Unload plug-in. */
        if (0 != harnessclient_plugins_unload(harnesses[index],
                                              handles[index])) {
          errno2 = (0 == errno2)?errno:errno2;
          RMIX_LOG((RMIX_WARN(unable to unload plug-in)))
          /* Check fail flag. */
          if (1 == fail) {
            errno = errno2;
            return unloaded;
          }
        } else {
          unloaded++;
          /* Set plug-in handle to zero. */
          handles[index] = 0;
        }
      }
    }
  }
  errno = errno2;
  return unloaded;
}

/*
 * Invokes harness_plugins_unload (client-side method stub).
 * Unloads a plug-in at multiple remote harness instances in parallel using
 * asynchonous method invocations.
 * The plug-in may be different for every remote harness instance.
 *
 * count     = The number of remote harness instances.
 * harnesses = The array of RMIX remote object references of harness instances.
 * handles   = The plug-in handle array (also return). 0 handles are ignored.
 *             Plug-in handles for successful unloading attempts are set to 0.
 * return    = The number of unloaded plug-ins with errno set appropriately.
 */
#undef  __FUNC__
#define __FUNC__ "harnessclient_plugins_punload"
int harnessclient_plugins_punload (unsigned int             count,
                                   const rmix_remoteref_t **harnesses,
                                   unsigned int            *handles) {
  int                errno2;
  unsigned int       unloaded, index;
  rmix_invokeref_t **invocations;
#ifdef DEBUG
  /* Check harnesses parameter. */
  if ((NULL == harnesses)&&(0 != count)) {
    errno = EINVAL;
    RMIX_LOG((RMIX_WARN(harnesses parameter is null)))
    errno = EINVAL;
    return -1;
  }
  /* Check count parameter. */
  if ((0 == count)&&(NULL != harnesses)) {
    errno = EINVAL;
    RMIX_LOG((RMIX_WARN(count parameter is zero)))
    errno = EINVAL;
    return -1;
  }
  /* Check handles parameter. */
  if ((NULL == handles)&&(0 != count)) {
    errno = EINVAL;
    RMIX_LOG((RMIX_WARN(handles parameter is null)))
    errno = EINVAL;
    return -1;
  }
  /* Check count parameter. */
  if ((0 == count)&&(NULL != handles)) {
    errno = EINVAL;
    RMIX_LOG((RMIX_WARN(count parameter is zero)))
    errno = EINVAL;
    return -1;
  }
#endif /* DEBUG */
  /* Allocate invocation reference array. */
  if (NULL == (invocations = malloc(count * sizeof(rmix_invokeref_t*)))) {
    errno2 = errno;
    RMIX_LOG((RMIX_WARN(unable to allocate invocation reference array)))
    errno = errno2;
    return 0;
  }
  /* Initialize errno and unloaded counter.*/
  errno2   = 0;
  unloaded = 0;
  /* Unload all plug-ins in forward order (send part). */
  for (index = 0; index < count; index++) {
    /* Check for non-zero plug-in handle. */
    if (0 != handles[index]) {
      /* Unload plug-in (send part). */
      if (0 != harnessclient_plugins_unload_send(&invocations[index],
                                                 harnesses[index],
                                                 handles[index])) {
        errno2 = (0 == errno2)?errno:errno2;
        RMIX_LOG((RMIX_WARN(unable to unload plug-in (send part))))
        invocations[index] = NULL;
      }
    }
  }
  /* Unload all plug-ins in forward order (retrieve part). */
  for (index = 0; index < count; index++) {
    /* Check for non-zero plug-in handle. */
    if (0 != handles[index]) {
      /* Check for successful send. */
      if (NULL != invocations[index]) {
        /* Unload plug-in (retrieve part). */
        if (0 != harnessclient_plugins_unload_retrieve(&invocations[index])) {
          errno2 = (0 == errno2)?errno:errno2;
          RMIX_LOG((RMIX_WARN(unable to unload plug-in (retrieve part))))
        } else {
          unloaded++;
          /* Set plug-in handle to zero. */
          handles[index] = 0;
        }
      }
    }
  }
  /* Deallocate invocation reference array. */
  free(invocations);
  errno = errno2;
  return unloaded;
}

/*
 * Invokes harnessclient_processes_execute (client-side method stub).
 * Starts a new process and executes a program at a remote harness instance.
 *
 * harness   = The RMIX remote object reference of harness.
 * handle    = The process handle (return).
 * arguments = The program arguments (first argument is program name).
 * variables = The program variables (may be NULL).
 * return    = 0 on success or -1 on any error with errno set appropriately.
 */
#undef  __FUNC__
#define __FUNC__ "harnessclient_processes_execute"
int harnessclient_processes_execute (const rmix_remoteref_t *harness,
                                     unsigned int           *handle,
                                     const char             *arguments,
                                     const char             *variables) {
  int         result;
  void       *outary[2];
  const void *inary[2];
#ifdef DEBUG
  /* Check harness parameter. */
  if (NULL == harness) {
    errno = EINVAL;
    HARNESS_LOG((HARNESS_WARN(harness parameter is null)))
    errno = EINVAL;
    return -1;
  }
  /* Check handle parameter. */
  if (NULL == handle) {
    errno = EINVAL;
    HARNESS_LOG((HARNESS_WARN(handle parameter is null)))
    errno = EINVAL;
    return -1;
  }
  /* Check arguments parameter. */
  if (NULL == arguments) {
    errno = EINVAL;
    HARNESS_LOG((HARNESS_WARN(arguments parameter is null)))
    errno = EINVAL;
    return -1;
  }
#endif /* DEBUG */
  /* Prepare input. */
  inary[0] = arguments;
  inary[1] = variables;
  /* Prepare output. */
  outary[0] = &result;
  outary[1] =  handle;
  /* Invoke remote object method. */
  if (0 != rmix_invoke(harness, &harnessclient_interface,
                       HARNESS_RMIX_METHODS_PROCESSES_EXECUTE_INDEX,
                       outary, 2, inary, 2)) {
    int errno2 = errno;
    RMIX_LOG((RMIX_WARN(unable to invoke remote object method)))
    errno = errno2;
    return -1;
  }
  return result;
}

/*
 * Invokes harnessclient_processes_execute (client-side method stub).
 * Starts a new process and executes a program at a remote harness instance.
 *
 * harness   = The RMIX remote object reference of harness.
 * provider  = The preloaded RMIX provider plug-in handle.
 * handle    = The process handle (return).
 * arguments = The program arguments (first argument is program name).
 * variables = The program variables (may be NULL).
 * return    = 0 on success or -1 on any error with errno set appropriately.
 */
#undef  __FUNC__
#define __FUNC__ "harnessclient_processes_execute_p"
int harnessclient_processes_execute_p (const rmix_remoteref_t *harness,
                                       unsigned int            provider,
                                       unsigned int           *handle,
                                       const char             *arguments,
                                       const char             *variables) {
  int         result;
  void       *outary[2];
  const void *inary[2];
#ifdef DEBUG
  /* Check harness parameter. */
  if (NULL == harness) {
    errno = EINVAL;
    HARNESS_LOG((HARNESS_WARN(harness parameter is null)))
    errno = EINVAL;
    return -1;
  }
  /* Check handle parameter. */
  if (NULL == handle) {
    errno = EINVAL;
    HARNESS_LOG((HARNESS_WARN(handle parameter is null)))
    errno = EINVAL;
    return -1;
  }
  /* Check arguments parameter. */
  if (NULL == arguments) {
    errno = EINVAL;
    HARNESS_LOG((HARNESS_WARN(arguments parameter is null)))
    errno = EINVAL;
    return -1;
  }
#endif /* DEBUG */
  /* Prepare input. */
  inary[0] = arguments;
  inary[1] = variables;
  /* Prepare output. */
  outary[0] = &result;
  outary[1] =  handle;
  /* Invoke remote object method. */
  if (0 != rmix_invoke_p(harness, provider, &harnessclient_interface,
                         HARNESS_RMIX_METHODS_PROCESSES_EXECUTE_INDEX,
                         outary, 2, inary, 2)) {
    int errno2 = errno;
    RMIX_LOG((RMIX_WARN(unable to invoke remote object method)))
    errno = errno2;
    return -1;
  }
  return result;
}

/*
 * Invokes harness_processes_execute (client-side method stub).
 * Starts a new process and executes a program at a remote harness instance
 * using a oneway invocation.
 *
 * harness   = The RMIX remote object reference of harness.
 * arguments = The program arguments (first argument is program name).
 * variables = The program variables (may be NULL).
 * return    = 0 on success or -1 on any error with errno set appropriately.
 */
#undef  __FUNC__
#define __FUNC__ "harnessclient_processes_execute_oneway"
int harnessclient_processes_execute_oneway (const rmix_remoteref_t *harness,
                                            const char             *arguments,
                                            const char             *variables) {
  const void *inary[2];
#ifdef DEBUG
  /* Check harness parameter. */
  if (NULL == harness) {
    errno = EINVAL;
    HARNESS_LOG((HARNESS_WARN(harness parameter is null)))
    errno = EINVAL;
    return -1;
  }
  /* Check arguments parameter. */
  if (NULL == arguments) {
    errno = EINVAL;
    HARNESS_LOG((HARNESS_WARN(arguments parameter is null)))
    errno = EINVAL;
    return -1;
  }
#endif /* DEBUG */
  /* Prepare input. */
  inary[0] = arguments;
  inary[1] = variables;
  /* Invoke remote object method. */
  if (0 != rmix_oneway(harness, &harnessclient_interface,
                       HARNESS_RMIX_METHODS_PROCESSES_EXECUTE_INDEX,
                       inary, 2)) {
    int errno2 = errno;
    RMIX_LOG((RMIX_WARN(unable to invoke remote object method)))
    errno = errno2;
    return -1;
  }
  return 0;
}

/*
 * Invokes harness_processes_execute (client-side method stub).
 * Starts a new process and executes a program at a remote harness instance
 * using a oneway invocation without loading the provider plug-in.
 *
 * harness   = The RMIX remote object reference of harness.
 * provider  = The preloaded RMIX provider plug-in handle.
 * arguments = The program arguments (first argument is program name).
 * variables = The program variables (may be NULL).
 * return    = 0 on success or -1 on any error with errno set appropriately.
 */
#undef  __FUNC__
#define __FUNC__ "harnessclient_processes_execute_oneway_p"
int harnessclient_processes_execute_oneway_p
  (const rmix_remoteref_t  *harness,
   unsigned int             provider,
   const char              *arguments,
   const char              *variables) {
  const void *inary[2];
#ifdef DEBUG
  /* Check harness parameter. */
  if (NULL == harness) {
    errno = EINVAL;
    HARNESS_LOG((HARNESS_WARN(harness parameter is null)))
    errno = EINVAL;
    return -1;
  }
  /* Check arguments parameter. */
  if (NULL == arguments) {
    errno = EINVAL;
    HARNESS_LOG((HARNESS_WARN(arguments parameter is null)))
    errno = EINVAL;
    return -1;
  }
#endif /* DEBUG */
  /* Prepare input. */
  inary[0] = arguments;
  inary[1] = variables;
  /* Invoke remote object method. */
  if (0 != rmix_oneway_p(harness, provider, &harnessclient_interface,
                         HARNESS_RMIX_METHODS_PROCESSES_EXECUTE_INDEX,
                         inary, 2)) {
    int errno2 = errno;
    RMIX_LOG((RMIX_WARN(unable to invoke remote object method)))
    errno = errno2;
    return -1;
  }
  return 0;
}

/*
 * Invokes harness_processes_execute (client-side method stub).
 * Starts a new process and executes a program at a remote harness instance
 * using an asynchronous invocation.
 *
 * invokeref = The invocation reference (return).
 * harness   = The RMIX remote object reference of harness.
 * arguments = The program arguments (first argument is program name).
 * variables = The program variables (may be NULL).
 * return    = 0 on success or -1 on any error with errno set appropriately.
 */
#undef  __FUNC__
#define __FUNC__ "harnessclient_processes_execute_send"
int harnessclient_processes_execute_send (rmix_invokeref_t       **invokeref,
                                          const rmix_remoteref_t  *harness,
                                          const char              *arguments,
                                          const char              *variables) {
  const void *inary[2];
#ifdef DEBUG
  /* Check invokeref parameter. */
  if (NULL == invokeref) {
    errno = EINVAL;
    HARNESS_LOG((HARNESS_WARN(invokeref parameter is null)))
    errno = EINVAL;
    return -1;
  }
  /* Check harness parameter. */
  if (NULL == harness) {
    errno = EINVAL;
    HARNESS_LOG((HARNESS_WARN(harness parameter is null)))
    errno = EINVAL;
    return -1;
  }
  /* Check arguments parameter. */
  if (NULL == arguments) {
    errno = EINVAL;
    HARNESS_LOG((HARNESS_WARN(arguments parameter is null)))
    errno = EINVAL;
    return -1;
  }
#endif /* DEBUG */
  /* Prepare input. */
  inary[0] = arguments;
  inary[1] = variables;
  /* Invoke remote object method. */
  if (0 != rmix_send(invokeref, harness, &harnessclient_interface,
                     HARNESS_RMIX_METHODS_PROCESSES_EXECUTE_INDEX,
                     inary, 2)) {
    int errno2 = errno;
    RMIX_LOG((RMIX_WARN(unable to invoke remote object method)))
    errno = errno2;
    return -1;
  }
  return 0;
}

/*
 * Invokes harness_processes_execute (client-side method stub).
 * Starts a new process and executes a program at a remote harness instance
 * using an asynchronous invocation without loading the provider plug-in.
 *
 * invokeref = The invocation reference (return).
 * harness   = The RMIX remote object reference of harness.
 * provider  = The preloaded RMIX provider plug-in handle.
 * arguments = The program arguments (first argument is program name).
 * variables = The program variables (may be NULL).
 * return    = 0 on success or -1 on any error with errno set appropriately.
 */
#undef  __FUNC__
#define __FUNC__ "harnessclient_processes_execute_send_p"
int harnessclient_processes_execute_send_p
  (rmix_invokeref_t       **invokeref,
   const rmix_remoteref_t  *harness,
   unsigned int             provider,
   const char              *arguments,
   const char              *variables) {
  const void *inary[2];
#ifdef DEBUG
  /* Check invokeref parameter. */
  if (NULL == invokeref) {
    errno = EINVAL;
    HARNESS_LOG((HARNESS_WARN(invokeref parameter is null)))
    errno = EINVAL;
    return -1;
  }
  /* Check harness parameter. */
  if (NULL == harness) {
    errno = EINVAL;
    HARNESS_LOG((HARNESS_WARN(harness parameter is null)))
    errno = EINVAL;
    return -1;
  }
  /* Check arguments parameter. */
  if (NULL == arguments) {
    errno = EINVAL;
    HARNESS_LOG((HARNESS_WARN(arguments parameter is null)))
    errno = EINVAL;
    return -1;
  }
#endif /* DEBUG */
  /* Prepare input. */
  inary[0] = arguments;
  inary[1] = variables;
  /* Invoke remote object method. */
  if (0 != rmix_send_p(invokeref, harness, provider, &harnessclient_interface,
                       HARNESS_RMIX_METHODS_PROCESSES_EXECUTE_INDEX,
                       inary, 2)) {
    int errno2 = errno;
    RMIX_LOG((RMIX_WARN(unable to invoke remote object method)))
    errno = errno2;
    return -1;
  }
  return 0;
}

/*
 * Retrieves harness_processes_execute output (client-side method stub).
 * Returns the output of a previous asynchronous invocation that starts a new
 * process and executes a program at a remote harness instance.
 *
 * invokeref = The invocation reference (also return).
 * handle    = The process handle (return).
 * return    = 0 on success or -1 on any error with errno set appropriately.
 */
#undef  __FUNC__
#define __FUNC__ "harnessclient_processes_execute_retrieve"
int harnessclient_processes_execute_retrieve (rmix_invokeref_t **invokeref,
                                              unsigned int      *handle) {
  int   result;
  void *outary[2];
#ifdef DEBUG
  /* Check invokeref parameter. */
  if (NULL == invokeref) {
    errno = EINVAL;
    HARNESS_LOG((HARNESS_WARN(invokeref parameter is null)))
    errno = EINVAL;
    return -1;
  }
  /* Check *invokeref parameter. */
  if (NULL == *invokeref) {
    errno = EINVAL;
    HARNESS_LOG((HARNESS_WARN(*invokeref parameter is null)))
    errno = EINVAL;
    return -1;
  }
  /* Check handle parameter. */
  if (NULL == handle) {
    errno = EINVAL;
    HARNESS_LOG((HARNESS_WARN(handle parameter is null)))
    errno = EINVAL;
    return -1;
  }
#endif /* DEBUG */
  /* Prepare output. */
  outary[0] = &result;
  outary[1] =  handle;
  /* Invoke remote object method. */
  if (0 != rmix_retrieve(invokeref, outary, 2)) {
    int errno2 = errno;
    RMIX_LOG((RMIX_WARN(unable to invoke remote object method)))
    errno = errno2;
    return -1;
  }
  return result;
}

/*
 * Retrieves harness_processes_execute output (client-side method stub).
 * Returns the output of a previous asynchronous invocation that starts a new
 * process and executes a program at a remote harness instance without
 * unloading the provider plug-in.
 *
 * invokeref = The invocation reference (also return).
 * handle    = The process handle (return).
 * return    = 0 on success or -1 on any error with errno set appropriately.
 */
#undef  __FUNC__
#define __FUNC__ "harnessclient_processes_execute_retrieve_p"
int harnessclient_processes_execute_retrieve_p (rmix_invokeref_t **invokeref,
                                                unsigned int      *handle) {
  int   result;
  void *outary[2];
#ifdef DEBUG
  /* Check invokeref parameter. */
  if (NULL == invokeref) {
    errno = EINVAL;
    HARNESS_LOG((HARNESS_WARN(invokeref parameter is null)))
    errno = EINVAL;
    return -1;
  }
  /* Check *invokeref parameter. */
  if (NULL == *invokeref) {
    errno = EINVAL;
    HARNESS_LOG((HARNESS_WARN(*invokeref parameter is null)))
    errno = EINVAL;
    return -1;
  }
  /* Check handle parameter. */
  if (NULL == handle) {
    errno = EINVAL;
    HARNESS_LOG((HARNESS_WARN(handle parameter is null)))
    errno = EINVAL;
    return -1;
  }
#endif /* DEBUG */
  /* Prepare output. */
  outary[0] = &result;
  outary[1] =  handle;
  /* Invoke remote object method. */ 
  if (0 != rmix_retrieve_p(invokeref, outary, 2)) {
    int errno2 = errno;
    RMIX_LOG((RMIX_WARN(unable to invoke remote object method)))
    errno = errno2;
    return -1;
  }
  return result;
}

/*
 * Invokes harness_processes_execute (client-side method stub).
 * Executes a new process and executes a program at multiple remote harness
 * instances. The program may be different for every remote harness instance.
 *
 * count     = The number of remote harness instances.
 * harnesses = The array of RMIX remote object references of harness instances.
 * handles   = The process handle array (also return). Process handles are
 *             unique, and 0 for unsuccessful execution attempts.
 * arguments = The program arguments array (first argument is program name).
 * variables = The program variables array (may be NULL).
 * order     = The order flag. 0 for forward and 1 for reverse.
 * fail      = The fail flag. 0 to continue and 1 to fail on any error.
 * return    = The number of executed programs with errno set appropriately.
 */
#undef  __FUNC__
#define __FUNC__ "harnessclient_processes_sexecute_mp"
unsigned int harnessclient_processes_sexecute_mp
  (unsigned int             count,
   const rmix_remoteref_t **harnesses,
   unsigned int            *handles,
   const char             **arguments,
   const char             **variables,
   unsigned int             order,
   unsigned int             fail) {
  int          errno2;
  unsigned int executed, index;
#ifdef DEBUG
  /* Check harnesses parameter. */
  if ((NULL == harnesses)&&(0 != count)) {
    errno = EINVAL;
    RMIX_LOG((RMIX_WARN(harnesses parameter is null)))
    errno = EINVAL;
    return -1;
  }
  /* Check count parameter. */
  if ((0 == count)&&(NULL != harnesses)) {
    errno = EINVAL;
    RMIX_LOG((RMIX_WARN(count parameter is zero)))
    errno = EINVAL;
    return -1;
  }
  /* Check handles parameter. */
  if ((NULL == handles)&&(0 != count)) {
    errno = EINVAL;
    RMIX_LOG((RMIX_WARN(handles parameter is null)))
    errno = EINVAL;
    return -1;
  }
  /* Check count parameter. */
  if ((0 == count)&&(NULL != handles)) {
    errno = EINVAL;
    RMIX_LOG((RMIX_WARN(count parameter is zero)))
    errno = EINVAL;
    return -1;
  }
  /* Check arguments parameter. */
  if ((NULL == arguments)&&(0 != count)) {
    errno = EINVAL;
    RMIX_LOG((RMIX_WARN(arguments parameter is null)))
    errno = EINVAL;
    return -1;
  }
  /* Check count parameter. */
  if ((0 == count)&&(NULL != arguments)) {
    errno = EINVAL;
    RMIX_LOG((RMIX_WARN(count parameter is zero)))
    errno = EINVAL;
    return -1;
  }
#endif /* DEBUG */
  /* Initialize errno and executed counter.*/
  errno2 = 0;
  executed = 0;
  /* Initialize all handles. */
  memset(handles, 0, count * sizeof(*handles));
  /* Check order flag. */
  if (0 == order) {
    /* Execute all programs in forward order. */
    for (index = 0; index < count; index++) {
      /* Execute program. */
      if (0 != harnessclient_processes_execute(harnesses[index],
                                               &handles[index],
                                               arguments[index],
                                               (NULL != variables)?
                                               variables[index]:NULL)) {
        errno2 = (0 == errno2)?errno:errno2;
        RMIX_LOG((RMIX_WARN(unable to execute program)))
        /* Check fail flag. */
        if (1 == fail) {
          errno = errno2;
          return executed;
        }
      } else {
        executed++;
      }
    }
  } else {
    /* Execute all programs in reverse order. */
    for (index = count; 0 < index--;) {
      /* Execute program. */
      if (0 != harnessclient_processes_execute(harnesses[index],
                                               &handles[index],
                                               arguments[index],
                                               (NULL != variables)?
                                               variables[index]:NULL)) {
        errno2 = (0 == errno2)?errno:errno2;
        RMIX_LOG((RMIX_WARN(unable to execute program)))
        /* Check fail flag. */
        if (1 == fail) {
          errno = errno2;
          return executed;
        }
      } else {
        executed++;
      }
    }
  }
  errno = errno2;
  return executed;
}

/*
 * Invokes harness_processes_execute (client-side method stub).
 * Executes a new process and executes a program at multiple remote harness
 * instances. The program is the same for every remote harness instance.
 *
 * count     = The number of remote harness instances.
 * harnesses = The array of RMIX remote object references of harness instances.
 * handles   = The process handle array (also return). Process handles are
 *             unique, and 0 for unsuccessful execution attempts.
 * arguments = The program arguments (first argument is program name).
 * variables = The program variables (may be NULL).
 * order     = The order flag. 0 for forward and 1 for reverse.
 * fail      = The fail flag. 0 to continue and 1 to fail on any error.
 * return    = The number of executed programs with errno set appropriately.
 */
#undef  __FUNC__
#define __FUNC__ "harnessclient_processes_sexecute_sp"
unsigned int harnessclient_processes_sexecute_sp
  (unsigned int             count,
   const rmix_remoteref_t **harnesses,
   unsigned int            *handles,
   const char              *arguments,
   const char              *variables,
   unsigned int             order,
   unsigned int             fail) {
  int          errno2;
  unsigned int executed, index;
#ifdef DEBUG
  /* Check harnesses parameter. */
  if ((NULL == harnesses)&&(0 != count)) {
    errno = EINVAL;
    RMIX_LOG((RMIX_WARN(harnesses parameter is null)))
    errno = EINVAL;
    return -1;
  }
  /* Check count parameter. */
  if ((0 == count)&&(NULL != harnesses)) {
    errno = EINVAL;
    RMIX_LOG((RMIX_WARN(count parameter is zero)))
    errno = EINVAL;
    return -1;
  }
  /* Check handles parameter. */
  if ((NULL == handles)&&(0 != count)) {
    errno = EINVAL;
    RMIX_LOG((RMIX_WARN(handles parameter is null)))
    errno = EINVAL;
    return -1;
  }
  /* Check count parameter. */
  if ((0 == count)&&(NULL != handles)) {
    errno = EINVAL;
    RMIX_LOG((RMIX_WARN(count parameter is zero)))
    errno = EINVAL;
    return -1;
  }
  /* Check arguments parameter. */
  if (NULL == arguments) {
    errno = EINVAL;
    HARNESS_LOG((HARNESS_WARN(arguments parameter is null)))
    errno = EINVAL;
    return -1;
  }
#endif /* DEBUG */
  /* Initialize errno and executed counter.*/
  errno2 = 0;
  executed = 0;
  /* Initialize all handles. */
  memset(handles, 0, count * sizeof(*handles));
  /* Check order flag. */
  if (0 == order) {
    /* Execute all programs in forward order. */
    for (index = 0; index < count; index++) {
      /* Execute program. */
      if (0 != harnessclient_processes_execute(harnesses[index],
                                               &handles[index],
                                               arguments, variables)) {
        errno2 = (0 == errno2)?errno:errno2;
        RMIX_LOG((RMIX_WARN(unable to execute program)))
        /* Check fail flag. */
        if (1 == fail) {
          errno = errno2;
          return executed;
        }
      } else {
        executed++;
      }
    }
  } else {
    /* Execute all programs in reverse order. */
    for (index = count; 0 < index--;) {
      /* Execute program. */
      if (0 != harnessclient_processes_execute(harnesses[index],
                                               &handles[index],
                                               arguments, variables)) {
        errno2 = (0 == errno2)?errno:errno2;
        RMIX_LOG((RMIX_WARN(unable to execute program)))
        /* Check fail flag. */
        if (1 == fail) {
          errno = errno2;
          return executed;
        }
      } else {
        executed++;
      }
    }
  }
  errno = errno2;
  return executed;
}

/*
 * Invokes harness_processes_execute (client-side method stub).
 * Executes a new process and executes a program at multiple remote harness
 * instances in parallel using asynchonous method invocations. The program may
 * be different for every remote harness instance.
 *
 * count     = The number of remote harness instances.
 * harnesses = The array of RMIX remote object references of harness instances.
 * handles   = The process handle array (also return). Process handles are
 *             unique, and 0 for unsuccessful execution attempts.
 * arguments = The program arguments array (first argument is program name).
 * variables = The program variables array (may be NULL).
 * return    = The number of executed programs with errno set appropriately.
 */
#undef  __FUNC__
#define __FUNC__ "harnessclient_processes_pexecute_mp"
unsigned int harnessclient_processes_pexecute_mp
  (unsigned int             count,
   const rmix_remoteref_t **harnesses,
   unsigned int            *handles,
   const char             **arguments,
   const char             **variables) {
  int                errno2;
  unsigned int       executed, index;
  rmix_invokeref_t **invocations;
#ifdef DEBUG
  /* Check harnesses parameter. */
  if ((NULL == harnesses)&&(0 != count)) {
    errno = EINVAL;
    RMIX_LOG((RMIX_WARN(harnesses parameter is null)))
    errno = EINVAL;
    return -1;
  }
  /* Check count parameter. */
  if ((0 == count)&&(NULL != harnesses)) {
    errno = EINVAL;
    RMIX_LOG((RMIX_WARN(count parameter is zero)))
    errno = EINVAL;
    return -1;
  }
  /* Check handles parameter. */
  if ((NULL == handles)&&(0 != count)) {
    errno = EINVAL;
    RMIX_LOG((RMIX_WARN(handles parameter is null)))
    errno = EINVAL;
    return -1;
  }
  /* Check count parameter. */
  if ((0 == count)&&(NULL != handles)) {
    errno = EINVAL;
    RMIX_LOG((RMIX_WARN(count parameter is zero)))
    errno = EINVAL;
    return -1;
  }
  /* Check arguments parameter. */
  if ((NULL == arguments)&&(0 != count)) {
    errno = EINVAL;
    RMIX_LOG((RMIX_WARN(arguments parameter is null)))
    errno = EINVAL;
    return -1;
  }
  /* Check count parameter. */
  if ((0 == count)&&(NULL != arguments)) {
    errno = EINVAL;
    RMIX_LOG((RMIX_WARN(count parameter is zero)))
    errno = EINVAL;
    return -1;
  }
#endif /* DEBUG */
  /* Allocate invocation reference array. */
  if (NULL == (invocations = malloc(count * sizeof(rmix_invokeref_t*)))) {
    errno2 = errno;
    RMIX_LOG((RMIX_WARN(unable to allocate invocation reference array)))
    errno = errno2;
    return 0;
  }
  /* Initialize errno and executed counter.*/
  errno2 = 0;
  executed = 0;
  /* Initialize all handles. */
  memset(handles, 0, count * sizeof(*handles));
  /* Execute all programs in forward order (send part). */
  for (index = 0; index < count; index++) {
    /* Execute program (send part). */
    if (0 != harnessclient_processes_execute_send(&invocations[index],
                                                  harnesses[index],
                                                  arguments[index],
                                                  (NULL != variables)?
                                                  variables[index]:NULL)) {
      errno2 = (0 == errno2)?errno:errno2;
      RMIX_LOG((RMIX_WARN(unable to execute program (send part))))
      invocations[index] = NULL;
    }
  }
  /* Execute all programs in forward order (retrieve part). */
  for (index = 0; index < count; index++) {
    /* Check for successful send. */
    if (NULL != invocations[index]) {
      /* Execute program (retrieve part). */
      if (0 != harnessclient_processes_execute_retrieve(&invocations[index],
                                                        &handles[index])) {
        errno2 = (0 == errno2)?errno:errno2;
        RMIX_LOG((RMIX_WARN(unable to execute program (retrieve part))))
      } else {
        executed++;
      }
    }
  }
  /* Deallocate invocation reference array. */
  free(invocations);
  errno = errno2;
  return executed;
}

/*
 * Invokes harness_processes_execute (client-side method stub).
 * Executes a new process and executes a program at multiple remote harness
 * instances in parallel using asynchonous method invocations. The program is
 * the same for every remote harness instance.
 *
 * count     = The number of remote harness instances.
 * harnesses = The array of RMIX remote object references of harness instances.
 * handles   = The process handle array (also return). Process handles are
 *             unique, and 0 for unsuccessful execution attempts.
 * arguments = The program arguments array (first argument is program name).
 * variables = The program variables array (may be NULL).
 * return    = The number of executed programs with errno set appropriately.
 */
#undef  __FUNC__
#define __FUNC__ "harnessclient_processes_pexecute_sp"
unsigned int harnessclient_processes_pexecute_sp
  (unsigned int             count,
   const rmix_remoteref_t **harnesses,
   unsigned int            *handles,
   const char              *arguments,
   const char              *variables) {
  int                errno2;
  unsigned int       executed, index;
  rmix_invokeref_t **invocations;
#ifdef DEBUG
  /* Check harnesses parameter. */
  if ((NULL == harnesses)&&(0 != count)) {
    errno = EINVAL;
    RMIX_LOG((RMIX_WARN(harnesses parameter is null)))
    errno = EINVAL;
    return -1;
  }
  /* Check count parameter. */
  if ((0 == count)&&(NULL != harnesses)) {
    errno = EINVAL;
    RMIX_LOG((RMIX_WARN(count parameter is zero)))
    errno = EINVAL;
    return -1;
  }
  /* Check handles parameter. */
  if ((NULL == handles)&&(0 != count)) {
    errno = EINVAL;
    RMIX_LOG((RMIX_WARN(handles parameter is null)))
    errno = EINVAL;
    return -1;
  }
  /* Check count parameter. */
  if ((0 == count)&&(NULL != handles)) {
    errno = EINVAL;
    RMIX_LOG((RMIX_WARN(count parameter is zero)))
    errno = EINVAL;
    return -1;
  }
  /* Check arguments parameter. */
  if (NULL == arguments) {
    errno = EINVAL;
    HARNESS_LOG((HARNESS_WARN(arguments parameter is null)))
    errno = EINVAL;
    return -1;
  }
#endif /* DEBUG */
  /* Allocate invocation reference array. */
  if (NULL == (invocations = malloc(count * sizeof(rmix_invokeref_t*)))) {
    errno2 = errno;
    RMIX_LOG((RMIX_WARN(unable to allocate invocation reference array)))
    errno = errno2;
    return 0;
  }
  /* Initialize errno and executed counter.*/
  errno2 = 0;
  executed = 0;
  /* Initialize all handles. */
  memset(handles, 0, count * sizeof(*handles));
  /* Execute all programs in forward order (send part). */
  for (index = 0; index < count; index++) {
    /* Execute program (send part). */
    if (0 != harnessclient_processes_execute_send(&invocations[index],
                                                  harnesses[index],
                                                  arguments, variables)) {
      errno2 = (0 == errno2)?errno:errno2;
      RMIX_LOG((RMIX_WARN(unable to execute program (send part))))
      invocations[index] = NULL;
    }
  }
  /* Execute all programs in forward order (retrieve part). */
  for (index = 0; index < count; index++) {
    /* Check for successful send. */
    if (NULL != invocations[index]) {
      /* Execute program (retrieve part). */
      if (0 != harnessclient_processes_execute_retrieve(&invocations[index],
                                                        &handles[index])) {
        errno2 = (0 == errno2)?errno:errno2;
        RMIX_LOG((RMIX_WARN(unable to execute program (retrieve part))))
      } else {
        executed++;
      }
    }
  }
  /* Deallocate invocation reference array. */
  free(invocations);
  errno = errno2;
  return executed;
}

/*
 * Invokes harnessclient_processes_execute2 (client-side method stub).
 * Starts a new process and executes a program at a remote harness instance.
 *
 * harness = The RMIX remote object reference of harness.
 * handle  = The process handle (return).
 * argc    = The number of arguments.
 * envc    = The number of variables.
 * argv    = The array of arguments (argv[0] is program name).
 * envv    = The array of variables.
 * return  = 0 on success or -1 on any error with errno set appropriately.
 */
#undef  __FUNC__
#define __FUNC__ "harnessclient_processes_execute2"
int harnessclient_processes_execute2 (const rmix_remoteref_t *harness,
                                      unsigned int           *handle,
                                      unsigned int            argc,
                                      unsigned int            envc,
                                      const char            **argv,
                                      const char            **envv) {
  int         result;
  void       *outary[2];
  const void *inary[4];
#ifdef DEBUG
  /* Check harness parameter. */
  if (NULL == harness) {
    errno = EINVAL;
    HARNESS_LOG((HARNESS_WARN(harness parameter is null)))
    errno = EINVAL;
    return -1;
  }
  /* Check handle parameter. */
  if (NULL == handle) {
    errno = EINVAL;
    HARNESS_LOG((HARNESS_WARN(handle parameter is null)))
    errno = EINVAL;
    return -1;
  }
  /* Check argv parameter. */
  if ((NULL == argv)&&(0 != argc)) {
    errno = EINVAL;
    RMIX_LOG((RMIX_WARN(argv parameter is null)))
    errno = EINVAL;
    return -1;
  }
  /* Check argc parameter. */
  if ((0 == argc)&&(NULL != argv)) {
    errno = EINVAL;
    RMIX_LOG((RMIX_WARN(argc parameter is zero)))
    errno = EINVAL;
    return -1;
  }
  /* Check envv parameter. */
  if ((NULL == envv)&&(0 != envc)) {
    errno = EINVAL;
    RMIX_LOG((RMIX_WARN(envv parameter is null)))
    errno = EINVAL;
    return -1;
  }
  /* Check envc parameter. */
  if ((0 == envc)&&(NULL != envv)) {
    errno = EINVAL;
    RMIX_LOG((RMIX_WARN(envc parameter is zero)))
    errno = EINVAL;
    return -1;
  }
#endif /* DEBUG */
  /* Prepare input. */
  inary[0] = &argc;
  inary[1] =  argv;
  inary[2] = &envc;
  inary[3] =  envv;
  /* Prepare output. */
  outary[0] = &result;
  outary[1] =  handle;
  /* Invoke remote object method. */
  if (0 != rmix_invoke(harness, &harnessclient_interface,
                       HARNESS_RMIX_METHODS_PROCESSES_EXECUTE2_INDEX,
                       outary, 2, inary, 4)) {
    int errno2 = errno;
    RMIX_LOG((RMIX_WARN(unable to invoke remote object method)))
    errno = errno2;
    return -1;
  }
  return result;
}

/*
 * Invokes harnessclient_processes_execute2 (client-side method stub).
 * Starts a new process and executes a program at a remote harness instance.
 *
 * harness  = The RMIX remote object reference of harness.
 * provider = The preloaded RMIX provider plug-in handle.
 * handle   = The process handle (return).
 * argc     = The number of arguments.
 * envc     = The number of variables.
 * argv     = The array of arguments (argv[0] is program name).
 * envv     = The array of variables.
 * return   = 0 on success or -1 on any error with errno set appropriately.
 */
#undef  __FUNC__
#define __FUNC__ "harnessclient_processes_execute2_p"
int harnessclient_processes_execute2_p (const rmix_remoteref_t *harness,
                                        unsigned int            provider,
                                        unsigned int           *handle,
                                        unsigned int            argc,
                                        unsigned int            envc,
                                        const char            **argv,
                                        const char            **envv) {
  int         result;
  void       *outary[2];
  const void *inary[4];
#ifdef DEBUG
  /* Check harness parameter. */
  if (NULL == harness) {
    errno = EINVAL;
    HARNESS_LOG((HARNESS_WARN(harness parameter is null)))
    errno = EINVAL;
    return -1;
  }
  /* Check handle parameter. */
  if (NULL == handle) {
    errno = EINVAL;
    HARNESS_LOG((HARNESS_WARN(handle parameter is null)))
    errno = EINVAL;
    return -1;
  }
  /* Check argv parameter. */
  if ((NULL == argv)&&(0 != argc)) {
    errno = EINVAL;
    RMIX_LOG((RMIX_WARN(argv parameter is null)))
    errno = EINVAL;
    return -1;
  }
  /* Check argc parameter. */
  if ((0 == argc)&&(NULL != argv)) {
    errno = EINVAL;
    RMIX_LOG((RMIX_WARN(argc parameter is zero)))
    errno = EINVAL;
    return -1;
  }
  /* Check envv parameter. */
  if ((NULL == envv)&&(0 != envc)) {
    errno = EINVAL;
    RMIX_LOG((RMIX_WARN(envv parameter is null)))
    errno = EINVAL;
    return -1;
  }
  /* Check envc parameter. */
  if ((0 == envc)&&(NULL != envv)) {
    errno = EINVAL;
    RMIX_LOG((RMIX_WARN(envc parameter is zero)))
    errno = EINVAL;
    return -1;
  }
#endif /* DEBUG */
  /* Prepare input. */
  inary[0] = &argc;
  inary[1] =  argv;
  inary[2] = &envc;
  inary[3] =  envv;
  /* Prepare output. */
  outary[0] = &result;
  outary[1] =  handle;
  /* Invoke remote object method. */
  if (0 != rmix_invoke_p(harness, provider, &harnessclient_interface,
                         HARNESS_RMIX_METHODS_PROCESSES_EXECUTE2_INDEX,
                         outary, 2, inary, 4)) {
    int errno2 = errno;
    RMIX_LOG((RMIX_WARN(unable to invoke remote object method)))
    errno = errno2;
    return -1;
  }
  return result;
}

/*
 * Invokes harness_processes_execute2 (client-side method stub).
 * Starts a new process and executes a program at a remote harness instance
 * using a oneway invocation.
 *
 * harness = The RMIX remote object reference of harness.
 * argc    = The number of arguments.
 * envc    = The number of variables.
 * argv    = The array of arguments (argv[0] is program name).
 * envv    = The array of variables.
 * return  = 0 on success or -1 on any error with errno set appropriately.
 */
#undef  __FUNC__
#define __FUNC__ "harnessclient_processes_execute2_oneway"
int harnessclient_processes_execute2_oneway (const rmix_remoteref_t  *harness,
                                             unsigned int             argc,
                                             unsigned int             envc,
                                             const char             **argv,
                                             const char             **envv) {
  const void *inary[4];
#ifdef DEBUG
  /* Check harness parameter. */
  if (NULL == harness) {
    errno = EINVAL;
    HARNESS_LOG((HARNESS_WARN(harness parameter is null)))
    errno = EINVAL;
    return -1;
  }
  /* Check argv parameter. */
  if ((NULL == argv)&&(0 != argc)) {
    errno = EINVAL;
    RMIX_LOG((RMIX_WARN(argv parameter is null)))
    errno = EINVAL;
    return -1;
  }
  /* Check argc parameter. */
  if ((0 == argc)&&(NULL != argv)) {
    errno = EINVAL;
    RMIX_LOG((RMIX_WARN(argc parameter is zero)))
    errno = EINVAL;
    return -1;
  }
  /* Check envv parameter. */
  if ((NULL == envv)&&(0 != envc)) {
    errno = EINVAL;
    RMIX_LOG((RMIX_WARN(envv parameter is null)))
    errno = EINVAL;
    return -1;
  }
  /* Check envc parameter. */
  if ((0 == envc)&&(NULL != envv)) {
    errno = EINVAL;
    RMIX_LOG((RMIX_WARN(envc parameter is zero)))
    errno = EINVAL;
    return -1;
  }
#endif /* DEBUG */
  /* Prepare input. */
  inary[0] = &argc;
  inary[1] =  argv;
  inary[2] = &envc;
  inary[3] =  envv;
  /* Invoke remote object method. */
  if (0 != rmix_oneway(harness, &harnessclient_interface,
                       HARNESS_RMIX_METHODS_PROCESSES_EXECUTE2_INDEX,
                       inary, 4)) {
    int errno2 = errno;
    RMIX_LOG((RMIX_WARN(unable to invoke remote object method)))
    errno = errno2;
    return -1;
  }
  return 0;
}

/*
 * Invokes harness_processes_execute2 (client-side method stub).
 * Starts a new process and executes a program at a remote harness instance
 * using a oneway invocation without loading the provider plug-in.
 *
 * harness  = The RMIX remote object reference of harness.
 * provider = The preloaded RMIX provider plug-in handle.
 * argc     = The number of arguments.
 * envc     = The number of variables.
 * argv     = The array of arguments (argv[0] is program name).
 * envv     = The array of variables.
 * return   = 0 on success or -1 on any error with errno set appropriately.
 */
#undef  __FUNC__
#define __FUNC__ "harnessclient_processes_execute2_oneway_p"
int harnessclient_processes_execute2_oneway_p
  (const rmix_remoteref_t  *harness,
   unsigned int             provider,
   unsigned int             argc,
   unsigned int             envc,
   const char             **argv,
   const char             **envv) {
  const void *inary[4];
#ifdef DEBUG
  /* Check harness parameter. */
  if (NULL == harness) {
    errno = EINVAL;
    HARNESS_LOG((HARNESS_WARN(harness parameter is null)))
    errno = EINVAL;
    return -1;
  }
  /* Check argv parameter. */
  if ((NULL == argv)&&(0 != argc)) {
    errno = EINVAL;
    RMIX_LOG((RMIX_WARN(argv parameter is null)))
    errno = EINVAL;
    return -1;
  }
  /* Check argc parameter. */
  if ((0 == argc)&&(NULL != argv)) {
    errno = EINVAL;
    RMIX_LOG((RMIX_WARN(argc parameter is zero)))
    errno = EINVAL;
    return -1;
  }
  /* Check envv parameter. */
  if ((NULL == envv)&&(0 != envc)) {
    errno = EINVAL;
    RMIX_LOG((RMIX_WARN(envv parameter is null)))
    errno = EINVAL;
    return -1;
  }
  /* Check envc parameter. */
  if ((0 == envc)&&(NULL != envv)) {
    errno = EINVAL;
    RMIX_LOG((RMIX_WARN(envc parameter is zero)))
    errno = EINVAL;
    return -1;
  }
#endif /* DEBUG */
  /* Prepare input. */
  inary[0] = &argc;
  inary[1] =  argv;
  inary[2] = &envc;
  inary[3] =  envv;
  /* Invoke remote object method. */
  if (0 != rmix_oneway_p(harness, provider, &harnessclient_interface,
                         HARNESS_RMIX_METHODS_PROCESSES_EXECUTE2_INDEX,
                         inary, 4)) {
    int errno2 = errno;
    RMIX_LOG((RMIX_WARN(unable to invoke remote object method)))
    errno = errno2;
    return -1;
  }
  return 0;
}

/*
 * Invokes harness_processes_execute2 (client-side method stub).
 * Starts a new process and executes a program at a remote harness instance
 * using an asynchronous invocation.
 *
 * invokeref = The invocation reference (return).
 * harness   = The RMIX remote object reference of harness.
 * argc      = The number of arguments.
 * envc      = The number of variables.
 * argv      = The array of arguments (argv[0] is program name).
 * envv      = The array of variables.
 * return    = 0 on success or -1 on any error with errno set appropriately.
 */
#undef  __FUNC__
#define __FUNC__ "harnessclient_processes_execute2_send"
int harnessclient_processes_execute2_send (rmix_invokeref_t       **invokeref,
                                           const rmix_remoteref_t  *harness,
                                           unsigned int             argc,
                                           unsigned int             envc,
                                           const char             **argv,
                                           const char             **envv) {
  const void *inary[4];
#ifdef DEBUG
  /* Check invokeref parameter. */
  if (NULL == invokeref) {
    errno = EINVAL;
    HARNESS_LOG((HARNESS_WARN(invokeref parameter is null)))
    errno = EINVAL;
    return -1;
  }
  /* Check harness parameter. */
  if (NULL == harness) {
    errno = EINVAL;
    HARNESS_LOG((HARNESS_WARN(harness parameter is null)))
    errno = EINVAL;
    return -1;
  }
  /* Check argv parameter. */
  if ((NULL == argv)&&(0 != argc)) {
    errno = EINVAL;
    RMIX_LOG((RMIX_WARN(argv parameter is null)))
    errno = EINVAL;
    return -1;
  }
  /* Check argc parameter. */
  if ((0 == argc)&&(NULL != argv)) {
    errno = EINVAL;
    RMIX_LOG((RMIX_WARN(argc parameter is zero)))
    errno = EINVAL;
    return -1;
  }
  /* Check envv parameter. */
  if ((NULL == envv)&&(0 != envc)) {
    errno = EINVAL;
    RMIX_LOG((RMIX_WARN(envv parameter is null)))
    errno = EINVAL;
    return -1;
  }
  /* Check envc parameter. */
  if ((0 == envc)&&(NULL != envv)) {
    errno = EINVAL;
    RMIX_LOG((RMIX_WARN(envc parameter is zero)))
    errno = EINVAL;
    return -1;
  }
#endif /* DEBUG */
  /* Prepare input. */
  inary[0] = &argc;
  inary[1] =  argv;
  inary[2] = &envc;
  inary[3] =  envv;
  /* Invoke remote object method. */
  if (0 != rmix_send(invokeref, harness, &harnessclient_interface,
                     HARNESS_RMIX_METHODS_PROCESSES_EXECUTE2_INDEX,
                     inary, 4)) {
    int errno2 = errno;
    RMIX_LOG((RMIX_WARN(unable to invoke remote object method)))
    errno = errno2;
    return -1;
  }
  return 0;
}

/*
 * Invokes harness_processes_execute2 (client-side method stub).
 * Starts a new process and executes a program at a remote harness instance
 * using an asynchronous invocation without loading the provider plug-in.
 *
 * invokeref = The invocation reference (return).
 * harness   = The RMIX remote object reference of harness.
 * provider  = The preloaded RMIX provider plug-in handle.
 * argc      = The number of arguments.
 * envc      = The number of variables.
 * argv      = The array of arguments (argv[0] is program name).
 * envv      = The array of variables.
 * return    = 0 on success or -1 on any error with errno set appropriately.
 */
#undef  __FUNC__
#define __FUNC__ "harnessclient_processes_execute2_send_p"
int harnessclient_processes_execute2_send_p
  (rmix_invokeref_t       **invokeref,
   const rmix_remoteref_t  *harness,
   unsigned int             provider,
   unsigned int             argc,
   unsigned int             envc,
   const char             **argv,
   const char             **envv) {
  const void *inary[4];
#ifdef DEBUG
  /* Check invokeref parameter. */
  if (NULL == invokeref) {
    errno = EINVAL;
    HARNESS_LOG((HARNESS_WARN(invokeref parameter is null)))
    errno = EINVAL;
    return -1;
  }
  /* Check harness parameter. */
  if (NULL == harness) {
    errno = EINVAL;
    HARNESS_LOG((HARNESS_WARN(harness parameter is null)))
    errno = EINVAL;
    return -1;
  }
  /* Check argv parameter. */
  if ((NULL == argv)&&(0 != argc)) {
    errno = EINVAL;
    RMIX_LOG((RMIX_WARN(argv parameter is null)))
    errno = EINVAL;
    return -1;
  }
  /* Check argc parameter. */
  if ((0 == argc)&&(NULL != argv)) {
    errno = EINVAL;
    RMIX_LOG((RMIX_WARN(argc parameter is zero)))
    errno = EINVAL;
    return -1;
  }
  /* Check envv parameter. */
  if ((NULL == envv)&&(0 != envc)) {
    errno = EINVAL;
    RMIX_LOG((RMIX_WARN(envv parameter is null)))
    errno = EINVAL;
    return -1;
  }
  /* Check envc parameter. */
  if ((0 == envc)&&(NULL != envv)) {
    errno = EINVAL;
    RMIX_LOG((RMIX_WARN(envc parameter is zero)))
    errno = EINVAL;
    return -1;
  }
#endif /* DEBUG */
  /* Prepare input. */
  inary[0] = &argc;
  inary[1] =  argv;
  inary[2] = &envc;
  inary[3] =  envv;
  /* Invoke remote object method. */
  if (0 != rmix_send_p(invokeref, harness, provider, &harnessclient_interface,
                       HARNESS_RMIX_METHODS_PROCESSES_EXECUTE2_INDEX,
                       inary, 4)) {
    int errno2 = errno;
    RMIX_LOG((RMIX_WARN(unable to invoke remote object method)))
    errno = errno2;
    return -1;
  }
  return 0;
}

/*
 * Retrieves harness_processes_execute2 output (client-side method stub).
 * Returns the output of a previous asynchronous invocation that starts a new
 * process and executes a program at a remote harness instance.
 *
 * invokeref = The invocation reference (also return).
 * handle    = The process handle (return).
 * return    = 0 on success or -1 on any error with errno set appropriately.
 */
#undef  __FUNC__
#define __FUNC__ "harnessclient_processes_execute2_retrieve"
int harnessclient_processes_execute2_retrieve (rmix_invokeref_t **invokeref,
                                               unsigned int      *handle) {
  int   result;
  void *outary[2];
#ifdef DEBUG
  /* Check invokeref parameter. */
  if (NULL == invokeref) {
    errno = EINVAL;
    HARNESS_LOG((HARNESS_WARN(invokeref parameter is null)))
    errno = EINVAL;
    return -1;
  }
  /* Check *invokeref parameter. */
  if (NULL == *invokeref) {
    errno = EINVAL;
    HARNESS_LOG((HARNESS_WARN(*invokeref parameter is null)))
    errno = EINVAL;
    return -1;
  }
  /* Check handle parameter. */
  if (NULL == handle) {
    errno = EINVAL;
    HARNESS_LOG((HARNESS_WARN(handle parameter is null)))
    errno = EINVAL;
    return -1;
  }
#endif /* DEBUG */
  /* Prepare output. */
  outary[0] = &result;
  outary[1] =  handle;
  /* Invoke remote object method. */
  if (0 != rmix_retrieve(invokeref, outary, 2)) {
    int errno2 = errno;
    RMIX_LOG((RMIX_WARN(unable to invoke remote object method)))
    errno = errno2;
    return -1;
  }
  return result;
}

/*
 * Retrieves harness_processes_execute2 output (client-side method stub).
 * Returns the output of a previous asynchronous invocation that starts a new
 * process and executes a program at a remote harness instance without
 * unloading the provider plug-in.
 *
 * invokeref = The invocation reference (also return).
 * handle    = The process handle (return).
 * return    = 0 on success or -1 on any error with errno set appropriately.
 */
#undef  __FUNC__
#define __FUNC__ "harnessclient_processes_execute2_retrieve_p"
int harnessclient_processes_execute2_retrieve_p (rmix_invokeref_t **invokeref,
                                                 unsigned int      *handle) {
  int   result;
  void *outary[2];
#ifdef DEBUG
  /* Check invokeref parameter. */
  if (NULL == invokeref) {
    errno = EINVAL;
    HARNESS_LOG((HARNESS_WARN(invokeref parameter is null)))
    errno = EINVAL;
    return -1;
  }
  /* Check *invokeref parameter. */
  if (NULL == *invokeref) {
    errno = EINVAL;
    HARNESS_LOG((HARNESS_WARN(*invokeref parameter is null)))
    errno = EINVAL;
    return -1;
  }
  /* Check handle parameter. */
  if (NULL == handle) {
    errno = EINVAL;
    HARNESS_LOG((HARNESS_WARN(handle parameter is null)))
    errno = EINVAL;
    return -1;
  }
#endif /* DEBUG */
  /* Prepare output. */
  outary[0] = &result;
  outary[1] =  handle;
  /* Invoke remote object method. */ 
  if (0 != rmix_retrieve_p(invokeref, outary, 2)) {
    int errno2 = errno;
    RMIX_LOG((RMIX_WARN(unable to invoke remote object method)))
    errno = errno2;
    return -1;
  }
  return result;
}

/*
 * Invokes harness_processes_execute2 (client-side method stub).
 * Executes a new process and executes a program at multiple remote harness
 * instances. The program may be different for every remote harness instance.
 *
 * count     = The number of remote harness instances.
 * harnesses = The array of RMIX remote object references of harness instances.
 * handles   = The process handle array (also return). Process handles are
 *             unique, and 0 for unsuccessful execution attempts.
 * argcs     = The number of arguments array.
 * envcs     = The number of variables array (may be NULL).
 * argvs     = The array of arguments (argv[0] is program name) array.
 * envvs     = The array of variables array (may be NULL).
 * order     = The order flag. 0 for forward and 1 for reverse.
 * fail      = The fail flag. 0 to continue and 1 to fail on any error.
 * return    = The number of executed programs with errno set appropriately.
 */
#undef  __FUNC__
#define __FUNC__ "harnessclient_processes_sexecute2_mp"
unsigned int harnessclient_processes_sexecute2_mp
  (unsigned int              count,
   const rmix_remoteref_t  **harnesses,
   unsigned int             *handles,
   const unsigned int       *argcs,
   const unsigned int       *envcs,
   const char             ***argvs,
   const char             ***envvs,
   unsigned int              order,
   unsigned int              fail) {
  int          errno2;
  unsigned int executed, index;
#ifdef DEBUG
  /* Check harnesses parameter. */
  if ((NULL == harnesses)&&(0 != count)) {
    errno = EINVAL;
    RMIX_LOG((RMIX_WARN(harnesses parameter is null)))
    errno = EINVAL;
    return -1;
  }
  /* Check count parameter. */
  if ((0 == count)&&(NULL != harnesses)) {
    errno = EINVAL;
    RMIX_LOG((RMIX_WARN(count parameter is zero)))
    errno = EINVAL;
    return -1;
  }
  /* Check handles parameter. */
  if ((NULL == handles)&&(0 != count)) {
    errno = EINVAL;
    RMIX_LOG((RMIX_WARN(handles parameter is null)))
    errno = EINVAL;
    return -1;
  }
  /* Check count parameter. */
  if ((0 == count)&&(NULL != handles)) {
    errno = EINVAL;
    RMIX_LOG((RMIX_WARN(count parameter is zero)))
    errno = EINVAL;
    return -1;
  }
  /* Check argcs parameter. */
  if ((NULL == argcs)&&(0 != count)) {
    errno = EINVAL;
    RMIX_LOG((RMIX_WARN(argcs parameter is null)))
    errno = EINVAL;
    return -1;
  }
  /* Check count parameter. */
  if ((0 == count)&&(NULL != argcs)) {
    errno = EINVAL;
    RMIX_LOG((RMIX_WARN(count parameter is zero)))
    errno = EINVAL;
    return -1;
  }
  /* Check argvs parameter. */
  if ((NULL == argvs)&&(0 != count)) {
    errno = EINVAL;
    RMIX_LOG((RMIX_WARN(argvs parameter is null)))
    errno = EINVAL;
    return -1;
  }
  /* Check count parameter. */
  if ((0 == count)&&(NULL != argvs)) {
    errno = EINVAL;
    RMIX_LOG((RMIX_WARN(count parameter is zero)))
    errno = EINVAL;
    return -1;
  }
  /* Check envvs parameter. */
  if ((NULL == envvs)&&(NULL != envcs)) {
    errno = EINVAL;
    RMIX_LOG((RMIX_WARN(envvs parameter is null)))
    errno = EINVAL;
    return -1;
  }
  /* Check envcs parameter. */
  if ((NULL == envcs)&&(NULL != envvs)) {
    errno = EINVAL;
    RMIX_LOG((RMIX_WARN(envcs parameter is null)))
    errno = EINVAL;
    return -1;
  }
#endif /* DEBUG */
  /* Initialize errno and executed counter.*/
  errno2 = 0;
  executed = 0;
  /* Initialize all handles. */
  memset(handles, 0, count * sizeof(*handles));
  /* Check order flag. */
  if (0 == order) {
    /* Execute all programs in forward order. */
    for (index = 0; index < count; index++) {
      /* Execute program. */
      if (0 != harnessclient_processes_execute2(harnesses[index],
                                               &handles[index],
                                               argcs[index],
                                               (NULL != envcs)?
                                               envcs[index]:0,
                                               argvs[index],
                                               (NULL != envvs)?
                                               envvs[index]:NULL)) {
        errno2 = (0 == errno2)?errno:errno2;
        RMIX_LOG((RMIX_WARN(unable to execute program)))
        /* Check fail flag. */
        if (1 == fail) {
          errno = errno2;
          return executed;
        }
      } else {
        executed++;
      }
    }
  } else {
    /* Execute all programs in reverse order. */
    for (index = count; 0 < index--;) {
      /* Execute program. */
      if (0 != harnessclient_processes_execute2(harnesses[index],
                                               &handles[index],
                                               argcs[index],
                                               (NULL != envcs)?
                                               envcs[index]:0,
                                               argvs[index],
                                               (NULL != envvs)?
                                               envvs[index]:NULL)) {
        errno2 = (0 == errno2)?errno:errno2;
        RMIX_LOG((RMIX_WARN(unable to execute program)))
        /* Check fail flag. */
        if (1 == fail) {
          errno = errno2;
          return executed;
        }
      } else {
        executed++;
      }
    }
  }
  errno = errno2;
  return executed;
}

/*
 * Invokes harness_processes_execute2 (client-side method stub).
 * Executes a new process and executes a program at multiple remote harness
 * instances. The program is the same for every remote harness instance.
 *
 * count     = The number of remote harness instances.
 * harnesses = The array of RMIX remote object references of harness instances.
 * handles   = The process handle array (also return). Process handles are
 *             unique, and 0 for unsuccessful execution attempts.
 * argc      = The number of arguments.
 * envc      = The number of variables.
 * argv      = The array of arguments (argv[0] is program name).
 * envv      = The array of variables.
 * order     = The order flag. 0 for forward and 1 for reverse.
 * fail      = The fail flag. 0 to continue and 1 to fail on any error.
 * return    = The number of executed programs with errno set appropriately.
 */
#undef  __FUNC__
#define __FUNC__ "harnessclient_processes_sexecute2_sp"
unsigned int harnessclient_processes_sexecute2_sp
  (unsigned int             count,
   const rmix_remoteref_t **harnesses,
   unsigned int            *handles,
   unsigned int             argc,
   unsigned int             envc,
   const char             **argv,
   const char             **envv,
   unsigned int             order,
   unsigned int             fail) {
  int          errno2;
  unsigned int executed, index;
#ifdef DEBUG
  /* Check harnesses parameter. */
  if ((NULL == harnesses)&&(0 != count)) {
    errno = EINVAL;
    RMIX_LOG((RMIX_WARN(harnesses parameter is null)))
    errno = EINVAL;
    return -1;
  }
  /* Check count parameter. */
  if ((0 == count)&&(NULL != harnesses)) {
    errno = EINVAL;
    RMIX_LOG((RMIX_WARN(count parameter is zero)))
    errno = EINVAL;
    return -1;
  }
  /* Check handles parameter. */
  if ((NULL == handles)&&(0 != count)) {
    errno = EINVAL;
    RMIX_LOG((RMIX_WARN(handles parameter is null)))
    errno = EINVAL;
    return -1;
  }
  /* Check count parameter. */
  if ((0 == count)&&(NULL != handles)) {
    errno = EINVAL;
    RMIX_LOG((RMIX_WARN(count parameter is zero)))
    errno = EINVAL;
    return -1;
  }
  /* Check argv parameter. */
  if ((NULL == argv)&&(0 != argc)) {
    errno = EINVAL;
    RMIX_LOG((RMIX_WARN(argv parameter is null)))
    errno = EINVAL;
    return -1;
  }
  /* Check argc parameter. */
  if ((0 == argc)&&(NULL != argv)) {
    errno = EINVAL;
    RMIX_LOG((RMIX_WARN(argc parameter is zero)))
    errno = EINVAL;
    return -1;
  }
  /* Check envv parameter. */
  if ((NULL == envv)&&(0 != envc)) {
    errno = EINVAL;
    RMIX_LOG((RMIX_WARN(envv parameter is null)))
    errno = EINVAL;
    return -1;
  }
  /* Check envc parameter. */
  if ((0 == envc)&&(NULL != envv)) {
    errno = EINVAL;
    RMIX_LOG((RMIX_WARN(envc parameter is zero)))
    errno = EINVAL;
    return -1;
  }
#endif /* DEBUG */
  /* Initialize errno and executed counter.*/
  errno2 = 0;
  executed = 0;
  /* Initialize all handles. */
  memset(handles, 0, count * sizeof(*handles));
  /* Check order flag. */
  if (0 == order) {
    /* Execute all programs in forward order. */
    for (index = 0; index < count; index++) {
      /* Execute program. */
      if (0 != harnessclient_processes_execute2(harnesses[index],
                                               &handles[index],
                                               argc, envc, argv, envv)) {
        errno2 = (0 == errno2)?errno:errno2;
        RMIX_LOG((RMIX_WARN(unable to execute program)))
        /* Check fail flag. */
        if (1 == fail) {
          errno = errno2;
          return executed;
        }
      } else {
        executed++;
      }
    }
  } else {
    /* Execute all programs in reverse order. */
    for (index = count; 0 < index--;) {
      /* Execute program. */
      if (0 != harnessclient_processes_execute2(harnesses[index],
                                               &handles[index],
                                               argc, envc, argv, envv)) {
        errno2 = (0 == errno2)?errno:errno2;
        RMIX_LOG((RMIX_WARN(unable to execute program)))
        /* Check fail flag. */
        if (1 == fail) {
          errno = errno2;
          return executed;
        }
      } else {
        executed++;
      }
    }
  }
  errno = errno2;
  return executed;
}

/*
 * Invokes harness_processes_execute2 (client-side method stub).
 * Executes a new process and executes a program at multiple remote harness
 * instances in parallel using asynchonous method invocations. The program may
 * be different for every remote harness instance.
 *
 * count     = The number of remote harness instances.
 * harnesses = The array of RMIX remote object references of harness instances.
 * handles   = The process handle array (also return). Process handles are
 *             unique, and 0 for unsuccessful execution attempts.
 * argcs     = The number of arguments array.
 * envcs     = The number of variables array (may be NULL).
 * argvs     = The array of arguments (argv[0] is program name) array.
 * envvs     = The array of variables array (may be NULL).
 * return    = The number of executed programs with errno set appropriately.
 */
#undef  __FUNC__
#define __FUNC__ "harnessclient_processes_pexecute2_mp"
unsigned int harnessclient_processes_pexecute2_mp
  (unsigned int              count,
   const rmix_remoteref_t  **harnesses,
   unsigned int             *handles,
   const unsigned int       *argcs,
   const unsigned int       *envcs,
   const char             ***argvs,
   const char             ***envvs) {
  int                errno2;
  unsigned int       executed, index;
  rmix_invokeref_t **invocations;
#ifdef DEBUG
  /* Check harnesses parameter. */
  if ((NULL == harnesses)&&(0 != count)) {
    errno = EINVAL;
    RMIX_LOG((RMIX_WARN(harnesses parameter is null)))
    errno = EINVAL;
    return -1;
  }
  /* Check count parameter. */
  if ((0 == count)&&(NULL != harnesses)) {
    errno = EINVAL;
    RMIX_LOG((RMIX_WARN(count parameter is zero)))
    errno = EINVAL;
    return -1;
  }
  /* Check handles parameter. */
  if ((NULL == handles)&&(0 != count)) {
    errno = EINVAL;
    RMIX_LOG((RMIX_WARN(handles parameter is null)))
    errno = EINVAL;
    return -1;
  }
  /* Check count parameter. */
  if ((0 == count)&&(NULL != handles)) {
    errno = EINVAL;
    RMIX_LOG((RMIX_WARN(count parameter is zero)))
    errno = EINVAL;
    return -1;
  }
  /* Check argvs parameter. */
  if ((NULL == argvs)&&(0 != count)) {
    errno = EINVAL;
    RMIX_LOG((RMIX_WARN(argvs parameter is null)))
    errno = EINVAL;
    return -1;
  }
  /* Check count parameter. */
  if ((0 == count)&&(NULL != argvs)) {
    errno = EINVAL;
    RMIX_LOG((RMIX_WARN(count parameter is zero)))
    errno = EINVAL;
    return -1;
  }
  /* Check envvs parameter. */
  if ((NULL == envvs)&&(NULL != envcs)) {
    errno = EINVAL;
    RMIX_LOG((RMIX_WARN(envvs parameter is null)))
    errno = EINVAL;
    return -1;
  }
  /* Check envcs parameter. */
  if ((NULL == envcs)&&(NULL != envvs)) {
    errno = EINVAL;
    RMIX_LOG((RMIX_WARN(envcs parameter is null)))
    errno = EINVAL;
    return -1;
  }
#endif /* DEBUG */
  /* Allocate invocation reference array. */
  if (NULL == (invocations = malloc(count * sizeof(rmix_invokeref_t*)))) {
    errno2 = errno;
    RMIX_LOG((RMIX_WARN(unable to allocate invocation reference array)))
    errno = errno2;
    return 0;
  }
  /* Initialize errno and executed counter.*/
  errno2 = 0;
  executed = 0;
  /* Initialize all handles. */
  memset(handles, 0, count * sizeof(*handles));
  /* Execute all programs in forward order (send part). */
  for (index = 0; index < count; index++) {
    /* Execute program (send part). */
    if (0 != harnessclient_processes_execute2_send(&invocations[index],
                                                   harnesses[index],
                                                   argcs[index],
                                                   (NULL != envcs)?
                                                   envcs[index]:0,
                                                   argvs[index],
                                                   (NULL != envvs)?
                                                   envvs[index]:NULL)) {
      errno2 = (0 == errno2)?errno:errno2;
      RMIX_LOG((RMIX_WARN(unable to execute program (send part))))
      invocations[index] = NULL;
    }
  }
  /* Execute all programs in forward order (retrieve part). */
  for (index = 0; index < count; index++) {
    /* Check for successful send. */
    if (NULL != invocations[index]) {
      /* Execute program (retrieve part). */
      if (0 != harnessclient_processes_execute2_retrieve(&invocations[index],
                                                         &handles[index])) {
        errno2 = (0 == errno2)?errno:errno2;
        RMIX_LOG((RMIX_WARN(unable to execute program (retrieve part))))
      } else {
        executed++;
      }
    }
  }
  /* Deallocate invocation reference array. */
  free(invocations);
  errno = errno2;
  return executed;
}

/*
 * Invokes harness_processes_execute2 (client-side method stub).
 * Executes a new process and executes a program at multiple remote harness
 * instances in parallel using asynchonous method invocations. The program is
 * the same for every remote harness instance.
 *
 * count     = The number of remote harness instances.
 * harnesses = The array of RMIX remote object references of harness instances.
 * handles   = The process handle array (also return). Process handles are
 *             unique, and 0 for unsuccessful execution attempts.
 * argc      = The number of arguments.
 * envc      = The number of variables.
 * argv      = The array of arguments (argv[0] is program name).
 * envv      = The array of variables.
 * return    = The number of executed programs with errno set appropriately.
 */
#undef  __FUNC__
#define __FUNC__ "harnessclient_processes_pexecute2_sp"
unsigned int harnessclient_processes_pexecute2_sp
  (unsigned int             count,
   const rmix_remoteref_t **harnesses,
   unsigned int            *handles,
   unsigned int             argc,
   unsigned int             envc,
   const char             **argv,
   const char             **envv) {
  int                errno2;
  unsigned int       executed, index;
  rmix_invokeref_t **invocations;
#ifdef DEBUG
  /* Check harnesses parameter. */
  if ((NULL == harnesses)&&(0 != count)) {
    errno = EINVAL;
    RMIX_LOG((RMIX_WARN(harnesses parameter is null)))
    errno = EINVAL;
    return -1;
  }
  /* Check count parameter. */
  if ((0 == count)&&(NULL != harnesses)) {
    errno = EINVAL;
    RMIX_LOG((RMIX_WARN(count parameter is zero)))
    errno = EINVAL;
    return -1;
  }
  /* Check handles parameter. */
  if ((NULL == handles)&&(0 != count)) {
    errno = EINVAL;
    RMIX_LOG((RMIX_WARN(handles parameter is null)))
    errno = EINVAL;
    return -1;
  }
  /* Check count parameter. */
  if ((0 == count)&&(NULL != handles)) {
    errno = EINVAL;
    RMIX_LOG((RMIX_WARN(count parameter is zero)))
    errno = EINVAL;
    return -1;
  }
  /* Check argv parameter. */
  if ((NULL == argv)&&(0 != argc)) {
    errno = EINVAL;
    RMIX_LOG((RMIX_WARN(argv parameter is null)))
    errno = EINVAL;
    return -1;
  }
  /* Check argc parameter. */
  if ((0 == argc)&&(NULL != argv)) {
    errno = EINVAL;
    RMIX_LOG((RMIX_WARN(argc parameter is zero)))
    errno = EINVAL;
    return -1;
  }
  /* Check envv parameter. */
  if ((NULL == envv)&&(0 != envc)) {
    errno = EINVAL;
    RMIX_LOG((RMIX_WARN(envv parameter is null)))
    errno = EINVAL;
    return -1;
  }
  /* Check envc parameter. */
  if ((0 == envc)&&(NULL != envv)) {
    errno = EINVAL;
    RMIX_LOG((RMIX_WARN(envc parameter is zero)))
    errno = EINVAL;
    return -1;
  }
#endif /* DEBUG */
  /* Allocate invocation reference array. */
  if (NULL == (invocations = malloc(count * sizeof(rmix_invokeref_t*)))) {
    errno2 = errno;
    RMIX_LOG((RMIX_WARN(unable to allocate invocation reference array)))
    errno = errno2;
    return 0;
  }
  /* Initialize errno and executed counter.*/
  errno2 = 0;
  executed = 0;
  /* Initialize all handles. */
  memset(handles, 0, count * sizeof(*handles));
  /* Execute all programs in forward order (send part). */
  for (index = 0; index < count; index++) {
    /* Execute program (send part). */
    if (0 != harnessclient_processes_execute2_send(&invocations[index],
                                                   harnesses[index],
                                                   argc, envc, argv, envv)) {
      errno2 = (0 == errno2)?errno:errno2;
      RMIX_LOG((RMIX_WARN(unable to execute program (send part))))
      invocations[index] = NULL;
    }
  }
  /* Execute all programs in forward order (retrieve part). */
  for (index = 0; index < count; index++) {
    /* Check for successful send. */
    if (NULL != invocations[index]) {
      /* Execute program (retrieve part). */
      if (0 != harnessclient_processes_execute2_retrieve(&invocations[index],
                                                         &handles[index])) {
        errno2 = (0 == errno2)?errno:errno2;
        RMIX_LOG((RMIX_WARN(unable to execute program (retrieve part))))
      } else {
        executed++;
      }
    }
  }
  /* Deallocate invocation reference array. */
  free(invocations);
  errno = errno2;
  return executed;
}

/*
 * Invokes harnessclient_processes_read (client-side method stub).
 * Reads the standard output of a process into a buffer (non-/blocking) at a
 * remote harness instance.
 *
 * harness = The RMIX remote object reference of harness.
 * handle  = The process handle.
 * length  = The maximum buffer length to read.
 * buffer  = The buffer to read into.
 * block   = The (non-)blocking flag (0 = non-blocking, 1 = blocking).
 * return  = number of bytes read on success, -2 on closed pipe or -1 on any
 *           other error with errno set appropriately.
 */
#undef  __FUNC__
#define __FUNC__ "harnessclient_processes_read"
int harnessclient_processes_read (const rmix_remoteref_t *harness,
                                  unsigned int            handle,
                                  unsigned int            length,
                                  void                   *buffer,
                                  unsigned int            block) {
  int           result;
  void         *outary[3];
  const void   *inary[3];
  unsigned int  length2;
#ifdef DEBUG
  /* Check harness parameter. */
  if (NULL == harness) {
    errno = EINVAL;
    HARNESS_LOG((HARNESS_WARN(harness parameter is null)))
    errno = EINVAL;
    return -1;
  }
  /* Check buffer parameter. */
  if ((NULL == buffer)&&(0 != length)) {
    errno = EINVAL;
    RMIX_LOG((RMIX_WARN(buffer parameter is null)))
    errno = EINVAL;
    return -1;
  }
  /* Check length parameter. */
  if ((0 == length)&&(NULL != buffer)) {
    errno = EINVAL;
    RMIX_LOG((RMIX_WARN(length parameter is zero)))
    errno = EINVAL;
    return -1;
  }
#endif /* DEBUG */
  /* Prepare input. */
  inary[0] = &handle;
  inary[1] = &length;
  inary[2] = &block;
  /* Prepare output. */
  outary[0] = &result;
  outary[1] = &length2;
  outary[2] = NULL;
  /* Invoke remote object method. */
  if (0 != rmix_invoke(harness, &harnessclient_interface,
                       HARNESS_RMIX_METHODS_PROCESSES_READ_INDEX,
                       outary, 3, inary, 3)) {
    int errno2 = errno;
    RMIX_LOG((RMIX_WARN(unable to invoke remote object method)))
    errno = errno2;
    return -1;
  }
  /* Prepare output. */
  if (NULL != outary[2]) {
    memcpy(buffer, outary[2], (length2 < length)?length2:length);
    free(outary[2]);
  }
  return result;
}

/*
 * Invokes harnessclient_processes_read (client-side method stub).
 * Reads the standard output of a process into a buffer (non-/blocking) at a
 * remote harness instance.
 *
 * harness  = The RMIX remote object reference of harness.
 * provider = The preloaded RMIX provider plug-in handle.
 * handle   = The process handle.
 * length   = The maximum buffer length to read.
 * buffer   = The buffer to read into.
 * block    = The (non-)blocking flag (0 = non-blocking, 1 = blocking).
 * return   = number of bytes read on success, -2 on closed pipe or -1 on any
 *            other error with errno set appropriately.
 */
#undef  __FUNC__
#define __FUNC__ "harnessclient_processes_read_p"
int harnessclient_processes_read_p (const rmix_remoteref_t *harness,
                                    unsigned int            provider,
                                    unsigned int            handle,
                                    unsigned int            length,
                                    void                   *buffer,
                                    unsigned int            block) {
  int           result;
  void         *outary[3];
  const void   *inary[3];
  unsigned int  length2;
#ifdef DEBUG
  /* Check harness parameter. */
  if (NULL == harness) {
    errno = EINVAL;
    HARNESS_LOG((HARNESS_WARN(harness parameter is null)))
    errno = EINVAL;
    return -1;
  }
  /* Check buffer parameter. */
  if ((NULL == buffer)&&(0 != length)) {
    errno = EINVAL;
    RMIX_LOG((RMIX_WARN(buffer parameter is null)))
    errno = EINVAL;
    return -1;
  }
  /* Check length parameter. */
  if ((0 == length)&&(NULL != buffer)) {
    errno = EINVAL;
    RMIX_LOG((RMIX_WARN(length parameter is zero)))
    errno = EINVAL;
    return -1;
  }
#endif /* DEBUG */
  /* Prepare input. */
  inary[0] = &handle;
  inary[1] = &length;
  inary[2] = &block;
  /* Prepare output. */
  outary[0] = &result;
  outary[1] = &length2;
  outary[2] = NULL;
  /* Invoke remote object method. */
  if (0 != rmix_invoke_p(harness, provider, &harnessclient_interface,
                         HARNESS_RMIX_METHODS_PROCESSES_READ_INDEX,
                         outary, 3, inary, 3)) {
    int errno2 = errno;
    RMIX_LOG((RMIX_WARN(unable to invoke remote object method)))
    errno = errno2;
    return -1;
  }
  /* Prepare output. */
  if (NULL != outary[2]) {
    memcpy(buffer, outary[2], (length2 < length)?length2:length);
    free(outary[2]);
  }
  return result;
}

/*
 * Invokes harness_processes_read (client-side method stub).
 * Reads the standard output of a process into a buffer (non-/blocking) at a
 * remote harness instance using an asynchronous invocation.
 *
 * invokeref = The invocation reference (return).
 * harness   = The RMIX remote object reference of harness.
 * handle    = The process handle.
 * length    = The maximum buffer length to read.
 * block     = The (non-)blocking flag (0 = non-blocking, 1 = blocking).
 * return    = 0 on success or -1 on any error with errno set appropriately.
 */
#undef  __FUNC__
#define __FUNC__ "harnessclient_processes_read_send"
int harnessclient_processes_read_send (rmix_invokeref_t       **invokeref,
                                       const rmix_remoteref_t  *harness,
                                       unsigned int             handle,
                                       unsigned int             length,
                                       unsigned int             block) {
  const void *inary[3];
#ifdef DEBUG
  /* Check invokeref parameter. */
  if (NULL == invokeref) {
    errno = EINVAL;
    HARNESS_LOG((HARNESS_WARN(invokeref parameter is null)))
    errno = EINVAL;
    return -1;
  }
  /* Check harness parameter. */
  if (NULL == harness) {
    errno = EINVAL;
    HARNESS_LOG((HARNESS_WARN(harness parameter is null)))
    errno = EINVAL;
    return -1;
  }
#endif /* DEBUG */
  /* Prepare input. */
  inary[0] = &handle;
  inary[1] = &length;
  inary[2] = &block;
  /* Invoke remote object method. */
  if (0 != rmix_send(invokeref, harness, &harnessclient_interface,
                     HARNESS_RMIX_METHODS_PROCESSES_READ_INDEX,
                     inary, 3)) {
    int errno2 = errno;
    RMIX_LOG((RMIX_WARN(unable to invoke remote object method)))
    errno = errno2;
    return -1;
  }
  return 0;
}

/*
 * Invokes harness_processes_read (client-side method stub).
 * Reads the standard output of a process into a buffer (non-/blocking) at a
 * remote harness instance using an asynchronous invocation without loading the
 * provider plug-in.
 *
 * invokeref = The invocation reference (return).
 * harness   = The RMIX remote object reference of harness.
 * provider  = The preloaded RMIX provider plug-in handle.
 * handle    = The process handle.
 * length    = The maximum buffer length to read.
 * block     = The (non-)blocking flag (0 = non-blocking, 1 = blocking).
 * return    = 0 on success or -1 on any error with errno set appropriately.
 */
#undef  __FUNC__
#define __FUNC__ "harnessclient_processes_read_send_p"
int harnessclient_processes_read_send_p (rmix_invokeref_t       **invokeref,
                                         const rmix_remoteref_t  *harness,
                                         unsigned int             provider,
                                         unsigned int             handle,
                                         unsigned int             length,
                                         unsigned int             block) {
  const void *inary[3];
#ifdef DEBUG
  /* Check invokeref parameter. */
  if (NULL == invokeref) {
    errno = EINVAL;
    HARNESS_LOG((HARNESS_WARN(invokeref parameter is null)))
    errno = EINVAL;
    return -1;
  }
  /* Check harness parameter. */
  if (NULL == harness) {
    errno = EINVAL;
    HARNESS_LOG((HARNESS_WARN(harness parameter is null)))
    errno = EINVAL;
    return -1;
  }
#endif /* DEBUG */
  /* Prepare input. */
  inary[0] = &handle;
  inary[1] = &length;
  inary[2] = &block;
  /* Invoke remote object method. */
  if (0 != rmix_send_p(invokeref, harness, provider, &harnessclient_interface,
                       HARNESS_RMIX_METHODS_PROCESSES_READ_INDEX,
                       inary, 3)) {
    int errno2 = errno;
    RMIX_LOG((RMIX_WARN(unable to invoke remote object method)))
    errno = errno2;
    return -1;
  }
  return 0;
}

/*
 * Retrieves harness_processes_read output (client-side method stub).
 * Returns the output of a previous asynchronous invocation that reads the
 * standard output of a process into a buffer (non-/blocking) at a remote
 * harness instance.
 *
 * invokeref = The invocation reference (also return).
 * length    = The maximum buffer length to read.
 * buffer    = The buffer to read into.
 * return    = 0 on success or -1 on any error with errno set appropriately.
 */
#undef  __FUNC__
#define __FUNC__ "harnessclient_processes_read_retrieve"
int harnessclient_processes_read_retrieve (rmix_invokeref_t **invokeref,
                                           unsigned int       length,
                                           void              *buffer) {
  int           result;
  void         *outary[3];
  unsigned int  length2;
#ifdef DEBUG
  /* Check invokeref parameter. */
  if (NULL == invokeref) {
    errno = EINVAL;
    HARNESS_LOG((HARNESS_WARN(invokeref parameter is null)))
    errno = EINVAL;
    return -1;
  }
  /* Check *invokeref parameter. */
  if (NULL == *invokeref) {
    errno = EINVAL;
    HARNESS_LOG((HARNESS_WARN(*invokeref parameter is null)))
    errno = EINVAL;
    return -1;
  }
  /* Check buffer parameter. */
  if ((NULL == buffer)&&(0 != length)) {
    errno = EINVAL;
    RMIX_LOG((RMIX_WARN(buffer parameter is null)))
    errno = EINVAL;
    return -1;
  }
  /* Check length parameter. */
  if ((0 == length)&&(NULL != buffer)) {
    errno = EINVAL;
    RMIX_LOG((RMIX_WARN(length parameter is zero)))
    errno = EINVAL;
    return -1;
  }
#endif /* DEBUG */
  /* Prepare output. */
  outary[0] = &result;
  outary[1] = &length2;
  outary[2] = NULL;
  /* Invoke remote object method. */
  if (0 != rmix_retrieve(invokeref, outary, 3)) {
    int errno2 = errno;
    RMIX_LOG((RMIX_WARN(unable to invoke remote object method)))
    errno = errno2;
    return -1;
  }
  /* Prepare output. */
  if (NULL != outary[2]) {
    memcpy(buffer, outary[2], (length2 < length)?length2:length);
    free(outary[2]);
  }
  return result;
}

/*
 * Retrieves harness_processes_read output (client-side method stub).
 * Returns the output of a previous asynchronous invocation that reads the
 * standard output of a process into a buffer (non-/blocking) at a remote
 * harness instance without unloading the provider plug-in.
 *
 * invokeref = The invocation reference (also return).
 * length    = The maximum buffer length to read.
 * buffer    = The buffer to read into.
 * return    = 0 on success or -1 on any error with errno set appropriately.
 */
#undef  __FUNC__
#define __FUNC__ "harnessclient_processes_read_retrieve_p"
int harnessclient_processes_read_retrieve_p (rmix_invokeref_t **invokeref,
                                             unsigned int       length,
                                             void              *buffer) {
  int           result;
  void         *outary[3];
  unsigned int  length2;
#ifdef DEBUG
  /* Check invokeref parameter. */
  if (NULL == invokeref) {
    errno = EINVAL;
    HARNESS_LOG((HARNESS_WARN(invokeref parameter is null)))
    errno = EINVAL;
    return -1;
  }
  /* Check *invokeref parameter. */
  if (NULL == *invokeref) {
    errno = EINVAL;
    HARNESS_LOG((HARNESS_WARN(*invokeref parameter is null)))
    errno = EINVAL;
    return -1;
  }
  /* Check buffer parameter. */
  if ((NULL == buffer)&&(0 != length)) {
    errno = EINVAL;
    RMIX_LOG((RMIX_WARN(buffer parameter is null)))
    errno = EINVAL;
    return -1;
  }
  /* Check length parameter. */
  if ((0 == length)&&(NULL != buffer)) {
    errno = EINVAL;
    RMIX_LOG((RMIX_WARN(length parameter is zero)))
    errno = EINVAL;
    return -1;
  }
#endif /* DEBUG */
  /* Prepare output. */
  outary[0] = &result;
  outary[1] = &length2;
  outary[2] = NULL;
  /* Invoke remote object method. */ 
  if (0 != rmix_retrieve_p(invokeref, outary, 3)) {
    int errno2 = errno;
    RMIX_LOG((RMIX_WARN(unable to invoke remote object method)))
    errno = errno2;
    return -1;
  }
  /* Prepare output. */
  if (NULL != outary[2]) {
    memcpy(buffer, outary[2], (length2 < length)?length2:length);
    free(outary[2]);
  }
  return result;
}

/*
 * Invokes harnessclient_processes_write (client-side method stub).
 * Writes a buffer into the standard input of a process (blocking) at a remote
 * harness instance.
 *
 * harness = The RMIX remote object reference of harness.
 * handle  = The process handle.
 * length  = The length to write.
 * buffer  = The buffer to write.
 * return  = 0 on success, -2 on closed pipe or -1 on any
 *          other error with errno set appropriately.
 */
#undef  __FUNC__
#define __FUNC__ "harnessclient_processes_write"
int harnessclient_processes_write (const rmix_remoteref_t *harness,
                                   unsigned int            handle,
                                   unsigned int            length,
                                   const void             *buffer) {
  int         result;
  void       *outary[1];
  const void *inary[3];
#ifdef DEBUG
  /* Check harness parameter. */
  if (NULL == harness) {
    errno = EINVAL;
    HARNESS_LOG((HARNESS_WARN(harness parameter is null)))
    errno = EINVAL;
    return -1;
  }
  /* Check buffer parameter. */
  if ((NULL == buffer)&&(0 != length)) {
    errno = EINVAL;
    RMIX_LOG((RMIX_WARN(buffer parameter is null)))
    errno = EINVAL;
    return -1;
  }
  /* Check length parameter. */
  if ((0 == length)&&(NULL != buffer)) {
    errno = EINVAL;
    RMIX_LOG((RMIX_WARN(length parameter is zero)))
    errno = EINVAL;
    return -1;
  }
#endif /* DEBUG */
  /* Prepare input. */
  inary[0] = &handle;
  inary[1] = &length;
  inary[2] =  buffer;
  /* Prepare output. */
  outary[0] = &result;
  /* Invoke remote object method. */
  if (0 != rmix_invoke(harness, &harnessclient_interface,
                       HARNESS_RMIX_METHODS_PROCESSES_WRITE_INDEX,
                       outary, 1, inary, 3)) {
    int errno2 = errno;
    RMIX_LOG((RMIX_WARN(unable to invoke remote object method)))
    errno = errno2;
    return -1;
  }
  return result;
}

/*
 * Invokes harnessclient_processes_write (client-side method stub).
 * Writes a buffer into the standard input of a process (blocking) at a remote
 * harness instance.
 *
 * harness  = The RMIX remote object reference of harness.
 * provider = The preloaded RMIX provider plug-in handle.
 * handle   = The process handle.
 * length   = The length to write.
 * buffer   = The buffer to write.
 * return   = 0 on success, -2 on closed pipe or -1 on any
 *            other error with errno set appropriately.
 */
#undef  __FUNC__
#define __FUNC__ "harnessclient_processes_write_p"
int harnessclient_processes_write_p (const rmix_remoteref_t *harness,
                                     unsigned int            provider,
                                     unsigned int            handle,
                                     unsigned int            length,
                                     const void             *buffer) {
  int         result;
  void       *outary[1];
  const void *inary[3];
#ifdef DEBUG
  /* Check harness parameter. */
  if (NULL == harness) {
    errno = EINVAL;
    HARNESS_LOG((HARNESS_WARN(harness parameter is null)))
    errno = EINVAL;
    return -1;
  }
  /* Check buffer parameter. */
  if ((NULL == buffer)&&(0 != length)) {
    errno = EINVAL;
    RMIX_LOG((RMIX_WARN(buffer parameter is null)))
    errno = EINVAL;
    return -1;
  }
  /* Check length parameter. */
  if ((0 == length)&&(NULL != buffer)) {
    errno = EINVAL;
    RMIX_LOG((RMIX_WARN(length parameter is zero)))
    errno = EINVAL;
    return -1;
  }
#endif /* DEBUG */
  /* Prepare input. */
  inary[0] = &handle;
  inary[1] = &length;
  inary[2] =  buffer;
  /* Prepare output. */
  outary[0] = &result;
  /* Invoke remote object method. */
  if (0 != rmix_invoke_p(harness, provider, &harnessclient_interface,
                         HARNESS_RMIX_METHODS_PROCESSES_WRITE_INDEX,
                         outary, 1, inary, 3)) {
    int errno2 = errno;
    RMIX_LOG((RMIX_WARN(unable to invoke remote object method)))
    errno = errno2;
    return -1;
  }
  return result;
}

/*
 * Invokes harness_processes_write (client-side method stub).
 * Writes a buffer into the standard input of a process (blocking) at a remote
 * harness instance using a oneway invocation.
 *
 * harness = The RMIX remote object reference of harness.
 * handle  = The process handle.
 * length  = The length to write.
 * buffer  = The buffer to write.
 * return  = 0 on success or -1 on any error with errno set appropriately.
 */
#undef  __FUNC__
#define __FUNC__ "harnessclient_processes_write_oneway"
int harnessclient_processes_write_oneway (const rmix_remoteref_t *harness,
                                          unsigned int            handle,
                                          unsigned int            length,
                                          const void             *buffer) {
  const void *inary[3];
#ifdef DEBUG
  /* Check harness parameter. */
  if (NULL == harness) {
    errno = EINVAL;
    HARNESS_LOG((HARNESS_WARN(harness parameter is null)))
    errno = EINVAL;
    return -1;
  }
  /* Check buffer parameter. */
  if ((NULL == buffer)&&(0 != length)) {
    errno = EINVAL;
    RMIX_LOG((RMIX_WARN(buffer parameter is null)))
    errno = EINVAL;
    return -1;
  }
  /* Check length parameter. */
  if ((0 == length)&&(NULL != buffer)) {
    errno = EINVAL;
    RMIX_LOG((RMIX_WARN(length parameter is zero)))
    errno = EINVAL;
    return -1;
  }
#endif /* DEBUG */
  /* Prepare input. */
  inary[0] = &handle;
  inary[1] = &length;
  inary[2] =  buffer;
  /* Invoke remote object method. */
  if (0 != rmix_oneway(harness, &harnessclient_interface,
                       HARNESS_RMIX_METHODS_PROCESSES_WRITE_INDEX,
                       inary, 3)) {
    int errno2 = errno;
    RMIX_LOG((RMIX_WARN(unable to invoke remote object method)))
    errno = errno2;
    return -1;
  }
  return 0;
}

/*
 * Invokes harness_processes_write (client-side method stub).
 * Writes a buffer into the standard input of a process (blocking) at a remote
 * harness instance using a oneway invocation without loading the
 * provider plug-in.
 *
 * harness  = The RMIX remote object reference of harness.
 * provider = The preloaded RMIX provider plug-in handle.
 * handle   = The process handle.
 * length   = The length to write.
 * buffer   = The buffer to write.
 * return   = 0 on success or -1 on any error with errno set appropriately.
 */
#undef  __FUNC__
#define __FUNC__ "harnessclient_processes_write_oneway_p"
int harnessclient_processes_write_oneway_p (const rmix_remoteref_t *harness,
                                            unsigned int            provider,
                                            unsigned int            handle,
                                            unsigned int            length,
                                            const void             *buffer) {
  const void *inary[3];
#ifdef DEBUG
  /* Check harness parameter. */
  if (NULL == harness) {
    errno = EINVAL;
    HARNESS_LOG((HARNESS_WARN(harness parameter is null)))
    errno = EINVAL;
    return -1;
  }
  /* Check buffer parameter. */
  if ((NULL == buffer)&&(0 != length)) {
    errno = EINVAL;
    RMIX_LOG((RMIX_WARN(buffer parameter is null)))
    errno = EINVAL;
    return -1;
  }
  /* Check length parameter. */
  if ((0 == length)&&(NULL != buffer)) {
    errno = EINVAL;
    RMIX_LOG((RMIX_WARN(length parameter is zero)))
    errno = EINVAL;
    return -1;
  }
#endif /* DEBUG */
  /* Prepare input. */
  inary[0] = &handle;
  inary[1] = &length;
  inary[2] =  buffer;
  /* Invoke remote object method. */
  if (0 != rmix_oneway_p(harness, provider, &harnessclient_interface,
                         HARNESS_RMIX_METHODS_PROCESSES_WRITE_INDEX,
                         inary, 3)) {
    int errno2 = errno;
    RMIX_LOG((RMIX_WARN(unable to invoke remote object method)))
    errno = errno2;
    return -1;
  }
  return 0;
}

/*
 * Invokes harness_processes_write (client-side method stub).
 * Writes a buffer into the standard input of a process (blocking) at a remote
 * harness instance using an asynchronous invocation.
 *
 * invokeref = The invocation reference (return).
 * harness   = The RMIX remote object reference of harness.
 * handle    = The process handle.
 * length    = The length to write.
 * buffer    = The buffer to write.
 * return    = 0 on success or -1 on any error with errno set appropriately.
 */
#undef  __FUNC__
#define __FUNC__ "harnessclient_processes_write_send"
int harnessclient_processes_write_send (rmix_invokeref_t       **invokeref,
                                        const rmix_remoteref_t  *harness,
                                        unsigned int             handle,
                                        unsigned int             length,
                                        const void              *buffer) {
  const void *inary[3];
#ifdef DEBUG
  /* Check invokeref parameter. */
  if (NULL == invokeref) {
    errno = EINVAL;
    HARNESS_LOG((HARNESS_WARN(invokeref parameter is null)))
    errno = EINVAL;
    return -1;
  }
  /* Check harness parameter. */
  if (NULL == harness) {
    errno = EINVAL;
    HARNESS_LOG((HARNESS_WARN(harness parameter is null)))
    errno = EINVAL;
    return -1;
  }
  /* Check buffer parameter. */
  if ((NULL == buffer)&&(0 != length)) {
    errno = EINVAL;
    RMIX_LOG((RMIX_WARN(buffer parameter is null)))
    errno = EINVAL;
    return -1;
  }
  /* Check length parameter. */
  if ((0 == length)&&(NULL != buffer)) {
    errno = EINVAL;
    RMIX_LOG((RMIX_WARN(length parameter is zero)))
    errno = EINVAL;
    return -1;
  }
#endif /* DEBUG */
  /* Prepare input. */
  inary[0] = &handle;
  inary[1] = &length;
  inary[2] =  buffer;
  /* Invoke remote object method. */
  if (0 != rmix_send(invokeref, harness, &harnessclient_interface,
                     HARNESS_RMIX_METHODS_PROCESSES_WRITE_INDEX,
                     inary, 3)) {
    int errno2 = errno;
    RMIX_LOG((RMIX_WARN(unable to invoke remote object method)))
    errno = errno2;
    return -1;
  }
  return 0;
}

/*
 * Invokes harness_processes_write (client-side method stub).
 * Writes a buffer into the standard input of a process (blocking) at a remote
 * harness instance using an asynchronous invocation without loading the
 * provider plug-in.
 *
 * invokeref = The invocation reference (return).
 * harness   = The RMIX remote object reference of harness.
 * provider  = The preloaded RMIX provider plug-in handle.
 * handle    = The process handle.
 * length    = The length to write.
 * buffer    = The buffer to write.
 * return    = 0 on success or -1 on any error with errno set appropriately.
 */
#undef  __FUNC__
#define __FUNC__ "harnessclient_processes_write_send_p"
int harnessclient_processes_write_send_p (rmix_invokeref_t       **invokeref,
                                          const rmix_remoteref_t  *harness,
                                          unsigned int             provider,
                                          unsigned int             handle,
                                          unsigned int             length,
                                          const void              *buffer) {
  const void *inary[3];
#ifdef DEBUG
  /* Check invokeref parameter. */
  if (NULL == invokeref) {
    errno = EINVAL;
    HARNESS_LOG((HARNESS_WARN(invokeref parameter is null)))
    errno = EINVAL;
    return -1;
  }
  /* Check harness parameter. */
  if (NULL == harness) {
    errno = EINVAL;
    HARNESS_LOG((HARNESS_WARN(harness parameter is null)))
    errno = EINVAL;
    return -1;
  }
  /* Check buffer parameter. */
  if ((NULL == buffer)&&(0 != length)) {
    errno = EINVAL;
    RMIX_LOG((RMIX_WARN(buffer parameter is null)))
    errno = EINVAL;
    return -1;
  }
  /* Check length parameter. */
  if ((0 == length)&&(NULL != buffer)) {
    errno = EINVAL;
    RMIX_LOG((RMIX_WARN(length parameter is zero)))
    errno = EINVAL;
    return -1;
  }
#endif /* DEBUG */
  /* Prepare input. */
  inary[0] = &handle;
  inary[1] = &length;
  inary[2] =  buffer;
  /* Invoke remote object method. */
  if (0 != rmix_send_p(invokeref, harness, provider, &harnessclient_interface,
                       HARNESS_RMIX_METHODS_PROCESSES_WRITE_INDEX,
                       inary, 3)) {
    int errno2 = errno;
    RMIX_LOG((RMIX_WARN(unable to invoke remote object method)))
    errno = errno2;
    return -1;
  }
  return 0;
}

/*
 * Retrieves harness_processes_write output (client-side method stub).
 * Returns the output of a previous asynchronous invocation that writes a buffer
 * into the standard input of a process (blocking) at a remote harness instance.
 *
 * invokeref = The invocation reference (also return).
 * return    = 0 on success or -1 on any error with errno set appropriately.
 */
#undef  __FUNC__
#define __FUNC__ "harnessclient_processes_write_retrieve"
int harnessclient_processes_write_retrieve (rmix_invokeref_t **invokeref) {
  int   result;
  void *outary[1];
#ifdef DEBUG
  /* Check invokeref parameter. */
  if (NULL == invokeref) {
    errno = EINVAL;
    HARNESS_LOG((HARNESS_WARN(invokeref parameter is null)))
    errno = EINVAL;
    return -1;
  }
  /* Check *invokeref parameter. */
  if (NULL == *invokeref) {
    errno = EINVAL;
    HARNESS_LOG((HARNESS_WARN(*invokeref parameter is null)))
    errno = EINVAL;
    return -1;
  }
#endif /* DEBUG */
  /* Prepare output. */
  outary[0] = &result;
  /* Invoke remote object method. */
  if (0 != rmix_retrieve(invokeref, outary, 1)) {
    int errno2 = errno;
    RMIX_LOG((RMIX_WARN(unable to invoke remote object method)))
    errno = errno2;
    return -1;
  }
  return result;
}

/*
 * Retrieves harness_processes_write output (client-side method stub).
 * Returns the output of a previous asynchronous invocation that writes a buffer
 * into the standard input of a process (blocking) at a remote harness instance
 * without unloading the provider plug-in.
 *
 * invokeref = The invocation reference (also return).
 * return    = 0 on success or -1 on any error with errno set appropriately.
 */
#undef  __FUNC__
#define __FUNC__ "harnessclient_processes_write_retrieve_p"
int harnessclient_processes_write_retrieve_p (rmix_invokeref_t **invokeref) {
  int   result;
  void *outary[1];
#ifdef DEBUG
  /* Check invokeref parameter. */
  if (NULL == invokeref) {
    errno = EINVAL;
    HARNESS_LOG((HARNESS_WARN(invokeref parameter is null)))
    errno = EINVAL;
    return -1;
  }
  /* Check *invokeref parameter. */
  if (NULL == *invokeref) {
    errno = EINVAL;
    HARNESS_LOG((HARNESS_WARN(*invokeref parameter is null)))
    errno = EINVAL;
    return -1;
  }
#endif /* DEBUG */
  /* Prepare output. */
  outary[0] = &result;
  /* Invoke remote object method. */ 
  if (0 != rmix_retrieve_p(invokeref, outary, 1)) {
    int errno2 = errno;
    RMIX_LOG((RMIX_WARN(unable to invoke remote object method)))
    errno = errno2;
    return -1;
  }
  return result;
}

/*
 * Invokes harnessclient_processes_signal (client-side method stub).
 * Sends a signal to a process at a remote harness instance.
 *
 * harness = The RMIX remote object reference of harness.
 * handle  = The process handle.
 * signum  = The signal to send.
 * return  = 0 on success or -1 on any error with errno set appropriately.
 */
#undef  __FUNC__
#define __FUNC__ "harnessclient_processes_signal"
int harnessclient_processes_signal (const rmix_remoteref_t *harness,
                                    unsigned int            handle,
                                    int                     signum) {
  int         result;
  void       *outary[1];
  const void *inary[2];
#ifdef DEBUG
  /* Check harness parameter. */
  if (NULL == harness) {
    errno = EINVAL;
    HARNESS_LOG((HARNESS_WARN(harness parameter is null)))
    errno = EINVAL;
    return -1;
  }
#endif /* DEBUG */
  /* Prepare input. */
  inary[0] = &handle;
  inary[1] = &signum;
  /* Prepare output. */
  outary[0] = &result;
  /* Invoke remote object method. */
  if (0 != rmix_invoke(harness, &harnessclient_interface,
                       HARNESS_RMIX_METHODS_PROCESSES_SIGNAL_INDEX,
                       outary, 1, inary, 2)) {
    int errno2 = errno;
    RMIX_LOG((RMIX_WARN(unable to invoke remote object method)))
    errno = errno2;
    return -1;
  }
  return result;
}

/*
 * Invokes harnessclient_processes_signal (client-side method stub).
 * Sends a signal to a process at a remote harness instance.
 *
 * harness  = The RMIX remote object reference of harness.
 * provider = The preloaded RMIX provider plug-in handle.
 * handle   = The process handle.
 * signum   = The signal to send.
 * return   = 0 on success or -1 on any error with errno set appropriately.
 */
#undef  __FUNC__
#define __FUNC__ "harnessclient_processes_signal_p"
int harnessclient_processes_signal_p (const rmix_remoteref_t *harness,
                                      unsigned int            provider,
                                      unsigned int            handle,
                                      int                     signum) {
  int         result;
  void       *outary[1];
  const void *inary[2];
#ifdef DEBUG
  /* Check harness parameter. */
  if (NULL == harness) {
    errno = EINVAL;
    HARNESS_LOG((HARNESS_WARN(harness parameter is null)))
    errno = EINVAL;
    return -1;
  }
#endif /* DEBUG */
  /* Prepare input. */
  inary[0] = &handle;
  inary[1] = &signum;
  /* Prepare output. */
  outary[0] = &result;
  /* Invoke remote object method. */
  if (0 != rmix_invoke_p(harness, provider, &harnessclient_interface,
                         HARNESS_RMIX_METHODS_PROCESSES_SIGNAL_INDEX,
                         outary, 1, inary, 2)) {
    int errno2 = errno;
    RMIX_LOG((RMIX_WARN(unable to invoke remote object method)))
    errno = errno2;
    return -1;
  }
  return result;
}

/*
 * Invokes harness_processes_signal (client-side method stub).
 * Sends a signal to a process at a remote harness instance using a one-way
 * invocation.
 *
 * harness = The RMIX remote object reference of harness.
 * handle  = The process handle.
 * signum  = The signal to send.
 * return  = 0 on success or -1 on any error with errno set appropriately.
 */
#undef  __FUNC__
#define __FUNC__ "harnessclient_processes_signal_oneway"
int harnessclient_processes_signal_oneway (const rmix_remoteref_t *harness,
                                           unsigned int            handle,
                                           int                     signum) {
  const void *inary[2];
#ifdef DEBUG
  /* Check harness parameter. */
  if (NULL == harness) {
    errno = EINVAL;
    HARNESS_LOG((HARNESS_WARN(harness parameter is null)))
    errno = EINVAL;
    return -1;
  }
#endif /* DEBUG */
  /* Prepare input. */
  inary[0] = &handle;
  inary[1] = &signum;
  /* Invoke remote object method. */
  if (0 != rmix_oneway(harness, &harnessclient_interface,
                       HARNESS_RMIX_METHODS_PROCESSES_SIGNAL_INDEX,
                       inary, 2)) {
    int errno2 = errno;
    RMIX_LOG((RMIX_WARN(unable to invoke remote object method)))
    errno = errno2;
    return -1;
  }
  return 0;
}

/*
 * Invokes harness_processes_signal (client-side method stub).
 * Sends a signal to a process at a remote harness instance at using a one-way
 * invocation without loading the provider plug-in.
 *
 * harness  = The RMIX remote object reference of harness.
 * provider = The preloaded RMIX provider plug-in handle.
 * handle   = The process handle.
 * signum   = The signal to send.
 * return   = 0 on success or -1 on any error with errno set appropriately.
 */
#undef  __FUNC__
#define __FUNC__ "harnessclient_processes_signal_oneway_p"
int harnessclient_processes_signal_oneway_p (const rmix_remoteref_t *harness,
                                             unsigned int            provider,
                                             unsigned int            handle,
                                             int                     signum) {
  const void *inary[2];
#ifdef DEBUG
  /* Check harness parameter. */
  if (NULL == harness) {
    errno = EINVAL;
    HARNESS_LOG((HARNESS_WARN(harness parameter is null)))
    errno = EINVAL;
    return -1;
  }
#endif /* DEBUG */
  /* Prepare input. */
  inary[0] = &handle;
  inary[1] = &signum;
  /* Invoke remote object method. */
  if (0 != rmix_oneway_p(harness, provider, &harnessclient_interface,
                         HARNESS_RMIX_METHODS_PROCESSES_SIGNAL_INDEX,
                         inary, 2)) {
    int errno2 = errno;
    RMIX_LOG((RMIX_WARN(unable to invoke remote object method)))
    errno = errno2;
    return -1;
  }
  return 0;
}

/*
 * Invokes harness_processes_signal (client-side method stub).
 * Sends a signal to a process at a remote harness instance using an
 * asynchronous invocation.
 *
 * invokeref = The invocation reference (return).
 * harness   = The RMIX remote object reference of harness.
 * handle    = The process handle.
 * signum    = The signal to send.
 * return    = 0 on success or -1 on any error with errno set appropriately.
 */
#undef  __FUNC__
#define __FUNC__ "harnessclient_processes_signal_send"
int harnessclient_processes_signal_send (rmix_invokeref_t       **invokeref,
                                         const rmix_remoteref_t  *harness,
                                         unsigned int             handle,
                                         int                      signum) {
  const void *inary[2];
#ifdef DEBUG
  /* Check invokeref parameter. */
  if (NULL == invokeref) {
    errno = EINVAL;
    HARNESS_LOG((HARNESS_WARN(invokeref parameter is null)))
    errno = EINVAL;
    return -1;
  }
  /* Check harness parameter. */
  if (NULL == harness) {
    errno = EINVAL;
    HARNESS_LOG((HARNESS_WARN(harness parameter is null)))
    errno = EINVAL;
    return -1;
  }
#endif /* DEBUG */
  /* Prepare input. */
  inary[0] = &handle;
  inary[1] = &signum;
  /* Invoke remote object method. */
  if (0 != rmix_send(invokeref, harness, &harnessclient_interface,
                     HARNESS_RMIX_METHODS_PROCESSES_SIGNAL_INDEX,
                     inary, 2)) {
    int errno2 = errno;
    RMIX_LOG((RMIX_WARN(unable to invoke remote object method)))
    errno = errno2;
    return -1;
  }
  return 0;
}

/*
 * Invokes harness_processes_signal (client-side method stub).
 * Sends a signal to a process at a remote harness instance at using an
 * asynchronous invocation without loading the provider plug-in.
 *
 * invokeref = The invocation reference (return).
 * harness   = The RMIX remote object reference of harness.
 * provider  = The preloaded RMIX provider plug-in handle.
 * handle    = The process handle.
 * signum    = The signal to send.
 * return    = 0 on success or -1 on any error with errno set appropriately.
 */
#undef  __FUNC__
#define __FUNC__ "harnessclient_processes_signal_send_p"
int harnessclient_processes_signal_send_p (rmix_invokeref_t       **invokeref,
                                           const rmix_remoteref_t  *harness,
                                           unsigned int             provider,
                                           unsigned int             handle,
                                           int                      signum) {
  const void *inary[2];
#ifdef DEBUG
  /* Check invokeref parameter. */
  if (NULL == invokeref) {
    errno = EINVAL;
    HARNESS_LOG((HARNESS_WARN(invokeref parameter is null)))
    errno = EINVAL;
    return -1;
  }
  /* Check harness parameter. */
  if (NULL == harness) {
    errno = EINVAL;
    HARNESS_LOG((HARNESS_WARN(harness parameter is null)))
    errno = EINVAL;
    return -1;
  }
#endif /* DEBUG */
  /* Prepare input. */
  inary[0] = &handle;
  inary[1] = &signum;
  /* Invoke remote object method. */
  if (0 != rmix_send_p(invokeref, harness, provider, &harnessclient_interface,
                       HARNESS_RMIX_METHODS_PROCESSES_SIGNAL_INDEX,
                       inary, 2)) {
    int errno2 = errno;
    RMIX_LOG((RMIX_WARN(unable to invoke remote object method)))
    errno = errno2;
    return -1;
  }
  return 0;
}

/*
 * Retrieves harness_processes_signal output (client-side method stub).
 * Returns the output of a previous asynchronous invocation that sends a signal
 * to a process at a remote harness instance.
 *
 * invokeref = The invocation reference (also return).
 * return    = 0 on success or -1 on any error with errno set appropriately.
 */
#undef  __FUNC__
#define __FUNC__ "harnessclient_processes_signal_retrieve"
int harnessclient_processes_signal_retrieve (rmix_invokeref_t **invokeref) {
  int   result;
  void *outary[1];
#ifdef DEBUG
  /* Check invokeref parameter. */
  if (NULL == invokeref) {
    errno = EINVAL;
    HARNESS_LOG((HARNESS_WARN(invokeref parameter is null)))
    errno = EINVAL;
    return -1;
  }
  /* Check *invokeref parameter. */
  if (NULL == *invokeref) {
    errno = EINVAL;
    HARNESS_LOG((HARNESS_WARN(*invokeref parameter is null)))
    errno = EINVAL;
    return -1;
  }
#endif /* DEBUG */
  /* Prepare output. */
  outary[0] = &result;
  /* Invoke remote object method. */
  if (0 != rmix_retrieve(invokeref, outary, 1)) {
    int errno2 = errno;
    RMIX_LOG((RMIX_WARN(unable to invoke remote object method)))
    errno = errno2;
    return -1;
  }
  return result;
}

/*
 * Retrieves harness_processes_signal output (client-side method stub).
 * Returns the output of a previous asynchronous invocation that sends a signal
 * to a process at a remote harness instance without unloading the provider
 * plug-in.
 *
 * invokeref = The invocation reference (also return).
 * return    = 0 on success or -1 on any error with errno set appropriately.
 */
#undef  __FUNC__
#define __FUNC__ "harnessclient_processes_signal_retrieve_p"
int harnessclient_processes_signal_retrieve_p (rmix_invokeref_t **invokeref) {
  int   result;
  void *outary[1];
#ifdef DEBUG
  /* Check invokeref parameter. */
  if (NULL == invokeref) {
    errno = EINVAL;
    HARNESS_LOG((HARNESS_WARN(invokeref parameter is null)))
    errno = EINVAL;
    return -1;
  }
  /* Check *invokeref parameter. */
  if (NULL == *invokeref) {
    errno = EINVAL;
    HARNESS_LOG((HARNESS_WARN(*invokeref parameter is null)))
    errno = EINVAL;
    return -1;
  }
#endif /* DEBUG */
  /* Prepare output. */
  outary[0] = &result;
  /* Invoke remote object method. */ 
  if (0 != rmix_retrieve_p(invokeref, outary, 1)) {
    int errno2 = errno;
    RMIX_LOG((RMIX_WARN(unable to invoke remote object method)))
    errno = errno2;
    return -1;
  }
  return result;
}

/*
 * Invokes harness_processes_signal (client-side method stub).
 * Sends the same signal to a process at multiple remote harness instances.
 *
 * count     = The number of remote harness instances.
 * harnesses = The array of RMIX remote object references of harness instances.
 * handles   = The processes handle array. 0 handles are ignored.
 * return    = The number of signaled signals with errno set appropriately.
 */
#undef  __FUNC__
#define __FUNC__ "harnessclient_processes_psignal"
unsigned int harnessclient_processes_psignal
  (unsigned int             count,
   const rmix_remoteref_t **harnesses,
   const unsigned int      *handles,
   int                      signum) {
  int                errno2;
  unsigned int       signaled, index;
  rmix_invokeref_t **invocations;
#ifdef DEBUG
  /* Check harnesses parameter. */
  if ((NULL == harnesses)&&(0 != count)) {
    errno = EINVAL;
    RMIX_LOG((RMIX_WARN(harnesses parameter is null)))
    errno = EINVAL;
    return -1;
  }
  /* Check count parameter. */
  if ((0 == count)&&(NULL != harnesses)) {
    errno = EINVAL;
    RMIX_LOG((RMIX_WARN(count parameter is zero)))
    errno = EINVAL;
    return -1;
  }
  /* Check handles parameter. */
  if ((NULL == handles)&&(0 != count)) {
    errno = EINVAL;
    RMIX_LOG((RMIX_WARN(handles parameter is null)))
    errno = EINVAL;
    return -1;
  }
  /* Check count parameter. */
  if ((0 == count)&&(NULL != handles)) {
    errno = EINVAL;
    RMIX_LOG((RMIX_WARN(count parameter is zero)))
    errno = EINVAL;
    return -1;
  }
#endif /* DEBUG */
  /* Allocate invocation reference array. */
  if (NULL == (invocations = malloc(count * sizeof(rmix_invokeref_t*)))) {
    errno2 = errno;
    RMIX_LOG((RMIX_WARN(unable to allocate invocation reference array)))
    errno = errno2;
    return 0;
  }
  /* Initialize errno and signaled counter.*/
  errno2 = 0;
  signaled = 0;
  /* Signal all processs in forward order (send part). */
  for (index = 0; index < count; index++) {
    /* Check for non-zero processes handle. */
    if (0 != handles[index]) {
      /* Signal process (send part). */
      if (0 != harnessclient_processes_signal_send(&invocations[index],
                                                   harnesses[index],
                                                   handles[index], signum)) {
        errno2 = (0 == errno2)?errno:errno2;
        RMIX_LOG((RMIX_WARN(unable to signal process (send part))))
        invocations[index] = NULL;
      }
    }
  }
  /* Signal all processs in forward order (retrieve part). */
  for (index = 0; index < count; index++) {
    /* Check for successful send. */
    if (NULL != invocations[index]) {
      /* Signal process (retrieve part). */
      if (0 != harnessclient_processes_signal_retrieve(&invocations[index])) {
        errno2 = (0 == errno2)?errno:errno2;
        RMIX_LOG((RMIX_WARN(unable to signal process (retrieve part))))
      } else {
        signaled++;
      }
    }
  }
  /* Deallocate invocation reference array. */
  free(invocations);
  errno = errno2;
  return signaled;
}

/*
 * Invokes harnessclient_processes_wait (client-side method stub).
 * Waits for the exit of a process, but does not remove the entry in the kernel
 * process table at a remote harness instance.
 *
 * harness = The RMIX remote object reference of harness.
 * handle  = The process handle.
 * status  = The process exit status (return).
 * return  = 0 on success or -1 on any error with errno set appropriately.
 */
#undef  __FUNC__
#define __FUNC__ "harnessclient_processes_wait"
int harnessclient_processes_wait (const rmix_remoteref_t *harness,
                                  unsigned int            handle,
                                  int                    *status) {
  int         result;
  void       *outary[2];
  const void *inary[1];
#ifdef DEBUG
  /* Check harness parameter. */
  if (NULL == harness) {
    errno = EINVAL;
    HARNESS_LOG((HARNESS_WARN(harness parameter is null)))
    errno = EINVAL;
    return -1;
  }
  /* Check status parameter. */
  if (NULL == status) {
    errno = EINVAL;
    HARNESS_LOG((HARNESS_WARN(status parameter is null)))
    errno = EINVAL;
    return -1;
  }
#endif /* DEBUG */
  /* Prepare input. */
  inary[0] = &handle;
  /* Prepare output. */
  outary[0] = &result;
  outary[1] =  status;
  /* Invoke remote object method. */
  if (0 != rmix_invoke(harness, &harnessclient_interface,
                       HARNESS_RMIX_METHODS_PROCESSES_WAIT_INDEX,
                       outary, 2, inary, 1)) {
    int errno2 = errno;
    RMIX_LOG((RMIX_WARN(unable to invoke remote object method)))
    errno = errno2;
    return -1;
  }
  return result;
}

/*
 * Invokes harnessclient_processes_wait (client-side method stub).
 * Waits for the exit of a process, but does not remove the entry in the kernel
 * process table at a remote harness instance.
 *
 * harness  = The RMIX remote object reference of harness.
 * provider = The preloaded RMIX provider plug-in handle.
 * handle   = The process handle.
 * status   = The process exit status (return).
 * return   = 0 on success or -1 on any error with errno set appropriately.
 */
#undef  __FUNC__
#define __FUNC__ "harnessclient_processes_wait_p"
int harnessclient_processes_wait_p (const rmix_remoteref_t *harness,
                                    unsigned int            provider,
                                    unsigned int            handle,
                                    int                    *status) {
  int         result;
  void       *outary[2];
  const void *inary[1];
#ifdef DEBUG
  /* Check harness parameter. */
  if (NULL == harness) {
    errno = EINVAL;
    HARNESS_LOG((HARNESS_WARN(harness parameter is null)))
    errno = EINVAL;
    return -1;
  }
  /* Check status parameter. */
  if (NULL == status) {
    errno = EINVAL;
    HARNESS_LOG((HARNESS_WARN(status parameter is null)))
    errno = EINVAL;
    return -1;
  }
#endif /* DEBUG */
  /* Prepare input. */
  inary[0] = &handle;
  /* Prepare output. */
  outary[0] = &result;
  outary[1] =  status;
  /* Invoke remote object method. */
  if (0 != rmix_invoke_p(harness, provider, &harnessclient_interface,
                         HARNESS_RMIX_METHODS_PROCESSES_WAIT_INDEX,
                         outary, 2, inary, 1)) {
    int errno2 = errno;
    RMIX_LOG((RMIX_WARN(unable to invoke remote object method)))
    errno = errno2;
    return -1;
  }
  return result;
}

/*
 * Invokes harness_processes_wait (client-side method stub).
 * Waits for the exit of a process, but does not remove the entry in the kernel
 * process table at a remote harness instance using an asynchronous invocation.
 *
 * invokeref = The invocation reference (return).
 * harness   = The RMIX remote object reference of harness.
 * handle    = The process handle.
 * return    = 0 on success or -1 on any error with errno set appropriately.
 */
#undef  __FUNC__
#define __FUNC__ "harnessclient_processes_wait_send"
int harnessclient_processes_wait_send (rmix_invokeref_t       **invokeref,
                                       const rmix_remoteref_t  *harness,
                                       unsigned int             handle) {
  const void *inary[1];
#ifdef DEBUG
  /* Check invokeref parameter. */
  if (NULL == invokeref) {
    errno = EINVAL;
    HARNESS_LOG((HARNESS_WARN(invokeref parameter is null)))
    errno = EINVAL;
    return -1;
  }
  /* Check harness parameter. */
  if (NULL == harness) {
    errno = EINVAL;
    HARNESS_LOG((HARNESS_WARN(harness parameter is null)))
    errno = EINVAL;
    return -1;
  }
#endif /* DEBUG */
  /* Prepare input. */
  inary[0] = &handle;
  /* Invoke remote object method. */
  if (0 != rmix_send(invokeref, harness, &harnessclient_interface,
                     HARNESS_RMIX_METHODS_PROCESSES_WAIT_INDEX,
                     inary, 1)) {
    int errno2 = errno;
    RMIX_LOG((RMIX_WARN(unable to invoke remote object method)))
    errno = errno2;
    return -1;
  }
  return 0;
}

/*
 * Invokes harness_processes_wait (client-side method stub).
 * Waits for the exit of a process, but does not remove the entry in the kernel
 * process table at a remote harness instance using an asynchronous invocation
 * without loading the provider plug-in.
 *
 * invokeref = The invocation reference (return).
 * harness   = The RMIX remote object reference of harness.
 * provider  = The preloaded RMIX provider plug-in handle.
 * handle    = The process handle.
 * return    = 0 on success or -1 on any error with errno set appropriately.
 */
#undef  __FUNC__
#define __FUNC__ "harnessclient_processes_wait_send_p"
int harnessclient_processes_wait_send_p (rmix_invokeref_t       **invokeref,
                                         const rmix_remoteref_t  *harness,
                                         unsigned int             provider,
                                         unsigned int             handle) {
  const void *inary[1];
#ifdef DEBUG
  /* Check invokeref parameter. */
  if (NULL == invokeref) {
    errno = EINVAL;
    HARNESS_LOG((HARNESS_WARN(invokeref parameter is null)))
    errno = EINVAL;
    return -1;
  }
  /* Check harness parameter. */
  if (NULL == harness) {
    errno = EINVAL;
    HARNESS_LOG((HARNESS_WARN(harness parameter is null)))
    errno = EINVAL;
    return -1;
  }
#endif /* DEBUG */
  /* Prepare input. */
  inary[0] = &handle;
  /* Invoke remote object method. */
  if (0 != rmix_send_p(invokeref, harness, provider, &harnessclient_interface,
                       HARNESS_RMIX_METHODS_PROCESSES_WAIT_INDEX,
                       inary, 1)) {
    int errno2 = errno;
    RMIX_LOG((RMIX_WARN(unable to invoke remote object method)))
    errno = errno2;
    return -1;
  }
  return 0;
}

/*
 * Retrieves harness_processes_wait output (client-side method stub).
 * Returns the output of a previous asynchronous invocation that waits for the
 * exit of a process, but does not remove the entry in the kernel process table
 * at a remote harness instance.
 *
 * invokeref = The invocation reference (also return).
 * status    = The process exit status (return).
 * return    = 0 on success or -1 on any error with errno set appropriately.
 */
#undef  __FUNC__
#define __FUNC__ "harnessclient_processes_wait_retrieve"
int harnessclient_processes_wait_retrieve (rmix_invokeref_t **invokeref,
                                           int               *status) {
  int   result;
  void *outary[2];
#ifdef DEBUG
  /* Check invokeref parameter. */
  if (NULL == invokeref) {
    errno = EINVAL;
    HARNESS_LOG((HARNESS_WARN(invokeref parameter is null)))
    errno = EINVAL;
    return -1;
  }
  /* Check *invokeref parameter. */
  if (NULL == *invokeref) {
    errno = EINVAL;
    HARNESS_LOG((HARNESS_WARN(*invokeref parameter is null)))
    errno = EINVAL;
    return -1;
  }
  /* Check status parameter. */
  if (NULL == status) {
    errno = EINVAL;
    HARNESS_LOG((HARNESS_WARN(status parameter is null)))
    errno = EINVAL;
    return -1;
  }
#endif /* DEBUG */
  /* Prepare output. */
  outary[0] = &result;
  outary[1] =  status;
  /* Invoke remote object method. */
  if (0 != rmix_retrieve(invokeref, outary, 2)) {
    int errno2 = errno;
    RMIX_LOG((RMIX_WARN(unable to invoke remote object method)))
    errno = errno2;
    return -1;
  }
  return result;
}

/*
 * Retrieves harness_processes_wait output (client-side method stub).
 * Returns the output of a previous asynchronous invocation that waits for the
 * exit of a process, but does not remove the entry in the kernel process table
 * at a remote harness instance without unloading the provider plug-in.
 *
 * invokeref = The invocation reference (also return).
 * status    = The process exit status (return).
 * return    = 0 on success or -1 on any error with errno set appropriately.
 */
#undef  __FUNC__
#define __FUNC__ "harnessclient_processes_wait_retrieve_p"
int harnessclient_processes_wait_retrieve_p (rmix_invokeref_t **invokeref,
                                             int               *status) {
  int   result;
  void *outary[2];
#ifdef DEBUG
  /* Check invokeref parameter. */
  if (NULL == invokeref) {
    errno = EINVAL;
    HARNESS_LOG((HARNESS_WARN(invokeref parameter is null)))
    errno = EINVAL;
    return -1;
  }
  /* Check *invokeref parameter. */
  if (NULL == *invokeref) {
    errno = EINVAL;
    HARNESS_LOG((HARNESS_WARN(*invokeref parameter is null)))
    errno = EINVAL;
    return -1;
  }
  /* Check status parameter. */
  if (NULL == status) {
    errno = EINVAL;
    HARNESS_LOG((HARNESS_WARN(status parameter is null)))
    errno = EINVAL;
    return -1;
  }
#endif /* DEBUG */
  /* Prepare output. */
  outary[0] = &result;
  outary[1] =  status;
  /* Invoke remote object method. */ 
  if (0 != rmix_retrieve_p(invokeref, outary, 2)) {
    int errno2 = errno;
    RMIX_LOG((RMIX_WARN(unable to invoke remote object method)))
    errno = errno2;
    return -1;
  }
  return result;
}

/*
 * Invokes harnessclient_processes_wait (client-side method stub).
 * Waits for the exit of a process at multiple remote harness instances, but
 * does not remove the entry in the kernel process table.
 * The processes may be different for every remote harness instance.
 *
 * count     = The number of remote harness instances.
 * harnesses = The array of RMIX remote object references of harness instances.
 * handles   = The processes handle array. 0 handles are ignored.
 * statuses  = The process exit status array (also return). The status of a
 *             process is set to -1 on error.
 * order     = The order flag. 0 for forward and 1 for reverse.
 * fail      = The fail flag. 0 to continue and 1 to fail on any error.
 * return    = The number of waited processess with errno set appropriately.
 */
#undef  __FUNC__
#define __FUNC__ "harnessclient_processes_swait"
int harnessclient_processes_swait (unsigned int             count,
                                   const rmix_remoteref_t **harnesses,
                                   const unsigned int      *handles,
                                   int                     *statuses,
                                   unsigned int             order,
                                   unsigned int             fail) {
  int          errno2;
  unsigned int waited, index;
#ifdef DEBUG
  /* Check harnesses parameter. */
  if ((NULL == harnesses)&&(0 != count)) {
    errno = EINVAL;
    RMIX_LOG((RMIX_WARN(harnesses parameter is null)))
    errno = EINVAL;
    return -1;
  }
  /* Check count parameter. */
  if ((0 == count)&&(NULL != harnesses)) {
    errno = EINVAL;
    RMIX_LOG((RMIX_WARN(count parameter is zero)))
    errno = EINVAL;
    return -1;
  }
  /* Check handles parameter. */
  if ((NULL == handles)&&(0 != count)) {
    errno = EINVAL;
    RMIX_LOG((RMIX_WARN(handles parameter is null)))
    errno = EINVAL;
    return -1;
  }
  /* Check count parameter. */
  if ((0 == count)&&(NULL != handles)) {
    errno = EINVAL;
    RMIX_LOG((RMIX_WARN(count parameter is zero)))
    errno = EINVAL;
    return -1;
  }
  /* Check statuses parameter. */
  if ((NULL == statuses)&&(0 != count)) {
    errno = EINVAL;
    RMIX_LOG((RMIX_WARN(statuses parameter is null)))
    errno = EINVAL;
    return -1;
  }
  /* Check count parameter. */
  if ((0 == count)&&(NULL != statuses)) {
    errno = EINVAL;
    RMIX_LOG((RMIX_WARN(count parameter is zero)))
    errno = EINVAL;
    return -1;
  }
#endif /* DEBUG */
  /* Initialize errno and waited counter.*/
  errno2 = 0;
  waited = 0;
  /* Check order flag. */
  if (0 == order) {
    /* Wait for all processess in forward order. */
    for (index = 0; index < count; index++) {
      /* Check for non-zero processes handle. */
      if (0 != handles[index]) {
        /* Wait for processes. */
        if (0 != harnessclient_processes_wait(harnesses[index],
                                              handles[index],
                                              &statuses[index])) {
          errno2 = (0 == errno2)?errno:errno2;
          RMIX_LOG((RMIX_WARN(unable to wait for processes)))
          /* Set exit status. */
          statuses[index] = -1;
          /* Check fail flag. */
          if (1 == fail) {
            errno = errno2;
            return waited;
          }
        } else {
          waited++;
        }
      }
    }
  } else {
    /* Wait for all processess in reverse order. */
    for (index = count; 0 < index--;) {
      /* Check for non-zero processes handle. */
      if (0 != handles[index]) {
        /* Wait for processes. */
        if (0 != harnessclient_processes_wait(harnesses[index],
                                              handles[index],
                                              &statuses[index])) {
          errno2 = (0 == errno2)?errno:errno2;
          RMIX_LOG((RMIX_WARN(unable to wait for processes)))
          /* Set exit status. */
          statuses[index] = -1;
          /* Check fail flag. */
          if (1 == fail) {
            errno = errno2;
            return waited;
          }
        } else {
          waited++;
        }
      }
    }
  }
  errno = errno2;
  return waited;
}

/*
 * Invokes harnessclient_processes_wait (client-side method stub).
 * Waits for the exit of a process at multiple remote harness instances in
 * parallel using asynchonous method invocations, but does not remove the entry
 * in the kernel process table.
 * The processes may be different for every remote harness instance.
 *
 * count     = The number of remote harness instances.
 * harnesses = The array of RMIX remote object references of harness instances.
 * handles   = The processes handle array. 0 handles are ignored.
 * statuses  = The process exit status array (also return). The status of a
 *             process is set to -1 on error.
 * return    = The number of waited processess with errno set appropriately.
 */
#undef  __FUNC__
#define __FUNC__ "harnessclient_processes_pwait"
int harnessclient_processes_pwait (unsigned int             count,
                                   const rmix_remoteref_t **harnesses,
                                   const unsigned int      *handles,
                                   int                     *statuses) {
  int                errno2;
  unsigned int       waited, index;
  rmix_invokeref_t **invocations;
#ifdef DEBUG
  /* Check harnesses parameter. */
  if ((NULL == harnesses)&&(0 != count)) {
    errno = EINVAL;
    RMIX_LOG((RMIX_WARN(harnesses parameter is null)))
    errno = EINVAL;
    return -1;
  }
  /* Check count parameter. */
  if ((0 == count)&&(NULL != harnesses)) {
    errno = EINVAL;
    RMIX_LOG((RMIX_WARN(count parameter is zero)))
    errno = EINVAL;
    return -1;
  }
  /* Check handles parameter. */
  if ((NULL == handles)&&(0 != count)) {
    errno = EINVAL;
    RMIX_LOG((RMIX_WARN(handles parameter is null)))
    errno = EINVAL;
    return -1;
  }
  /* Check count parameter. */
  if ((0 == count)&&(NULL != handles)) {
    errno = EINVAL;
    RMIX_LOG((RMIX_WARN(count parameter is zero)))
    errno = EINVAL;
    return -1;
  }
  /* Check statuses parameter. */
  if ((NULL == statuses)&&(0 != count)) {
    errno = EINVAL;
    RMIX_LOG((RMIX_WARN(statuses parameter is null)))
    errno = EINVAL;
    return -1;
  }
  /* Check count parameter. */
  if ((0 == count)&&(NULL != statuses)) {
    errno = EINVAL;
    RMIX_LOG((RMIX_WARN(count parameter is zero)))
    errno = EINVAL;
    return -1;
  }
#endif /* DEBUG */
  /* Allocate invocation reference array. */
  if (NULL == (invocations = malloc(count * sizeof(rmix_invokeref_t*)))) {
    errno2 = errno;
    RMIX_LOG((RMIX_WARN(unable to allocate invocation reference array)))
    errno = errno2;
    return 0;
  }
  /* Initialize errno and waited counter.*/
  errno2 = 0;
  waited = 0;
  /* Wait for all processess in forward order (send part). */
  for (index = 0; index < count; index++) {
    /* Check for non-zero processes handle. */
    if (0 != handles[index]) {
      /* Wait for processes (send part). */
      if (0 != harnessclient_processes_wait_send(&invocations[index],
                                                 harnesses[index],
                                                 handles[index])) {
        errno2 = (0 == errno2)?errno:errno2;
        RMIX_LOG((RMIX_WARN(unable to wait for processes (send part))))
        invocations[index] = NULL;
        /* Set exit status. */
        statuses[index] = -1;
      }
    }
  }
  /* Wait for all processess in forward order (retrieve part). */
  for (index = 0; index < count; index++) {
    /* Check for non-zero processes handle. */
    if (0 != handles[index]) {
      /* Check for successful send. */
      if (NULL != invocations[index]) {
        /* Wait for processes (retrieve part). */
        if (0 != harnessclient_processes_wait_retrieve(&invocations[index],
                                                       &statuses[index])) {
          errno2 = (0 == errno2)?errno:errno2;
          RMIX_LOG((RMIX_WARN(unable to wait for processes (retrieve part))))
          /* Set exit status. */
          statuses[index] = -1;
        } else {
          waited++;
        }
      }
    }
  }
  /* Deallocate invocation reference array. */
  free(invocations);
  errno = errno2;
  return waited;
}

/*
 * Invokes harnessclient_processes_terminate (client-side method stub).
 * Terminates a process, waits for its exit and removes the entry in the kernel
 * process table at a remote harness instance.
 *
 * harness = The RMIX remote object reference of harness.
 * handle  = The process handle.
 * status  = The process exit status (return).
 * return  = 0 on success or -1 on any error with errno set appropriately.
 */
#undef  __FUNC__
#define __FUNC__ "harnessclient_processes_terminate"
int harnessclient_processes_terminate (const rmix_remoteref_t *harness,
                                       unsigned int            handle,
                                       int                    *status) {
  int         result;
  void       *outary[2];
  const void *inary[1];
#ifdef DEBUG
  /* Check harness parameter. */
  if (NULL == harness) {
    errno = EINVAL;
    HARNESS_LOG((HARNESS_WARN(harness parameter is null)))
    errno = EINVAL;
    return -1;
  }
  /* Check status parameter. */
  if (NULL == status) {
    errno = EINVAL;
    HARNESS_LOG((HARNESS_WARN(status parameter is null)))
    errno = EINVAL;
    return -1;
  }
#endif /* DEBUG */
  /* Prepare input. */
  inary[0] = &handle;
  /* Prepare output. */
  outary[0] = &result;
  outary[1] =  status;
  /* Invoke remote object method. */
  if (0 != rmix_invoke(harness, &harnessclient_interface,
                       HARNESS_RMIX_METHODS_PROCESSES_TERMINATE_INDEX,
                       outary, 2, inary, 1)) {
    int errno2 = errno;
    RMIX_LOG((RMIX_WARN(unable to invoke remote object method)))
    errno = errno2;
    return -1;
  }
  return result;
}

/*
 * Invokes harnessclient_processes_terminate (client-side method stub).
 * Terminates a process, waits for its exit and removes the entry in the kernel
 * process table at a remote harness instance.
 *
 * harness  = The RMIX remote object reference of harness.
 * provider = The preloaded RMIX provider plug-in handle.
 * handle   = The process handle.
 * status   = The process exit status (return).
 * return   = 0 on success or -1 on any error with errno set appropriately.
 */
#undef  __FUNC__
#define __FUNC__ "harnessclient_processes_terminate_p"
int harnessclient_processes_terminate_p (const rmix_remoteref_t *harness,
                                         unsigned int            provider,
                                         unsigned int            handle,
                                         int                    *status) {
  int         result;
  void       *outary[2];
  const void *inary[1];
#ifdef DEBUG
  /* Check harness parameter. */
  if (NULL == harness) {
    errno = EINVAL;
    HARNESS_LOG((HARNESS_WARN(harness parameter is null)))
    errno = EINVAL;
    return -1;
  }
  /* Check status parameter. */
  if (NULL == status) {
    errno = EINVAL;
    HARNESS_LOG((HARNESS_WARN(status parameter is null)))
    errno = EINVAL;
    return -1;
  }
#endif /* DEBUG */
  /* Prepare input. */
  inary[0] = &handle;
  /* Prepare output. */
  outary[0] = &result;
  outary[1] =  status;
  /* Invoke remote object method. */
  if (0 != rmix_invoke_p(harness, provider, &harnessclient_interface,
                         HARNESS_RMIX_METHODS_PROCESSES_TERMINATE_INDEX,
                         outary, 2, inary, 1)) {
    int errno2 = errno;
    RMIX_LOG((RMIX_WARN(unable to invoke remote object method)))
    errno = errno2;
    return -1;
  }
  return result;
}

/*
 * Invokes harness_processes_terminate (client-side method stub).
 * Terminates a process, waits for its exit and removes the entry in the kernel
 * process table at a remote harness instance using a oneway invocation.
 *
 * harness = The RMIX remote object reference of harness.
 * handle  = The process handle.
 * return  = 0 on success or -1 on any error with errno set appropriately.
 */
#undef  __FUNC__
#define __FUNC__ "harnessclient_processes_terminate_oneway"
int harnessclient_processes_terminate_oneway (const rmix_remoteref_t *harness,
                                              unsigned int            handle) {
  const void *inary[1];
#ifdef DEBUG
  /* Check harness parameter. */
  if (NULL == harness) {
    errno = EINVAL;
    HARNESS_LOG((HARNESS_WARN(harness parameter is null)))
    errno = EINVAL;
    return -1;
  }
#endif /* DEBUG */
  /* Prepare input. */
  inary[0] = &handle;
  /* Invoke remote object method. */
  if (0 != rmix_oneway(harness, &harnessclient_interface,
                       HARNESS_RMIX_METHODS_PROCESSES_TERMINATE_INDEX,
                       inary, 1)) {
    int errno2 = errno;
    RMIX_LOG((RMIX_WARN(unable to invoke remote object method)))
    errno = errno2;
    return -1;
  }
  return 0;
}

/*
 * Invokes harness_processes_terminate (client-side method stub).
 * Terminates a process, waits for its exit and removes the entry in the kernel
 * process table at a remote harness instance using a oneway invocation
 * without loading the provider plug-in.
 *
 * harness  = The RMIX remote object reference of harness.
 * provider = The preloaded RMIX provider plug-in handle.
 * handle   = The process handle.
 * return   = 0 on success or -1 on any error with errno set appropriately.
 */
#undef  __FUNC__
#define __FUNC__ "harnessclient_processes_terminate_oneway_p"
int harnessclient_processes_terminate_oneway_p
  (const rmix_remoteref_t *harness,
   unsigned int            provider,
   unsigned int            handle) {
  const void *inary[1];
#ifdef DEBUG
  /* Check harness parameter. */
  if (NULL == harness) {
    errno = EINVAL;
    HARNESS_LOG((HARNESS_WARN(harness parameter is null)))
    errno = EINVAL;
    return -1;
  }
#endif /* DEBUG */
  /* Prepare input. */
  inary[0] = &handle;
  /* Invoke remote object method. */
  if (0 != rmix_oneway_p(harness, provider, &harnessclient_interface,
                         HARNESS_RMIX_METHODS_PROCESSES_TERMINATE_INDEX,
                         inary, 1)) {
    int errno2 = errno;
    RMIX_LOG((RMIX_WARN(unable to invoke remote object method)))
    errno = errno2;
    return -1;
  }
  return 0;
}

/*
 * Invokes harness_processes_terminate (client-side method stub).
 * Terminates a process, waits for its exit and removes the entry in the kernel
 * process table at a remote harness instance using an asynchronous invocation.
 *
 * invokeref = The invocation reference (return).
 * harness   = The RMIX remote object reference of harness.
 * handle    = The process handle.
 * return    = 0 on success or -1 on any error with errno set appropriately.
 */
#undef  __FUNC__
#define __FUNC__ "harnessclient_processes_terminate_send"
int harnessclient_processes_terminate_send (rmix_invokeref_t       **invokeref,
                                            const rmix_remoteref_t  *harness,
                                            unsigned int             handle) {
  const void *inary[1];
#ifdef DEBUG
  /* Check invokeref parameter. */
  if (NULL == invokeref) {
    errno = EINVAL;
    HARNESS_LOG((HARNESS_WARN(invokeref parameter is null)))
    errno = EINVAL;
    return -1;
  }
  /* Check harness parameter. */
  if (NULL == harness) {
    errno = EINVAL;
    HARNESS_LOG((HARNESS_WARN(harness parameter is null)))
    errno = EINVAL;
    return -1;
  }
#endif /* DEBUG */
  /* Prepare input. */
  inary[0] = &handle;
  /* Invoke remote object method. */
  if (0 != rmix_send(invokeref, harness, &harnessclient_interface,
                     HARNESS_RMIX_METHODS_PROCESSES_TERMINATE_INDEX,
                     inary, 1)) {
    int errno2 = errno;
    RMIX_LOG((RMIX_WARN(unable to invoke remote object method)))
    errno = errno2;
    return -1;
  }
  return 0;
}

/*
 * Invokes harness_processes_terminate (client-side method stub).
 * Terminates a process, waits for its exit and removes the entry in the kernel
 * process table at a remote harness instance using an asynchronous invocation
 * without loading the provider plug-in.
 *
 * invokeref = The invocation reference (return).
 * harness   = The RMIX remote object reference of harness.
 * provider  = The preloaded RMIX provider plug-in handle.
 * handle    = The process handle.
 * return    = 0 on success or -1 on any error with errno set appropriately.
 */
#undef  __FUNC__
#define __FUNC__ "harnessclient_processes_terminate_send_p"
int harnessclient_processes_terminate_send_p
  (rmix_invokeref_t       **invokeref,
   const rmix_remoteref_t  *harness,
   unsigned int             provider,
   unsigned int             handle) {
  const void *inary[1];
#ifdef DEBUG
  /* Check invokeref parameter. */
  if (NULL == invokeref) {
    errno = EINVAL;
    HARNESS_LOG((HARNESS_WARN(invokeref parameter is null)))
    errno = EINVAL;
    return -1;
  }
  /* Check harness parameter. */
  if (NULL == harness) {
    errno = EINVAL;
    HARNESS_LOG((HARNESS_WARN(harness parameter is null)))
    errno = EINVAL;
    return -1;
  }
#endif /* DEBUG */
  /* Prepare input. */
  inary[0] = &handle;
  /* Invoke remote object method. */
  if (0 != rmix_send_p(invokeref, harness, provider, &harnessclient_interface,
                       HARNESS_RMIX_METHODS_PROCESSES_TERMINATE_INDEX,
                       inary, 1)) {
    int errno2 = errno;
    RMIX_LOG((RMIX_WARN(unable to invoke remote object method)))
    errno = errno2;
    return -1;
  }
  return 0;
}

/*
 * Retrieves harness_processes_terminate output (client-side method stub).
 * Returns the output of a previous asynchronous invocation that terminates a
 * process, waits for its exit and removes the entry in the kernel process table
 * at a remote harness instance.
 *
 * invokeref = The invocation reference (also return).
 * status    = The process exit status (return).
 * return    = 0 on success or -1 on any error with errno set appropriately.
 */
#undef  __FUNC__
#define __FUNC__ "harnessclient_processes_terminate_retrieve"
int harnessclient_processes_terminate_retrieve (rmix_invokeref_t **invokeref,
                                                int               *status) {
  int   result;
  void *outary[2];
#ifdef DEBUG
  /* Check invokeref parameter. */
  if (NULL == invokeref) {
    errno = EINVAL;
    HARNESS_LOG((HARNESS_WARN(invokeref parameter is null)))
    errno = EINVAL;
    return -1;
  }
  /* Check *invokeref parameter. */
  if (NULL == *invokeref) {
    errno = EINVAL;
    HARNESS_LOG((HARNESS_WARN(*invokeref parameter is null)))
    errno = EINVAL;
    return -1;
  }
  /* Check status parameter. */
  if (NULL == status) {
    errno = EINVAL;
    HARNESS_LOG((HARNESS_WARN(status parameter is null)))
    errno = EINVAL;
    return -1;
  }
#endif /* DEBUG */
  /* Prepare output. */
  outary[0] = &result;
  outary[1] =  status;
  /* Invoke remote object method. */
  if (0 != rmix_retrieve(invokeref, outary, 2)) {
    int errno2 = errno;
    RMIX_LOG((RMIX_WARN(unable to invoke remote object method)))
    errno = errno2;
    return -1;
  }
  return result;
}

/*
 * Retrieves harness_processes_terminate output (client-side method stub).
 * Returns the output of a previous asynchronous invocation that terminates a
 * process, waits for its exit and removes the entry in the kernel process table
 * at a remote harness instance without unloading the provider plug-in.
 *
 * invokeref = The invocation reference (also return).
 * status    = The process exit status (return).
 * return    = 0 on success or -1 on any error with errno set appropriately.
 */
#undef  __FUNC__
#define __FUNC__ "harnessclient_processes_terminate_retrieve_p"
int harnessclient_processes_terminate_retrieve_p (rmix_invokeref_t **invokeref,
                                                  int               *status) {
  int   result;
  void *outary[2];
#ifdef DEBUG
  /* Check invokeref parameter. */
  if (NULL == invokeref) {
    errno = EINVAL;
    HARNESS_LOG((HARNESS_WARN(invokeref parameter is null)))
    errno = EINVAL;
    return -1;
  }
  /* Check *invokeref parameter. */
  if (NULL == *invokeref) {
    errno = EINVAL;
    HARNESS_LOG((HARNESS_WARN(*invokeref parameter is null)))
    errno = EINVAL;
    return -1;
  }
  /* Check status parameter. */
  if (NULL == status) {
    errno = EINVAL;
    HARNESS_LOG((HARNESS_WARN(status parameter is null)))
    errno = EINVAL;
    return -1;
  }
#endif /* DEBUG */
  /* Prepare output. */
  outary[0] = &result;
  outary[1] =  status;
  /* Invoke remote object method. */ 
  if (0 != rmix_retrieve_p(invokeref, outary, 2)) {
    int errno2 = errno;
    RMIX_LOG((RMIX_WARN(unable to invoke remote object method)))
    errno = errno2;
    return -1;
  }
  return result;
}

/*
 * Invokes harnessclient_processes_terminate (client-side method stub).
 * Terminates a process, waits for its exit and removes the entry in the kernel
 * process table at multiple remote harness instances.
 * The processes may be different for every remote harness instance.
 *
 * count     = The number of remote harness instances.
 * harnesses = The array of RMIX remote object references of harness instances.
 * handles   = The processes handle array. 0 handles are ignored.
 * statuses  = The process exit status array (also return). The status of a
 *             process is set to -1 on error.
 * order     = The order flag. 0 for forward and 1 for reverse.
 * fail      = The fail flag. 0 to continue and 1 to fail on any error.
 * return    = The number of terminated processess with errno set appropriately.
 */
#undef  __FUNC__
#define __FUNC__ "harnessclient_processes_sterminate"
int harnessclient_processes_sterminate (unsigned int             count,
                                        const rmix_remoteref_t **harnesses,
                                        const unsigned int      *handles,
                                        int                     *statuses,
                                        unsigned int             order,
                                        unsigned int             fail) {
  int          errno2;
  unsigned int terminated, index;
#ifdef DEBUG
  /* Check harnesses parameter. */
  if ((NULL == harnesses)&&(0 != count)) {
    errno = EINVAL;
    RMIX_LOG((RMIX_WARN(harnesses parameter is null)))
    errno = EINVAL;
    return -1;
  }
  /* Check count parameter. */
  if ((0 == count)&&(NULL != harnesses)) {
    errno = EINVAL;
    RMIX_LOG((RMIX_WARN(count parameter is zero)))
    errno = EINVAL;
    return -1;
  }
  /* Check handles parameter. */
  if ((NULL == handles)&&(0 != count)) {
    errno = EINVAL;
    RMIX_LOG((RMIX_WARN(handles parameter is null)))
    errno = EINVAL;
    return -1;
  }
  /* Check count parameter. */
  if ((0 == count)&&(NULL != handles)) {
    errno = EINVAL;
    RMIX_LOG((RMIX_WARN(count parameter is zero)))
    errno = EINVAL;
    return -1;
  }
  /* Check statuses parameter. */
  if ((NULL == statuses)&&(0 != count)) {
    errno = EINVAL;
    RMIX_LOG((RMIX_WARN(statuses parameter is null)))
    errno = EINVAL;
    return -1;
  }
  /* Check count parameter. */
  if ((0 == count)&&(NULL != statuses)) {
    errno = EINVAL;
    RMIX_LOG((RMIX_WARN(count parameter is zero)))
    errno = EINVAL;
    return -1;
  }
#endif /* DEBUG */
  /* Initialize errno and terminated counter.*/
  errno2 = 0;
  terminated = 0;
  /* Check order flag. */
  if (0 == order) {
    /* Terminate all processess in forward order. */
    for (index = 0; index < count; index++) {
      /* Check for non-zero processes handle. */
      if (0 != handles[index]) {
        /* Terminate processes. */
        if (0 != harnessclient_processes_terminate(harnesses[index],
                                                   handles[index],
                                                   &statuses[index])) {
          errno2 = (0 == errno2)?errno:errno2;
          RMIX_LOG((RMIX_WARN(unable to terminate processes)))
          /* Set exit status. */
          statuses[index] = -1;
          /* Check fail flag. */
          if (1 == fail) {
            errno = errno2;
            return terminated;
          }
        } else {
          terminated++;
        }
      }
    }
  } else {
    /* Terminate all processess in reverse order. */
    for (index = count; 0 < index--;) {
      /* Check for non-zero processes handle. */
      if (0 != handles[index]) {
        /* Terminate processes. */
        if (0 != harnessclient_processes_terminate(harnesses[index],
                                                   handles[index],
                                                   &statuses[index])) {
          errno2 = (0 == errno2)?errno:errno2;
          RMIX_LOG((RMIX_WARN(unable to wait for processes)))
          /* Set exit status. */
          statuses[index] = -1;
          /* Check fail flag. */
          if (1 == fail) {
            errno = errno2;
            return terminated;
          }
        } else {
          terminated++;
        }
      }
    }
  }
  errno = errno2;
  return terminated;
}

/*
 * Terminates a process, waits for its exit and removes the entry in the kernel
 * process table at multiple remote harness instances in parallel using
 * asynchonous method invocations.
 * The processes may be different for every remote harness instance.
 *
 * count     = The number of remote harness instances.
 * harnesses = The array of RMIX remote object references of harness instances.
 * handles   = The processes handle array. 0 handles are ignored.
 * statuses  = The process exit status array (also return). The status of a
 *             process is set to -1 on error.
 * return    = The number of terminated processess with errno set appropriately.
 */
#undef  __FUNC__
#define __FUNC__ "harnessclient_processes_pterminate"
int harnessclient_processes_pterminate (unsigned int             count,
                                        const rmix_remoteref_t **harnesses,
                                        const unsigned int      *handles,
                                        int                     *statuses) {
  int                errno2;
  unsigned int       terminated, index;
  rmix_invokeref_t **invocations;
#ifdef DEBUG
  /* Check harnesses parameter. */
  if ((NULL == harnesses)&&(0 != count)) {
    errno = EINVAL;
    RMIX_LOG((RMIX_WARN(harnesses parameter is null)))
    errno = EINVAL;
    return -1;
  }
  /* Check count parameter. */
  if ((0 == count)&&(NULL != harnesses)) {
    errno = EINVAL;
    RMIX_LOG((RMIX_WARN(count parameter is zero)))
    errno = EINVAL;
    return -1;
  }
  /* Check handles parameter. */
  if ((NULL == handles)&&(0 != count)) {
    errno = EINVAL;
    RMIX_LOG((RMIX_WARN(handles parameter is null)))
    errno = EINVAL;
    return -1;
  }
  /* Check count parameter. */
  if ((0 == count)&&(NULL != handles)) {
    errno = EINVAL;
    RMIX_LOG((RMIX_WARN(count parameter is zero)))
    errno = EINVAL;
    return -1;
  }
  /* Check statuses parameter. */
  if ((NULL == statuses)&&(0 != count)) {
    errno = EINVAL;
    RMIX_LOG((RMIX_WARN(statuses parameter is null)))
    errno = EINVAL;
    return -1;
  }
  /* Check count parameter. */
  if ((0 == count)&&(NULL != statuses)) {
    errno = EINVAL;
    RMIX_LOG((RMIX_WARN(count parameter is zero)))
    errno = EINVAL;
    return -1;
  }
#endif /* DEBUG */
  /* Allocate invocation reference array. */
  if (NULL == (invocations = malloc(count * sizeof(rmix_invokeref_t*)))) {
    errno2 = errno;
    RMIX_LOG((RMIX_WARN(unable to allocate invocation reference array)))
    errno = errno2;
    return 0;
  }
  /* Initialize errno and terminated counter.*/
  errno2 = 0;
  terminated = 0;
  /* Terminate all processess in forward order (send part). */
  for (index = 0; index < count; index++) {
    /* Check for non-zero processes handle. */
    if (0 != handles[index]) {
      /* Terminate processes (send part). */
      if (0 != harnessclient_processes_terminate_send(&invocations[index],
                                                      harnesses[index],
                                                      handles[index])) {
        errno2 = (0 == errno2)?errno:errno2;
        RMIX_LOG((RMIX_WARN(unable to terminate processes (send part))))
        invocations[index] = NULL;
        /* Set exit status. */
        statuses[index] = -1;
      }
    }
  }
  /* Terminate all processess in forward order (retrieve part). */
  for (index = 0; index < count; index++) {
    /* Check for non-zero processes handle. */
    if (0 != handles[index]) {
      /* Check for successful send. */
      if (NULL != invocations[index]) {
        /* Terminate processes (retrieve part). */
        if (0 != harnessclient_processes_terminate_retrieve(&invocations[index],
                                                            &statuses[index])) {
          errno2 = (0 == errno2)?errno:errno2;
          RMIX_LOG((
            RMIX_WARN(unable to terminate processes (retrieve part))))
          /* Set exit status. */
          statuses[index] = -1;
        } else {
          terminated++;
        }
      }
    }
  }
  /* Deallocate invocation reference array. */
  free(invocations);
  errno = errno2;
  return terminated;
}

/*
 * Invokes harness_kernel_shutdown (client-side method stub).
 * Asynchonously shuts down the kernel and subsequently the forker at a remote
 * harness instance.
 *
 * harness = The RMIX remote object reference of harness.
 * return  = 0 on success or -1 on any error with errno set appropriately.
 */
#undef  __FUNC__
#define __FUNC__ "harnessclient_kernel_shutdown"
int harnessclient_kernel_shutdown (const rmix_remoteref_t *harness) {
#ifdef DEBUG
  /* Check harness parameter. */
  if (NULL == harness) {
    errno = EINVAL;
    HARNESS_LOG((HARNESS_WARN(harness parameter is null)))
    errno = EINVAL;
    return -1;
  }
#endif /* DEBUG */
  /* Invoke remote object method. */
  if (0 != rmix_invoke(harness, &harnessclient_interface,
                       HARNESS_RMIX_METHODS_KERNEL_SHUTDOWN_INDEX,
                       NULL, 0, NULL, 0)) {
    int errno2 = errno;
    RMIX_LOG((RMIX_WARN(unable to invoke remote object method)))
    errno = errno2;
    return -1;
  }
  return 0;
}

/*
 * Invokes harness_kernel_shutdown (client-side method stub).
 * Asynchonously shuts down the kernel and subsequently the forker at a remote
 * harness instance.
 *
 * harness  = The RMIX remote object reference of harness.
 * provider = The preloaded RMIX provider plug-in handle.
 * return   = 0 on success or -1 on any error with errno set appropriately.
 */
#undef  __FUNC__
#define __FUNC__ "harnessclient_kernel_shutdown_p"
int harnessclient_kernel_shutdown_p (const rmix_remoteref_t *harness,
                                     unsigned int            provider) {
#ifdef DEBUG
  /* Check harness parameter. */
  if (NULL == harness) {
    errno = EINVAL;
    HARNESS_LOG((HARNESS_WARN(harness parameter is null)))
    errno = EINVAL;
    return -1;
  }
#endif /* DEBUG */
  /* Invoke remote object method. */
  if (0 != rmix_invoke_p(harness, provider, &harnessclient_interface,
                         HARNESS_RMIX_METHODS_KERNEL_SHUTDOWN_INDEX,
                         NULL, 0, NULL, 0)) {
    int errno2 = errno;
    RMIX_LOG((RMIX_WARN(unable to invoke remote object method)))
    errno = errno2;
    return -1;
  }
  return 0;
}

/*
 * Invokes harness_kernel_shutdown (client-side method stub).
 * Asynchonously shuts down the kernel and subsequently the forker a remote
 * harness instance using a one-way invocation.
 *
 * harness = The RMIX remote object reference of harness.
 * return  = 0 on success or -1 on any error with errno set appropriately.
 */
#undef  __FUNC__
#define __FUNC__ "harnessclient_kernel_shutdown_oneway"
int harnessclient_kernel_shutdown_oneway (const rmix_remoteref_t *harness) {
#ifdef DEBUG
  /* Check harness parameter. */
  if (NULL == harness) {
    errno = EINVAL;
    HARNESS_LOG((HARNESS_WARN(harness parameter is null)))
    errno = EINVAL;
    return -1;
  }
#endif /* DEBUG */
  /* Invoke remote object method. */
  if (0 != rmix_oneway(harness, &harnessclient_interface,
                       HARNESS_RMIX_METHODS_KERNEL_SHUTDOWN_INDEX,
                       NULL, 0)) {
    int errno2 = errno;
    RMIX_LOG((RMIX_WARN(unable to invoke remote object method)))
    errno = errno2;
    return -1;
  }
  return 0;
}

/*
 * Invokes harness_kernel_shutdown (client-side method stub).
 * Asynchonously shuts down the kernel and subsequently the forker at a remote
 * harness instance using a one-way invocation without loading the provider
 * plug-in.
 *
 * harness  = The RMIX remote object reference of harness.
 * provider = The preloaded RMIX provider plug-in handle.
 * return   = 0 on success or -1 on any error with errno set appropriately.
 */
#undef  __FUNC__
#define __FUNC__ "harnessclient_kernel_shutdown_oneway_p"
int harnessclient_kernel_shutdown_oneway_p (const rmix_remoteref_t *harness,
                                            unsigned int            provider) {
#ifdef DEBUG
  /* Check harness parameter. */
  if (NULL == harness) {
    errno = EINVAL;
    HARNESS_LOG((HARNESS_WARN(harness parameter is null)))
    errno = EINVAL;
    return -1;
  }
#endif /* DEBUG */
  /* Invoke remote object method. */
  if (0 != rmix_oneway_p(harness, provider, &harnessclient_interface,
                         HARNESS_RMIX_METHODS_KERNEL_SHUTDOWN_INDEX,
                         NULL, 0)) {
    int errno2 = errno;
    RMIX_LOG((RMIX_WARN(unable to invoke remote object method)))
    errno = errno2;
    return -1;
  }
  return 0;
}

/*
 * Invokes harness_kernel_shutdown (client-side method stub).
 * Asynchonously shuts down the kernel and subsequently the forker a remote
 * harness instance using an asynchronous invocation.
 *
 * invokeref = The invocation reference (return).
 * harness   = The RMIX remote object reference of harness.
 * return    = 0 on success or -1 on any error with errno set appropriately.
 */
#undef  __FUNC__
#define __FUNC__ "harnessclient_kernel_shutdown_send"
int harnessclient_kernel_shutdown_send (rmix_invokeref_t       **invokeref,
                                        const rmix_remoteref_t  *harness) {
#ifdef DEBUG
  /* Check invokeref parameter. */
  if (NULL == invokeref) {
    errno = EINVAL;
    HARNESS_LOG((HARNESS_WARN(invokeref parameter is null)))
    errno = EINVAL;
    return -1;
  }
  /* Check harness parameter. */
  if (NULL == harness) {
    errno = EINVAL;
    HARNESS_LOG((HARNESS_WARN(harness parameter is null)))
    errno = EINVAL;
    return -1;
  }
#endif /* DEBUG */
  /* Invoke remote object method. */
  if (0 != rmix_send(invokeref, harness, &harnessclient_interface,
                     HARNESS_RMIX_METHODS_KERNEL_SHUTDOWN_INDEX,
                     NULL, 0)) {
    int errno2 = errno;
    RMIX_LOG((RMIX_WARN(unable to invoke remote object method)))
    errno = errno2;
    return -1;
  }
  return 0;
}

/*
 * Invokes harness_kernel_shutdown (client-side method stub).
 * Asynchonously shuts down the kernel and subsequently the forker at a remote
 * harness instance using an asynchronous invocation without loading the
 * provider plug-in.
 *
 * invokeref = The invocation reference (return).
 * harness   = The RMIX remote object reference of harness.
 * provider  = The preloaded RMIX provider plug-in handle.
 * return    = 0 on success or -1 on any error with errno set appropriately.
 */
#undef  __FUNC__
#define __FUNC__ "harnessclient_kernel_shutdown_send_p"
int harnessclient_kernel_shutdown_send_p (rmix_invokeref_t       **invokeref,
                                          const rmix_remoteref_t  *harness,
                                          unsigned int             provider) {
#ifdef DEBUG
  /* Check invokeref parameter. */
  if (NULL == invokeref) {
    errno = EINVAL;
    HARNESS_LOG((HARNESS_WARN(invokeref parameter is null)))
    errno = EINVAL;
    return -1;
  }
  /* Check harness parameter. */
  if (NULL == harness) {
    errno = EINVAL;
    HARNESS_LOG((HARNESS_WARN(harness parameter is null)))
    errno = EINVAL;
    return -1;
  }
#endif /* DEBUG */
  /* Invoke remote object method. */
  if (0 != rmix_send_p(invokeref, harness, provider, &harnessclient_interface,
                       HARNESS_RMIX_METHODS_KERNEL_SHUTDOWN_INDEX,
                       NULL, 0)) {
    int errno2 = errno;
    RMIX_LOG((RMIX_WARN(unable to invoke remote object method)))
    errno = errno2;
    return -1;
  }
  return 0;
}

/*
 * Retrieves harness_kernel_shutdown output (client-side method stub).
 * Returns the output of a previous asynchronous invocation that asynchronously
 * shuts down the kernel and subsequently the forker a remote harness instance.
 *
 * invokeref = The invocation reference (also return).
 * return    = 0 on success or -1 on any error with errno set appropriately.
 */
#undef  __FUNC__
#define __FUNC__ "harnessclient_kernel_shutdown_retrieve"
int harnessclient_kernel_shutdown_retrieve (rmix_invokeref_t **invokeref) {
#ifdef DEBUG
  /* Check invokeref parameter. */
  if (NULL == invokeref) {
    errno = EINVAL;
    HARNESS_LOG((HARNESS_WARN(invokeref parameter is null)))
    errno = EINVAL;
    return -1;
  }
  /* Check *invokeref parameter. */
  if (NULL == *invokeref) {
    errno = EINVAL;
    HARNESS_LOG((HARNESS_WARN(*invokeref parameter is null)))
    errno = EINVAL;
    return -1;
  }
#endif /* DEBUG */
  /* Invoke remote object method. */
  if (0 != rmix_retrieve(invokeref, NULL, 0)) {
    int errno2 = errno;
    RMIX_LOG((RMIX_WARN(unable to invoke remote object method)))
    errno = errno2;
    return -1;
  }
  return 0;
}

/*
 * Retrieves harness_kernel_shutdown output (client-side method stub).
 * Returns the output of a previous asynchronous invocation that asynchronously
 * shuts down the kernel and subsequently the forker a remote harness instance
 * without unloading the provider plug-in.
 *
 * invokeref = The invocation reference (also return).
 * return    = 0 on success or -1 on any error with errno set appropriately.
 */
#undef  __FUNC__
#define __FUNC__ "harnessclient_kernel_shutdown_retrieve_p"
int harnessclient_kernel_shutdown_retrieve_p (rmix_invokeref_t **invokeref) {
#ifdef DEBUG
  /* Check invokeref parameter. */
  if (NULL == invokeref) {
    errno = EINVAL;
    HARNESS_LOG((HARNESS_WARN(invokeref parameter is null)))
    errno = EINVAL;
    return -1;
  }
  /* Check *invokeref parameter. */
  if (NULL == *invokeref) {
    errno = EINVAL;
    HARNESS_LOG((HARNESS_WARN(*invokeref parameter is null)))
    errno = EINVAL;
    return -1;
  }
#endif /* DEBUG */
  /* Invoke remote object method. */
  if (0 != rmix_retrieve_p(invokeref, NULL, 0)) {
    int errno2 = errno;
    RMIX_LOG((RMIX_WARN(unable to invoke remote object method)))
    errno = errno2;
    return -1;
  }
  return 0;
}

/*
 * Invokes harness_kernel_shutdown (client-side method stub).
 * Asynchonously shuts down the kernel and subsequently the forker at multiple
 * remote harness instances in parallel using asynchonous method invocations.
 *
 * count     = The number of remote harness instances.
 * harnesses = The array of RMIX remote object references of harness instances.
 * return    = The number of initiated shutdowns with errno set appropriately.
 */
#undef  __FUNC__
#define __FUNC__ "harnessclient_kernel_pshutdown"
unsigned int harnessclient_kernel_pshutdown
  (unsigned int             count,
   const rmix_remoteref_t **harnesses) {
  int                errno2;
  unsigned int       initiated, index;
  rmix_invokeref_t **invocations;
#ifdef DEBUG
  /* Check harnesses parameter. */
  if ((NULL == harnesses)&&(0 != count)) {
    errno = EINVAL;
    RMIX_LOG((RMIX_WARN(harnesses parameter is null)))
    errno = EINVAL;
    return -1;
  }
  /* Check count parameter. */
  if ((0 == count)&&(NULL != harnesses)) {
    errno = EINVAL;
    RMIX_LOG((RMIX_WARN(count parameter is zero)))
    errno = EINVAL;
    return -1;
  }
#endif /* DEBUG */
  /* Allocate invocation reference array. */
  if (NULL == (invocations = malloc(count * sizeof(rmix_invokeref_t*)))) {
    errno2 = errno;
    RMIX_LOG((RMIX_WARN(unable to allocate invocation reference array)))
    errno = errno2;
    return 0;
  }
  /* Initialize errno and initiated counter.*/
  errno2 = 0;
  initiated = 0;
  /* Shutdown all kernels in forward order (send part). */
  for (index = 0; index < count; index++) {
    /* Shutdown kernel (send part). */
    if (0 != harnessclient_kernel_shutdown_send(&invocations[index],
                                                harnesses[index])) {
      errno2 = (0 == errno2)?errno:errno2;
      RMIX_LOG((RMIX_WARN(unable to shutdown kernel (send part))))
      invocations[index] = NULL;
    }
  }
  /* Shutdown all kernels in forward order (retrieve part). */
  for (index = 0; index < count; index++) {
    /* Check for successful send. */
    if (NULL != invocations[index]) {
      /* Shutdown kernel (retrieve part). */
      if (0 != harnessclient_kernel_shutdown_retrieve(&invocations[index])) {
        errno2 = (0 == errno2)?errno:errno2;
        RMIX_LOG((RMIX_WARN(unable to shutdown kernel (retrieve part))))
      } else {
        initiated++;
      }
    }
  }
  /* Deallocate invocation reference array. */
  free(invocations);
  errno = errno2;
  return initiated;
}

/*
 * Executes a command line remotely using the current HARNESS_RSH command.
 * Helper function to start programs remotely without having Harness. This
 * function will fork-exec HARNESS_RSH with host and command line as argument.
 * It produces a SIGCHLD as result of the exiting HARNESS_RSH command.
 *
 * NOTE: Do NOT use this function within a multi-threaded environment, such as
 *       Harness. If you are inside a Harness run time environment, use the
 *       Harness process control function instead. Fork-exec in a multi-
 *       threaded environment is not very well defined. You may cause harm to
 *       your run time environment and to data. This function is intended for
 *       Harness client tools to bootstrap Harness run time environments.
 *
 * host   = The host to execute at (may include user: user@host).
 * cmd    = The command line to execute.
 * return = 0 on success or -1 on any error with errno set appropriately.
 */
#undef  __FUNC__
#define __FUNC__ "harnessclient_rexec"
int harnessclient_rexec (const char *host,
                         const char *cmd) {
  int          status, error;
  pid_t        pid;
  const char  *rsh;
#ifdef DEBUG
  /* Check host parameter. */
  if (NULL == host) {
    errno = EINVAL;
    HARNESS_LOG((HARNESS_WARN(host parameter is null)))
    errno = EINVAL;
    return -1;
  }
  /* Check cmd parameter. */
  if (NULL == cmd) {
    errno = EINVAL;
    HARNESS_LOG((HARNESS_WARN(cmd parameter is null)))
    errno = EINVAL;
    return -1;
  }
#endif /* DEBUG */
  /* Get HARNESS_RSH command. */
  rsh = getenv("HARNESS_RSH");
  /* Ckeck HARNESS_RSH command. */
  if (NULL == rsh) {
    errno = EINVAL;
    HARNESS_LOG((HARNESS_WARN(HARNESS_RSH not set)))
    errno = EINVAL;
    return -1;
  }
  /* Fork child process. */
  switch (pid = fork()) {
    /* Check for error. */
    case -1: {
      int errno2 = errno;
      HARNESS_LOG((HARNESS_WARN(unable to fork process)))
      errno = errno2;
      return -1;
    }
    /* Check for child. */
    case 0: {
      /* Execute HARNESS_RSH command with host and command lina as parameter. */
      execl(rsh, rsh, host, cmd, NULL);
      HARNESS_LOG((HARNESS_WARN(unable to execute remote command)))
      /* Exit unsuccessfully. */
      exit(EXIT_FAILURE);
    }
    /* Check for parent. */
    default: {
      /* Wait for child exit. */
      while ((-1 == (error = waitpid(pid, &status, 0)))&&(EINTR == errno));
      /* Check for error. */
      if (pid != error) {
        int errno2 = errno;
        HARNESS_LOG((HARNESS_WARN(unable to wait for child process)))
        errno = errno2;
        return -1;
      }
      /* Check for error in clild. */
      if ((!WIFEXITED(status))||(EXIT_SUCCESS != WEXITSTATUS(status))) {
        errno = EINVAL;
        HARNESS_LOG((HARNESS_WARN(unable to execute remote command)))
        errno = EINVAL;
        return -1;
      }
      return 0;
    }
  }
}

/*
 * Executes a command line remotely at multiple locations using the current
 * HARNESS_RSH command. Helper function to start programs remotely without
 * having Harness. This function will fork-exec HARNESS_RSH with host and
 * command line as argument for every entry in the host and command arrays.
 * It produces multiple SIGCHLDs as result of the exiting HARNESS_RSH command.
 * Returns immediately upon failure. The command to execute may be different
 * for every host.
 *
 * NOTE: Do NOT use this function within a multi-threaded environment, such as
 *       Harness. If you are inside a Harness run time environment, use the
 *       Harness process control function instead. Fork-exec in a multi-
 *       threaded environment is not very well defined. You may cause harm to
 *       your run time environment and to data. This function is intended for
 *       Harness client tools to bootstrap Harness run time environments.
 *
 * count  = The number of hosts to execute at and commands to execute.
 * host   = The host array to execute at (may include user: user@host).
 * cmds   = The command line array to execute.
 * return = The number of successful executions with errno set appropriately.
 */
#undef  __FUNC__
#define __FUNC__ "harnessclient_rexec_mp"
unsigned int harnessclient_rexec_mp (unsigned int   count,
                                     const char   **hosts,
                                     const char   **cmds) {
  unsigned int index;
#ifdef DEBUG
  /* Check hosts parameter. */
  if ((NULL == hosts)&&(0 != count)) {
    errno = EINVAL;
    RMIX_LOG((RMIX_WARN(hosts parameter is null)))
    errno = EINVAL;
    return -1;
  }
  /* Check count parameter. */
  if ((0 == count)&&(NULL != hosts)) {
    errno = EINVAL;
    RMIX_LOG((RMIX_WARN(count parameter is zero)))
    errno = EINVAL;
    return -1;
  }
  /* Check cmds parameter. */
  if ((NULL == cmds)&&(0 != count)) {
    errno = EINVAL;
    RMIX_LOG((RMIX_WARN(hosts parameter is null)))
    errno = EINVAL;
    return -1;
  }
  /* Check count parameter. */
  if ((0 == count)&&(NULL != cmds)) {
    errno = EINVAL;
    RMIX_LOG((RMIX_WARN(count parameter is zero)))
    errno = EINVAL;
    return -1;
  }
#endif /* DEBUG */
  /* Execute all remote commands. */
  for (index = 0; index < count; index++) {
    /* Execute remote command. */
    if (0 != harnessclient_rexec(hosts[index], cmds[index])) {
      int errno2 = errno;
      HARNESS_LOG((HARNESS_WARN(unable to execute remote command)))
      errno = errno2;
      return index;
    }
  }
  return index;
}

/*
 * Executes a command line remotely at multiple locations using the current
 * HARNESS_RSH command. Helper function to start programs remotely without
 * having Harness. This function will fork-exec HARNESS_RSH with host and
 * command line as argument for every entry in the host and command arrays.
 * It produces multiple SIGCHLDs as result of the exiting HARNESS_RSH command.
 * Returns immediately upon failure. The command to execute is the same for
 * every host.
 *
 * NOTE: Do NOT use this function within a multi-threaded environment, such as
 *       Harness. If you are inside a Harness run time environment, use the
 *       Harness process control function instead. Fork-exec in a multi-
 *       threaded environment is not very well defined. You may cause harm to
 *       your run time environment and to data. This function is intended for
 *       Harness client tools to bootstrap Harness run time environments.
 *
 * count  = The number of hosts to execute at and commands to execute.
 * host   = The host array to execute at (may include user: user@host).
 * cmd    = The command line to execute.
 * return = The number of successful executions with errno set appropriately.
 */
#undef  __FUNC__
#define __FUNC__ "harnessclient_rexec_sp"
unsigned int harnessclient_rexec_sp (unsigned int   count,
                                     const char   **hosts,
                                     const char    *cmd) {
  unsigned int index;
#ifdef DEBUG
  /* Check hosts parameter. */
  if ((NULL == hosts)&&(0 != count)) {
    errno = EINVAL;
    RMIX_LOG((RMIX_WARN(hosts parameter is null)))
    errno = EINVAL;
    return -1;
  }
  /* Check count parameter. */
  if ((0 == count)&&(NULL != hosts)) {
    errno = EINVAL;
    RMIX_LOG((RMIX_WARN(count parameter is zero)))
    errno = EINVAL;
    return -1;
  }
  /* Check cmd parameter. */
  if (NULL == cmd) {
    errno = EINVAL;
    HARNESS_LOG((HARNESS_WARN(cmd parameter is null)))
    errno = EINVAL;
    return -1;
  }
#endif /* DEBUG */
  /* Execute all remote commands. */
  for (index = 0; index < count; index++) {
    /* Execute remote command. */
    if (0 != harnessclient_rexec(hosts[index], cmd)) {
      int errno2 = errno;
      HARNESS_LOG((HARNESS_WARN(unable to execute remote command)))
      errno = errno2;
      return index;
    }
  }
  return index;
}

/*
 * Executes a command line remotely using the current HARNESS_RSH command.
 * Helper function to start programs remotely without having Harness. This
 * function will fork-exec HARNESS_RSH with host and command line as argument.
 * It produces a SIGCHLD as result of the exiting HARNESS_RSH command.
 *
 * NOTE: Do NOT use this function within a multi-threaded environment, such as
 *       Harness. If you are inside a Harness run time environment, use the
 *       Harness process control function instead. Fork-exec in a multi-
 *       threaded environment is not very well defined. You may cause harm to
 *       your run time environment and to data. This function is intended for
 *       Harness client tools to bootstrap Harness run time environments.
 *
 * parameters = The export parameters of the host to execute at (needs to
 *              contain HOST=... and may contain USER=... ).
 * cmd        = The command line to execute.
 * return     = 0 on success or -1 on any error with errno set appropriately.
 */
#undef  __FUNC__
#define __FUNC__ "harnessclient_rexec2"
int harnessclient_rexec2 (const rmix_parameters_t *parameters,
                          const char              *cmd) {
  char       *host2;
  const char *host, *user;
#ifdef DEBUG
  /* Check parameters parameter. */
  if (NULL == parameters) {
    errno = EINVAL;
    HARNESS_LOG((HARNESS_WARN(parameters parameter is null)))
    errno = EINVAL;
    return -1;
  }
  /* Check cmd parameter. */
  if (NULL == cmd) {
    errno = EINVAL;
    HARNESS_LOG((HARNESS_WARN(cmd parameter is null)))
    errno = EINVAL;
    return -1;
  }
#endif /* DEBUG */
  /* Search for host parameter. */
  if (0 != rmix_parameters_search(parameters, &host, "HOST")) {
    if (ESRCH == errno) {
      errno = EINVAL;
      RMIX_LOG((RMIX_WARN(no host specified)))
      errno = EINVAL;
      return -1;
    } else {
      int errno2 = errno;
      RMIX_LOG((RMIX_WARN(unable to search for host parameter)))
      errno = errno2;
      return -1;
    }
  }
  /* Search for additional user parameter. */
  if (0 != rmix_parameters_search(parameters, &user, "USER")) {
    if (ESRCH != errno) {
      int errno2 = errno;
      RMIX_LOG((RMIX_WARN(unable to search for user parameter)))
      errno = errno2;
      return -1;
    }
    user = NULL;
  }
  /* Check for user parameter. */
  if (NULL != user) {
    /* Combine host and user parameter. */
    if (-1 == asprintf(&host2, "%s@%s", user, host)) {
      int errno2 = errno;
      RMIX_LOG((RMIX_WARN(unable to combine host and user parameter)))
      errno = errno2;
      return -1;
    }
    host = host2;
  }
  /* Execute remote command. */
  if (0 != harnessclient_rexec(host, cmd)) {
    int errno2 = errno;
    HARNESS_LOG((HARNESS_WARN(unable to execute remote command)))
    errno = errno2;
    return -1;
  }
  /* Check for user. */
  if (NULL != user) {
    /* Deallocate host and user combination. */
    free(host2);
  }
  return 0;
}

/*
 * Executes a command line remotely at multiple locations using the current
 * HARNESS_RSH command. Helper function to start programs remotely without
 * having Harness. This function will fork-exec HARNESS_RSH with host and
 * command line as argument for every entry in the host and command arrays.
 * It produces multiple SIGCHLDs as result of the exiting HARNESS_RSH command.
 * Returns immediately upon failure. The command to execute may be different
 * for every host.
 *
 * NOTE: Do NOT use this function within a multi-threaded environment, such as
 *       Harness. If you are inside a Harness run time environment, use the
 *       Harness process control function instead. Fork-exec in a multi-
 *       threaded environment is not very well defined. You may cause harm to
 *       your run time environment and to data. This function is intended for
 *       Harness client tools to bootstrap Harness run time environments.
 *
 * count      = The number of hosts to execute at and commands to execute.
 * parameters = The export parameters array of the hosts to execute at (needs to
 *              contain HOST=... and may contain USER=... ).
 * cmds       = The command line array to execute.
 * return     = The number of successful executions with errno set
 *              appropriately.
 */
#undef  __FUNC__
#define __FUNC__ "harnessclient_rexec2_mp"
unsigned int harnessclient_rexec2_mp (unsigned int              count,
                                      const rmix_parameters_t **parameters,
                                      const char              **cmds) {
  unsigned int index;
#ifdef DEBUG
  /* Check parameters parameter. */
  if ((NULL == parameters)&&(0 != count)) {
    errno = EINVAL;
    RMIX_LOG((RMIX_WARN(parameters parameter is null)))
    errno = EINVAL;
    return -1;
  }
  /* Check count parameter. */
  if ((0 == count)&&(NULL != parameters)) {
    errno = EINVAL;
    RMIX_LOG((RMIX_WARN(count parameter is zero)))
    errno = EINVAL;
    return -1;
  }
  /* Check cmds parameter. */
  if ((NULL == cmds)&&(0 != count)) {
    errno = EINVAL;
    RMIX_LOG((RMIX_WARN(hosts parameter is null)))
    errno = EINVAL;
    return -1;
  }
  /* Check count parameter. */
  if ((0 == count)&&(NULL != cmds)) {
    errno = EINVAL;
    RMIX_LOG((RMIX_WARN(count parameter is zero)))
    errno = EINVAL;
    return -1;
  }
#endif /* DEBUG */
  /* Execute all remote commands. */
  for (index = 0; index < count; index++) {
    /* Execute remote command. */
    if (0 != harnessclient_rexec2(parameters[index], cmds[index])) {
      int errno2 = errno;
      HARNESS_LOG((HARNESS_WARN(unable to execute remote command)))
      errno = errno2;
      return index;
    }
  }
  return index;
}

/*
 * Executes a command line remotely at multiple locations using the current
 * HARNESS_RSH command. Helper function to start programs remotely without
 * having Harness. This function will fork-exec HARNESS_RSH with host and
 * command line as argument for every entry in the host and command arrays.
 * It produces multiple SIGCHLDs as result of the exiting HARNESS_RSH command.
 * Returns immediately upon failure. The command to execute is the same for
 * every host.
 *
 * NOTE: Do NOT use this function within a multi-threaded environment, such as
 *       Harness. If you are inside a Harness run time environment, use the
 *       Harness process control function instead. Fork-exec in a multi-
 *       threaded environment is not very well defined. You may cause harm to
 *       your run time environment and to data. This function is intended for
 *       Harness client tools to bootstrap Harness run time environments.
 *
 * count      = The number of hosts to execute at and commands to execute.
 * parameters = The export parameters array of the hosts to execute at (needs to
 *              contain HOST=... and may contain USER=... ).
 * cmd        = The command line to execute.
 * return     = The number of successful executions with errno set
 *              appropriately.
 */
#undef  __FUNC__
#define __FUNC__ "harnessclient_rexec2_sp"
unsigned int harnessclient_rexec2_sp (unsigned int              count,
                                      const rmix_parameters_t **parameters,
                                      const char               *cmd) {
  unsigned int index;
#ifdef DEBUG
  /* Check parameters parameter. */
  if ((NULL == parameters)&&(0 != count)) {
    errno = EINVAL;
    RMIX_LOG((RMIX_WARN(parameters parameter is null)))
    errno = EINVAL;
    return -1;
  }
  /* Check count parameter. */
  if ((0 == count)&&(NULL != parameters)) {
    errno = EINVAL;
    RMIX_LOG((RMIX_WARN(count parameter is zero)))
    errno = EINVAL;
    return -1;
  }
  /* Check cmd parameter. */
  if (NULL == cmd) {
    errno = EINVAL;
    HARNESS_LOG((HARNESS_WARN(cmd parameter is null)))
    errno = EINVAL;
    return -1;
  }
#endif /* DEBUG */
  /* Execute all remote commands. */
  for (index = 0; index < count; index++) {
    /* Execute remote command. */
    if (0 != harnessclient_rexec2(parameters[index], cmd)) {
      int errno2 = errno;
      HARNESS_LOG((HARNESS_WARN(unable to execute remote command)))
      errno = errno2;
      return index;
    }
  }
  return index;
}

/*
 * Executes a command line remotely using the current HARNESS_RSH command.
 * Helper function to start programs remotely without having Harness. This
 * function will fork-exec HARNESS_RSH with host and command line as argument.
 * It produces a SIGCHLD as result of the exiting HARNESS_RSH command.
 *
 * NOTE: Do NOT use this function within a multi-threaded environment, such as
 *       Harness. If you are inside a Harness run time environment, use the
 *       Harness process control function instead. Fork-exec in a multi-
 *       threaded environment is not very well defined. You may cause harm to
 *       your run time environment and to data. This function is intended for
 *       Harness client tools to bootstrap Harness run time environments.
 *
 * parameters = The export parameters of the host to execute at (needs to
 *              contain HOST=... and may contain USER=... ).
 * cmd        = The command line to execute.
 * return     = 0 on success or -1 on any error with errno set appropriately.
 */
#undef  __FUNC__
#define __FUNC__ "harnessclient_rexec3"
int harnessclient_rexec3 (const char *parameters,
                          const char *cmd) {
  rmix_parameters_t *parameters2;
#ifdef DEBUG
  /* Check parameters parameter. */
  if (NULL == parameters) {
    errno = EINVAL;
    HARNESS_LOG((HARNESS_WARN(parameters parameter is null)))
    errno = EINVAL;
    return -1;
  }
  /* Check cmd parameter. */
  if (NULL == cmd) {
    errno = EINVAL;
    HARNESS_LOG((HARNESS_WARN(cmd parameter is null)))
    errno = EINVAL;
    return -1;
  }
#endif /* DEBUG */
  /* Create papameters. */
  if (0 != rmix_parameters_create4(&parameters2, parameters)) {
    int errno2 = errno;
    RMIX_LOG((RMIX_WARN(unable to create parameters)))
    errno = errno2;
    return -1;
  }
  /* Execute remote command. */
  if (0 != harnessclient_rexec2(parameters2, cmd)) {
    int errno2 = errno;
    HARNESS_LOG((HARNESS_WARN(unable to execute remote command)))
    /* Destroy papameters. */
    if (0 != rmix_parameters_destroy(&parameters2)) {
      RMIX_LOG((RMIX_WARN(unable to destroy parameters)))
    }
    errno = errno2;
    return -1;
  }
  /* Destroy papameters. */
  if (0 != rmix_parameters_destroy(&parameters2)) {
    int errno2 = errno;
    RMIX_LOG((RMIX_WARN(unable to destroy parameters)))
    errno = errno2;
    return -1;
  }
  return 0;
}

/*
 * Executes a command line remotely at multiple locations using the current
 * HARNESS_RSH command. Helper function to start programs remotely without
 * having Harness. This function will fork-exec HARNESS_RSH with host and
 * command line as argument for every entry in the host and command arrays.
 * It produces multiple SIGCHLDs as result of the exiting HARNESS_RSH command.
 * Returns immediately upon failure. The command to execute may be different
 * for every host.
 *
 * NOTE: Do NOT use this function within a multi-threaded environment, such as
 *       Harness. If you are inside a Harness run time environment, use the
 *       Harness process control function instead. Fork-exec in a multi-
 *       threaded environment is not very well defined. You may cause harm to
 *       your run time environment and to data. This function is intended for
 *       Harness client tools to bootstrap Harness run time environments.
 *
 * count      = The number of hosts to execute at and commands to execute.
 * parameters = The export parameters array of the hosts to execute at (needs to
 *              contain HOST=... and may contain USER=... ).
 * cmds       = The command line array to execute.
 * return     = The number of successful executions with errno set
 *              appropriately.
 */
#undef  __FUNC__
#define __FUNC__ "harnessclient_rexec3_mp"
unsigned int harnessclient_rexec3_mp (unsigned int   count,
                                      const char   **parameters,
                                      const char   **cmds) {
  unsigned int index;
#ifdef DEBUG
  /* Check parameters parameter. */
  if ((NULL == parameters)&&(0 != count)) {
    errno = EINVAL;
    RMIX_LOG((RMIX_WARN(parameters parameter is null)))
    errno = EINVAL;
    return -1;
  }
  /* Check count parameter. */
  if ((0 == count)&&(NULL != parameters)) {
    errno = EINVAL;
    RMIX_LOG((RMIX_WARN(count parameter is zero)))
    errno = EINVAL;
    return -1;
  }
  /* Check cmds parameter. */
  if ((NULL == cmds)&&(0 != count)) {
    errno = EINVAL;
    RMIX_LOG((RMIX_WARN(hosts parameter is null)))
    errno = EINVAL;
    return -1;
  }
  /* Check count parameter. */
  if ((0 == count)&&(NULL != cmds)) {
    errno = EINVAL;
    RMIX_LOG((RMIX_WARN(count parameter is zero)))
    errno = EINVAL;
    return -1;
  }
#endif /* DEBUG */
  /* Execute all remote commands. */
  for (index = 0; index < count; index++) {
    /* Execute remote command. */
    if (0 != harnessclient_rexec3(parameters[index], cmds[index])) {
      int errno2 = errno;
      HARNESS_LOG((HARNESS_WARN(unable to execute remote command)))
      errno = errno2;
      return index;
    }
  }
  return index;
}

/*
 * Executes a command line remotely at multiple locations using the current
 * HARNESS_RSH command. Helper function to start programs remotely without
 * having Harness. This function will fork-exec HARNESS_RSH with host and
 * command line as argument for every entry in the host and command arrays.
 * It produces multiple SIGCHLDs as result of the exiting HARNESS_RSH command.
 * Returns immediately upon failure. The command to execute is the same for
 * every host.
 *
 * NOTE: Do NOT use this function within a multi-threaded environment, such as
 *       Harness. If you are inside a Harness run time environment, use the
 *       Harness process control function instead. Fork-exec in a multi-
 *       threaded environment is not very well defined. You may cause harm to
 *       your run time environment and to data. This function is intended for
 *       Harness client tools to bootstrap Harness run time environments.
 *
 * count      = The number of hosts to execute at and commands to execute.
 * parameters = The export parameters array of the hosts to execute at (needs to
 *              contain HOST=... and may contain USER=... ).
 * cmd        = The command line to execute.
 * return     = The number of successful executions with errno set
 *              appropriately.
 */
#undef  __FUNC__
#define __FUNC__ "harnessclient_rexec3_sp"
unsigned int harnessclient_rexec3_sp (unsigned int   count,
                                      const char   **parameters,
                                      const char    *cmd) {
  unsigned int index;
#ifdef DEBUG
  /* Check parameters parameter. */
  if ((NULL == parameters)&&(0 != count)) {
    errno = EINVAL;
    RMIX_LOG((RMIX_WARN(parameters parameter is null)))
    errno = EINVAL;
    return -1;
  }
  /* Check count parameter. */
  if ((0 == count)&&(NULL != parameters)) {
    errno = EINVAL;
    RMIX_LOG((RMIX_WARN(count parameter is zero)))
    errno = EINVAL;
    return -1;
  }
  /* Check cmd parameter. */
  if (NULL == cmd) {
    errno = EINVAL;
    HARNESS_LOG((HARNESS_WARN(cmd parameter is null)))
    errno = EINVAL;
    return -1;
  }
#endif /* DEBUG */
  /* Execute all remote commands. */
  for (index = 0; index < count; index++) {
    /* Execute remote command. */
    if (0 != harnessclient_rexec3(parameters[index], cmd)) {
      int errno2 = errno;
      HARNESS_LOG((HARNESS_WARN(unable to execute remote command)))
      errno = errno2;
      return index;
    }
  }
  return index;
}

/******************************************************************************
 *
 * END OF FILE
 *
 ******************************************************************************/
